/*  V.4000 Functions - #include "4.000_changes.h"
     ---------------------------------------------
  - shiftTo()
  - clutchOffset()
  - CANretard()
  - afterShiftRevs()
  - resetShiftRequests()
  - solOffset()
  - getGear()
  - SetLEDs()
  - setOut3PWM()
  - PNRset_outputs()
  - getLeverDigital(); // in getLever.h
  - getLeverADC();     // in getLever.h
  - SetSparePort();    // in sparePorts.h 
  
  SPI
  ---
  - 
 */

#pragma CODE_SEG MAIN_ROM

// ---------------------------------------------------------------------------------------------

signed char shiftTo(signed char new_gear)  
{
// This function performs the shift by switching the solenoids. It takes the target gear as an argument, 
// and returns the gear shifted to.
// It calculates whether the shift is an upshift or downshift.
// For example, a shift to second can be done like this:
//
// outpc.current_gear = shiftTo(2);
//
// and a shift from the auto table can be done like this:
//
// outpc.current_gear = shiftTo(outpc.target_gear);
//
// where the target gear is set in a separate function:
//
// outpc.target_gear = auto_gear_lookup(outpc.speedo,outpc.LOAD_short,NUM_MPH,NUM_LOADS,&inpram.vss_table1[0],&inpram.LOAD_table1[0],&inpram.auto_table1[0][0]);

unsigned int tmpU16;
unsigned char ix;
unsigned int offset;
unsigned char cltchoffset;

if ((waiting[0]) || (shifting_flag) || (outpc.hyst_timer))  // if shifting or waiting during a shift or in hysteresis return current gear
  {	 // do nothing and return
  return outpc.current_gear;
  }
	  
offset = solOffset(new_gear);

if (new_gear == outpc.current_gear) return outpc.current_gear;

if (afterShiftRevs(new_gear) > 0) // Check if shift will under/over-rev engine, if not, proceed with shift
                                  // Function checks if rpm checking is enabled
  {
  // Reset PC PWM slip parameters so we will start again after shift
  Slip_msec = 0;
  outpc.Slip_Adjustment = 0; // reset adjustment
  
   // set solC shift PWM value if the shift value is not zero
   // will set to final value after shift is completed
   // if the user value is not zero and PWM is enabled
   if (inpram.PWM_enable & 0x04)
     {
     if (new_gear > outpc.current_gear)
       { // upshift
       if (inpram.solCupshiftPWM[new_gear-2]) OUTPUT3pwm = (char) ((OUTPUT3_ON * inpram.solCupshiftPWM[new_gear-2] + 50)/100);
       }
       else  // downshift
       {
       if (inpram.solCupshiftPWM[new_gear-1]) OUTPUT3pwm = (char) ((OUTPUT3_ON * inpram.solCdwnshiftPWM[new_gear-1] + 50)/100); 
       }
       Output3(0); // set state indicator (dummy variable when not in NAG mode)
     }
    
   // Set flags      
   shifting_flag = 1; // set shifting flag
   if ((outpc.auto_mode > 0) && (rev_shift_flag == 0))  // set flag for hysteresis - auto shift mode only
   	{
   	outpc.hyst_timer = in2ram.hyst_delay; // set delay, will count down in 0.100 timer
   	}
   	else
   	{
   	outpc.hyst_timer = 0; // clear flag for hysteresis - manual shift mode or rpm forced shift only
   	}
   		
   // Send shift timing retard to MS-II
   if (inpram.CAN_enabled)
     {
     if (outpc.LOAD > inpram.Tretard_load)
       {
       if (new_gear < outpc.current_gear) outpc.SpkAdj = inpram.dwnshift_retard;  // down shift
       if (new_gear > outpc.current_gear) outpc.SpkAdj = inpram.upshift_retard;   // up shift
       outpc.FuelAdj = 0; // reset in case we set for slip reasons
       sendCANAdj(); 
       } // end if (outpc.LOAD > ...
       else	 // LOAD less than Tretard_load
       {  // reset spark and fuel adjustments in case we have set them for slip reasons
       outpc.SpkAdj = 0;
       outpc.FuelAdj = 0;
       sendCANAdj();
       }
     } // end if (inpram.CAN_enabled)

    // If spare port 0 used for PWM
	if (in2ram.sp0_mode == 1)
	 {  // sp1 PWM mode									    
     if (in2ram.sp0_active == 1) // de-activate sp1 pwm if ON except during shift
      {
       spr0(0);
      } // end if (inpram.sp1_mode == 1) ...
     if (in2ram.sp0_active == 2)
      { // activate sp1 pwm if ON only while shifting
	   spr0(1);	   
      } 
     } // end if (inpram.sp0_mode == 1) ...
    
    // If spare port 1 or spare port 2 used for PWM
	for (ix=0; ix < 2; ix++)
		{
		if (inpram.sp_mode[ix] == 1)
		 {  // sp1 PWM mode									    
	     if (inpram.sp_active[ix] == 1) // de-activate sp1 pwm if ON except during shift
	      {
	      if (ix == 0) spr1(0); else spr2(0);
	      } // end if (inpram.sp1_mode == 1) ...
	     if (inpram.sp_active[ix] == 2)
	      { // activate sp1 pwm if ON only while shifting
		  if (ix == 0) spr1(1); else spr2(1);	   
	      } 
	     } // end if (inpram.sp1_mode == 1) ...
  		}	 // end for (ix=0;
	     
	     
	    // If spare port 0 used for PWM
	if (inpram.sp_mode[0] == 1)
	 {  // sp1 PWM mode									    
     if (inpram.sp_active[0] == 1) // de-activate sp1 pwm if ON except during shift
      {
       spr1(0);
      } // end if (inpram.sp1_mode == 1) ...
     if (inpram.sp_active[0] == 2)
      { // activate sp1 pwm if ON only while shifting
	   spr1(1);	   
      } 
     } // end if (inpram.sp0_mode == 1) ...
     
         // If spare port 0 used for PWM
	if (inpram.sp_mode[1] == 1)
	 {  // sp1 PWM mode									    
     if (inpram.sp_active[1] == 1) // de-activate sp1 pwm if ON except during shift
      {
       spr2(0);
      } // end if (inpram.sp1_mode == 1) ...
     if (inpram.sp_active[1] == 2)
      { // activate sp1 pwm if ON only while shifting
	   spr2(1);	   
      } 
     } // end if (inpram.sp0_mode == 1) ...
	         
   // Adjust line pressure
   if (in2ram.sp3_mode == 0) tmpU16 = outpc.PC_duty; else tmpU16 = 100 - outpc.sp3PWM; // Grab the appropriate PWM% for testing against shift limits
   if (((new_gear > outpc.current_gear) && (outpc.current_gear > 1)) && ((in2ram.sp3_mode == 0) || ((in2ram.sp3_mode == 1) && (inpram.LPshiftOptn != 0))))  // if upshifting
     {
     // maxUPpressure[0] is not allowed,
     // maxUPpressure[1] is value for upshift to 2nd,
     // maxUPpressure[2] is value for upshift to 3rd,
     // maxUPpressure[3] is value for upshift to 4th,
     // ...
     if ((in2ram.sp3_mode == 0) || ((in2ram.sp3_mode == 1) && (inpram.LPshiftOptn == 2)))
     	{  // if line pressure mode, or if spare port PWM mode and use limit during shift
		if (tmpU16 > (100-inpram.maxUPpressure[new_gear-1]))
			{
			// Set the pressure
			PC_DUTY = (unsigned char)((((int) inpram.maxUPpressure[new_gear-1] * (int) PC_PER) + (PC_PER/2))/100);
			// set value for datalog
			outpc.sp3PWM = (unsigned char)(((int) PC_DUTY*100 + (PC_PER/2))/(int) PC_PER); // set datalog value (100-value to have datalog value same as PC table)
			outpc.PC_duty = 100 - outpc.sp3PWM; // set datalog value
			// wait for pressure to adjust
			DelayTimer1 = inpram.pressure_delay;
			while (DelayTimer1)
				{  // do something pointless...
				checkCANbus();
				// do the Timer details
				last0001 = Timer0001();
				lastADC = TimerADC();
				last0100 = Timer0100();
				// last1000 = Timer1000();
				}
			} // end if PC_duty
     	}
     	else if ((in2ram.sp3_mode == 1) && (inpram.LPshiftOptn == 1))
     	{
     	// Set the pressure to the OFF value
		PC_DUTY = (unsigned char)((((int) in2ram.sp3_off * (int) PC_PER) + (PC_PER/2))/100);
		// set value for datalog
		outpc.sp3PWM = (unsigned char)(((int) PC_DUTY*100 + (PC_PER/2))/(int) PC_PER); // set datalog value (100-value to have datalog value same as PC table)
		outpc.PC_duty = 100 - outpc.sp3PWM; // set datalog value
		DelayTimer1 = inpram.pressure_delay;
		while (DelayTimer1)
			{  // do something pointless...
			checkCANbus();
			// do the Timer details (will also do in delay loops)
			last0001 = Timer0001();
			lastADC = TimerADC();
			last0100 = Timer0100();
			// last1000 = Timer1000();
			}
		}
     }
     else if ((new_gear < outpc.current_gear) && (new_gear > 1)) // if downshifting
     {
     // maxDWNpressure[0] is value for downshift to 1st,
     // maxDWNpressure[1] is value for downshift to 2nd,
     // maxDWNpressure[2] is value for downshift to 3rd,
     // maxDWNpressure[3] is value for downshift to 4th,
     // ...
     if ((in2ram.sp3_mode == 0) || ((in2ram.sp3_mode == 1) && (inpram.LPshiftOptn == 2)))
     	{  // if line pressure mode, or if spare port PWM mode and use limit during shift
		if (tmpU16 > (100-inpram.maxDWNpressure[new_gear-1]))
			// if the PC table value is greater than the limit
			{
			// Set the pressure
			PC_DUTY = (unsigned char)(((int) inpram.maxDWNpressure[new_gear-1] * (int) PC_PER + (PC_PER/2))/100);
			// set value for datalog
			outpc.sp3PWM = (unsigned char)(((int) PC_DUTY*100 + (PC_PER/2))/(int) PC_PER); // set datalog value (100-value to have datalog value same as PC table)
			outpc.PC_duty =  100 - outpc.sp3PWM; // set datalog value
			// wait for pressure to adjust
  			DelayTimer1 = inpram.pressure_delay;
			while (DelayTimer1)
				{  // check CANbus comms if necessary
		        checkCANbus();
		        // do the Timer details (will also do in delay loops)
		        last0001 = Timer0001();
				last0010 = Timer0010();
				lastADC = TimerADC();
				last0100 = Timer0100();
				// last1000 = Timer1000();
			    }
			} // end if (outpc.PC_duty < ...
		}
		else if ((in2ram.sp3_mode == 1) && (inpram.LPshiftOptn == 1))
     	{
     	// Set the pressure to the OFF value
		PC_DUTY = (unsigned char)((((int) in2ram.sp3_off * (int) PC_PER) + (PC_PER/2))/100);
		// set value for datalog
		outpc.sp3PWM = (unsigned char)(((int) PC_DUTY*100 + (PC_PER/2))/(int) PC_PER); // set datalog value (100-value to have datalog value same as PC table)
		outpc.PC_duty = 100 - outpc.sp3PWM; // set datalog value
		// wait for pressure to adjust
		DelayTimer1 = inpram.pressure_delay;
		while (DelayTimer1)
			{  // check CANbus comms if necessary
			checkCANbus();
			// do the Timer details
			last0001 = Timer0001();
			last0010 = Timer0010();
			lastADC = TimerADC();
			last0100 = Timer0100();
			// last1000 = Timer1000();
			}
     	}
     } // end else if (new_gear < outpc.current-gear)

   // Set clutch output 0
   if (in2ram.sp0_mode == 4) 
     {
     // Get clutch offset
     cltchoffset = clutchOffset(new_gear);

     // Set the outputs
     if (new_gear > outpc.current_gear)
       {
       // if bit set in clutchNup, turn on 
       if (in2ram.clutch0up & cltchoffset) spr0(1);
       }
       else
       {
       // if bit set in clutchNdwn, turn on 
       if (in2ram.clutch0dwn & cltchoffset) spr0(1);             
       } // end if (new_gear > outpc.current_gear)

     } // end if (inpram.sp0_mode == 4) 

   // Set clutch outputs 1
   if (inpram.sp_mode[0] == 4) 
     {
     // Reset flags
     TCC_follow_flag1  = 0;
     TCC_follow_count1 = 0;
     
     // Get clutch offset
     cltchoffset = clutchOffset(new_gear);

     // Set the outputs
     if (new_gear > outpc.current_gear)
       {
       // if bit set in clutchNup, turn on 
       if (inpram.clutch1up & cltchoffset) spr1(1); else spr1(0);
       }
       else
       {
       // if bit set in clutchNdwn, turn on 
       if (inpram.clutch1dwn & cltchoffset) spr1(1); else spr1(0);             
       } // end if (new_gear > outpc.current_gear)

     } // end if (inpram.sp1_mode == 4) 

   // Set clutch output 2
   if (inpram.sp_mode[1] == 4) 
     {
     // Reset flags
     TCC_follow_flag2  = 0;
     TCC_follow_count2 = 0;
     
     // Get clutch offset
     cltchoffset = clutchOffset(new_gear);

     // Set the outputs
     if (new_gear > outpc.current_gear)
       {
       // if bit set in clutchNup, turn on 
       if (inpram.clutch2up & cltchoffset) spr2(1); else spr2(0); 
       }
       else
       {
       // if bit set in clutchNdwn, turn on 
       if (inpram.clutch2dwn & cltchoffset) spr2(1); else spr2(0);             
       } // end if (new_gear > outpc.current_gear)

     } // end if (inpram.sp2_mode == 4) 

   // Set clutch output 3
   if (in2ram.sp3_mode == 4) 
     {
     // Get clutch offset
     cltchoffset = clutchOffset(new_gear);

     // Set the outputs
     if (new_gear > outpc.current_gear)
       {
       // if bit set in clutchNup, turn on 
       if (in2ram.clutch3up & cltchoffset) spr3(1);
       }
       else
       {
       // if bit set in clutchNdwn, turn on 
       if (in2ram.clutch3dwn & cltchoffset) spr3(1);             
       } // end if (new_gear > outpc.current_gear)

     } // end if (inpram.sp3_mode == 4)
              
   // switch Output 1
   Output1((char) ((offset & inpram.Output1ptrn)/offset));  // if 1 in Output1, then send 1 to turn solA on
           
   // switch Output 2
   Output2((char) ((offset & inpram.Output2ptrn)/offset));  // if 1 in Output2, then send 1 to turn solB on
   
   // switch Output 3 if shift PWM not 0
   if (!inpram.solCupshiftPWM[new_gear-1])
     {
     if (inpram.PWM_enable & 0x04)
       {
       OUTPUT3pwm = dc_32 * ((char) ((offset & inpram.Output3ptrn)/offset));
       }
       else
       {
       OUTPUT3pwm = OUTPUT3_ON * ((char) ((offset & inpram.Output3ptrn)/offset));  
       }
     Output3(0); // set state indicator (dummy argument when not in NAG mode)
     }
            
   // switch Output 4
   if (inpram.sp_mode[1] == 3)   // shift solenoid
   	{
   	Output4((char) ((offset & inpram.Output4ptrn)/offset));  // if 1 in Output4, then send 1 to turn solD on
   	}        
   // switch Output 5
   if ((inpram.mlever_mode == 1) && (inpram.spareInputs > 0)) Output5((char) ((offset & inpram.Output5ptrn)/offset));  // if 1 in Output5, then send 1 to turn solE on

   // switch Output 6
   if ((inpram.mlever_mode == 1) && (inpram.spareInputs == 2)) Output6((char) ((offset & inpram.Output6ptrn)/offset));  // if 1 in Output6, then send 1 to turn solF on
   
   // switch Output 7
   if (inpram.sp_mode[0] == 3) Output7((char) ((offset & inpram.Output7ptrn)/offset));  // if 1 in Output7, then send 1 to turn Output 7 on

   // switch Output 8
   if (in2ram.sp0_mode == 3) Output8((char) ((offset & in2ram.Output8ptrn)/offset));  // if 1 in Output8, then send 1 to turn Output 8 on   

   // switch Output 9
   if (in2ram.sp3_mode == 3) Output9((char) ((offset & in2ram.Output9ptrn)/offset));  // if 1 in Output9, then send 1 to turn Output 9 on   
             
   // int intrp_1dctable(	char sgnx, 
   //						int x, 
   //						unsigned char n, 
   //						int * x_table, 
   //						char * z_table)  
   // wait
	DelayTimer2 = 10 * intrp_1dctable(0,outpc.LOAD/10,(uchar)8,in2ram.shiftCompBins,in2ram.shiftCompValues);
	while (DelayTimer2)
		{  // check CANbus comms if necessary
		checkCANbus();
		// do the Timer details
		last0001 = Timer0001();
		last0010 = Timer0010();
		lastADC = TimerADC();
		last0100 = Timer0100();
		// last1000 = Timer1000();
		}
   // set solC
   if (inpram.solCupshiftPWM[new_gear-1])
     {
     if (inpram.PWM_enable & 0x04)
       {
       OUTPUT3pwm = dc_32 * ((char) ((offset & inpram.Output3ptrn)/offset));
       }
       else
       {
       OUTPUT3pwm = OUTPUT3_ON * ((char) ((offset & inpram.Output3ptrn)/offset));  
       }
     Output3(0); // set state indicator	 (dummy argument when not in NAG mode)
     }
            
   // Send gear retard timing to MS-II
   if (inpram.CAN_enabled) 
     {
     if (outpc.LOAD > inpram.Tretard_load)
       {
       outpc.SpkAdj = CANretard(new_gear); 
       sendCANAdj();
       sentSpkAdj = new_gear;
       }
       else
       {
       outpc.SpkAdj = 0;
       sendCANAdj();
       sentSpkAdj = 0;              
       }
     } // end if (inpram.CAN_enabled)    
         
    // Reset clutch outputs, 0 then 1 then 2
    if (in2ram.sp0_mode == 4) spr0(0);
    if (inpram.sp_mode[0] == 4)     
    	{
    	spr1(0);
    	sp[0] = 0;
    	TCC_follow_count1 = 0;
    	TCC_follow_flag1 = 0;
    	}
    if (inpram.sp_mode[1] == 4)
        {
    	spr2(0);
    	sp[1] = 0;
    	TCC_follow_count2 = 0;
    	TCC_follow_flag2 = 0;
    	}
    if (in2ram.sp3_mode == 4) spr3(0);           

    resetShiftRequests();
    shifting_flag = 0; 		// clear shifting flag
    rev_shift_flag = 0;	 	// clear the rev shift flags
  	outpc.miscStatus &= ~0x18;  // clear forced shift flags
    
   // Reset PWM spare ports if spX_active mode == 1 or 2
   
    // If spare port 0 used for PWM 
	if (in2ram.sp0_mode == 1)
		{  // sp0 PWM mode
    	if (in2ram.sp0_active == 2)
      	{ // de-activate sp0 pwm if ON only while shift
	   	spr0(0);
      	} // end if (inpram.sp1_mode == 1) ...
    	if (in2ram.sp0_active == 1)
      	{ // activate sp0 pwm if ON except during shifting
	   spr0(1);
      	} 
     } // end if (inpram.sp0_mode == 1) ...
   
   // If spare port 1/2 used for PWM 
	for (ix=0; ix < 2; ix++)
		{	
		if (inpram.sp_mode[ix] == 1)
			{  // sp1 PWM mode
    		if (inpram.sp_active[ix] == 2)
    	  		{ // de-activate sp1 pwm if ON only while shift
		   		spr1(0);
    	  		} // end if (inpram.sp1_mode == 1) ...
    		if (inpram.sp_active[ix] == 1)
    	  		{ // activate sp1 pwm if ON except during shifting
		   		spr1(1);
    		  	} 
     		} // end if (inpram.sp1_mode == 1) ...
		} // end for (ix=0 ....
		
  } // end if (afterShiftRevs(char new_gear) > 0)
  else 
  { // will over-rev, don't shift
  new_gear = outpc.current_gear;
  }

  outpc.cruise_flag = 0; // reset cruise flag 

if (inpram.NAGmode == 1) outpc.current_gear = new_gear;

if ((outpc.LOAD_short < outpc.LOAD) && (inpram.load_adj & 0x04)) outpc.LOAD_short = outpc.LOAD;

last_mph = outpc.speedo;      // capture speed after shift
last_kpa = outpc.LOAD_short;  // capture load
last_tps = outpc.rawTPS;

return new_gear;

} // end ShiftTo() function

// ------------------------------------------------------------------------------

unsigned char clutchOffset(unsigned char gear)
{
// This function takes the target gear as an argument, and returns the clutch parameter offset

unsigned char offset;

offset = 0; // initialize

if (gear > outpc.current_gear)
  {
  offset = (uchar) (1 << (gear - 2)); 	// 2 raised to the power of gear minus 2
  }
  else if (gear < outpc.current_gear)
  {
  offset = (uchar) (1 << (gear - 1)); 	// 2 raised to the power of gear minus 2
  } // end else if (gear < outpc.current_gear)

return offset;

} // end function clutchOffset()


// ------------------------------------------------------------------------------

signed int CANretard(char new_gear)
{
unsigned int degrees;

// This function returns the CAN gear retard in degrees (x10)
switch (new_gear)
  {
  case -1:
  case 0:
  case 1:
  degrees = 0;
  break;
  
  case 2: 
  degrees = inpram.gear2_retard; 
  break;

  case 3: 
  degrees = inpram.gear3_retard; 
  break;

  case 4: 
  degrees = inpram.gear4_retard; 
  break;

  case 5: 
  degrees = inpram.gear5_retard; 
  break;

  case 6: 
  degrees = inpram.gear6_retard; 
  break;

  case 7: 
  degrees = inpram.gear7_retard; 
  break;

  case 8: 
  degrees = inpram.gear8_retard; 
  break;

  } // end switch (new_gear)

return degrees;

} // end function CANretard(char degrees)

// ------------------------------------------------------------------------------

unsigned char afterShiftRevs(char new_gear)
{
// This function checks if the engine will be over-reved on a downshift
// or under-reved on an upshift.
// It returns 0 if there will be an under/over-rev, and returns 1 otherwise (or if no error checking).

unsigned char okay;

// Check to make sure we are not PNR
if ((new_gear < 1) || (outpc.manual_gear < 1)) 
	{
	okay = 1;  	// allow shift
	return okay;
	}
	
// Check if we are out of range
if (new_gear > inpram.num_gears) 
	{
	okay = 0;	// block shift
	return okay;
	}
	
// If burnout mode then return
// No rpm checking in follow mode
if ((outpc.manual_gear <= inpram.follow_maxgear) && (inpram.follow_mode)) 
	{
	okay = 1;
	return okay;
	}
	
if (inpram.rpm_check)  // if doing rpm checks
  { 
  // calculate after-shift revs
  outpc.ashiftrevs = (outpc.engine_rpm/20) * (inpram.gear_table[new_gear]/10)/ (inpram.gear_table[outpc.current_gear]/10)*20;
  
  if (new_gear < outpc.current_gear)  // downshift, so check for over-rev
  	{
	if (outpc.ashiftrevs > in2ram.rpm_limit[outpc.current_gear - 2])
		// example: if in third, want to check if we would over rev in 2nd 
		// (because we are checking downshifts)
		// so we check 
		// rpm_limit[current gear - 2]
		// rpm_limit[3-2] == rpm_limt[1] == rpm_limit[2nd] since [0] is first, [1] is 2nd, etc.
    	{
      	outpc.error |= 0x04; // set error flag
      	outpc.shiftStat = 5; // shift blocked by revs
      	okay = 0;
     	}
     	else
     	{  // revs okay
     	outpc.error &= ~0x04; // clear error flag
     	okay = 1;  // shift is okay
     	outpc.shiftStat = 3; // rev shift	
     	}
  	}  // end if (new_gear < outpc.current_gear)
  	else if (new_gear > outpc.current_gear)  // upshift, so check for under-rev
   	{
  	if (outpc.ashiftrevs < in2ram.under_rev_limit[outpc.current_gear - 1])
  		// example: if in third, want to check if we would under rev in 4th 
		// (because we are checking upshifts)
		// so we check 
		// under_rev_limit[current gear - 1]
		// under_rev_limit[3-1] == rpm_limt[2] == rpm_limit[2nd] since [0] is 2nd, [1] is 3rd, etc.
 
   		{
    	outpc.error |= 0x04; // set error flag
      	outpc.shiftStat = 5; // shift blocked by revs
      	okay = 0;
     	}
     	else
     	{ // revs okay
     	outpc.error &= ~0x04; // clear error flag
     	okay = 1;  // shift is okay
     	outpc.shiftStat = 3; // rev shift
     	}
    } // end else if (new_gear > outpc.current_gear)
  }
else
  { // no rpm checking
  outpc.error &= ~0x04; // clear error flag
  okay = 1;
  }
  
  return okay;
}  // end afterShiftRevs()

// ------------------------------------------------------------------------------

void resetShiftRequests(void)
{
    // reset shift requests
    outpc.shift_action &= ~0x0C; // clear all requests

    // reset shift button counters
    UP_avg   = 500;
    DWN_avg  = 500;
    MODE_avg = 500;
    shift_release = (inpram.shftr_accum + 2)/4;
    
    return;
} // end function resetShiftRequests()

// -------------------------------------------------------------------------------

unsigned int solOffset(signed char gear)
{
// This is a function to return the correct offset in 
// the inpram.OutputX parameter when passed a valid gear

unsigned int offset;

if (gear > inpram.num_gears) gear = inpram.num_gears;
if (gear <= -1)
	{
	offset = 0x04;
	}
	else if (gear == 0)
	{
	offset = 0x01;    // Park; Neutral is 0x02
	}
    else  // first or higher
    {	
	offset = (uchar) (1 << (gear + 2)); // 2^(gear+2), ex. 2^4 for 2nd = 2*2*2*2 = 16 = 0x10
    }
return offset;

} // end function solOffset()

signed char getGear(void)
  {
  // This function checks the current output solenoid states and manual gear lever position,
  // and returns the current_gear.
  // The intended usage is:
  //    outpc.current_gear = getGear();
   
  signed char found_gear;
  signed char i;
  unsigned int j[9];
      
  // initialize
  found_gear = outpc.current_gear;
  
  for (i=1;i<=8;i++)
  	{
  	j[i] = (uchar) (1 << (i+2));
  	}
  // j[0] not used
  /*
  j[1] = 0x08;  // 1st
  j[2] = 0x10;  // 2nd
  j[3] = 0x20;  // 3rd
  j[4] = 0x40;  // 4th
  j[5] = 0x80;  // 5th
  j[6] = 0x100; // 6th
  j[7] = 0x200; // 7th
  j[8] = 0x400; // 8th
   */  
  // One for PNR else for 1st-8th
  if (outpc.manual_gear < 1)
    {
    if (outpc.manual_gear != -99)
    	{
    	// set outputs
    	if (outpc.manual_gear == -1) PNRset_outputs(0x04); 
    		else if (parkNeutral == 1) PNRset_outputs(0x01); // gear == 0, parkNeutral === 1
    		else PNRset_outputs(0x02); // gear = 0, ParkNeutral == 2
		// PNRset_outputs() function sets the outputs for Park, Neutral and Reverse based
		// on the passed offset (and the output patterns for each output). 
		// offset = 0x01 is Park
		// offset = 0x02 is Neutral
		// offset = 0x04 is Reverse
    	// if (outpc.speedo < 50) outpc.target_gear = 1;	
		// set the target gear for when we shift to drive
    	}
		else
		{
		// manual gear == -99
		return outpc.current_gear;	 // gear error
		}
    }
    else // in Drive
    {
    for (i=1; i <= inpram.num_gears; i++)
      { 
        // find a matching pattern
        if (	((Output1state*j[i]) == (j[i] & inpram.Output1ptrn)) 
			&& 	((Output2state*j[i]) == (j[i] & inpram.Output2ptrn)) 
			&&	((Output3state*j[i]) == (j[i] & inpram.Output3ptrn)) 
			&&	((inpram.sp_mode[1] != 3) || ((Output4state*j[i]) == (j[i] & inpram.Output4ptrn)))
			&&	((Output5state*j[i]) == (j[i] & inpram.Output5ptrn))
			&&	(((inpram.mlever_mode != 1) || (inpram.spareInputs != 2)) || ((Output6state*j[i]) ==(j[i] & inpram.Output6ptrn)))
			&&	((inpram.sp_mode[0] != 3) || ((Output7state*j[i]) == (j[i] & inpram.Output7ptrn)))
			&&	((in2ram.sp0_mode != 3) || ((Output8state*j[i]) == (j[i] & in2ram.Output8ptrn)))
			)
          {
          return i;
          }  // end if ((Output1state*i)==(0x200 ...
      } // end for (i=1; ...  
    }  // end if outpc.manual_gear < 1
    
    // Make sure gear is not out of range
	// if out-of-range, use current gear
    if ((outpc.current_gear <= inpram.num_gears) && (outpc.current_gear >= -1))
    	{
    	if (found_gear > inpram.num_gears) found_gear = outpc.current_gear;
    	if (found_gear < -1) found_gear = outpc.current_gear;
    	}
    
   return found_gear;
  
 } // end function getGear()
 
void setLEDs(signed char gear)
  {
   // This function sets the LEDs based on the passed gear and the LED user patterns
   unsigned char offset;

   offset = 0;

   if (gear > 0)
     {      
	 offset = (uchar) (1 << (gear-1)); // ex. for 5th, return 2^(5-1) = 2^4 = 16 = 0x10
     if (millisec < 488)
     	{ // for first half of second, set LEDs	
     	// switch LED1
     	if (offset & inpram.LED1) LED1_OUT |= LED1pin; else LED1_OUT &= ~LED1pin;
     	// switch LED2
     	if (offset & inpram.LED2) LED2_OUT |= LED2pin; else LED2_OUT &= ~LED2pin;
     	// switch LED3
     	if (offset & inpram.LED3) LED3_OUT |= LED3pin; else LED3_OUT &= ~LED3pin;
     	// switch LED4
     	if (offset & inpram.LED4) LED4_OUT |= LED4pin; else LED4_OUT &= ~LED4pin;
     	}
     	else // millisec >= 488
     	{
     	if ((in2ram.LEDflash == 1) && (outpc.auto_mode == 0))
     		{
     		// if set to flash and shift mode is manual
     		// switch all LEDs off for second half of second
     		LED1_OUT &= ~LED1pin;
     		// switch LED2 off
     		LED2_OUT &= ~LED2pin;
     		// switch LED3 off
     		LED3_OUT &= ~LED3pin;
     		// switch LED4 off
     		LED4_OUT &= ~LED4pin;	
     		}
     	}  // end else if (millisec ....
     }// end if (gear > 0)
     else if (gear == 0)
     {
     if (parkNeutral == 2)	  // Neutral
      	{  
      	// needs 488 msec switching for P & N
      	if (millisec < 488)   // in first half of a second   
      	    {											  	
        	// switch LED1
        	if (inpram.LEDN & 0x01) LED1_OUT |= LED1pin; else LED1_OUT &= ~LED1pin;
        	// switch LED2
        	if (inpram.LEDN & 0x02) LED2_OUT |= LED2pin; else LED2_OUT &= ~LED2pin;
        	// switch LED3
        	if (inpram.LEDN & 0x04) LED3_OUT |= LED3pin; else LED3_OUT &= ~LED3pin;
        	// switch LED4
        	if (inpram.LEDN & 0x08) LED4_OUT |= LED4pin; else LED4_OUT &= ~LED4pin; 
      	    }
      	    else   // 2nd half of second
      	    {
      	    // switch LED1
        	if (inpram.LEDN & 0x10) LED1_OUT |= LED1pin; else LED1_OUT &= ~LED1pin;
        	// switch LED2
        	if (inpram.LEDN & 0x20) LED2_OUT |= LED2pin; else LED2_OUT &= ~LED2pin;
        	// switch LED3
        	if (inpram.LEDN & 0x40) LED3_OUT |= LED3pin; else LED3_OUT &= ~LED3pin;
        	// switch LED4
        	if (inpram.LEDN & 0x80) LED4_OUT |= LED4pin; else LED4_OUT &= ~LED4pin; 	
      	    } // end if (millisec < 488)
      	}  // end if (parkNeutral == 2)	
      	else if (parkNeutral == 1)	  // Park
      	{
      	if (millisec < 488)   // in first half of a second
			{
        	// switch LED1
        	if (inpram.LEDP & 0x01) LED1_OUT |= LED1pin; else LED1_OUT &= ~LED1pin;
        	// switch LED2
        	if (inpram.LEDP & 0x02) LED2_OUT |= LED2pin; else LED2_OUT &= ~LED2pin;
        	// switch LED3
        	if (inpram.LEDP & 0x04) LED3_OUT |= LED3pin; else LED3_OUT &= ~LED3pin;
        	// switch LED4
        	if (inpram.LEDP & 0x08) LED4_OUT |= LED4pin; else LED4_OUT &= ~LED4pin;
      	    } // end if (millisec < 488) 
      	    else   
      	    {
        	// switch LED1
        	if (inpram.LEDP & 0x10) LED1_OUT |= LED1pin; else LED1_OUT &= ~LED1pin;
        	// switch LED2
        	if (inpram.LEDP & 0x20) LED2_OUT |= LED2pin; else LED2_OUT &= ~LED2pin;
        	// switch LED3
        	if (inpram.LEDP & 0x40) LED3_OUT |= LED3pin; else LED3_OUT &= ~LED3pin;
        	// switch LED4
        	if (inpram.LEDP & 0x80) LED4_OUT |= LED4pin; else LED4_OUT &= ~LED4pin;
      	    } // end if (millisec < 488) 
      	} // end else if (parkNeutral == 1)
     } // end else if (gear == 0)
     else // reverse
     {
      // set up to allow flashing of LEDs in reverse (@ 1 Hertz)
      if (millisec < 488)   // in first half of a second
        {
        // switch LED1
        if (inpram.LEDR & 0x01) LED1_OUT |= LED1pin; else LED1_OUT &= ~LED1pin;
        // switch LED2
        if (inpram.LEDR & 0x02) LED2_OUT |= LED2pin; else LED2_OUT &= ~LED2pin;
        // switch LED3
        if (inpram.LEDR & 0x04) LED3_OUT |= LED3pin; else LED3_OUT &= ~LED3pin;
        // switch LED4
        if (inpram.LEDR & 0x08) LED4_OUT |= LED4pin; else LED4_OUT &= ~LED4pin;        
                 
        }
        else   // in 2nd half of second 
        {
        if (inpram.LEDR & 0x10) LED1_OUT |= LED1pin; else LED1_OUT &= ~LED1pin;
        // switch LED2
        if (inpram.LEDR & 0x20) LED2_OUT |= LED2pin; else LED2_OUT &= ~LED2pin;
        // switch LED3
        if (inpram.LEDR & 0x40) LED3_OUT |= LED3pin; else LED3_OUT &= ~LED3pin;
        // switch LED4
        if (inpram.LEDR & 0x80) LED4_OUT |= LED4pin; else LED4_OUT &= ~LED4pin;
                  
        } 
        return; 
      }
   return;  
  
  }  // end setLEDs function
  
#pragma CODE_SEG ROM_7000
   
#pragma CODE_SEG MAIN_ROM

void setOut3PWM(signed char gear)
{
// This is a function to set the PWM duty cycle for Output3 (aka. "SolC" or "Sol32")
// and it also sends the gear retard amount over CAN to MS-II
 
signed int offset;
offset = solOffset(gear);

  if (inpram.PWM_enable & 0x04)
    {
    OUTPUT3pwm = dc_32*((offset & inpram.Output3ptrn)/offset);
    }
    else
    {
    OUTPUT3pwm = OUTPUT3_ON*((offset & inpram.Output3ptrn)/offset);  
    }
  Output3(0); // set state indicator  (dummy argument when not in NAG mode)
  
  // send CAN retard for each forward gear (>1)
  if (outpc.LOAD > inpram.Tretard_load)
    {
    switch (gear)
      {
      	case 2:
        outpc.SpkAdj = inpram.gear2_retard;
        if (sentSpkAdj != 2 && inpram.CAN_enabled) sendCANAdj();
        sentSpkAdj = 2;
        break;
    
       	case 3:
        outpc.SpkAdj = inpram.gear3_retard;
        if (sentSpkAdj != 3 && inpram.CAN_enabled) sendCANAdj();
        sentSpkAdj = 3;
        break;
        
        case 4:
        outpc.SpkAdj = inpram.gear4_retard;
        if (sentSpkAdj != 4 && inpram.CAN_enabled) sendCANAdj();
        sentSpkAdj = 4;
        break;
        
       	case 5:
        outpc.SpkAdj = inpram.gear5_retard;
        if (sentSpkAdj != 5 && inpram.CAN_enabled) sendCANAdj();
        sentSpkAdj = 5;
        break;
        
        case 6:
        outpc.SpkAdj = inpram.gear6_retard;
        if (sentSpkAdj != 6 && inpram.CAN_enabled) sendCANAdj();
        sentSpkAdj = 6;
        break;
        
        case 7:
        outpc.SpkAdj = inpram.gear7_retard;
        if (sentSpkAdj != 7 && inpram.CAN_enabled) sendCANAdj();
        sentSpkAdj = 7;
        break;   
        
        case 8:
        outpc.SpkAdj = inpram.gear8_retard;
        if (sentSpkAdj != 8 && inpram.CAN_enabled) sendCANAdj();
        sentSpkAdj = 8;
        break; 
        
        default:
        outpc.SpkAdj = 0;
        if (sentSpkAdj != 7 && inpram.CAN_enabled) sendCANAdj();
        sentSpkAdj = 0;
        break;    
      }
    }
    else
    {
    outpc.SpkAdj = 0;
    if (sentSpkAdj != 0 && inpram.CAN_enabled) sendCANAdj();
    sentSpkAdj = 0;           
    }
      
    return;

} // end setOut3PWM()

void PNRset_outputs(unsigned char offset)
	{
	// This function sets the outputs for Park, Neutral and Reverse based
	// on the passed offset (and the output patterns for each output). 

	// offset = 0x01 is Park
	// offset = 0x02 is Neutral
	// offset = 0x04 is Reverse
	
	if (offset > 0x04) return; // return if out of range
	
	 // switch Output 1
     Output1((char) ((offset & inpram.Output1ptrn)/offset));  // if 1 in Output1, then send 1 to turn solA on
           
     // switch Output 2
     Output2((char) ((offset & inpram.Output2ptrn)/offset));  // if 1 in Output2, then send 1 to turn solB on
   
     // switch Output 3 if shift PWM not 0
      if (inpram.PWM_enable & 0x04)
        {
        OUTPUT3pwm = dc_32 * ((char) ((offset & inpram.Output3ptrn)/offset));
        }
        else
        {
        OUTPUT3pwm = OUTPUT3_ON * ((char) ((offset & inpram.Output3ptrn)/offset));  
        }
      Output3(0); // set state indicator (dummy argument when not in NAG mode)
            
     // switch Output 4
     if (inpram.sp_mode[1] == 3)   // shift solenoid
   	 	{
   	 	Output4((char) ((offset & inpram.Output4ptrn)/offset));  // if 1 in Output4, then send 1 to turn solD on
   	 	}        
     // switch Output 5
     if ((inpram.mlever_mode == 1) && (inpram.spareInputs > 0)) Output5((char) ((offset & inpram.Output5ptrn)/offset));  // if 1 in Output5, then send 1 to turn solE on

     // switch Output 6
     if ((inpram.mlever_mode == 1) && (inpram.spareInputs == 2)) Output6((char) ((offset & inpram.Output6ptrn)/offset));  // if 1 in Output6, then send 1 to turn solF on
   
     // switch Output 7
     if (inpram.sp_mode[0] == 3) Output7((char) ((offset & inpram.Output7ptrn)/offset));  // if 1 in Output7, then send 1 to turn Output 7 on

     // switch Output 8
     if (in2ram.sp0_mode == 3) Output8((char) ((offset & in2ram.Output8ptrn)/offset));  // if 1 in Output8, then send 1 to turn Output 8 on   

	 // switch Output 9
     if (in2ram.sp3_mode == 3) Output9((char) ((offset & in2ram.Output9ptrn)/offset));  // if 1 in Output9, then send 1 to turn Output 9 on   
       
     // End set outputs
	return;
	}  // end void PNRset_outputs()...
	
#include "getLever.h"

#include "sparePorts.h"

#include "odometer.h"