
// Initialize PWM Dither and Refresh Variables
dither_flag = 0;          // reset PWM dither and refresh flags
dither_counter = 0;       // dither counter to zero
refresh_counter = 0;      // refresh counter to zero  

// initialize odometer
ODOlastSpeed = 0;
outpc.odo = 0;

/******************* Initialize IC_period array for VSS/ISS ****************************/
// So we start at zero speed until array is filled (which will happen very fast at any real 
// speed, but is never updated when stopped)
// If we have:
// - engine rpm, OR
// - VSS count is incrementing 
// then assume this is a 'hot' reset, not starting from scratch

// fill ic2 array
if ((outpc.engine_rpm > 500) || (ic1_count > 3))
{
// running reset
for (ix=1;ix<21;ix++)
	{
	ic1_period[ix] = (minTics*((ix/4)+1)); // set the VSS period high (speed low)  
	ic2_period[ix] = 0x00FFF;   // set the ISS period high (speed low)
	}
} 
else 
{
// not running - no/low rpm AND not P or N AND no VSS interrupts  
for (ix=1;ix<21;ix++)
	{
	ic1_period[ix] = 0x0FFFF; // set the VSS period high (speed low)  
	ic2_period[ix] = 0x0FFFF; // set the ISS period high (speed low)
	}
} // end if (outpc.engine_rpm > 500)

// Initialize Idle Flag
IdleFlag = 0;

// Initialize outpc. variables
last_LOAD = 1000;         // kpa x 10
outpc.LOAD_short = 1000;  // kpa x 100
LOAD_long = outpc.LOAD_short;
outpc.manual_gear = inpram.num_gears;    // Drive, as PNR normally controlled mechanically
outpc.target_gear = inpram.num_gears;
outpc.lock_TCC = tcc_unlock();
jake_brake = 0;

// Clear MS-II adjustments 
// (in case this is a hot restart and adjustments have already been sent but not cleared)
outpc.FuelAdj = 0;
outpc.SpkAdj = 0;
outpc.IdleAdj = 0;
outpc.SprAdj = 0;
if(inpram.CAN_enabled) sendCANAdj();

vss_teeth = 0;
loop_count = 0;
outpc.aux_volts = 100;
os_rpm = 0;         // output shaft rpm = 0

// Initialize Switch Manifold Variables
// set to 3rd gear == 110, we will poll the switches immediately so this isn't crucial
outpc.Inadc[0] = 1000;   // (1000/1024)*5.0 Volts
outpc.Inadc[1] = 1000;   // (1000/1024)*5.0 Volts
outpc.Inadc[2] =  100;    //  (100/1024)*5.0 Volts

// set inputs to 3rd gear until we have measured them
Inputstate[0] = (inpram.Inputptrn[0] & 0x20)/0x20;
Inputstate[1] = (inpram.Inputptrn[1] & 0x20)/0x20;
Inputstate[2] = (inpram.Inputptrn[2] & 0x20)/0x20;
if (Inpt4_active) Inputstate[3] = (inpram.Inputptrn[3] & 0x20)/0x20;

// Initialize debounce parameters (reinitialize when shifting gears)
Inputacc[0] = inpram.debounce/2;
Inputacc[1] = inpram.debounce/2;
Inputacc[2] = inpram.debounce/2;
Inputacc[3] = inpram.debounce/2;

effective_axle_ratio = ((long) inpram.axle_ratio * (long) in2ram.inter_ratio) / 1000;  

// Calculate VSS divide factor
// do it here to reduce computations in main loop - changes require restart
// VSSdivide is the number of inches travel for 20 outputs shaft revs (x100)
// This is:
// 20*num_teeth teeth
// and 20/axle_ratio of a tire revolution
// or  20/axle_ratio*(tire_diam*3.1416) linear distance (in inches x100)
// tire diam is inchesx100, axle ratio is x1000
// therefore 20*1000*PI * (tire_diam)/(axle_ratio)                                            
if (inpram.VSS_location == 0) // VSS on output shaft
		{
		vss_divide = 62832 * ((long) inpram.tire_diam)/((long) effective_axle_ratio); // inches per 20 revs
		}
		else // VSS on wheel
		{
		vss_divide = 62832 * ((long) inpram.tire_diam)/1000; // inches per 20 revs
		}

// Ex. for 26" tire, 3.08 axle, VSSdivide = 53040 = 530.40 inches per 20*num_teeth 
// vss_factor1 = vss_divide/inpram.num_teeth;

rev_distance = (int)(2016811/(long) inpram.tire_diam);

// Calculate Maximum/Minimum Number of VSS tics based on inpram.maxSpeed, etc.
// actual factor is 53, but we use 40 to allow some tolerance for short but valid pulses
minTics = (int) (40*(vss_divide/inpram.num_teeth)/(inpram.maxSpeed));
vssTics = minTics/4;

// ex. vss_divide = 53040, num_teeth = 40, max speed= 200 (x10), min speed  = 30 (x10)
// minTics = 53*(53040/40)/(2000)
//         = 35.1
// at 5.333 microsec/tic = 187 microseconds = 0.000187 seconds =  5335 Hz 
// (since 60 mph is 1584 Hz, this is reasonable)

// Initialize local variables
scount =   0;
lcount =   0;

outpc.auto_mode = inpram.shift_mode; // start in user specified shift mode 

// Set pulse/mile speedo counters
//
// number of VSS pulses/mile is:
//
//   rev_distance * axle_ratio * num_teeth
//
//
// the speedo output total count is:
//
//   (VSS pulses/mile)/(speedo pulse mile)
//
//   
ultmp1 = ((long)rev_distance * (long) effective_axle_ratio)/1000;
ultmp2 = (ultmp1 * (long)inpram.num_teeth)/(long)inpram.pulse_mile;
speedo_total = (int) (ultmp2);
// these are used in the VSS interrupt routine VSS_timer0 
speedo_toggle = ((speedo_total*in2ram.speedoutDC)+50)/100; // we will create user duty cycle signal
speedo_counter = 0;             // initialize speedo-counter
// make sure toggle will generate a signal
if (speedo_toggle == 0) speedo_toggle = 1; else if (speedo_toggle >= speedo_total) speedo_toggle = speedo_total - 1;

// Ex. for 26" tire, 3.08 gears and 40 teeth, with a 2002 pulse/mile output, this is:
//   speedo_total = (775 * 3.080 * 40)/2002 =  47.692 => 47
//   speedo_toggle = 23

UP_avg = 500; // will move up and down according to button status
DWN_avg = 500;
MODE_avg = 500;
high_mph = 0;

// Calculate Output3 (SolC) duty cycle in PWM ticks 
dc_32 = (uchar) ((inpram.out3dc * PWMPER1)/100);

outpc.mloop = 0; // initialize main loop counter

outpc.TCC_count = inpram.TCC_hyst; // initialize first TCC hysteresis cycle

restart = 1; // restart flag set