unsigned int odometer(unsigned int newSpeed)
{	
// This function calculates the distance travelled since the last time the function was called.
// Current time (seconds, milliseconds) and speed (x10) is passed to the function.
// Delta distance is returned from function (to be added to outpc.odo).
// This function uses three global variables ODOlastSec, ODOlastMsec, 
// and ODOlastSpeed to track last time delta distance was computed.
//
// example usage:

// outpc.odo = outpc.odo + odometer(outpc.seconds, millisec, outpc.speedo);

// Delta distance is speed * time
// (speed is x10) 

// the shorter the time between calls, the more 'up to date' the display will be,
// but the longer the time the more accurate the accumulated value will be 
// (at a steady speed, at least)

// don't need to calculate if distance is less than 0.01 (km or mph) since third digit isn't displayed
// @ 10  (mph or kph) this takes 3.6 sec (0.01 mile = 52.8 ft @ 14.67 ft/sec)
// @ 30  (mph or kph) this takes 1.2 sec (0.01 mile = 52.8 ft @ 44 ft/sec)
// @ 60  (mph or kph) this takes 0.6 sec (0.01 mile = 52.8 ft @ 88 ft/sec) 
// @ 100 (mph or kph) this takes 0.36 sec (0.01 mile = 52.8 ft @ 146.7 ft/sec)
// @ 200 (mph or kph) this takes 0.18 sec (0.01 mile = 52.8 ft @ 293.3 ft/sec)
//
// i.e.
//
// not more often than deltaTime = 36/((newSpeed+oldSpeed)/2)  = 72/(newSpeed+oldSpeed)

unsigned int deltaDistance; // returned distance in interval since last call
unsigned int ODOmilliseconds;

ODOmilliseconds = ODOmsec; // capture the value so it doesn't change

if (ODOmilliseconds < (72000/(newSpeed+ODOlastSpeed))) return(0);   // ex. 600 msec (0.600 sec) at 60mph/kph

if (((newSpeed + ODOlastSpeed + 1)/2) == 0) 
	{
	ODOmilliseconds = 0;  // re-initialize value
	return(0); // skip if not moving
	}
	
// compute the distance travelled since the last function call
// as average of old and new speed multiplied by net elapsed time
// deltaDistance = ((newSpeed + ODOlastSpeed)/2 * deltaTime/10);
// ODOmsec has to be converted to hours
// so multiply by 1/3600*1000
// deltaDistance = (((newSpeed + ODOlastSpeed) * ODOmsec/3600)/20); // speed is x10, time is x1000, distance is x1000

deltaDistance = ((newSpeed + ODOlastSpeed + 6)/12 * ((ODOmilliseconds/6)+3)+500)/1000; // distance (x100); speed is x10, time is x1000, distance is x1000
// +6, +3, and +500 are to make truncation into rounding

ODOlastSpeed = newSpeed;
	
ODOmsec = 0;  // re-initialize value, will increment in TimerClockISR()
	
// reset the global values

return(deltaDistance);  // return distance travelled (x1000) since last function call

} // end odometer() ...

