// ****** Set Up OUTPC. Structure for Serial Communications with TunerStudio *******
// These are the rs232 outputs to PC via the serial port, and the corresponding 
// values should be in the TunerStudio INI [OutputChannels] section.

typedef struct {
unsigned int seconds;                     // clock
unsigned int speedo;                      // vehicle speed (x10) 
                                          // (in mph or kph, depending on Metric_Units)
unsigned long odo;                        // trip odometer (x1000)
unsigned char auto_mode;                  // 0 = manual, 1 = VSS-Load auto (sequential), 2 = VSS-Load auto (skip), 3 = Tach-Load based
unsigned char shift_action;               // Shift requests:
                                          // bit 0: downshift button pressed
                                          // bit 1: upshift button pressed
                                          // bit 2: downshift requested
                                          // bit 3: upshift requested
unsigned char error;                      // error status codes
                                          /*
                                          bit --  & with  --  meaning
                                          ---     ------      -------
                                           0  --   0x01   --  1 = okay, 0 = serial comms error
                                           1  --   0x02   --  0 = okay, 1 = no CAN comms (load = 0.0)
                                           2  --   0x04   --  0 = okay, 1 = under/over rev shift attempted
                                           3  --   0x08   --  0 = okay, 1 = VSS reset occurred
                                           4  --   0x10   --  0 = okay, 1 = other VSS error - no signal, excessive noisy/drop-outs (check tooth error counter)
                                           5  --   0x20   --  0 = okay, 1 = gear out-of-range
                                           6  --   0x40   --  0 = okay, 1 = low voltage (<10)
                                           7  --   0x80   --  0 = okay, 1 = no brake signal  
                                          */
signed char manual_gear;                  // manual valve position determined from switch manifold
signed char current_gear;                 // current gear determined from state of shift solenoids
signed char target_gear;                  // next gear to select
unsigned int engine_rpm;                  // engine rpm (from CAN or estimated from VSS and gear)
signed int LOAD;                          // LOAD from MS-II via CAN - (kpa x 10)
signed int clt;                           // trans temperature deg(C/F)x 10
signed int line_pressure;                 // line pressure
unsigned char lock_TCC;                   // 0=unlocked torque converter clutch, 1=locked clutch
unsigned char miscStatus;                 // bit 0 (0x01): brake,  1=brake on, 0=brake off
                                          // bit 1 (0x02): burnstatus, 1 = burning, 0 = done
                                          // bit 2 (0x04): FWD, 1=2wd, 0=4wd (grounded = table 2)
                                          // bit 3 (0x08): upshift forced
                                          // bit 4 (0x10): downshift forced
                                          // bit 5 (0x20): Jake brake active
                                          // bit 6 (0x40): baud rate error (less than 300 or more than 921600)
unsigned int LOAD_short;                  // LOAD averages
unsigned int aux_volts;					  // axillary data channel (used for load if no CAN)
unsigned char PC_duty;                    // Pressure Control valve duty cycle (% for PWM)
unsigned char slip;                       // torque converter slip or internal slip (%)
unsigned int dbug;                        // spare debugging variable brought out in TunerStudio (default is VSS tooth error count)       
unsigned int Inadc[3];					  // ADCs for Input 1, 2, & 3

unsigned int solst;                       // 0x00000 = 0 - Output1 off, Output2 off, Output3 off, Output4 off, ...
                                          // 0x00001 = 1 - Output1 on,  Output2 off, Output3 off, Output4 off, ...
                                          // 0x00010 = 2 - Output1 off, Output2 on,  Output3 off, Output4 off, ...
                                          // 0x00011 = 3 - Output1 on,  Output2 on,  Output3 off, Output4 off, ...
                                          // 0x00100 = 4 - Output1 off, Output2 off, Output3 on,  Output4 off, ...
                                          // 0x00101 = 5 - Output1 on,  Output2 off, Output3 on,  Output4 off, ...
                                          // 0x00110 = 6 - Output1 off, Output2 on,  Output3 on,  Output4 off, ...
                                          // 0x00111 = 7 - Output1 on,  Output2 on,  Output3 on,  Output4 off, ...
                                          // 0x01000 = 8 - Output1 off, Output2 off, Output3 off, Output4 on,  ...
unsigned char cksum;
unsigned int mileage;                     // fuel efficiency (mpg X 100)
short
   FuelAdj,SpkAdj,	                      // Adjustments for fuel (%x10 +/- to VE table), spark (degx10),
                                          // idle step, spare 
   IdleAdj,SprAdj;			              // Send to MS-II inpram input table.
unsigned char switches;                   // bit 0 - 0x01 - In1, 
										  // bit 1 - 0x02 - In2, 
										  // bit 2 - 0x04 - In3, 
										  // bit 3 - 0x08 - In4,
										  // bit 4 - 0x10 -
										  // bit 5 - 0x20 -
										  // bit 6 - 0x40 -
										  // bit 7 = gear lever error
unsigned char sp1PWM;                     // spare port 1 PWM %
unsigned int is_rpm;                      // calculated/measured input shaft rpm
unsigned int butADC;                      // shift button ADC count
unsigned char sp2PWM;                     // spare port 2 PWM %
unsigned char TCC_count;                  // TCC hysteresis counter
unsigned char sp0PWM;                     // spare port 0 PWM % 
signed int sp09inx;                       // spare port 0 9-bin index value 
signed int sp016inx;                      // spare port 0 12-bin index value
unsigned char sp3PWM;                     // spare port 2 PWM %
signed int sp39inx;                       // spare port 0 9-bin index value 
signed int sp316inx;                      // spare port 0 12-bin index value
unsigned int rawMAP;                      // unadjusted MAP value
unsigned int rawTPS;             		  // TPS value from MS-II over CAN
unsigned char hyst_timer;                 // Counter for hyst flag (set to inpram.hyst_delay initially)
unsigned long mloop;                      // main loop counter
unsigned char Slip_Adjustment;            // slip adjustment (%) when slipping
unsigned char LEDstatus;                  // LED status
unsigned char porta;
unsigned char porte;
signed char estimated_gear;
unsigned int RPM_short;				      // average RPM
unsigned int cruise_flag;				  // cruise status (cruising time is seconds)
unsigned int ashiftrevs;				  // after shift revs
unsigned char decel_flag;				  // deceleration flag (to delay downshifts)
unsigned long VSS_per;					  // measured VSS period
unsigned int instant_ratio;               // current effective gear ratio

unsigned int vBatt;                       // battery voltage over CANbus
unsigned int testing_inputs;			  // inputs values testing
unsigned char shiftStat;         		  // shift status
                                 		  // 0 = no data
                                 		  // 1 = shift completed - shift table
                                 		  // 2 = shift completed - manual shift buttons
                                 		  // 3 = shift completed - rev limit shift
                                 		  // 4 = shift completed - manual gear lever position change
                                 		  // 5 = shift blocked - under/over rev
                                 		  // 6 = shift blocked - hysteresis not met
                                 		  // 7 = shift blocked - manual lever position not found
                                 		  // 8 = lower gear not available
                                 		  // 9 = higher gear not available
unsigned int rawSpeedo;          		  // raw vehicle speed (x10) before corrections 
                                 		  // (in mph or kph, depending on Metric_Units)
unsigned char lever_count;        		  // lever error timer (in 10ths of second)
								  		  // = 0 if not set (valid gear found)
								  		  // = 1 if no valid gear found, but timer expired so report -99
								  		  // = 2...255 if counting down and reporting previous manual gear
                                                                  
} variables;                              // end outpc. structure


variables outpc;