// ----- Spare Port Switching -------------------------------------------------------
// This trans code has 2 spare port (PT7 and PA0, on VB1 and VB2, AMP 11 and 12).
// PT7 -> VB1 must be jumpered from one end of 25x2 header to near the other end.
// The code lets these be set by speed, load, and gear. All conditions are 'AND'ed.

void SetSparePort(unsigned char ix) 
{	
// Pass ix=0 for spare port 1, ix=1 for spare port 2	

char tmp,iy;
unsigned char condOp;
signed int outpcValue;
signed int condValue; 	// The condition operator is:
						// - 00 (=0) = no condition (do not apply thresholds, etc., just ignore), 
						// - 01 (=1) = greater than or equal, 
						// - 10 (=2) = less than or equal,
						// - 11 (=3) = equal
signed int hystValue; 
//
// For spare ports spr1 and spr2 only 
// (spr0 an spr3 do not have conditional abilities - use 16x9 table instead)
if ((ix != 0) || (ix != 1)) return;
 
tmp = 0;

// for each of four conditions
// turn ports on if conditions met
if ( ((ix == 0) && (sp[0] == 0)) || ((ix == 1) && (sp[1] == 0)) )  // if not already on
 	{		
	for (iy=0; iy <= 3; iy++)
		{
		// iy = 0 = speed
		// iy = 1 = rpm
		// iy = 2 = load
		// iy = 3 = gear
		// Compare the condition to the current value 
		switch (iy)
			{
			// might need to cast values
			case 0:  // speed
				condOp = (unsigned char) inpram.spspeed_cond[ix];
				condValue = (signed int) inpram.spspeed[ix];
				outpcValue =(signed int) outpc.speedo;
				break; 
			case 1:  // rpm
				condOp = (unsigned char) inpram.sprpm_cond[ix];
				condValue = (signed int) inpram.sprpm[ix];
				outpcValue = (signed int) outpc.engine_rpm;
				break;
			case 2:  // load
				condOp = (unsigned char) inpram.spload_cond[ix];
				condValue = (signed int) inpram.spload[ix];
				outpcValue = (signed int) outpc.LOAD;
				break;
			case 3:  // gear
				condOp = (unsigned char) inpram.spgear_cond[ix];
				condValue = (signed int) inpram.spgear[ix];
				outpcValue = (signed int) outpc.current_gear;
				break;
			default:
				iy = 4;
				break;
			} // end switch(...
			
		// check each of 4 boolean operators for a given threshold parameter	
		if (	((condOp == 3) && (condValue == outpcValue))    // Equal
			||	((condOp == 2) && (condValue >= outpcValue))    // Less than or equal
			||	((condOp == 1) && (condValue <= outpcValue))    // Greater than or equal
			||	 (condOp == 0)	  // no condition - so always 'true'
			)
			{
			// Condition met
			tmp++;
			}
			else
			{
			// Condition not met
			tmp = 0; 
			iy = 4; // didn't meet condition, since we are ANDing, we abort the remaining conditions
			}
		} // end for for (iy=0; ....
    	
		if (tmp > 0) 	// at least one of the conditions was true 
			{
			if (ix == 0) // sp1
				{
				spr1(1); // no value returned
				}
				else // ix == 1; sp2
				{
				spr2(1); // no value returned
				}
			tmp = 0;		   
			} // end if tmp1 = ...
	} // end if (((ix==0) && (spr1 == 0)) ...

// if port already on
if (((ix == 0) && (sp[0] == 1)) || ((ix == 1) && (sp[1] == 1))) // if already on
	{	
	// Check hysteresis
	// turn ports off if hysteresis exceeded
	// for each of four conditions
	for (iy=0; iy < 4; iy++)
		{
		// iy == 0 = speed
		// iy == 1 = rpm
		// iy == 2 = load
		// iy == 3 = gear
		// Compare the condition to the current value 
		switch (iy)
			{
			// might need to cast values
			case 0:  // speed
				condOp = (unsigned char) inpram.spspeed_cond[ix];
				condValue = (signed int) inpram.spspeed[ix];
				hystValue = (signed int) inpram.spspeed_hyst[ix];
				outpcValue = outpc.speedo;
				break; 
			case 1:  // rpm
				condOp = (unsigned char) inpram.sprpm_cond[ix];
				condValue = (signed int) inpram.sprpm[ix];
				hystValue = (signed int) inpram.sprpm_hyst[ix];
				outpcValue = outpc.engine_rpm;
				break;
			case 2:  // load
				condOp = (unsigned char) inpram.spload_cond[ix];
				condValue = (signed int) inpram.spload[ix];
				hystValue = (signed int) inpram.spload_hyst[ix];
				outpcValue = outpc.LOAD;
				break;
			case 3:  // gear
				condOp = (unsigned char) inpram.spgear_cond[ix];
				condValue = (signed int) inpram.spgear[ix];
				hystValue = 0;	   // no hysteresis in gears
				outpcValue = (signed int) outpc.current_gear;
				break;
			default:
				break;
			}  // end switch( ....
		
		if ((condOp == 1) && (condValue > (outpcValue + hystValue))) // greater than or equal
			{
			if (ix == 0) spr1(0); else spr2(0);
			iy = 4;	  // done
			}
			else if ((condOp == 2) && (condValue < (outpcValue - hystValue))) // less than or equal
			{
			if (ix == 0) spr1(0); else spr2(0);
			iy = 4;	  // done
			}
			else if ((condOp == 3) && (condValue != (outpcValue))) //  equal
			{
			if (ix == 0) spr1(0); else spr2(0);
			iy = 4;	  // done
			}//
   	 } // end for(iy, ...
	} // end if (((ix==0) && (spr1 == 1)) ...

return;

} // end void SetSparePort(char ix);
    
