/*************************************************************************
**************************************************************************
**   M E G A S Q U I R T  GPIO - 2 0 0 7 - V1.000
**
**   (C) 2007 - B. A. Bowling And A. C. Grippo
**
**   This header must appear on all derivatives of this code.
**
***************************************************************************
**************************************************************************/

/*------------------------
  Version 1.020
    This first version of GPIO board software demonstrates various simple uses,
    including generic I/O, generic PWM, and generic ADC readbacks, along with 
    a simple fan control model written with generic user input instructions.
  Version 1.030
    Added pass-through mode so ECU can use gpio as an i/o extender. Can read 
    or set any pin on gpio, including io, adc, ic and pwm.
  Version 1.040-43
    Put all instructions in flash and expanded to 11.5 sectors (640  instr).
    Added CAN outpc message configuration.
  Version 1.100
    Modified CAN pass-through mode to request up to 80 bytes of data and write 
    up to 65K with one serial message, Remove T can/ serial mode (not needed).
  Version 1.200
    Made serial and CAN into semi-universal, standalone modules.
 -----------------------*/
#include "hcs12def.h"      /* common defines and macros */
#include "flash.h"         /* flashburner defines, structures */
#include <string.h>

//#define CAN_TEST
//#define CAN_TESTOUT

#define ENABLE_INTERRUPTS asm cli;
#define DISABLE_INTERRUPTS asm sei;
#define NEAR near
#define VECT_ATTR @0xFF80
#pragma CODE_SEG NON_BANKED /* Interrupt section for this module. 
                          Placement will be in NON_BANKED area. */

extern void near _Startup(void);       /* Startup routine */

interrupt void UnimplementedISR(void);
interrupt void TimerIC7Isr(void);
interrupt void TimerIC6Isr(void);
interrupt void TimerIC5Isr(void);
interrupt void TimerIC4Isr(void);
interrupt void TimerIC3Isr(void);
interrupt void TimerIC2Isr(void);
interrupt void TimerIC1Isr(void);
interrupt void TimerIC0Isr(void);
interrupt void ISR_Timer_Clock(void);

#include "..\..\serCANComm\serCAN0.h"
#include "..\..\serCANComm\serCAN1.h"
#include "..\..\serCANComm\serCAN2.h"

typedef void (* NEAR tIsrFunc)(void);
const tIsrFunc _vect[] VECT_ATTR = {      /* Interrupt table */
        UnimplementedISR,                 /* vector 63 */
        UnimplementedISR,                 /* vector 62 */
        UnimplementedISR,                 /* vector 61 */
        UnimplementedISR,                 /* vector 60 */
        UnimplementedISR,                 /* vector 59 */
        UnimplementedISR,                 /* vector 58 */
        UnimplementedISR,                 /* vector 57 */
        UnimplementedISR,                 /* vector 56 */
        UnimplementedISR,                 /* vector 55 */
        UnimplementedISR,                 /* vector 54 */
        UnimplementedISR,                 /* vector 53 */
        UnimplementedISR,                 /* vector 52 */
        UnimplementedISR,                 /* vector 51 */
        UnimplementedISR,                 /* vector 50 */
        UnimplementedISR,                 /* vector 49 */
        UnimplementedISR,                 /* vector 48 */
        UnimplementedISR,                 /* vector 47 */
        UnimplementedISR,                 /* vector 46 */
        UnimplementedISR,                 /* vector 45 */
        UnimplementedISR,                 /* vector 44 */
        UnimplementedISR,                 /* vector 43 */
        UnimplementedISR,                 /* vector 42 */
        UnimplementedISR,                 /* vector 41 */
        UnimplementedISR,                 /* vector 40 */
        CanTxIsr,                         /* vector 39 */
        CanRxIsr,                         /* vector 38 */
        CanRxIsr,                         /* vector 37 */
        UnimplementedISR,                 /* vector 36 */
        UnimplementedISR,                 /* vector 35 */
        UnimplementedISR,                 /* vector 34 */
        UnimplementedISR,                 /* vector 33 */
        UnimplementedISR,                 /* vector 32 */
        UnimplementedISR,                 /* vector 31 */
        UnimplementedISR,                 /* vector 30 */
        UnimplementedISR,                 /* vector 29 */
        UnimplementedISR,                 /* vector 28 */
        UnimplementedISR,                 /* vector 27 */
        UnimplementedISR,                 /* vector 26 */
        UnimplementedISR,                 /* vector 25 */
        UnimplementedISR,                 /* vector 24 */
        UnimplementedISR,                 /* vector 23 */
        UnimplementedISR,                 /* vector 22 */
        UnimplementedISR,                 /* vector 21 */
        ISR_SCI_Comm,                     /* vector 20 */
        UnimplementedISR,                 /* vector 19 */
        UnimplementedISR,                 /* vector 18 */
        UnimplementedISR,                 /* vector 17 */
        UnimplementedISR,                 /* vector 16 */
        TimerIC7Isr,                      /* vector 15 */
        TimerIC6Isr,                      /* vector 14 */
        TimerIC5Isr,                      /* vector 13 */
        TimerIC4Isr,                      /* vector 12 */
        TimerIC3Isr,                      /* vector 11 */
        TimerIC2Isr,                      /* vector 10 */
        TimerIC1Isr,                      /* vector 09 */
        TimerIC0Isr,                      /* vector 08 */
        ISR_Timer_Clock,                  /* vector 07 */
        UnimplementedISR,                 /* vector 06 */
        UnimplementedISR,                 /* vector 05 */
        UnimplementedISR,                 /* vector 04 */
        UnimplementedISR,                 /* vector 03 */
        UnimplementedISR,                 /* vector 02 */
        UnimplementedISR,                 /* vector 01 */
        _Startup                          /* Reset vector */
   };

#pragma CODE_SEG ROM_7400

void move_instr(short ix);
int intrp_1dctable(char sgnx, int x, unsigned char n, int * x_table, 
  char sgnz, unsigned char * z_table);
int intrp_2dctable(unsigned int x, int y, unsigned char nx, unsigned char ny,
  unsigned int * x_table, int * y_table, unsigned char * z_table);

#pragma CODE_SEG DEFAULT

#define NO_TBLES    16
#define NIOPIN    25
#define MAX_IC       8
#define MAX_PWM      5
#define MAX_PWMSCL   2
#define SER_TOUT  5
#define NADC      8
#define N_Varray  240
#define NINSTR    640
#define NO_TMGRP   5
#define NO_MSVAR_BYTES  112

// Structures for generic io instructions of form:
//   Logic: if( (VI1[ix] .op[ix]. VI2[ix]) then VO1[ix], else VO2[ix]
//   Arithmetic:  VO1[ix] = VI1[ix] .op[ix]. VI2[ix]
//                VO2[ix] not used or = VO1[ix]
typedef struct {
short offset;   // byte offset from start of ECU outpc array (used for type=0)
unsigned char
  sze;       // size in bytes of the outpc variable (for type=0)
} tblvar;

typedef struct {    // Structure for VI1,2
unsigned char
  type;      // specifies where value of VIn[ix] comes from:
             // 0=outpc variable from ECU, location specified by tbv.offset, sze
             // 1=variable from gen. purpose short array Var_array[varix]
             // 2=short constant, specified in varval below
             // 3=i/o pin[ipno], ipno is index.of PINTYPE array (0-(NIOPIN-1)).
             // 4=ADC pin[ipno], ipno is adc pin no on port AD(0-7).
             // 5=do nothing (only used for VI2, eg, if(VI1), instead of if(VI1 > VI2))
             // 6=gpio millisec timer
             // 7=outpc.seconds timer
             // 8=input capture period(ipno), ipno is pin no.on port T (0-7)
             // 9=input capture duty cycle(ipno), ipno is pin no.on port T (0-7)
             // 90=start time block: all following instructions until last instruction
             //   or next time block will execute every deltex ms 
             //  or VO = VI1, not (VI1 + VI2).
union uvtyp {   // use this union of structures to save > 30% more space for commands 
  tblvar tbv;
  short varval;           // variable to hold constant (for type = 2)
  unsigned short varix;   // index for Var_array (type = 1)
  unsigned short ipno;    // i/o or adc or inpt capture pin no (type =3,4,8,9).
  unsigned short deltex;  // following instruction block executes every deltex ms.
                          //   0 means execute approx every 0.25 ms. First instruction
                          //   after n_init_instr MUST start 1st time block, 
                          //   last instruction can't start a time block; 
                          //   consec time blks with no instr in between not allowed.
  } uv;
} inpgpio;

typedef struct {   // Structure for VO1,2
unsigned char
  type,      // 0=set gen. purpose short array Var_array[voix] 
             // 1=set I/O output pin[voix],voix is index.of PINTYPE array(0-(NIOPIN-1)).
             // 2=set PWM output pin[voix],voix is pin no.on port T (0-4)
             // 3=do nothing (only used for VO2, for example, if no else condition
             //    or VO1=(VI1 op VI2), but not VO2).
             // 4-7 = set Fuel,Spk,IAC,SprAdj, respectively (see below).
	           // 20(only used for VO1) = if cond True, execute VO2 + following 
	           //   voval instructions. If cond False, skip VO2 + following
	           //   voval instructions.
  voix;      // index for Var_array (type = 0) or io pin(0-(NIOPIN-1)) or PWM pin
             //   on port T(0-4).
short voval; // For logic operator only, = value to set the Var_array (type=0) or 
   // io pin (type=1) or pwm duty cycle (type=2) or Fuel,Spk,IAC,SprAdj(type=4-7).
   // Also number of instr to exec or skip (type = 20).
} outgpio;

typedef struct {   // structure for 1 instruction
inpgpio VI1;   // left side structure variable involved in evaluating i/o instruction.
// op ='<', '>','=','!' (latter means not equal) for a logic instruction
// op = ' ',  means if(VI1)then VO1; else VO2.
// op = '+','-','x','/' operators for arithmetic instruction.
// bw has following uses:
   // op = > or <,  if(v1 > or < v2 + bw)
   // op = '=' or '!',  if(v1 between or outside v2 - bw and v2 + bw)
   // op = + or -,  vo = (v1*bw) + or - v2
   // op = x or /,  vo = (v1 x or / v2) + bw
char op, bw;
inpgpio VI2;  // right side structure variable involved in evaluating i/o instruction.
outgpio VO1,VO2;  // structure variables specifying the output
                                // action when an io instruction is executed.
} gpio_instr;

// User inputs - 1 set in flash, 1 in ram
typedef struct {
char 
 PINTYPE[NIOPIN];   // Type of generic i/o pins:
                    //    -1 = pin not used
                    //     0 = gp input 
                    //     1 = gp output
                    //     2 = ADC
                    //     3 = high current o/p (vb921)
                    //     4 = vr/ ic input
                    //     5 = pwm o/p
unsigned char 
 PWM_TSCL[MAX_PWMSCL],  // PWM scaling - 1st element is for PWM pins T0,1,4, which are 
                        // controlled by scale clk A; 2nd for pins T2,3 on scale clk B
                      // 0 = configure scale clk for 1 us timer tic; this covers PWM
                      //  frequencies of 4 kHz to 100 kHz with good duty cycle resolution
                      // 1 = configure scale clk for 10 us timer tic; this covers PWM
                      //  frequencies of 400 Hz to 4 kHz with good duty cycle resolution
 PWMPd[MAX_PWM];      // PWM period(us or 10us tics) - (250 - 10 tics) with freq 
                      //  400Hz - 100kHz, depending on clk selection for the pin
unsigned char
 adcLF[NADC];     // Lag filter coefficients for 8 ADCs. Acts like
                  // averager: xnew = xold + xLF * (xmeas - xold), so xLF=0
                  // means value will never change, xLF=100 means no filtering.
int
  adclo[NADC],adchi[NADC];  // adc value in engineering units(e.g. deg C, psi) at
                            // adc low(0 V) and high(5 V) values. 
unsigned long baud;         // baud rate
unsigned char board_id_type,  // board type (1-255) of this board;
                  // type=0, reserved
                  // type=1, ECU (MS II)
                  // type=2. Router board
                  // type=3, Generic I/O controller
                  // type=4, Transmission controller
                  //   .........
 mycan_id;        // can_id (address) of this board (< MAX_CANBOARDS). Always 1 for non=ECU.
unsigned short can_var_rate;    // rate at which to get CAN vriables from ECU -
                             // 8 bytes (4 words) per can_var_rate, 0.128 ms tics
unsigned char debug_data,    // switch to turn on debug data for pass-thru mode 
   cpad2;                    // pad byte
unsigned char SetPin[NIOPIN],   // Pass through mode set pin values: =0, do nothing, 
 cpad3,                         // =1, set pin to 0, =2, set pin to 1; cpad byte.
 EnabPWM[MAX_PWM],     // EnabPWM enables pass thru mode for the PWM,
 SetPWMDty[MAX_PWM];   // SetPWMDty allows setting PWM duty factor (%).
// Generic I/O instruction parameters:
char gpio_halt,cpad1; // always initialized to 0 at startup so flash value doesn't
                     // matter. To change io instructions, user should make all
                     // changes, then when Burn Instructions hit, MT should send 
                     // halt=1, wait a few secs, then send all instructions to
                     // ram, then send normal 'b' command, then send command for 
                     // ECU to re-initialize('!x').(this will clear halt.) 
char n_instr, n_init_instr;  // total no. generic io instructions (<=NINSTR);  
               // the first n_init_instr instructions are executed only the
               // first time through the loop.
} inputs1;

#pragma CONST_SEG INP_VAR
// flash copy of inputs - initialized

#define SECTOR_BYTES         1024 // bytes
#define SECTOR_WORDS         ((int)(SECTOR_BYTES/2))
#define N_SECTORS(theStruct) ((int)((sizeof(theStruct)+SECTOR_BYTES-1)/SECTOR_BYTES))
#define N_PADDING(theStruct) ((int)(N_SECTORS(theStruct)*SECTOR_BYTES - sizeof(theStruct)))

const inputs1 in1flash = { 
             // PINTYPE[NIOPIN]: Type of generic i/o pins:
                    //    -1 = pin not used
                    //     0 = gp input 
                    //     1 = gp output
                    //     2 = ADC
                    //     3 = high current o/p (vb921)
                    //     4 = vr/ ic input
                    //     5 = pwm o/p
      //                       Port/pin     Default use
{ -1, //                   [0] =  A0            VB2  (set = 3 to use)
  -1, //                   [1] =  E0            GPI2 (set = 0 to use)
  -1, //                   [2] =  T4            PWM1 (set = 5 to use)
  -1, //                   [3] =  T3            PWM2 (set = 5 to use)
  -1, //                   [4] =  T2            PWM3 (set = 5 to use)
  -1, //                   [5] =  T1            PWM4 (set = 5 to use)
  -1, //                   [6] =  T0            VR/IC1 (set = 4 to use)
  -1, //                   [7] =  M2            VB4  (set = 3 to use)
  -1, //                   [8] =  M3            GPO2 (set = 1 to use)
   1, // Fan/old acc led   [9] =  M4            GPO1 (set = 1 to use)
  -1, //                   [10] = M5            GPO3 (set = 1 to use)
   2, //                   [11] = AD07          GPI4 (set = 1 to use)
   2, //                   [12] = AD06          GPI5 (set = 1 to use)
   2, //                   [13] = AD05          EGT1 (set = 2 to use)
   2, //                   [14] = AD04          EGT2 (set = 2 to use)
   2, //                   [15] = AD03          EGT3 (set = 2 to use)
   2, //                   [16] = AD02          GPI3 (set = 0 to use)
   2, //                   [17] = AD01          VB1  (set = 3 to use)
   2, //                   [18] = AD00          EGT4 (set = 2 to use)
  -1, //                   [19] = T6            VR/IC2 (set = 4 to use)
  -1, //                   [20] = T5            VR/IC3 (set = 4 to use)
  -1, //                   [21] = E1            GPI1 (set = 0 to use)
  -1, //                   [22] = E4            VB3  (set = 3 to use)
  -1, //                   [23] = B4            GPO4 (set = 1 to use)
  -1 }, //                 [24] = T7            VR/IC4 (set = 4 to use)

{0,0},    // PWM_TSCL[MAX_PWMSCL]: PWM scaling - 1st element for PWM pins T0,1,4, controlled
                      // by scale clk A; 2nd for pins T2,3 on scale clk B
                      // 0 = configure scale clk for 1 us timer tic; this covers PWM
                      //  frequencies of 4 kHz to 100 kHz with good duty cycle resolution
                      // 1 = configure scale clk for 10 us timer tic; this covers PWM
                      //  frequencies of 400 Hz to 4 kHz with good duty cycle resolution
{66,66,66,        // PWMPd[MAX_PWM]: PWM period (us or 10us tics) - (250 - 10 tics) with freq 
 66,66},          //  400Hz - 100kHz, depending on clk selection for the pin

{50,50,           // Lag filter coefficients (1-100%) for 8 ADCs. Acts like
 50,50,           // averager: xnew = xold + xLF * (xmeas - xold), so xLF=0
 50,50,           // means value will never change, xLF=100 means no filtering.
 50,50},
{0,0,             // ADC low values (in engineering units) at 0 V
 0,0,
 0,0,
 0,0}, 
{500,500,         // ADC hi values (in engineering units, default here is Vx100) at 5 V
 500,500,
 500,500,
 500,500},

115200,           // baud rate
3,                // board_type
1,                   // mycan_id (address) of this board.
110,                 // can_var_rate; rate at which to get CAN vriables from ECU -
                     // 8 bytes (4 words) per can_var_rate in 0.128 tics;
                     // can get all 112 bytes of outpc in 14 x canvarate x .128 ms.
0,                   // debug_data: switch to turn on debug data for pass-thru mode, 
0,                   // pad byte
{0},0,               // SetPin: Pass through mode set pin values: =0, do nothing, 
                     // =1, set pin to 0, =2, set pin to 1; cpad byte.
{0,0,0,0,0},         // EnabPWM: enables pass thru mode for the PWM,
{50,50,50,50,50},    // SetPWMDty allows setting PWM duty factor (%).
0,0,   // gpio_halt: always initialized to 0 at startup so flash value doesn't
                     // matter. used to change io instructions; cpad byte  
15,1    // n_instr, n_init_instr: total no. generic io instructions (<=NINSTR)
};
const unsigned char in1padding[N_PADDING(inputs1)] ={0}; 

#pragma CONST_SEG ROM_3D
// 18 bytes per instruction; 56.8888 instructions = 1 sector = 1024 bytes
//  640 instructions = 11.25 sectors. Can easily extend to 16 sectors, then
//  go on to another page..
const gpio_instr finstr[NINSTR] = 
// Generic I/O instruction parameters:
        // the first n_init_instr instructions are executed only the
        // first time through the loop.
// { type, offset,  sze,  
//         varval 
//         varix  
//         ipno                                         {type voix voval}
//         deltex      }                    
{    //  VI1          op   bw         VI2               VO1             VO2
    2,    1,     0,   ' ', 0,    5,    0,     0,     1,  9,   0,    3,  0,   0,  // init command-fan off
    90,   1000,  0,   ' ', 0,    0,    0,     0,     0,  0,   0,    0,  0,   0,  // exec every 1000 ms
    0,    6,     2,   '>', 0,    2,    300,   0,     20, 0,   2,    3,  0,   0,  // check rpm > cranking; if true, exec 2 instr
    0,    22,    2,   '>', 0,    2,    1850,  0,     1,  9,   1,    3,  0,   0,  // > hot clt degx10,fan on
    0,    22,    2,   '<', 5,    2,    1850,  0,     1,  9,   0,    3,  0,   0,  // < hot clt- h degx10, fan off,
    0,    6,     2,   '<', 1,    2,    300,   0,     1,  9,   0,    3,  0,   0,  // check rpm cranking, fan off
    2,    1,     0,   ' ', 0,    5,    0,     0,     0,  0,   0,    3,  0,   0,  // if(1), Var_arr ay[0]=0
    3,    9,     0,   '=', 0,    2,    1,     0,     20, 0,   5,    3,  0,   0,  // fan pin on,if true, exec 5 instr
    0,    24,    2,   '<', 0,    2,    100,   0,     20, 0,   4,    6,  0,   0,  // check idle tps %x10, if true, exec 4 instr
    0,    6,     2,   '<', 0,    2,    650,   0,     20, 0,   3,    3,  0,   0,  // check idle rpm, if true, exec 3 instr;
    2,    1,     0,   ' ', 0,    5,    0,     0,     6,  0,   4,    3,  0,   0,  //   IdleAdj = 4 steps
    2,    1,     0,   ' ', 0,    5,    0,     0,     5,  0,   30,   3,  0,   0,  //   SpkAdj = 30 degx10
    2,    1,     0,   ' ', 0,    5,    0,     0,     0,  0,   1,    3,  0,   0,  //   Var_array[0]=1
    1,    0,     0,   '=', 0,    2,    0,     0,     6,  0,   0,    3,  0,   0,  // if Var_array[0]=0, clear IdleAdj
    1,    0,     0,   '=', 0,    2,    0,     0,     5,  0,   0,    3,  0,   0,  // if Var_array[0]=0, clear SpkAdj, last instr.
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0,
    0,    0,     0,   ' ', 0,    0,    0,     0,     0,  0,   0,    3,  0,   0
};

// Note: *** Must set same outmsg data in ECU and CAN device, or at least have the same
//  no_bytes and offsets.
const can_outmsg outmsg[MAX_OUTMSGS] = { 
{ 156,    // xrate: =0 means the message is only transmitted when requested, 
          //  otherwise it indicates the clock speed in 0.128 tics at which
          //  a CAN message will be transmitted automatically after receiving 
          //  a request for data by setting the activate_xrate mth bit= 1,
          //  where m = msg no = 0 to MAX_OUTMSGS.
  18,     // no_bytes: the total number of bytes of all the variables in the message.                             
          //  Must be <= size of outpc.
  19,     // destination id = 1 << 4, varblk = 3.
 {6,7,   //rpm  // offset[MAX_BYTES_OUTMSG]: this holds the byte offsets of each of the
  8,9,   //adv  //  bytes in the message relative to start of  outpc. For a short, specify
  18,19, //map  //  2 consecutive offsets in the proper order for the requesting processor.
  20,21, //mat
  22,23, // clt
  24,25, // tps
  28,29, // afr1
  32,33, // knock
  64,65, // maf
  {0} }
}, 
{0}      //  next 7 nessages
};
#pragma CONST_SEG ROM_OVF
const char RevNum[20] =  {    // revision no:
 // only change for major rev and/or interface change.
  "GPIO Rev 1.20000   "
},
 Signature[32] = {            // program title.
 // Change this every time you tweak a feature.
  "** V1.20Embedded Code by B&G **"
 };

// ram copy of inputs
inputs1 inpram;

// ram buffer for single instruction
gpio_instr rinstr; 

// I/O variables
volatile unsigned char *pwmdty[MAX_PWM],*pwmper[MAX_PWM],*pinptr[NIOPIN];
unsigned char pinmask[NIOPIN],pinpos[NIOPIN],pwm_mask,ic_mask,iopin[NIOPIN];
short Var_array[N_Varray];
unsigned char no_tmgrp,first_instr;
unsigned long tmgrp[NO_TMGRP];
short ixl_tmgrp[NO_TMGRP],gpio_ms;
short
   // Adjustments for fuel(%x10 +/- to ve),spark(degx10),idle step,spare 
   // for GPIO boards. These are previous values, current are in outpc..
   FuelAdjl,SpkAdjl,IdleAdjl,SprAdjl; 
// General variables
unsigned int mms,millisec,burn_flag,can_mmsclk,can_varoff;
unsigned long lmms,ltch_lmms,adc_lmms;
unsigned char next_adc,first_adc,sendCANAdj=0;
char nadc,npwm,nic;

// pointers for spare ADC pins
volatile unsigned short *adcraw[NADC];

#include "..\..\serCANComm\serCAN3.h"

// allocate space in ram for flash burner core
volatile unsigned char RamBurnPgm[36]; 

// rs232 Outputs to pc
typedef struct {
unsigned short seconds; 
short adcval[NADC];                  // adc values in engineering units
unsigned char pinval[NIOPIN],cpad1;  // status of generic io pins.
unsigned char pwm_duty[MAX_PWM],cpad2;     // current pwm duty(%).
unsigned short ic_period[MAX_IC];    // Input capture period (us) read from IC.
unsigned short ic_duty[MAX_IC];      // Input capture duty (%) calculated from IC times.
short
   FuelAdj,SpkAdj,	     // Adjustments for fuel(%x10 +/- to ve),spark(degx10),idle step,spare 
   IdleAdj,SprAdj;			 // for GPIO boards. Send to MS ram input table.
short spare[10];
unsigned char ospare,cksum;
} variables;

variables outpc;

// buffer to hold the MS output variables
unsigned char msvar[NO_MSVAR_BYTES];

#pragma CONST_SEG ROM_OVF

typedef struct {
   unsigned int *addrRam;
   unsigned int *addrFlash;
   unsigned int  n_bytes;
   unsigned int  fpage;
} tableDescriptor;

const tableDescriptor tables[NO_TBLES] =  { 
  { (unsigned int *)&inpram, (unsigned int *)&in1flash,       sizeof(inputs1),    0x3C }, 
  { NULL,                    NULL,                            0,                  0x3C },
  { NULL,                    NULL,                            0,                  0x3C },
  { (unsigned int *)msvar,   NULL,                            sizeof(msvar),      0x3C },
  { NULL,                    (unsigned int *)&finstr,         sizeof(finstr),     0x3D },
  { NULL,                    (unsigned int *)&outmsg,         sizeof(can_outmsg), 0x3D },
  { (unsigned int *)txbuf,   NULL,                            sizeof(txbuf),      0x3C },
  { (unsigned int *)&outpc,  NULL,                            sizeof(outpc),      0x3C },
  { NULL,                    NULL,                            0,                  0x3C },
  { NULL,                    NULL,                            0,                  0x3C },
  { NULL,                    NULL,                            0,                  0x3C },
  { NULL,                    NULL,                            0,                  0x3C },
  { NULL,                    NULL,                            0,                  0x3C },
  { NULL,                    NULL,                            0,                  0x3C },
  { NULL,                    NULL,                            0,                  0x3C },
  { NULL,                    (unsigned int *)&RevNum,         sizeof(RevNum),     0x3C }
};

#define tableInit(iTable) (void)memcpy(tables[iTable].addrRam, tables[iTable].addrFlash, tables[iTable].n_bytes)
#define tableByteRam(iTable, iByte)   ((unsigned char *)tables[iTable].addrRam + iByte)
#define tableWordRam(iTable, iWord)   (tables[iTable].addrRam + iWord)
#define tableByteFlash(iTable, iByte) ((unsigned char *)tables[iTable].addrFlash + iByte)
#define tableWordFlash(iTable, iWord) (tables[iTable].addrFlash + iWord)
#define tableBytes(iTable)            (tables[iTable].n_bytes)
#define tableWords(iTable)            ((tables[iTable].n_bytes+1)/2) // Round up
#define tableFPage(iTable)            ((unsigned char)tables[iTable].fpage)

// Prototypes - Note: ISRs prototyped above.
void main(void);
short exec_rinstr(void);
void reset(void);
void get_adc(char chan1, char chan2);
void fburner(unsigned int* progAdr, unsigned int* bufferPtr, 
  					 unsigned int no_words);
int intrp_2ditable(unsigned int x, int y, unsigned char nx, unsigned char ny,
  unsigned int * x_table, int * y_table, int * z_table) ;
void Flash_Init(unsigned long oscclk) ;
void Flash_Erase_Sector(unsigned int *address) ;
void Flash_Write_Word(unsigned int *address, unsigned int data) ;

#pragma CONST_SEG ROM_OVF

#pragma CODE_SEG DEFAULT

extern void reboot(void) ;
extern void monitor(void) ;
extern void SpSub(void);
extern void NoOp(void);

void main(void) {
short ix,jx,ni_skip;
char tbexflg;
long ltmp;
unsigned long ultmp,ultmp2,dtjx;

  PPAGE = 0x3C;
  // initalize PLL - reset default is Oscillator clock
  // 8 MHz oscillator, PLL freq = 48 MHz, 24 MHz bus, 
  //  divide by 16 for timer of 2/3 usec tic
  PLLCTL &= 0xBF;     // Turn off PLL so can change freq
  SYNR = 0x02;        // set PLL/ Bus freq to 48/ 24 MHz
  REFDV = 0x00;
  PLLCTL |= 0x40;     // Turn on PLL
  // wait for PLL lock
  while (!(CRGFLG & 0x08));
  CLKSEL = 0x80;      // select PLL as clock
  // wait for clock transition to finish
  for (ix = 0; ix < 60; ix++);
    
  // open flash programming capability
  Flash_Init(8000);
  //inp_spare used to force inpflash, flashve_table into sectors.
  // Must use in program or it won't use up the entire sector. POS CW 
  // ignores the pragma making this stupid statement necessary.
  ix = in1padding[0];
  if((int)RamBurnPgm & 0x0001)	{   // odd address - cpy to even one
    (void)memcpy((void *)RamBurnPgm,NoOp,1);         // cpy noop to 1st location
    (void)memcpy((void *)&RamBurnPgm[1],SpSub,32);   // cpy flashburn core pgm to ram
  }
  else
    (void)memcpy((void *)RamBurnPgm,SpSub,32);       // cpy flashburn core pgm to ram

  // load all user inputs from Flash to RAM
  tableInit(0);
  
  // set all unused (even unbonded) ports to inputs with pullups
  DDRA = 0x00;
  DDRB = 0x00;
  DDRE = 0x00;
  DDRM = 0x00;
  PUCR |= 0x13;    // enable pullups for ports E, B and A
                   //  Note: this causes unattached pins to read high
  DDRP = 0x00;
  PERP = 0xFF;     // enable pullup resistance for port P 
  DDRJ &= 0x3F;
  PERJ |= 0xC0;    // enable pullup resistance for port J6,7
  DDRS &= 0xF3;
  PERS |= 0x0C;    // enable pullup resistance for port S2,3

  // set up all selected i/o ports based on following:
  //   PINTYPE     =0     =1     =2     =3     =4     =5
  //    ix        GPIN   GPOUT   ADC   VBOut  VR/IC   PWM
  // -----------------------------------------------------
  //    0 - A0                    X             X      X
  //    1 - E0             X      X      X      X      X
  //    2 - T4                    X
  //    3 - T3                    X
  //    4 - T2                    X
  //    5 - T1                    X
  //    6 - T0                    X
  //    7 - M2                    X             X      X
  //    8 - M3                    X             X      X
  //    9 - M4                    X             X      X
  //    10- M5                    X             X      X
  //    11- AD07                                X      X
  //    12- AD06                                X      X
  //    13- AD05                                X      X
  //    14- AD04                                X      X
  //    15- AD03                                X      X
  //    16- AD02                                X      X
  //    17- AD01                                X      X
  //    18- AD00                                X      X
  //    19- T6                    X                    X
  //    20- T5                    X                    X
  //    21- E1             X      X      X      X      X
  //    22- E4                    X             X      X
  //    23- B4                    X             X      X
  //    24- T7                    X                    X
  // -----------------------------------------------------
  //     X means port/pin won't support this pintype
  //  Also note board hardware may not support pintype 
  //  (for example hw for only 4 pwms, but 5 pins capable).
  // -----------------------------------------------------
  //  Set pointers to real port addresses and configure i/o
  nadc = 0;
  ATD0CTL2 = 0x00;     // turn off adc power initially
  npwm = 0;
  // clear all pwms
  MODRR = 0;
  PWME = 0;             // disable pwms initially
  pwm_mask = 0;
  nic = 0;
  ic_mask = 0;
  TIOS = 0;
  TCTL1 = 0;            // clear OC options
  TCTL2 = 0;
  TCTL3 = 0;            // clear IC options
  TCTL4 = 0;
  TIE = 0;

  for(ix = 0; ix < NIOPIN; ix++)  {
    iopin[ix] = 0;
    if(ix == 0)  {
      pinptr[0] = pPORTA;
      pinpos[0] = 0;       // A0
      pinmask[0] = 0x01;
      iopin[0] = 1;
      if(inpram.PINTYPE[0] <= 0)   // set as input (0)
        DDRA &= ~pinmask[0];
      else                         // set as output (1)
        DDRA |= pinmask[0];
    } 
    else if(ix == 1)  {
      pinptr[1] = pPORTE;
      pinpos[1] = 0;       // E0
      pinmask[1] = 0x01;
      iopin[1] = 1;
      DDRE &= ~pinmask[1];         // must be set as input (0)
    }
    else if(((ix >= 2) && (ix <= 6)) ||
            (ix == 19) || (ix == 20) || (ix == 24))  {
      // port T, pins 4 - 0, 6,5, 7
      pinptr[ix] = pPTT;
      if((ix >= 2) && (ix <= 6))  {
        pinpos[ix] = 6 - ix;
        pwmper[pinpos[ix]] = 0;
      } 
      else if((ix == 19) || (ix == 20))  {
        pinpos[ix] = 25 - ix;
      } 
      else  {
        pinpos[24] = 7;
      }
      pinmask[ix] = 0x01 << pinpos[ix];
      if(inpram.PINTYPE[ix] <= 0)  {          // set as input (0)
        DDRT &= ~pinmask[ix];
        PPST &= ~pinmask[ix];                 // set to pullup (0)
        PERT |= pinmask[ix];                  // enable pull ups (1)
        iopin[ix] = 1;
      } 
      else if((inpram.PINTYPE[ix] == 1) || 
              (inpram.PINTYPE[ix] == 3))  {   // set as output (1)
        DDRT |= pinmask[ix];
        TIOS |= pinmask[ix];
        iopin[ix] = 1;
      }
      else if(inpram.PINTYPE[ix] == 4)  {     // VR/IC input
        //  Set up generic Input Captures IC0-7
        ic_mask |= pinmask[ix];
        if(pinpos[ix] < 4)  {
          TCTL4 |= (0x03 << 2*pinpos[ix]);    // set for capture both edges
        }
        else  {
          TCTL3 |= (0x03 << 2*pinpos[ix]);    // set for capture both edges
        }
        nic++;
      }
      else if(inpram.PINTYPE[ix] == 5)  {     // PWM output
        //  Set up generic PWMs - PWM0-4
        pwm_mask |= 0x01 << pinpos[ix];  // PWMs on T0-T4 only
        if(npwm == 0)  {     // one-time setup on first PWM output
          // Set prescaler to 16. This divides bus clk (= PLLCLK/2 = 24 MHz)
          //   by 16. This gives 1.5 MHz timer, 1 tic = 2/3 us.
          TSCR2 = 0x04;
          PWMCLK = 0x0F;    // select scaled clocks SB, SA
          // configure 1st 2 PWMs on first (A) clock
          if(inpram.PWM_TSCL[0] == 0)  {   
            PWMPRCLK = 0x00;  // prescale A clock = bus = 24 MHz
            PWMSCLA = 0x0C;   // pwm clk = SA clk/(2*SCLA) = 24MHz/24 = 1 us clk
          } 
          else  {
            PWMPRCLK = 0x00;  // prescale A clock = bus = 24 MHz
            PWMSCLA = 0x78;   // pwm clk = SA clk/(2*SCLA) = 24MHz/240 = 10 us clk
          }
          // configure 2nd 2 PWMs on 2nd (B) clock
          if(inpram.PWM_TSCL[1] == 0)  {   
            PWMPRCLK = 0x00;  // prescale B clock = bus = 24 MHz
            PWMSCLB = 0x0C;   // pwm clk = SB clk/(2*SCLB) = 24MHz/24 = 1 us clk
          } 
          else  {
            PWMPRCLK = 0x00;  // prescale B clock = bus = 24 MHz
            PWMSCLB = 0x78;   // pwm clk = SB clk/(2*SCLB) = 24MHz/240 = 10 us clk
          }
          PWMCAE = 0x00;    // standard left align pulse:   -----
                            //                             |     |______
                            //                               duty
                            //                             <---period---> 
        }
        pwmper[pinpos[ix]] = (&PWMPER0 + pinpos[ix]);
        pwmdty[pinpos[ix]] = (&PWMDTY0 + pinpos[ix]);
        *pwmper[pinpos[ix]] = inpram.PWMPd[pinpos[ix]];  // set PWM period (us)
        *pwmdty[pinpos[ix]] = 0;                         //clear pwm duty cycle
        npwm++;                      
      }
    }    
    else if((ix >= 7) && (ix <= 10))  {
      // port M, pins 2 - 5
      pinptr[ix] = pPTM;
      pinpos[ix] = ix - 5;
      pinmask[ix] = 0x01 << pinpos[ix];
      iopin[ix] = 1;
      if(inpram.PINTYPE[ix] <= 0)   // set as input (0)
        DDRM &= ~pinmask[ix];
      else                          // set as output (1)
        DDRM |= pinmask[ix];
    }
    else if((ix >= 11) && (ix <= 18))  {
      // port AD0, pins 7 - 0
      pinpos[ix] = 18 - ix;
      pinmask[ix] = 0x01 << pinpos[ix];
      adcraw[pinpos[ix]] = 0;
      if(inpram.PINTYPE[ix] <= 0)  {       // set as input (0)
        ATD0DIEN |= pinmask[ix];           // enable PORTAD pin as input
        pinptr[ix] = pPTIAD;
        DDRAD &= ~pinmask[ix];             // set as input (0)
        PPSAD &= ~pinmask[ix];             // set to pullup
        PERAD |= pinmask[ix];              // enable pull ups
        iopin[ix] = 1;
      }
      else if((inpram.PINTYPE[ix] == 1) ||
              (inpram.PINTYPE[ix] == 3))  {  // set as output (1)
        ATD0DIEN |= pinmask[ix];             // enable PORTAD pin as input
        pinptr[ix] = pPTAD;
        DDRAD |= pinmask[ix];                // set as output (1)
        RDRAD &= ~pinmask[ix];               // full drive - not reduced
        iopin[ix] = 1;
      }
      else if(inpram.PINTYPE[ix] == 2)  {
        // set up ADC
        ATD0DIEN &= ~pinmask[ix];            // disable io
        pinptr[ix] = pPORTAD0;               // not used, values read in adcraw
        DDRAD &= ~pinmask[ix];               // set as input (0)
        if(nadc == 0)  {
          // set up ADC processing first time through
          // Set up ADCs so they continuously convert, then read result registers
          //  every millisecond
          ATD0CTL2 = 0x40;  // leave interrupt disabled, set fast flag clear
          ATD0CTL3 = 0x00;  // do 8 conversions/ sequence
          ATD0CTL4 = 0x67;  // 10-bit resoln, 16 tic cnvsn (max accuracy),
                    // prescaler divide by 16 => 2/3 us tic x 18 tics
          ATD0CTL5 = 0xB0;  // right justified,unsigned, continuous cnvsn,
                    // sample 8 channels starting with AN0
        }
        // set pointers to get ADC holding reg values
        nadc++;
        adcraw[pinpos[ix]] = (&ATD0DR0 + pinpos[ix]);
      }
    } 
    else if(ix == 21)  {
      // port E, pin 1
      pinptr[21] = pPORTE;
      pinpos[21] = 1;
      pinmask[21] = 0x01 << 1;
      iopin[21] = 1;
      DDRE &= ~pinmask[21];         // must be set as input (0)
    }
    else if(ix == 22)  {
      // port E, pin 4
      pinptr[22] = pPORTE;
      pinpos[22] = 4;
      pinmask[22] = 0x01 << 4;
      iopin[22] = 1;
      if(inpram.PINTYPE[22] <= 0)   // set as input (0)
        DDRE &= ~pinmask[22];
      else                          // set as output (1)
        DDRE |= pinmask[22];
    }
    else if(ix == 23)  {
      // port B, pin 4
      pinptr[23] = pPORTB;
      pinpos[23] = 4;
      pinmask[23] = 0x01 << 4;
      iopin[23] = 1;
      if(inpram.PINTYPE[23] <= 0)   // set as input (0)
        DDRB &= ~pinmask[23];
      else                          // set as output (1)
        DDRB |= pinmask[23];
    }
  }

  reinit_flag = 0;

  // set up CRG RTI Interrupt for .128 ms clock. CRG from 8MHz oscillator.
  mms = 0;        // .128 ms tics
  millisec = 0;   // 1.024 ms clock (8 tics) for adcs
  lmms = 0;
  ltch_lmms = 0;
  can_mmsclk = 0; // .128 ms tics
  outpc.seconds = 0;   // (1.0035) secs
  burn_flag = 0;
  RTICTL = 0x10;   // load timeout register for .128 ms (smallest possible)
  CRGINT |= 0x80;  // enable interrupt
  CRGFLG = 0x80;   // clear interrupt flag (0 writes have no effect)

  // Set up SCI (rs232): SCI BR reg= BusFreq(=24MHz)/16/baudrate
//inpram.baud = 115200; // Use this for debugging corrupted flash sectors.
  SCI0BDL = (unsigned char)(1500000/inpram.baud);
  ltmp = (150000000/inpram.baud) - ((long)SCI0BDL*100);
  if(ltmp > 50)SCI0BDL++;   // round up 
//inpram.baud = in1flash.baud; // See above.
  SCI0CR1 = 0x00;
  SCI0CR2 = 0x24;   // TIE=0,RIE = 1; TE=0,RE =1
  txcnt = 0;
  rxcnt = 0;
  txmode = 0;
  txgoal = 0;
  kill_ser = 0;
  vfy_flg = 0;
  rcv_timeout = 0xFFFFFFFF;

  // Finish setting up PWMs
  if(npwm > 0)  {  
    TIOS |= pwm_mask;    // Timer ch 0-4 = PWM/IC/IO, ch 5-7 = IC/IO. 
    MODRR = pwm_mask;    // Make Port T pins 0-4 be PWM or not
    PWMPOL = pwm_mask;   // if pwm, polarity = 1, => go hi when start
    PWME = pwm_mask;     // enable PWM pins
  }
  
  //  Finish etting up VR/ICs
  for(ix = 0;ix < MAX_IC;ix++)  {
    // clear ic stuff
    outpc.ic_period[ix] = 0;
    outpc.ic_duty[ix] = 0;                     
  }
  if(nic > 0)
    TIE |= ic_mask;              // enable IC interrupts
    
  // finish setting up ADC processing
  if(nadc > 0)  {
    ATD0CTL2 |= 0x80;  // turn on ADC0
    // wait for ADC engine charge-up or P/S ramp-up
    for(ix = 0; ix < 160; ix++)  {
      while(!(ATD0STAT0 >> 7));	 // wait til conversion complete
      ATD0STAT0 = 0x80;
    }
  }
  // get all adc values
  first_adc = 1;
  get_adc(0,7);
  next_adc = 0;	   	// specifies next adc channel to be read
  first_adc = 0;
  adc_lmms = lmms;

  reset();

  // Initialize variables
  flocker = 0;  

  /* Initialize CAN comms */
  can_reset = 0;
  can_id = inpram.mycan_id;
  can_varoff = 0;
  CanInit();

  // enable timer and global interrupts
  TSCR1 |= 0x80;
  ENABLE_INTERRUPTS
  
  // wait 1 sec to allow ECU to fully initialize
  while(outpc.seconds == 0);
  
  // Set up COP timeout for main loop
  COPCTL = 0x44;     // 131 ms timeout for reboot

  /***********************************************************
  //  main loop
  ***********************************************************/

  for (;;)  {
  
  // clear COP timeout counter
  ARMCOP = 0x55;
  ARMCOP = 0xAA;
    
  // Get ADC values
  DISABLE_INTERRUPTS
  ultmp = lmms;
  ENABLE_INTERRUPTS
  if((ultmp - adc_lmms) > 78)  {          // every 10 ms (78 x .128 ms clk)
    adc_lmms = ultmp;
    // read 10-bit ADC results, convert to engineering units and filter
    get_adc(next_adc,next_adc+1);    // get two channels on each pass, so 
    next_adc += 2;                   //  each channel updated every 40 ms
    if(next_adc >= nadc)next_adc = 0;
  }

  // Get i/o pin values
  for(ix = 0;ix < NIOPIN; ix++)  {
    if(iopin[ix])
      outpc.pinval[ix] = (unsigned char)((*pinptr[ix] & pinmask[ix]) >> pinpos[ix]);
    else
      outpc.pinval[ix] = 0;
  }
  // Get pwm duty values and convert to %
  for(ix = 0;ix < MAX_PWM; ix++)  {
    outpc.pwm_duty[ix] = (unsigned char)((*pwmdty[ix] * (unsigned short)100) 
                                                 / (*pwmper[ix]));
  }
    
  // Fan control (This logic done with gpio commands)
  /*
    Init Cmd:  Fan(port E4) Off
    if(rpm > crank_rpm)  {
      if(clt > clt_hot)
        Fan(port E4) = On
      if(clt < clt_hot - hysteresis )
        Fan(port E4) = Off
    }
    if(rpm < crank_rpm+1)
      Fan(port E4) = Off
    Var_array[0] = 0
    if(Fan(port E4) On)  {
      if(tps < idle_tps)  {
        if(rpm < idle_rpm)  {
          IdleAdj = idle_xtra
          SpkAdj = spk_xtra
          Var_array[0] = 1
        }
      }
    }
    if(Var_array[0] == 0)
      IdleAdj = 0
    if(Var_array[0] == 0)
      SpkAdj = 0
    }
  */
  
  // Turbo control (if can't be done with gpio instructions)
  
  // Nitrous control (if can't be done with gpio instructions)

/***************************************************************************
**
**    Implement pass through mode 
**
**************************************************************************/
    for(ix = 0;ix < NIOPIN; ix++)  {
      if(inpram.SetPin[ix] == 1)  {
        // set pin low 
        *pinptr[ix] = *pinptr[ix] & ~pinmask[ix];
      }
      else if(inpram.SetPin[ix] == 2)  {
        // set pin high
        *pinptr[ix] = *pinptr[ix] | pinmask[ix];
      }
    }
    for(ix = 0;ix < MAX_PWM; ix++)  {
      if(inpram.EnabPWM[ix])  {
        // set pwm_duty_cycle in register (us or 10 us tics)
        *pwmdty[ix] = (unsigned char)(((unsigned short)inpram.SetPWMDty[ix] * 
             (*pwmper[ix])) / 100);
      }
    }
          
/***************************************************************************
**
**    Implement general purpose i/o 
**
**************************************************************************/
#ifdef CAN_TEST
  outpc.spare[0] = (short)((msvar[0] << 8) + msvar[1]);  // secs
  outpc.spare[1] = (short)((msvar[2] << 8) + msvar[3]);  // pw1
  outpc.spare[2] = (short)((msvar[6] << 8) + msvar[7]);  // rpm
  outpc.spare[3] = (short)((msvar[8] << 8) + msvar[9]);  // adv
  outpc.spare[4] = (short)((msvar[18] << 8) + msvar[19]);  // map
  outpc.spare[5] = (short)((msvar[20] << 8) + msvar[21]);  // mat
  outpc.spare[6] = (short)((msvar[22] << 8) + msvar[23]);  // clt
  outpc.spare[7] = (short)((msvar[24] << 8) + msvar[25]);  // tps
  outpc.spare[8] = (short)((msvar[26] << 8) + msvar[27]);  // bat
#endif
  
if(inpram.debug_data)  {
  outpc.spare[0] = (short)((msvar[0] << 8) + msvar[1]);  // rpm
  outpc.spare[1] = (short)((msvar[2] << 8) + msvar[3]);  // adv
  outpc.spare[2] = (short)((msvar[4] << 8) + msvar[5]);  // map
  outpc.spare[3] = (short)((msvar[6] << 8) + msvar[7]);  // mat
  outpc.spare[4] = (short)((msvar[8] << 8) + msvar[9]);  // clt
  outpc.spare[5] = (short)((msvar[10] << 8) + msvar[11]);  // tps
  outpc.spare[6] = (short)((msvar[12] << 8) + msvar[13]);  // afr1,2
  outpc.spare[7] = (short)((msvar[14] << 8) + msvar[15]);  // knkV
  outpc.spare[8] = (short)((msvar[16] << 8) + msvar[17]);  // maf
}
  
  if(first_instr)  {
    // 1st time start with inital instructions executed 1x only
    for(ix = 0;ix < inpram.n_init_instr; ix++)  {
      // move flash instruction into ram buffer (rinstr)
      move_instr(ix);
      ni_skip = exec_rinstr();
    }
    first_instr = 0;
  }
  else  {
    ix = inpram.n_init_instr;  // this is first normal instr - MUST be type=90;
                               // last instr can NOT be type=90
  }
  
  //  This starts executing time blocks of instructions
  for(jx = 0;jx < no_tmgrp; jx++)  {
    // MT has commanded halting exec of instrs preparatory to modifying them
    if(inpram.gpio_halt)break;
    // move flash instruction into ram buffer (rinstr)
    move_instr(ix);
    // check if time for this time group to execute
    if(lmms > tmgrp[jx])  {
      // convert ms to .128 ms tics
      if(rinstr.VI1.uv.deltex == 0)
        dtjx = 2;      // 2  .128 ms tics (~0.25 ms)
      else
        dtjx = (rinstr.VI1.uv.deltex * (unsigned long)1000) / 128; 
      tmgrp[jx] = lmms + dtjx;    // last instruction in this time group done -  
                                  // update for next exec time
      tbexflg = 1;
    } 
    else
      tbexflg = 0;
    
// process block of instructions in this time group
BLK_EXEC:     
    if(tbexflg && (rinstr.VI1.type != 90))  {
      
      // execute instruction in ram
      ni_skip = exec_rinstr();
      
      // skip next ni_skip instructions
      ix += ni_skip;
    }
    // Is this instr (ix) the last one in this time group
    if(ix != ixl_tmgrp[jx])   {     // not last instr in this time block
      ix++;
      if(ix < inpram.n_instr)  {
        // move flash instruction into ram buffer (rinstr)
        move_instr(ix);
        goto BLK_EXEC;    // keep going til next time group
      }
      break;
    }
    // go to to next time group, on next instruction
    ix++;
    if(ix >= inpram.n_instr)
      break;
  }			 // end for jx=0, no_tmgrp loop
  // at this point have executed all instructions for all ready time groups
  
  if(sendCANAdj)  {
    // send new Fuel,Spk,Idle and Spare adjustments
    if((outpc.FuelAdj != FuelAdjl) || (outpc.SpkAdj != SpkAdjl) ||
       (outpc.IdleAdj != IdleAdjl) || (outpc.SprAdj != SprAdjl))  {
      // there has been a change - send data via CAN
      // load main loop ring buffer(can[1]) - send data
      jx = can[1].cxno_in;
      can[1].cx_msg_type[jx] = MSG_CMD; 
      can[1].cx_datbuf[jx][0] = *((char *)&outpc.FuelAdj);
      can[1].cx_datbuf[jx][1] = *((char *)&outpc.FuelAdj + 1);
      can[1].cx_datbuf[jx][2] = *((char *)&outpc.SpkAdj);
      can[1].cx_datbuf[jx][3] = *((char *)&outpc.SpkAdj + 1);
      can[1].cx_datbuf[jx][4] = *((char *)&outpc.IdleAdj);
      can[1].cx_datbuf[jx][5] = *((char *)&outpc.IdleAdj + 1);
      can[1].cx_datbuf[jx][6] = *((char *)&outpc.SprAdj);
      can[1].cx_datbuf[jx][7] = *((char *)&outpc.SprAdj + 1);
      // Put FuelAdj, etc in ECU, in2ram.FuelAdj, etc.
      can[1].cx_destvarblk[jx] = 5;
      // below is offset of FuelAdj, etc from start of in2ram in ECU
      can[1].cx_destvaroff[jx] = 946;
      can[1].cx_dest[jx] = 0;		// send to ECU (board 0)
      can[1].cx_varbyt[jx] = 8;	 // sending 8 bytes (4 words)
      send_can(1);
    }
    sendCANAdj = 0;
    FuelAdjl = outpc.FuelAdj;
    SpkAdjl = outpc.SpkAdj;
    IdleAdjl = outpc.IdleAdj;
    SprAdjl = outpc.SprAdj;
  }

  // Get ECU output variables.
  // Do below if want entire outpc array - very slow
  // Grab 8 bytes every can_var_rate(.128 ms tics)
#ifdef CAN_TEST  
  if(can_mmsclk > inpram.can_var_rate)  {
      // load xmt interrupt ring buffer(can[0]) - request data
      ix = can[1].cxno_in;
      can[1].cx_msg_type[ix] = MSG_REQ; 
      // Get all data from var. block outpc in ECU)      
      can[1].cx_destvarblk[ix] = 7;
      // Put data rcvd from ECU into var block msvar in GPIO)      
      can[1].cx_myvarblk[ix] = 3;
      can[1].cx_myvaroff[ix] = can_varoff;
      can[1].cx_dest[ix] = 0;		// send to ECU (board 0)
      // below is offset from start of outpc in ECU
      can[1].cx_destvaroff[ix] = can_varoff;
      // calculate no bytes, and offset for next time
      can_varoff += 8;
      if(can_varoff < NO_MSVAR_BYTES)
        can[1].cx_varbyt[ix] = 8;	 // get 8 bytes
      else  {
                                   // get remaining bytes
        can[1].cx_varbyt[ix] = NO_MSVAR_BYTES - (can_varoff - 8);
        can_varoff = 0; 
      }
      send_can(1);
      can_mmsclk = 0;
  }
#endif

  // Do below if want specific, non-contiguous vars from outpc array
  // Grab all bytes every can_var_rate(.128 ms tics)
  // Note: *** Must set same outmsg data in ECU and CAN device, or at 
  //  least have the same no_bytes and offsets.
#ifdef CAN_TESTOUT
  if(can_mmsclk > inpram.can_var_rate)  {
		  ix = can[1].cxno_in;
		  can[1].cx_msg_type[ix] = OUTMSG_REQ; 
		  // Put data from ECU in this gpio block      
		  can[1].cx_destvarblk[ix] = 3;  // from tableDescriptor
		     // above really myvarblk, destvarblk always outpc
		  can[1].cx_dest[ix] = 0;		// send to ECU (board 0)
		  // Below is offset from start of gpio block - it should
		  // always be 0. If more than 1 msg, ECU will send offset
		  // to count byte sent
		  can[1].cx_destvaroff[ix] = 0;
		  can[1].cx_outmsg_no[ix] = 0;  // 0 to < MAX_OUTMSGS
		  send_can(1);
      can_mmsclk = 0;
  }
#endif
  
/***************************************************************************
**
** Check whether to burn flash
**
**************************************************************************/
BURN_FLASH:
    if(burn_flag) {
      // burn flash 1024 byte(512 word) sector(s)
        fburner(tableWordFlash(burn_idx, 0), tableWordRam(burn_idx, 0), tableWords(burn_idx));
        if(burn_flag >= tableWords(burn_idx))  {
          burn_flag = 0;
          flocker = 0;
        }
        else
          burn_flag++;
    }
/***************************************************************************
**
** Check for reinit command
**
**************************************************************************/
    if(reinit_flag)  {
      reinit_flag = 0;
      // update initialization of variables/ registers dependent on
      // user config inputs
      reset();
      // Set up SCI (rs232): SCI BR reg= BusFreq(=24MHz)/16/baudrate
      SCI0BDL = (unsigned char)(1500000/inpram.baud);
      ltmp = (150000000/inpram.baud) - ((long)SCI0BDL*100);
      if(ltmp > 50)SCI0BDL++;   // round up
      can_id = inpram.mycan_id;
      can_reset = 1;     // do this to reset id in CAN message mask register
    }

/***************************************************************************
**
**  Check for serial, CAN receiver timeout
**
**************************************************************************/
    DISABLE_INTERRUPTS
    ultmp = lmms;
    ultmp2 = rcv_timeout;
    ENABLE_INTERRUPTS
    if(ultmp > ultmp2)  {
      txmode = 0;    // break out of current receive sequence
      rcv_timeout = 0xFFFFFFFF;
    }
    DISABLE_INTERRUPTS
    ultmp2 = ltch_CAN;
    ENABLE_INTERRUPTS
    if(ultmp > ultmp2)  {
      getCANdat = 0;    // break out of current receive sequence
      ltch_CAN = 0xFFFFFFFF;
    }
    if(kill_ser)  {
      if(outpc.seconds > kill_ser_t)  {
        kill_ser = 0;
        SCI0CR2 |= 0x24;     // rcv, rcvint re-enable
      }
    }
    if(vfy_flg)  {
      vfy_flg = 0;
      for(ix = 0; ix < tables[tble_idx].n_bytes; ix++)  {
        if(*tableByteFlash(tble_idx,ix) != *tableByteRam(tble_idx,ix))
        vfy_fail++;
      }
      txcnt = 0;
      txgoal = 2;
      SCI0DRL = *(char *)&vfy_fail;   
      SCI0CR2 &= ~0x24;   // rcv, rcvint disable
      SCI0CR2 |= 0x88;    // xmit enable & xmit interrupt enable
    }
  
/***************************************************************************
**
**  Check for CAN reset
**
**************************************************************************/
	  if(can_reset)  {
		  /* Re-initialize CAN comms */
		  CanInit();
		  can_reset = 0;
	  }
    
  }     //  END Main while(1) Loop
}

short exec_rinstr()  {
short jx,vieval[2],vo1eval,vo2eval,nskip;
char skipvi2,trueflg,cond,voskip;

    // evaluate the 2 input variables in each instruction
    switch(rinstr.VI1.type)  {
      case 0:       // set variable to outpc var from MS
        if(rinstr.VI1.uv.tbv.sze == 1)
          vieval[0] = *(msvar + rinstr.VI1.uv.tbv.offset);
        else
          vieval[0] = *( int *)(msvar + rinstr.VI1.uv.tbv.offset);
        break;
      case 1:       // set variable from gen purpose array
        vieval[0] = Var_array[rinstr.VI1.uv.varix];
        break;
      case 2:       // set variable to input variable
        vieval[0] = rinstr.VI1.uv.varval;
        break;
      case 3:       // set variable to I/O pin
        vieval[0] = outpc.pinval[rinstr.VI1.uv.ipno];
        break;
      case 4:       // set variable to ADC value
        vieval[0] = outpc.adcval[rinstr.VI1.uv.ipno];
        break;
      case 6:       // set variable to ms timer value
        vieval[0] = gpio_ms;
        break;
      case 7:       // set variable to sec timer value
        vieval[0] = outpc.seconds;
        break;
      case 8:       // set variable to IC period (us)
        vieval[0] = outpc.ic_period[rinstr.VI1.uv.ipno];
        break;
      case 9:       // set variable to IC duty (%)
        vieval[0] = outpc.ic_duty[rinstr.VI1.uv.ipno];
        break;
      default:
        vieval[0] = 0;
        break;
    }    // end VI1 input variable eval switch
    skipvi2 = 0;   // tells whether 2nd input used or not
    switch(rinstr.VI2.type)  {
      case 0:       // set variable to outpc var from MS
        if(rinstr.VI2.uv.tbv.sze == 1)   // if need long can put in later
          vieval[1] = *(msvar + rinstr.VI2.uv.tbv.offset);
        else
          vieval[1] = *( int *)(msvar + rinstr.VI2.uv.tbv.offset);
        break;
      case 1:       // set variable from gen purpose array
        vieval[1] = Var_array[rinstr.VI2.uv.varix];
        break;
      case 2:       // set variable to input variable
        vieval[1] = rinstr.VI2.uv.varval;
        break;
      case 3:       // set variable to I/O pin
        vieval[1] = outpc.pinval[rinstr.VI2.uv.ipno];
        break;
      case 4:       // set variable to ADC value
        vieval[1] = outpc.adcval[rinstr.VI2.uv.ipno];
        break;
      case 5:       // variable not used
        skipvi2 = 1;
        vieval[1] = 0;
        break;
      case 6:       // set variable to ms timer value
        vieval[1] = gpio_ms;
        break;
      case 7:       // set variable to sec timer value
        vieval[1] = outpc.seconds;
        break;
      case 8:       // set variable to IC period (us)
        vieval[1] = outpc.ic_period[rinstr.VI2.uv.ipno];
        break;
      case 9:       // set variable to IC duty (%)
        vieval[1] = outpc.ic_duty[rinstr.VI2.uv.ipno];
        break;
      default:
        vieval[1] = 0;
        break;
    }    // end VI2 input variable eval switch

    // evaluate the instruction
    trueflg = 0;   // tells whether logic cond T or F
    cond = 0;      // tells if op is logic condition(=1)
    switch(rinstr.op)  {
      // evaluate logic condition
      case '>':
        if(skipvi2)  {
          if(vieval[0])
            trueflg = 1;
        } 
        else  {          
          // here bw serves as hysteresis
          if(vieval[0] > (vieval[1] + rinstr.bw))
            trueflg = 1;
        }
        cond = 1;
        break;
      case '<':
        if(skipvi2)  {
          if(vieval[0])
            trueflg = 1;
        } 
        else  {          
          // here bw serves as hysteresis
          if(vieval[0] < (vieval[1] + rinstr.bw))
            trueflg = 1;
        }
        cond = 1;
        break;
      case '=':
        if(skipvi2)  {
          if(vieval[0])
            trueflg = 1;
        } 
        else  {          
          // here bw serves as tolerance bandwidth
          if((vieval[0] >= (vieval[1] - rinstr.bw)) &&
             (vieval[0] <= (vieval[1] + rinstr.bw)))
            trueflg = 1;
        }
        cond = 1;
        break;
      case '!':    // not equal
        if(skipvi2)  {
          if(vieval[0])
            trueflg = 1;
        } 
        else  {          
          // here bw serves as tolerance bandwidth
          if((vieval[0] < (vieval[1] - rinstr.bw)) ||
             (vieval[0] > (vieval[1] + rinstr.bw)))
            trueflg = 1;
        }
        cond = 1;
        break;
      case ' ':    // evaluate VI1 only
        trueflg = (char)vieval[0];
        cond = 1;
        break;
        
      // evaluate the arithmetic operation  
      case '+':
        vo1eval = vieval[0] * rinstr.bw;
        if(!skipvi2)
          vo1eval += vieval[1];
        break;
      case '-':
        vo1eval = vieval[0] * rinstr.bw;
        if(!skipvi2)
          vo1eval -= vieval[1];
        break;
      case 'x':
        vo1eval = vieval[0];
        if(!skipvi2)
          vo1eval *= vieval[1];
        vo1eval += rinstr.bw;
        break;
      case '/':
        vo1eval = vieval[0];
        if((!skipvi2) && (vieval[1] != 0))
          vo1eval /= vieval[1];
        vo1eval += rinstr.bw;
        break;
      default:
        vo1eval = 0;
        break;
    }    // end instruction eval switch
    
    // finish calculating outputs
    if(cond)  {    // logic condition operator
      if(trueflg)  {
        vo1eval = rinstr.VO1.voval;
        voskip = 2;    // if true, exec VO1 and skip VO2
      }
      else  {
        vo2eval = rinstr.VO2.voval;
        voskip = 1;    // if false, exec VO2 and skip VO1
      }
    }              
    else  {     // arithmetic operator - only 1 output value
                // but can set VO2 = VO1 if desired, or not
      vo2eval = vo1eval;
      voskip = 0;
    }
    
    // check for exec/ skip instructions
    nskip = 0;
    if(rinstr.VO1.type == 20)  {
      if(trueflg)  {
        voskip = 0;
        goto VO2_EXEC;           // exec VO2 and keep going
      }
      else  {
        nskip = rinstr.VO1.voval;   // skip VO2 + next voval instructions
        goto EXEC_DONE;
      }
    }
    // execute instruction by setting VOn outputs
    if(voskip == 1)
      goto VO2_EXEC;
    switch(rinstr.VO1.type)  {
      case 0:       // set Var_array[voix]
        Var_array[rinstr.VO1.voix] = vo1eval;
        break;
      case 1:       // set io pin[voix]
        jx = rinstr.VO1.voix;
        if(vo1eval)      // set high
          *pinptr[jx] = *pinptr[jx] | pinmask[jx];
        else             // set low
          *pinptr[jx] = *pinptr[jx] & ~pinmask[jx];
        break;
      case 2:       // set pwm_duty_cycle[voix]
        jx = rinstr.VO1.voix;
        *pwmdty[jx] = (unsigned char)((vo1eval * (*pwmper[jx])) / 100);
        break;
      case 4:       // adjust fuel (adds this % to VE)
        outpc.FuelAdj = vo1eval;
        sendCANAdj = 1;
        break;
      case 5:       // adjust spk (adds this degx10 to spk advance)
        outpc.SpkAdj = vo1eval;
        sendCANAdj = 1;
        break;
      case 6:       // adjust fuel (adds this no. steps to IAC).
        outpc.IdleAdj = vo1eval;    // a larger no. makes idle faster
        sendCANAdj = 1;
        break;
      case 7:       // spare adjustment
        outpc.SprAdj = vo1eval;
        sendCANAdj = 1;
        break;
      default:
        break;
    }    // end VO1 instruction exec switch
    
VO2_EXEC:
    if(voskip == 2)
      goto EXEC_DONE;
    switch(rinstr.VO2.type)  {
      case 0:       // set Var_array[voix]
        Var_array[rinstr.VO2.voix] = vo2eval;
        break;
      case 1:       // set io pin[voix]
        jx = rinstr.VO2.voix;
        if(vo2eval)      // set high
          *pinptr[jx] = *pinptr[jx] | pinmask[jx];
        else             // set low
          *pinptr[jx] = *pinptr[jx] & ~pinmask[jx];
        break;
      case 2:       // set pwm_duty_cycle[voix]
        jx = rinstr.VO2.voix;
        *pwmdty[jx] = (unsigned char)((vo2eval * (*pwmper[jx])) / 100);
        break;
      case 3:       // do nothing
        break;
      case 4:       // adjust fuel (adds this % to VE)
        outpc.FuelAdj = vo2eval;
        sendCANAdj = 1;
        break;
      case 5:       // adjust spk (adds this degx10 to spk advance)
        outpc.SpkAdj = vo2eval;
        sendCANAdj = 1;
        break;
      case 6:       // adjust fuel (adds this no. steps to IAC).
        outpc.IdleAdj = vo2eval;    // a larger no. makes idle faster
        sendCANAdj = 1;
        break;
      case 7:       // spare adjustment
        outpc.SprAdj = vo2eval;
        sendCANAdj = 1;
        break;
      default:
        break;
    }    // end VO2 instruction exec switch
EXEC_DONE:
    return(nskip);  // no. of subsequent instructions to skip
}

void reset(void)  {
short ix;
  no_tmgrp = 0;
  for(ix = inpram.n_init_instr;ix < inpram.n_instr; ix++)  {
    move_instr(ix);
    if(rinstr.VI1.type == 90)  {
      tmgrp[no_tmgrp] = 0;
      ixl_tmgrp[no_tmgrp] = ix - 1;  // index of instr just before this time blk starts
      no_tmgrp++;
    }
  }
  outpc.FuelAdj = 0;
  outpc.SpkAdj = 0;
  outpc.IdleAdj = 0;
  outpc.SprAdj = 0;
  FuelAdjl = 0;
  SpkAdjl = 0;
  IdleAdjl = 0;
  SprAdjl = 0;
  sendCANAdj = 0;
  first_instr = 1;
  gpio_ms = 0;
  inpram.gpio_halt = 0;
  return;
}

void get_adc(char chan1, char chan2)  {
char chan;
short adcvale,tmpx1,tmpx2;
  
  for (chan = chan1; chan <= chan2; chan++)  {
    if(adcraw[chan] == 0)  {
      outpc.adcval[chan] = 0;   // not used for adc
    }
    else  {
      // get value in engineering units:
      //  adc_engring = adclo + (adchi - adclo) * adcraw / 1023
      tmpx1 = inpram.adchi[chan] - inpram.adclo[chan];
      tmpx2 = *adcraw[chan];
      asm {
            ldy tmpx1 
            ldd tmpx2
            EMULS
            ldx  #1023
            EDIVS
            sty adcvale
      };
      adcvale += inpram.adclo[chan];  
      if(first_adc)
            outpc.adcval[chan] = adcvale;
      else    // lag filter
            outpc.adcval[chan] += ((inpram.adcLF[chan] * 
                            (adcvale - outpc.adcval[chan])) / 100);
    }
  }				 // end of for loop
  
  return;
}

#pragma CODE_SEG  ROM_7400

void move_instr(short ix)  {
short jx;
char save_page;
  save_page = PPAGE;
  PPAGE = 0x3D;
  for (jx = 0;jx < sizeof(rinstr)/2; jx++)  {
    // do word transfers for speed, but make sure sizeof(rinstr) stays even
    *((unsigned short *)&rinstr + jx) = *((unsigned short *)&finstr[ix] + jx);
  }
  PPAGE = save_page;
  return;
}

int intrp_1dctable(char sgnx, int x, unsigned char n, int * x_table, char sgnz, 
        unsigned char * z_table)  {
  int ix;
  long interp, interp3;
  // bound input arguments
  if((sgnx && (x > x_table[n-1])) ||
     (!sgnx && ((unsigned int)x > (unsigned int)x_table[n-1])))  {
    if(!sgnz)
      return((int)z_table[n -1]);
    else
      return((int)((char)z_table[n -1]));
  }
  if((sgnx && (x < x_table[0])) ||
     (!sgnx && ((unsigned int)x < (unsigned int)x_table[0])))  {
    if(!sgnz)
      return((int)z_table[0]);
    else
      return((int)((char)z_table[0]));
  }
  for(ix = n - 2; ix > -1; ix--)  { 
    if((sgnx && (x > x_table[ix])) ||
       (!sgnx && ((unsigned int)x > (unsigned int)x_table[ix])))  {
   		break;
  	}
  }
  if(ix < 0)ix = 0;
 
  if(sgnx)
    interp =	x_table[ix + 1] - x_table[ix];
  else
    interp =	(unsigned int)x_table[ix + 1] - (unsigned int)x_table[ix];
  if(interp != 0)  {
    if(sgnx)
      interp3 = x - x_table[ix];
    else
      interp3 = (unsigned int)x - (unsigned int)x_table[ix];
    interp3 = (100 * interp3);
    interp = interp3 / interp;
  }
  if(!sgnz)
    return((int)(z_table[ix] +
	    interp * (z_table[ix+1] - z_table[ix])/ 100));
	else
    return((int)((char)z_table[ix] +
	    interp * ((char)z_table[ix+1] - (char)z_table[ix])/ 100));
}

int intrp_2dctable(unsigned int x, int y, unsigned char nx, unsigned char ny,
  unsigned int * x_table, int * y_table, unsigned char * z_table)  {
  int ix,jx;
  long interp1, interp2, interp3;
  // bound input arguments
  if(x > x_table[nx-1])x = x_table[nx-1];
  else if(x < x_table[0])x = x_table[0];
  if(y > y_table[ny-1])y = y_table[ny-1];
  else if(y < y_table[0])y = y_table[0];
  // Find bounding indices in table
  for(ix = ny - 2; ix > -1; ix--)  {  // Start w highest index
	//  because will generally have least time for calculations at hi y
  	if(y > y_table[ix])  {
   		break;
  	}
  }
  if(ix < 0)ix = 0;
  for(jx = nx - 2; jx > -1; jx--)  {  // Start w highest index
	// because will generally have least time for calculations at hi x
	  if(x > x_table[jx])  {
	    break;
    }
  }
  if(jx < 0)jx = 0;
  // do 2D interpolate
  interp1 = y_table[ix + 1] - y_table[ix];
  if(interp1 != 0)  {
    interp3 = (y - y_table[ix]); 
    interp3 = (100 * interp3); 
    interp1 = interp3 / interp1; 
  }
  interp2 =	x_table[jx + 1] - x_table[jx];
  if(interp2 != 0)  {
    interp3 = (x - x_table[jx]); 
    interp3 = (100 * interp3); 
    interp2 = interp3 / interp2; 
  }
  return((int)(((100 - interp1) * (100 - interp2) * z_table[ix*nx+jx]
	  + interp1 * (100 - interp2) * z_table[(ix+1)*nx+jx]
	  + interp2 * (100 - interp1) * z_table[ix*nx+jx+1]
	  + interp1 * interp2 * z_table[(ix+1)*nx+jx+1]) / 10000));
}

#pragma CODE_SEG DEFAULT

#pragma CODE_SEG  NON_BANKED
interrupt void ISR_Timer_Clock(void)  {
  
  // .128 ms clock interrupt - clear flag immediately to resume count
  CRGFLG = 0x80;  // clear RTI interrupt flag
  // also generate 1.024 ms, .10035 sec and 1.0035 sec clocks
  lmms++;   // free running clock(.128 ms tics) good for ~ 110 hrs
  can_mmsclk++;
  mms++;    // in .128 ms tics - reset every 8 tics = 1.024 ms
      
  // check mms to generate other clocks
  if(mms < 8)goto CLK_DONE;
  mms = 0;
  millisec++;     // actually 1.024 ms, for seconds count
  gpio_ms++;       // gpio 1 ms timer
  
  if(millisec > 976) {
    millisec = 0;
    // update seconds to send back to PC
    outpc.seconds++;
  }
   
CLK_DONE:
  return;
}

interrupt void TimerIC0Isr(void)  {
  static unsigned short tcsav=0;
  unsigned short utmp;
  
  TFLG1 = 0x01;
  utmp = ((TC0 - tcsav) / 3) << 1;    // high time, us
  // check pin value
  if(PTIT & 0x01)  {   // start at rising edge
    //  calculate last period, us
    outpc.ic_period[0] = utmp;        // us
    tcsav = TC0;       // save time of new rising edge
  }
  else  {              // ic_pinval = 0, falling edge
    // calculate duty cycle,  %
    if(outpc.ic_period[0] != 0)  {
      asm {
        ldy utmp 
        ldd #100
        EMUL
        ldx  outpc.ic_period[0]
        EDIV
        sty outpc.ic_duty[0]
      };
    } 
    else
      outpc.ic_duty[0] = 0;
  }
  return;
}
interrupt void TimerIC1Isr(void)  {
  static unsigned short tcsav=0;
  unsigned short utmp;
  
  TFLG1 = 0x02;
  utmp = ((TC1 - tcsav) / 3) << 1;    // high time, us
  // check pin value
  if(PTIT & 0x02)  {   // start at rising edge
    //  calculate last period, us
    outpc.ic_period[1] = utmp;        // us
    tcsav = TC1;       // save time of new rising edge
  }
  else  {              // ic_pinval = 0, falling edge
    // calculate duty cycle,  %
    if(outpc.ic_period[1] != 0)  {
      asm {
        ldy utmp 
        ldd #100
        EMUL
        ldx  outpc.ic_period[1]
        EDIV
        sty outpc.ic_duty[1]
      };
    } 
    else
      outpc.ic_duty[1] = 0;
  }
  return;
}
interrupt void TimerIC2Isr(void)  {
  static unsigned short tcsav=0;
  unsigned short utmp;
  
  TFLG1 = 0x04;
  utmp = ((TC2 - tcsav) / 3) << 1;    // high time, us
  // check pin value
  if(PTIT & 0x04)  {   // start at rising edge
    //  calculate last period, us
    outpc.ic_period[2] = utmp;        // us
    tcsav = TC2;       // save time of new rising edge
  }
  else  {              // ic_pinval = 0, falling edge
    // calculate duty cycle,  %
    if(outpc.ic_period[2] != 0)  {
      asm {
        ldy utmp 
        ldd #100
        EMUL
        ldx  outpc.ic_period[2]
        EDIV
        sty outpc.ic_duty[2]
      };
    } 
    else
      outpc.ic_duty[2] = 0;
  }
  return;
}
interrupt void TimerIC3Isr(void)  {
  static unsigned short tcsav=0;
  unsigned short utmp;
  
  TFLG1 = 0x08;
  utmp = ((TC3 - tcsav) / 3) << 1;    // high time, us
  // check pin value
  if(PTIT & 0x08)  {   // start at rising edge
    //  calculate last period, us
    outpc.ic_period[3] = utmp;        // us
    tcsav = TC3;       // save time of new rising edge
  }
  else  {              // ic_pinval = 0, falling edge
    // calculate duty cycle,  %
    if(outpc.ic_period[3] != 0)  {
      asm {
        ldy utmp 
        ldd #100
        EMUL
        ldx  outpc.ic_period[3]
        EDIV
        sty outpc.ic_duty[3]
      };
    } 
    else
      outpc.ic_duty[3] = 0;
  }
  return;
}
interrupt void TimerIC4Isr(void)  {
  static unsigned short tcsav=0;
  unsigned short utmp;
  
  TFLG1 = 0x10;
  utmp = ((TC4 - tcsav) / 3) << 1;    // high time, us
  // check pin value
  if(PTIT & 0x10)  {   // start at rising edge
    //  calculate last period, us
    outpc.ic_period[4] = utmp;        // us
    tcsav = TC4;       // save time of new rising edge
  }
  else  {              // ic_pinval = 0, falling edge
    // calculate duty cycle,  %
    if(outpc.ic_period[4] != 0)  {
      asm {
        ldy utmp 
        ldd #100
        EMUL
        ldx  outpc.ic_period[4]
        EDIV
        sty outpc.ic_duty[4]
      };
    } 
    else
      outpc.ic_duty[4] = 0;
  }
  return;
}
interrupt void TimerIC5Isr(void)  {
  static unsigned short tcsav=0;
  unsigned short utmp;
  
  TFLG1 = 0x20;
  utmp = ((TC5 - tcsav) / 3) << 1;    // high time, us
  // check pin value
  if(PTIT & 0x20)  {   // start at rising edge
    //  calculate last period, us
    outpc.ic_period[5] = utmp;        // us
    tcsav = TC5;       // save time of new rising edge
  }
  else  {              // ic_pinval = 0, falling edge
    // calculate duty cycle,  %
    if(outpc.ic_period[5] != 0)  {
      asm {
        ldy utmp 
        ldd #100
        EMUL
        ldx  outpc.ic_period[5]
        EDIV
        sty outpc.ic_duty[5]
      };
    } 
    else
      outpc.ic_duty[5] = 0;
  }
  return;
}
interrupt void TimerIC6Isr(void)  {
  static unsigned short tcsav=0;
  unsigned short utmp;
  
  TFLG1 = 0x40;
  utmp = ((TC6 - tcsav) / 3) << 1;    // high time, us
  // check pin value
  if(PTIT & 0x40)  {   // start at rising edge
    //  calculate last period, us
    outpc.ic_period[6] = utmp;        // us
    tcsav = TC6;       // save time of new rising edge
  }
  else  {              // ic_pinval = 0, falling edge
    // calculate duty cycle,  %
    if(outpc.ic_period[6] != 0)  {
      asm {
        ldy utmp 
        ldd #100
        EMUL
        ldx  outpc.ic_period[6]
        EDIV
        sty outpc.ic_duty[6]
      };
    } 
    else
      outpc.ic_duty[6] = 0;
  }
  return;
}
interrupt void TimerIC7Isr(void)  {
  static unsigned short tcsav=0;
  unsigned short utmp;
  
  TFLG1 = 0x80;
  utmp = ((TC7 - tcsav) / 3) << 1;    // high time, us
  // check pin value
  if(PTIT & 0x80)  {   // start at rising edge
    //  calculate last period, us
    outpc.ic_period[7] = utmp;        // us
    tcsav = TC7;       // save time of new rising edge
  }
  else  {              // ic_pinval = 0, falling edge
    // calculate duty cycle,  %
    if(outpc.ic_period[7] != 0)  {
      asm {
        ldy utmp 
        ldd #100
        EMUL
        ldx  outpc.ic_period[7]
        EDIV
        sty outpc.ic_duty[7]
      };
    } 
    else
      outpc.ic_duty[7] = 0;
  }
  return;
}

#pragma CODE_SEG  NON_BANKED
#include "..\..\serCANComm\serComm.h"
#include "..\..\serCANComm\CANComm.h"

#pragma CODE_SEG DEFAULT        
int intrp_2ditable(unsigned int x, int y, unsigned char nx, unsigned char ny,
  unsigned int * x_table, int * y_table, int * z_table)  {
  int ix,jx;
  long interp1, interp2, interp3;
  // bound input arguments
  if(x > x_table[nx-1])x = x_table[nx-1];
  else if(x < x_table[0])x = x_table[0];
  if(y > y_table[ny-1])y = y_table[ny-1];
  else if(y < y_table[0])y = y_table[0];
  // Find bounding indices in table
  for(ix = ny - 2; ix > -1; ix--)  {  // Start w highest index
	//  because will generally have least time for calculations at hi y
  	if(y > y_table[ix])  {
   		break;
  	}
  }
  if(ix < 0)ix = 0;
  for(jx = nx - 2; jx > -1; jx--)  {  // Start w highest index
	// because will generally have least time for calculations at hi x
	  if(x > x_table[jx])  {
	    break;
    }
  }
  if(jx < 0)jx = 0;
  // do 2D interpolate
  interp1 = y_table[ix + 1] - y_table[ix];
  if(interp1 != 0)  {
    interp3 = (y - y_table[ix]); 
    interp3 = (100 * interp3); 
    interp1 = interp3 / interp1; 
  }
  interp2 =	x_table[jx + 1] - x_table[jx];
  if(interp2 != 0)  {
    interp3 = (x - x_table[jx]); 
    interp3 = (100 * interp3); 
    interp2 = interp3 / interp2; 
  }
  return((int)(((100 - interp1) * (100 - interp2) * z_table[ix*nx+jx]
	  + interp1 * (100 - interp2) * z_table[(ix+1)*nx+jx]
	  + interp2 * (100 - interp1) * z_table[ix*nx+jx+1]
	  + interp1 * interp2 * z_table[(ix+1)*nx+jx+1]) / 10000));
}

#pragma CODE_SEG  NON_BANKED
interrupt void UnimplementedISR(void) {
   /* Unimplemented ISRs trap.*/
   return;
}

#pragma CODE_SEG DEFAULT        

/******************************************************************************
Arguments:	

      progAdr		Pointer to the start of the destination 
			  Flash location to be programmed
      bufferPtr	Pointer to the start of the source data 
      size      Number of WORDS to be programmed
                  
    This function will program non-paged flash only 
******************************************************************************/
void fburner(unsigned int* progAdr, unsigned int* bufferPtr, 
                                          unsigned int no_words)
{
unsigned char sect,no_sectors;
																
  	if(flocker != 0xCC)return; // if not set, we got here unintentionally
  	// sector = 1024 bytes (512 words)
  	if(burn_flag == 1)  {
  	  no_sectors = (unsigned char)(no_words / SECTOR_WORDS);
  	  if(no_words > (no_sectors * SECTOR_WORDS))no_sectors++;
  	  for (sect = 0; sect < no_sectors; sect++)  {  	  
  	    Flash_Erase_Sector(progAdr + sect * SECTOR_WORDS);
  	  }
  	}
  	Flash_Write_Word(progAdr + burn_flag-1, *(bufferPtr + burn_flag-1));
    	
	return;
}

//*****************************************************************************
//* Function Name: Flash_Init
//* Description : Initialize Flash NVM for HCS12 by programming
//* FCLKDIV based on passed oscillator frequency, then
//* uprotect the array, and finally ensure PVIOL and
//* ACCERR are cleared by writing to them.
//*
//*****************************************************************************
void Flash_Init(unsigned long oscclk)  {
unsigned char fclk_val;
unsigned long temp;
  /* Next, initialize FCLKDIV register to ensure we can program/erase */
  temp = oscclk;
  if (oscclk >= 12000) {
    fclk_val = oscclk/8/200 - 1; /* FDIV8 set since above 12MHz clock */
    FCLKDIV = FCLKDIV | fclk_val | FDIV8;
  }
  else
  {
    fclk_val = oscclk/8/200 - 1;
    FCLKDIV = FCLKDIV | fclk_val;
  }
  FPROT = 0xFF; /* Disable all protection (only in special modes)*/
  FSTAT = FSTAT | (PVIOL|ACCERR);/* Clear any errors */
  return;
}
void Flash_Erase_Sector(unsigned int *address)  {
  FSTAT = (ACCERR | PVIOL); // clear errors
  (*address) = 0xFFFF;/* Dummy store to page to be erased */
  FCMD = ERASE;
/*********************************************************************
;* Allow final steps in a flash prog/erase command to execute out
;* of RAM while flash is out of the memory map
;* This routine can be used for flash word-program or erase commands
;*
;********************************************************************/
/* Execute the sub out of ram  */
   asm {
     ldaa  #CBEIF
     jsr  RamBurnPgm
   };
   return;
}

void Flash_Write_Word(unsigned int *address, unsigned int data)  {
  FSTAT = (ACCERR | PVIOL); // clear errors
  (*address) = data; // Store desired data to address being programmed
  FCMD = PROG; // Store programming command in FCMD
/* Execute the sub out of ram  */
   asm {
     ldaa  #CBEIF
     jsr  RamBurnPgm
   };
  return;
}