uint8_t CanInit(void)
{
uint8_t ix;
	/* Set up CAN communications */

	/* Configure CAN */
	CAN_A.MCR.R = 0x50000000 + 63;  // Freeze, Halt all buffs
#ifdef ECU_PPC5553
	CAN_A.CR.R = 0x0EDB2017;   // 132MHz/ (PRESDIV + 1) x No Tq = 
	        // 517,647 Hz bit rate with PRESDIV = 14, No Tq = 17
	        // RJW=3, PSEG1=3, PSEG2=3, CLK_SRC=1(sysclk), PROPSEG=7
	        // LBUF=1 (lowest no. buffer xmitted 1st)
#endif
#ifdef ECU_PPC5634M
	CAN_A.CR.R = 0x08DB2017;   // 80MHz/ ((PRESDIV + 1) x No Tq) = 
	        // 522,876 Hz bit rate with PRESDIV = 8, No Tq = 17
	        // RJW=3, PSEG1=3, PSEG2=3, CLK_SRC=1(sysclk), PROPSEG=7
	        // LBUF=1 (lowest no. buffer xmitted 1st)
#endif
	for(ix = 0;ix < 64;ix++)
	  CAN_A.BUF[ix].CS.B.CODE = 0;  // inactivate all msg buffers
	CAN_A.MCR.R |= (NO_XMT_BUF + NO_RCV_BUF - 1); // set max buffs to be used

	/* CAN message ID format:
	 Reg Bits: 
	      0 <-------------------- 7
	      |xxx-var_off(11 bits)---|  (xxx,M28 <-- M24)
	      |-----cont'd-----|-mtyp-|  (M23 <-- M16)
	      |-|--From ID--|--To ID--|  (M15 <-- M8)
	      |-|---var_blk(5)--|-spr-|  (M7 <-- M0)
	      
  |                |                |                |                |
   xxxxxx2827262524 2322212019181716 1514131211100908 0706050403020100
	     |------varoff----------|-mtyp-|src id-|-dst id-|-var blk-|-spr      
	*/  
	/* Set identifier mask registers to accept messages 
	     only for can_id. At present no way to send 
	     an all devices msg to device #15 */
	     
	can_status = 0;
	getCANdat = 0;
	RFlag = 0;
	ltch_CAN = 0xFFFFFFFF;
	
	// set up message buffers - only use first NO_XMT_BUF + NO_RCV_BUF
	for(ix = 0;ix < NO_XMT_BUF + NO_RCV_BUF;ix++)  {
	  CAN_A.BUF[ix].CS.B.SRR = 1;   // need for xmit in xtnded fmt
	  CAN_A.BUF[ix].CS.B.IDE = 1;   // extended length msg
	  CAN_A.BUF[ix].CS.B.RTR = 0;   // always data frame xmits
	  CAN_A.BUF[ix].CS.B.LENGTH = 0;
	  if(ix < NO_XMT_BUF)                // 1st NO_XMT_BUF buffers are xmit
	    CAN_A.BUF[ix].CS.B.CODE = 0x08;  // inactive xmit buffer 
	  else  {                       // next NO_RCV_BUF buffers are rcv
	  	// set my can_id into id field
	  	CAN_A.BUF[ix].ID.MSMSG.SRC = can_id;
	    CAN_A.BUF[ix].CS.B.CODE = 0x04;  // active, empty receive buffer
	  }
	}

	CAN_A.RXGMASK.R = 0x00000780;   // anything ok as long as msg dest
	CAN_A.RX14MASK.R = 0x00000780;  //  matches can_id
	CAN_A.RX15MASK.R = 0x00000780;
	
	// set interrupt masks and clear flags
	for(ix = 0;ix < NO_XMT_BUF + NO_RCV_BUF;ix++)  {
	  if(ix < NO_XMT_BUF)  {
	    if(ix < 32)	   // disable ints for xmt buffers
	      CAN_A.IMRL.R &= ~can_intmask[ix];
	    else
	      CAN_A.IMRH.R &= ~can_intmask[ix-32];
	  }
	  else  {
	    if(ix < 32)	   // enable ints for rcv buffers
	      CAN_A.IMRL.R |= can_intmask[ix];
	    else
	      CAN_A.IMRH.R |= can_intmask[ix-32];
	  }
	}
	SIU.PCR[83].R = 0x062C; // configure pad for CNTxDA function, open drain
	SIU.PCR[84].R = 0x0500; // configure pad for CNRxDA function
	
	CAN_A.MCR.R = 0x00000000 + (NO_XMT_BUF + NO_RCV_BUF - 1);  
	                          // clr DIS,FRZ, Halt states for used buffers
	// set up interrupts
	if(NO_XMT_BUF + NO_RCV_BUF <= 16)  {
	  for(ix = 0; ix < NO_XMT_BUF + NO_RCV_BUF; ix++)  {
	    INTC.PSR[155 + ix].R = 4;
	  }
	}
	else  {
	  for(ix = 0; ix < 17; ix++)  {
	    INTC.PSR[155 + ix].R = 4;
	  }
	  if(NO_XMT_BUF + NO_RCV_BUF > 32)
	    INTC.PSR[172].R = 4;
	}
	CAN_A.IFRL.R = 0xFFFFFFFF;   // clear flags for all buffers
	CAN_A.IFRH.R = 0xFFFFFFFF;
	return(0);
}

__declspec(interrupt nowarn SRR)
     void CanTxISR(void)  
{
uint8_t ibuf,ix,kx,msg_no,set_int,var_byt;
uint16_t destvaroff;
uint32_t ultmp;

  // we only get here if an MSG_RSP or OUTMSG_RSP msg was just xmitted 
  //  and a continuation is pending in Rbytes or canoutmsg. Find a 
  //  ready buffer, load it, check if another and if so xfer it to 
  //  msg buffer or canoutmsg
  
	// Find next ready xmit buffer
	ix = 0;
	for(ibuf = 0;ibuf < NO_XMT_BUF;ibuf++)  {
	  if(CAN_A.BUF[ibuf].CS.B.CODE == 0x08)  {
	    if(NO_XMT_BUF < 32)
	      ultmp = CAN_A.IFRL.R;
	    else
	      ultmp = CAN_A.IFRH.R;
	    if(!(ultmp & can_intmask[ibuf]))continue;
	    ix = 1;
	    break;	
	  } 
	}
	if( !ix )goto CANTX_DONE;  
	                    // no inactive xmt buffs - lose msg
	CANtx++;
	
	if(ibuf < 32)	   // clear int flag (wrt 1) for buffer ibuf
	  CAN_A.IFRL.R |= can_intmask[ibuf];
	else
	  CAN_A.IFRH.R |= can_intmask[ibuf-32];

	set_int = 0;      // *** NEVER set xmit interrupt except for  
	                  //  a continued MSG_RSP or outmsg_resp type message
	if(setCANTx_int[1])  {
	  //case OUTMSG_RSP:  
	      // ECU sending msg to gpio device in reply to gpio's request for 
		  // specific values from outpc + offset to be sent to gpio 
		  // destvarblk + destvaroff
		  msg_no = can_outbufmn[ibuf];
		  // update buffer ID, LENGTH and DATA fields
		  // first half of ID reg - just update destvaroff
		  destvaroff = (uint16_t)(outmsg[msg_no].no_bytes 
		                                 - can_outbytes[msg_no]);
		  CAN_A.BUF[ibuf].ID.MSMSG.VAROFF = destvaroff;
		  if(can_outbytes[msg_no] < 8)  {
		    CAN_A.BUF[ibuf].CS.B.LENGTH = can_outbytes[msg_no];
		    can_outbytes[msg_no] = 0;
		    setCANTx_int[1] = 0;
		  }  
		  else  {
		    CAN_A.BUF[ibuf].CS.B.LENGTH = 8;
		    can_outbytes[msg_no] -= 8;
		  }  
		  for(kx = 0;kx < CAN_A.BUF[ibuf].CS.B.LENGTH;kx++)  {
		    CAN_A.BUF[ibuf].DATA.B[kx] = *((int8_t *)&outpc +
		                  outmsg[msg_no].offset[destvaroff + kx]);
		  }
		  if(can_outbytes[msg_no])  // still more outmsg data after this
		    set_int = 1;      // need to set xmit interrupt for this type
		                      //  message so will continue sending
	}
	else if(setCANTx_int[0])  {
	  //case MSG_RSP:
		// put next set of variable value(s) in xmit buffer
		// mtype, dest var blk, offset, id same as last msg
		CAN_A.BUF[ibuf].ID.MSMSG.MTYP = MSG_RSP;
		CAN_A.BUF[ibuf].ID.MSMSG.VARBLK = CAN_A.BUF[Ribuf].ID.MSMSG.VARBLK;
		CAN_A.BUF[ibuf].ID.MSMSG.VAROFF = Rdvaroff;
		CAN_A.BUF[ibuf].ID.MSMSG.DEST = CAN_A.BUF[Ribuf].ID.MSMSG.DEST;
		if(RFlag && (CAN_A.BUF[ibuf].ID.MSMSG.DEST == 0))  {    
		                // keep responding til all requested data sent
		  Rbytes -= 8;  // Rbytes = 8 1st time thru, and if < 8 will quit
		  if(Rbytes <= 8)  {
		    // this last message - turn off RFlag
		    RFlag = 0;
		    setCANTx_int[0] = 0;
		    if(Rbytes)
		      var_byt = Rbytes;
		  } 
		  else  {
		    var_byt = 8;
		    set_int = 1; // set xmit for this so keep sending
		    Ribuf = ibuf;
		  }
		  Rvaroff += 8;  // will not get here if just sent < 8 bytes
		  Rdvaroff += 8;
		}
		CAN_A.BUF[ibuf].CS.B.LENGTH = var_byt;
		if(tables[Rvarblk].addrRam != MNULL)  {
		  for(kx = 0;kx < var_byt;kx++)  {
		    CAN_A.BUF[ibuf].DATA.B[kx] = 
                     *tableByteRam(Rvarblk, Rvaroff + kx);
		  }
		}
		else  {
		  for(kx = 0;kx < var_byt;kx++)  {
		    CAN_A.BUF[ibuf].DATA.B[kx] = 
                     *tableByteFlash(Rvarblk, Rvaroff + kx);
		  }
		}
	}
	
	if(set_int)  {
	  if(ibuf < 32)	   // enable int for buffer ibuf
	    CAN_A.IMRL.R |= can_intmask[ibuf];
	  else
	    CAN_A.IMRH.R |= can_intmask[ibuf-32];
	}
	else  {            // no MSG_RSP or OUTMSG_RSP data left
	  if(ibuf < 32)	   // disable int for buffer ibuf
	    CAN_A.IMRL.R &= ~can_intmask[ibuf];
	  else
	    CAN_A.IMRH.R &= ~can_intmask[ibuf-32];
	}
	
	CAN_A.BUF[ibuf].CS.B.CODE = 0x0C;   // this will result in xmit and
	                                    // return to inactive (CODE = 0x08)
	if(!set_int)
	  canTxBuf[ibuf] = 0;               // buffer no longer in use
	can_status &= CLR_XMT_ERR;
	if(CAN_A.ESR.R != 0)  {
	  // error in xmt
	  can_status |= XMT_ERR;
	  //can_reset = 1;
	}
CANTX_DONE:
	asm(mbar 0);   // done in sample epilogs to ensure interrupt flag cleared
	INTC.EOIR.R = 0;
	return;
}
  
void sendCanTx(canmsg *canmx)
{
uint8_t ibuf,ix,kx,mtype,msg_no,set_int;
  
	/*  set up a CAN Xmit */
	// Interrupts disabled if called from non-ISR
	// Find next ready xmit buffer
	ix = 0;
	for(ibuf = 0;ibuf < NO_XMT_BUF;ibuf++)  {
	  if(CAN_A.BUF[ibuf].CS.B.CODE == 0x08)  {
	    if(canTxBuf[ibuf])continue;    // buffer already in use
	    ix = 1;
	    break;	
	  } 
	}
	if( !ix )  {
	  INTC.CPR.B.PRI = 0;  // re-enable CAN, SCI interrupts
	  return;    // no inactive xmt buffs - lose msg
	}
	canTxBuf[ibuf] = 1;     // lock buff in case get interrupt
	INTC.CPR.B.PRI = 0;  // re-enable CAN, SCI interrupts

 	/* Set up identifier register */	  
	// first half of ID reg
	CAN_A.BUF[ibuf].ID.MSMSG.VAROFF = canmx->destvaroff;
	CAN_A.BUF[ibuf].ID.MSMSG.MTYP = canmx->msg_type;
	CAN_A.BUF[ibuf].ID.MSMSG.SRC = can_id;
	CAN_A.BUF[ibuf].ID.MSMSG.DEST = canmx->dest;
	CAN_A.BUF[ibuf].ID.MSMSG.VARBLK = canmx->destvarblk;
  
    CANtx++;

	/* set data length and data in buffer */
	mtype = canmx->msg_type;
	if(mtype == MSG_XTND)
	  mtype = canmx->datbuf[0];
	set_int = 0;      // don't want to set xmit interrupt except for 
	                  //  MSG_RSP, outmsg_resp type of message
	switch(mtype)  {
		case MSG_CMD:  // msg for dest ecu to set a variable to val in msg
		case MSG_RSP:  // msg in reply to dest ecu's request for a var val
		  CAN_A.BUF[ibuf].CS.B.LENGTH = canmx->varbyt;
		  for(kx = 0;kx < canmx->varbyt;kx++)  {
		    CAN_A.BUF[ibuf].DATA.B[kx] = canmx->datbuf[kx];
		  }
		  if(RFlag && (canmx->dest == 0))  {    
		    Ribuf = ibuf;
		    set_int = 1;          // set xmit int for this so keep sending
		    setCANTx_int[0] = 1;  // need to distinguish from outmsg_rsp
		  }
		  break;
		
		case MSG_REQ:  // msg to send back current value of variable(s)
		  // this 1st byte holds var blk for where to put rcvd data
		  // 2nd,3rd bytes hold var offset rel. to var blk and how
		  // many consecutive bytes to be sent back. If canmx->varbyt > 8,
		  // a 4th byte added for it, otherwise it is last 5 bits of 
		  // DATA.B[2].  NOTE: MSG_REQ w > 8 bytes meant only for TS.
		  // GPIOs should use outmsg for this purpose, otherwise need to
		  // have multiple RFlags = gpio ids w corresponding Rbytes, etc. 
		  CAN_A.BUF[ibuf].DATA.B[0] = (uint8_t)canmx->myvarblk;
		  CAN_A.BUF[ibuf].DATA.B[1] = (uint8_t)(canmx->myvaroff >> 3);
		  if(canmx->varbyt <= 8)  {
		    CAN_A.BUF[ibuf].CS.B.LENGTH = 3;
		    CAN_A.BUF[ibuf].DATA.B[2] = (uint8_t)(((canmx->myvaroff & 0x0007) 
		                                     << 5) | canmx->varbyt);
		  }
		  else  {
		    CAN_A.BUF[ibuf].CS.B.LENGTH = 4;
		    CAN_A.BUF[ibuf].DATA.B[2] = (uint8_t)((canmx->myvaroff & 0x0007) << 5);
		    CAN_A.BUF[ibuf].DATA.B[3] = canmx->varbyt; 
		  }
		  break;
		
		case OUTMSG_REQ:  // gpio msg to ECU to send back current values of specific 
		  // outpc variables. The buffer ID reg holds var blk, varoff for where in gpio 
		  //  to put the data to be sent back to gpio
		  CAN_A.BUF[ibuf].CS.B.LENGTH = 1;
		  CAN_A.BUF[ibuf].DATA.B[0] = canmx->outmsg_no;
		  break;

		case OUTMSG_RSP:  // ECU sending msg to gpio device in reply to gpio's request for 
		  // specific values from outpc + offset to be sent to gpio destvarblk + destvaroff
		  msg_no = canmx->outmsg_no;
		  can_outbufmn[ibuf] = msg_no;   // keep this ibuf until all bytes sent
		  if(can_outbytes[msg_no] < 8)  {
		    CAN_A.BUF[ibuf].CS.B.LENGTH = can_outbytes[msg_no];
		    can_outbytes[msg_no] = 0;
		  }  
		  else  {
		    CAN_A.BUF[ibuf].CS.B.LENGTH = 8;
		    can_outbytes[msg_no] -= 8;
		  }  
		  for(kx = 0;kx < CAN_A.BUF[ibuf].CS.B.LENGTH;kx++)  {
		    CAN_A.BUF[ibuf].DATA.B[kx] = canmx->datbuf[kx];
		  }
		  // check if have more bytes to send after this
		  if(can_outbytes[msg_no])  {  
		    set_int = 1;      // need to set xmit interrupt for this type
		    setCANTx_int[1] = 1;  //  message so will continue sending    
		  }                   
		  break;
		  
		case MSG_XSUB:
		  CAN_A.BUF[ibuf].CS.B.LENGTH = 0;
		  break;
		  
		case MSG_BURN:
		  CAN_A.BUF[ibuf].CS.B.LENGTH = 0;
		  break;
	}     // end mtype switch
	
	if(set_int)  {
	  if(ibuf < 32)	   // enable interrupt for buffer ibuf
	    CAN_A.IMRL.R |= can_intmask[ibuf];
	  else
	    CAN_A.IMRH.R |= can_intmask[ibuf-32];
	}
	else  {
	  if(ibuf < 32)	   // disable int for buffer ibuf
	    CAN_A.IMRL.R &= ~can_intmask[ibuf];
	  else
	    CAN_A.IMRH.R &= ~can_intmask[ibuf-32];
	}
	
	CAN_A.BUF[ibuf].CS.B.CODE = 0x0C;   // this will result in xmit and
	                                    // return to inactive (CODE = 0x08)
	if(!set_int)
	  canTxBuf[ibuf] = 0;               // buffer no longer in use
	can_status &= CLR_XMT_ERR;
	if(CAN_A.ESR.R != 0)  {
	  // error in xmt
	  can_status |= XMT_ERR;
	  //can_reset = 1;
	}
	
	return;
}

__declspec(interrupt nowarn SRR)
     void CanRxISR(void)
{
uint8_t rcv_id,mtype,var_blk,var_byt,ibuf,ix,jx,kx;
uint16_t var_off,ntword;
uint32_t ultmp;
static uint8_t tble_page[256];
uint8_t msg_no;
static uint8_t tbleCnt;
static uint16_t canrxbytes;
canmsg canmx;

    CANrx++;

    /* CAN Recv Interrupt */
    // Find next ready rcv buffer
    ix = 0;
    for(ibuf = NO_XMT_BUF;ibuf < NO_XMT_BUF + NO_RCV_BUF;ibuf++)  {
      if(CAN_A.BUF[ibuf].CS.B.CODE == 0x02)  {  // buffer full
        ix = 1;
        break;	
      } 
    }
    if(!ix)goto CANRX_DONE;
    /* clear flag to be ready for next rcv int */
	if(ibuf < 32)	   // clear int flag (wrt 1) for buffer ibuf
	  CAN_A.IFRL.R |= can_intmask[ibuf];
	else
	  CAN_A.IFRH.R |= can_intmask[ibuf-32];
    
    /* Decode identifier register */    
    var_off = (uint16_t)CAN_A.BUF[ibuf].ID.MSMSG.VAROFF;
    mtype  = (uint8_t)CAN_A.BUF[ibuf].ID.MSMSG.MTYP;
    if(mtype == MSG_XTND)
          mtype = CAN_A.BUF[ibuf].DATA.B[0];
    rcv_id = (uint8_t)CAN_A.BUF[ibuf].ID.MSMSG.SRC;  // msg from device rcv_id
    var_blk = (uint8_t)CAN_A.BUF[ibuf].ID.MSMSG.VARBLK;
    
    var_byt = (uint8_t)(CAN_A.BUF[ibuf].CS.B.LENGTH & 0x0F);

    switch(mtype)  {
          case MSG_CMD:  // msg for this ecu to set a variable to val in msg
          case MSG_RSP:  // msg in reply to this ecu's request for a var val
        	               // value in the data buffer in recvd msg
            if(getCANdat && (getCANdat == rcv_id))  {
              // set up for serial xmit of the recvd CAN bytes.
              //   TS getting back data requested from gpio board(pass-thru)
              for(jx = 0;jx < var_byt;jx++)  {
                txbuf[jx + ctxboff] = CAN_A.BUF[ibuf].DATA.B[jx];
              }
              if(ctxboff == 0)  {
                // set up to start xmitting data back to TS; this process
                // will be self-sustaining because CAN data arrives faster 
                // than serial data goes out; *** sizeof(txbuf) is max size  
                // of request. *** 
                txcnt = 0;
                txgoal = rxnbytes;
                txmode = 6;
                ESCI_A.DR.B.D = txbuf[0];
                cksum += ESCI_A.DR.B.D;
                ESCI_A.CR1.R |= 0x00000088;    // xmit enable & xmit interrupt enable
              }
              else  {
                if(txcnt >= ctxboff)  {
                  // serial was ahead - restart xmits
                  ESCI_A.DR.B.D = txbuf[ctxboff];
                  cksum += ESCI_A.DR.B.D;
                }
              }
              ctxboff += var_byt;
              if(ctxboff >= rxnbytes - 1)  {
                // all done receiving; remaining data will be auto-serial-xmitted  
                getCANdat = 0;
                ltch_CAN = 0xFFFFFFFF;
              }
              break;
            } 
            else  {     // not pass-thru mode
              if(tables[var_blk].addrRam != MNULL)  {
                // update ram variables with received data
                for(jx = 0;jx < var_byt;jx++)  {
                  *tableByteRam(var_blk, var_off + jx) = 
                                                   CAN_A.BUF[ibuf].DATA.B[jx];
                }
              } 
              else  {     // burn entire flash-only table with recvd data
                if(!burnCANdat)  {
                  // erase flash block
                  burnCANdat = 1; 
                  canrxbytes = 0;
                  // erase table
                  Flash_Erase_Block(tableWordFlash(var_blk, 0), tableBlk_no(var_blk));
                  tbleCnt = 0;
                }
                // update flash table (assume multiple of 256 bytes) with recvd data 
                for(jx = 0; jx < var_byt;jx++)  {
                  tble_page[tbleCnt + jx] = CAN_A.BUF[ibuf].DATA.B[jx];
                }
                tbleCnt += var_byt;
                canrxbytes += var_byt;
                if((tbleCnt >= 255) ||           // wait til a page full
                      (canrxbytes >= tableBytes(var_blk)))  {  // or no more data
                  // write table page (256 bytes)
                  ntword = (uint16_t)(canrxbytes >> 2);
                  Flash_Write_Page(tableWordFlash(var_blk, ntword), (uint32_t *)tble_page);
                  tbleCnt = 0;
                }
                if(canrxbytes >= tableBytes(var_blk))  {
                  // all done
                  burnCANdat = 0;
                  ltch_CAN = 0xFFFFFFFF;
                  Flash_End_Write(tableBlk_no(var_blk));
                } 
                else  {
                  // still more data - set 1 sec timeout
                  ltch_CAN = lmms + 10000;
                } 
              }
            }
            break;
        		
          case MSG_REQ:  // msg to send back current value of variable(s)
            // Update related parameters for xmt buffer
            canmx.msg_type = MSG_RSP;
            // destination var blk
            canmx.destvarblk = CAN_A.BUF[ibuf].DATA.B[0];
            canmx.destvaroff = 
                        (uint16_t)((CAN_A.BUF[ibuf].DATA.B[1] << 3) |
                                   ((CAN_A.BUF[ibuf].DATA.B[2] & 0xE0) >> 5));
            canmx.dest = rcv_id;
            if(var_byt <= 3)  {
              var_byt = (uint8_t)(CAN_A.BUF[ibuf].DATA.B[2] & 0x1F);
            } 
            else  {    // > 8 bytes requested (from ECU ser/TS)
              var_byt = 8;
              RFlag = 1;
              Rbytes = CAN_A.BUF[ibuf].DATA.B[3];
              Rvarblk = var_blk;
              Rdvaroff = canmx.destvaroff;
              Rvaroff = var_off;
            }
            canmx.varbyt = var_byt;
            // put variable value(s) in xmit buffer
            if(tables[var_blk].addrRam != MNULL)  {
              for(kx = 0;kx < canmx.varbyt;kx++)  {
                canmx.datbuf[kx] = 
                     *tableByteRam(var_blk, var_off + kx);
              }
            }
            else  {
              for(kx = 0;kx < canmx.varbyt;kx++)  {
                canmx.datbuf[kx] = 
                     *tableByteFlash(var_blk, var_off + kx);
              }
            }
            sendCanTx(&canmx);
            break;
        		
          case OUTMSG_REQ:   // ECU receiving msg from gpio device to send back 
               // various bytes in the outpc structure to destvarblk.
               // The bytes are as configured in outmsg[msg_no] where 
               // msg_no= CAN_RB_DSR0 
            // Update related parameters for xmt buffer
            canmx.msg_type = OUTMSG_RSP;
            msg_no = CAN_A.BUF[ibuf].DATA.B[0];
            canmx.outmsg_no = msg_no;
            can_outbytes[msg_no] = outmsg[msg_no].no_bytes; // init byte counter
            // destination var blk - for this msg the varblk is really the gpio
            //   varblk to return the data to
            canmx.destvarblk = var_blk;
            canmx.destvaroff = 0;   // store data starting at gpio varblk;
              // if > 8 bytes, will have subsequent messages, with varoff provided 
              // by ECU
            canmx.dest = rcv_id;
            if(can_outbytes[msg_no] >= 8)  {
              canmx.varbyt = 8;
            }
            else  {
              canmx.varbyt = can_outbytes[msg_no];
            }
            // put variable value(s) in xmit buffer as follows:
            for(kx = 0;kx < canmx.varbyt;kx++)  {
               canmx.datbuf[kx] = 
                   *((int8_t *)&outpc + outmsg[msg_no].offset[kx]);
            }
            sendCanTx(&canmx);
            break;

          case OUTMSG_RSP:  // msg from ECU in reply to this gpio device's request 
          	// for specific values from ECU outpc to be put in gpio varblk + varoff.
            // update ram var block with received data; don't need can_outbytes[msg_no]
            // counter - ECU provides it via varoff
            for(jx = 0;jx < var_byt;jx++)  {
              *tableByteRam(var_blk, var_off + jx) =
                                    CAN_A.BUF[ibuf].DATA.B[jx];
            }
            break;
        		
          case MSG_XSUB:  // msg to execute a subroutine
            switch(rcv_id)  {
              case 1:           // message to this ecu from device 1
                can_xsub01();		// execute sub immediately here (set
                                // flag if can execute in main loop) 
                break;
            }
            break;
        	  
          case MSG_BURN:  // msg to burn ram data table into flash data table
            burn_idx = (uint8_t)var_blk;
            burn_flag   = 1;
            break;
    }					 // end mtype switch
    CAN_A.BUF[ibuf].CS.B.CODE = 0x04;    // set buf back to empty, actv rcv
    ultmp = CAN_A.TIMER.R;      // read timer to unlock all msg buffers
    can_status &= CLR_RCV_ERR;
    if(CAN_A.ESR.R != 0)  {
      // Rcv error
      can_status |= RCV_ERR;
      //can_reset = 1;
    }
CANRX_DONE:
	asm(mbar 0);   // done in sample epilogs to ensure interrupt flag cleared
	INTC.EOIR.R = 0;
    return;
}

void can_xsub01(void)  {
  return;
}