typedef struct {
  /* CAN Xmt mssge ring buffer:
      msg_type = CMD,REQ,RESP,XSUB (= set value, request value, respond 
          to a request for value, execute a subroutine)
      varblk,varoffset,varbyte = blk no of data structure, byte offset 
          from start of structure, no. bytes of data
      datbuf = the actual data (max of 8 bytes)
      dest = id no. of device to which msg being sent. 
  */
  uint8_t msg_type, myvarblk, 
	  dest,varbyt,
	  outmsg_no;     // only applies to OUTMSGs:
      // msg_no is an arbitray number(0 to MAX_OUTMSGS-1) assigned to each message                              
      //  configured. When a request for this message number is received, 
      //  ECU determines the bytes and table offset needed from the outmsg structure 
      //  in flash.
  uint8_t destvarblk;
  uint8_t datbuf[8];  // max msg data = 8 bytes
  uint16_t myvaroff,destvaroff;
} canmsg;

#define  MAX_BYTES_OUTMSG  24    // enough for 3 8-byte packets.
#define  MAX_OUTMSGS  8    // 4 GPIOs can specify 2 outmessages each,
                           // and each outmsg can hold up to 24 bytes
typedef struct {
   uint16_t xrate;   // =0 means the message is only transmitted when requested, 
             //  otherwise it indicates the clock speed in 0.128 tics at which
             //  a CAN message will be transmitted automatically after receiving 
             //  a request for data by setting the activate_xrate mth bit= 1,
             //  where m = msg no = 0 to MAX_OUTMSGS.
   uint8_t dest_varblk;   // destination ID of message (msb, 4 bits) and variable block 
                                //  (lsb, 5 bits) in destination device where message is to be put.
   uint8_t no_bytes; // The total number of bytes of all the variables in the message.                             
                           //  Must be <= sizeof(outpc).
   uint8_t offset[MAX_BYTES_OUTMSG];    //  This holds the byte offsets of each of the    
                           //  bytes in the message relative to start of  outpc. For a short, 
                           //  specify 2 consecutive offsets in the proper order for the requesting 
                           //  processor.
} can_outmsg;
