void CanInit(void)
{
unsigned char ix;
	/* Set up CAN communications */
	/* Enable CAN, set Init mode so can change registers */
	CANCTL1 |= 0x80;
	CANCTL0 |= 0x01;
	
	/* clear ring buffers */
	for(ix = 0;ix < 2;ix++)  {
	  can[ix].cxno = 0;
	  can[ix].cxno_in = 0;
	  can[ix].cxno_out = 0;
	}
	can_status = 0;
	getCANdat = 0;
	
	while(!(CANCTL1 & 0x01));  // make sure in init mode
	
	/* Set Can enable, use IPBusclk (24 MHz),clear rest */
	CANCTL1 = 0xC0;  
	/* Set timing for .5Mbits/ sec */
	CANBTR0 = 0xC2;  /* SJW=4,BR Prescaler= 3(24MHz CAN clk) */
	CANBTR1 = 0x1C;  /* Set time quanta: tseg2 =2,tseg1=13 
	              (16 Tq total including sync seg (=1)) */
	CANIDAC = 0x00;   /* 2 32-bit acceptance filters */
	/* CAN message format:
	 Reg Bits: 7 <-------------------- 0
	  IDR0:    |---var_off(11 bits)----|  (Header bits 28 <-- 21)
	  IDR1:    |cont'd 1 1 --msg type--|  (Header bits 20 <-- 15)
	  IDR2:    |---From ID--|--To ID---|  (Header bits 14 <--  7)
	  IDR3:    |--var_blk-|--spare--rtr|  (Header bits  6 <-- 0,rtr)
	*/  
	CAN_TB0_IDR3 = 0;
	/* Set identifier acceptance and mask registers to accept 
	     messages only for can_id or device #15 (=> all devices) */
	/* 1st 32-bit filter bank-to mask filtering, set bit=1 */
	CANIDMR0 = 0xFF;           // anything ok in IDR0(var offset)
	CANIDAR1 = 0x18;           // 0,0,0,SRR=IDE=1
	CANIDMR1 = 0xE7;		       // anything ok for var_off cont'd, msgtype
	CANIDAR2 = can_id;         // rcv msg must be to can_id, but
	CANIDMR2 = 0xF0;			     // can be from any other device
	CANIDMR3 = 0xFF;           // any var_blk, spare, rtr
	/* 2nd 32-bit filter bank */
	CANIDMR4 = 0xFF;           // anything ok in IDR0(var offset)
	CANIDAR5 = 0x18;           // 0,0,0,SRR=IDE=1
	CANIDMR5 = 0xE7;		       // anything ok for var_off cont'd, msgtype
	CANIDAR6 = 0x0F;			     // rcv msg can be to everyone (id=15), and
	CANIDMR6 = 0xF0;			     // can be from any other device
	CANIDMR7 = 0xFF;           // any var_blk, spare, rtr
	
	/* clear init mode */
	CANCTL0 &= 0xFE;  
	/* wait for synch to bus */
	while(!(CANCTL0 & 0x10));
	
	/* no transmit yet */
	CANTIER = 0x00;
	/* clear RX flag to ready for CAN recv interrupt */
	CANRFLG = 0xC3;
	/* set CAN rcv full interrupt bit */
	CANRIER = 0x01;
	return;
}

interrupt void CanTxIsr(void)
{
unsigned char ix,jx,kx,mtype,dest_id,msg_no,var_blk,var_byt_tx;
unsigned short dvar_off_tx;

CANtx++;
  
/* CAN transmit Interrupt */
CANTBSEL = CANTFLG;    // select MSCAN transmit buffer

// Check ring buffers and transfer to MSCAN buffer
for(ix = 0;ix < 2;ix++)  
{
 if(can[ix].cxno)  
  {
 	jx = can[ix].cxno_out;
 	/* Set up identifier registers */
	CAN_TB0_IDR0 = (unsigned char)(can[ix].cx_destvaroff[jx] >> 3);
	                      // 8 high bits in IDR0, 3 low bits in IDR1
	CAN_TB0_IDR1 = 
	  (unsigned char)((can[ix].cx_destvaroff[jx] & 0x0007) << 5) | 
	               0x18 |           // SRR=IDE=1
                 can[ix].cx_msg_type[jx];			// 3 bits
	CAN_TB0_IDR2 = (can_id << 4) | can[ix].cx_dest[jx];
	CAN_TB0_IDR3 = (can[ix].cx_destvarblk[jx] << 4);
	/* Set xmt buffer priorities (lower is > priority) */
	CAN_TB0_TBPR = 0x02;
	
	/* set data in buffer */
	mtype = can[ix].cx_msg_type[jx];
	
	if(mtype == MSG_XTND) mtype = can[ix].cx_datbuf[jx][0];
	
	switch(mtype)  {
		case MSG_CMD:  // message for destination ecu to set a variable to value in msg
		case MSG_RSP:  // message in reply to destination ecu's request for a variable value
		  CAN_TB0_DLR = can[ix].cx_varbyt[jx];
		  for(kx = 0;kx < CAN_TB0_DLR;kx++)  
		    {
		    *(&CAN_TB0_DSR0 + kx) = can[ix].cx_datbuf[jx][kx];
		    }
		  break;
		
		case MSG_REQ:  // msg to send back current value of variable(s)
		               // this 1st byte holds var blk for where to put rcvd data
		               // 2nd, 3rd bytes hold var offset relative to var blk and how
		               // many consecutive bytes to be sent back
		  CAN_TB0_DLR = 3;
		  CAN_TB0_DSR0 = can[ix].cx_myvarblk[jx];
		  CAN_TB0_DSR1 = (unsigned char)(can[ix].cx_myvaroff[jx] >> 3);
		  CAN_TB0_DSR2 = (unsigned char)((can[ix].cx_myvaroff[jx] & 0x0007) << 5) | 
		                                  can[ix].cx_varbyt[jx];
		  break;
		
		case OUTMSG_REQ:  // gpio msg to ECU to send back current values of specific 
		               // outpc variables. IDR0-3(above) hold var blk, varoff for where in gpio to put
		               // the data to be sent back to gpio
		               /*   ****** sample message setup (placed in gpio mainloop or in a ISR, 
		                                                    eg timer)  ****** 
		               ix = can[0].cxno_in;
		               can[0].cx_msg_type[ix] = OUTMSG_REQ; 
		               // Put data from ECU in this gpio block      
		               can[0].cx_destvarblk[ix] = <varblk no.>;  // from tableDescriptor
		               can[0].cx_dest[ix] = 0;		// send to ECU (board 0)
		               // Below is offset from start of gpio block - it should
		               // always be 0. If more than 1 msg, ECU will send offset
		               // to count byte sent
		               can[0].cx_destvaroff[ix] = 0;
		               can[0].cx_outmsg_no[ix] = <msg no>;  // 0-MAX_OUTMSGS
		               send_can(0);
		               *******************************************************************/ 
		  CAN_TB0_DLR = 1;
		  CAN_TB0_DSR0 = can[ix].cx_outmsg_no[jx];
		  break;

		case OUTMSG_RSP:  // ECU sending msg to gpio device in reply 
		                  // to gpio's request for 
		                  // specific values from outpc + offset 
		                  // to be sent to gpio destvarblk + destvaroff
		  msg_no = can[ix].cx_outmsg_no[jx];
		  
		  if(n_outbytes[msg_no] < 8)  
		    {
		    CAN_TB0_DLR = n_outbytes[msg_no];  // number of bytes to go (less then 8)
		    n_outbytes[msg_no] = 0;            // this is the last message
		    }  
		   else  // n_outbytes[msg_no] >= 8
		    {                      
		    CAN_TB0_DLR = 8;           // send 8 bytes
		    n_outbytes[msg_no] -= 8;   // more messages to go, decrement counter by 8 bytes
		    }   // end if(n_outbytes[msg_no] < 8) 
		  
		  for(kx = 0;kx < CAN_TB0_DLR;kx++)  
		    {
		    *(&CAN_TB0_DSR0 + kx) = can[ix].cx_datbuf[jx][kx];
		    }   // end for(kx
		  
		  if(n_outbytes[msg_no]) // if there are still bytes left to transmit
		    {
		    dest_id = can[ix].cx_dest[jx];
		    var_blk = can[ix].cx_destvarblk[jx];
		    
		    // set up for next transmission
		    jx = can[0].cxno_in;
		    can[0].cx_msg_type[jx] = OUTMSG_RSP;
		    can[0].cx_outmsg_no[jx] = msg_no;
		    
		    // destination var blk
		    can[0].cx_destvarblk[jx] = var_blk;
		    dvar_off_tx = outmsg[msg_no].no_bytes - n_outbytes[msg_no];  // as n_outbytes counts down, dvar_off_tx counts up
		    can[0].cx_destvaroff[jx] = dvar_off_tx;
		    can[0].cx_dest[jx] = dest_id;
		       
		    if(n_outbytes[msg_no] >= 8)  
		      {
		      var_byt_tx = 8;
		      }
		     else // n_outbytes[msg_no] < 8  
		      {
		      var_byt_tx = n_outbytes[msg_no];   // number of remaining bytes
		      } // end  if(n_outbytes[msg_no] >= 8)
		      
		    can[0].cx_varbyt[jx] = var_byt_tx;
		    
		    // put variable value(s) in transmit ring buffer as follows:
		    for(kx = 0;kx < var_byt_tx;kx++)  
		      {
		      can[0].cx_datbuf[jx][kx] = 
		           *((char *)&outpc + outmsg[msg_no].offset[dvar_off_tx + kx]);
		      }   // end for(kx = 0...
		      
		    send_can(0);
		    
		    }  // end if(n_outbytes[msg_no]) 
		  break;
		  
		case MSG_XSUB:
		  CAN_TB0_DLR = 0;
		  break;
		  
		case MSG_BURN:
		  CAN_TB0_DLR = 0;
		  break;
	}     // end mtype switch
	
	// This is where (in xmt buffer) to get next outgoing message
	if(can[ix].cxno_out < (NO_CANMSG - 1))
		  can[ix].cxno_out++;
	  else
		  can[ix].cxno_out = 0;
	  
	if(can[ix].cxno > 0)
	 	can[ix].cxno--;    // TB buffer loaded for transmit - decrement ring buffer count
		
	can_status &= CLR_XMT_ERR;
	
	/* set CAN xmt interrupt bit and clear flag to initiate transmit */
	CANTIER |= CANTBSEL;  
	CANTFLG = CANTBSEL;  // 1 clears(buffer full), 0 is ignored
	break;               // only handle 1 buffer entry at time
 }   // end if(can[ix].cxno) ...
 
 if(ix == 1)  
  {       // nothing left in either ring buffer
  CANTIER = 0x00;     // leave CANTFLG as buffer empty, but disable interrupt
                      // Note: if this last transmit in the buffer, will re-enter ISR, but 
                      // then exit because can[0,1].cxno = 0. 
  break;  
  } // end  if(ix == 1) 
}		        // end for loop

if((CANRFLG & 0x0C) != 0)  
   {
   // Xmt error count
   can_status |= XMT_ERR;
   //can_reset = 1;
   } // end  if((CANRFLG & 0x0C) != 0)
return;
}

interrupt void CanRxIsr(void)
{
unsigned char rcv_id,mtype,var_blk,var_byt_rx,ix,jx,kx;
unsigned short var_off,dvar_off_rx,tble_word,ntword;
static unsigned short canrxbytes;
unsigned char save_page,sect,nsect,msg_no;

    CANrx++;

    /* CAN Recv Interrupt */
    if(CANRFLG & 0x01)  
       {
    
        var_off = ((unsigned short)CAN_RB_IDR0 << 3) |
                       ((CAN_RB_IDR1 & 0xE0) >> 5);
        
        mtype = (CAN_RB_IDR1 & 0x07);    // right most 3 bits: these are supposed to allways be zero?!?
        
        if(mtype == MSG_XTND)
           mtype = CAN_RB_DSR0;
        
        rcv_id     = CAN_RB_IDR2 >> 4;		   // message for device rcv_id (extended format identifier)
        var_blk    = CAN_RB_IDR3 >> 4;      // table id (extended identifier mapping)
        var_byt_rx = (CAN_RB_DLR & 0x0F);   // number of bytes (0-8)

        switch(mtype)  {
        
          case MSG_CMD:  // message for this ecu to set a variable to value in message
          case MSG_RSP:  // message in reply to this ecu's request for a variable value
        	               // value in the data buffer in received message
            if(getCANdat && (getCANdat == rcv_id))  
              {
              // set up for serial transmit of the received CAN bytes.
              // TunerStudio getting back data requested from gpio board
             for(jx = 0;jx < var_byt_rx;jx++)  
                 {
                 // Put bytes in txbuf
                 // ctxboff is a global variable
                 txbuf[jx + ctxboff] = *(&CAN_RB_DSR0 + jx);
                 } // end for (jx...
                 
              if(ctxboff == 0)  
                {
                // set up to start transmitting data back to TunerStudio; this process
                // will be self-sustaining because CAN data arrives faster 
                // than serial data goes out; *** sizeof(txbuf) is max size  
                // of request. *** 
                txcnt = 0;
                txgoal = rxnbytes;
                txmode = 6;
                SCI0DRL = txbuf[0];
                cksum += SCI0DRL;
                SCI0CR2 |= 0x88;        // transmit enable & transmit interrupt enable
                }
                else   // ctxboff != 0  
                {
                // ctxboff != 0
                if(txcnt >= ctxboff)  
                  {
                  // serial was ahead - restart transmits
                  SCI0DRL = txbuf[ctxboff];
                  cksum += SCI0DRL;
                  }  // end if (txcnt >= ctxboff) ...
                  
              }  // end else  if(ctxboff == 0)  
              
              ctxboff  += var_byt_rx;  // increment ctxboff (CAN transmit offset)
              rxoffset += var_byt_rx;  // increment rxoffset (serial transmit offset)
              rxnbytes -= var_byt_rx;  // decrement rxnbytes (serial number of bytes remainng)
              
              if(rxnbytes > 0)   // if not done yet
                {
                // set up single CAN message & forward to gpio board
                // TunerStudio requesting data from gpio board.
                // Note: TunerStudio must only deal with 1 board at a time and 
                // wait until data back or timeout.
                
                ix = can[0].cxno_in;
                can[0].cx_msg_type[ix] = MSG_REQ; 
                can[0].cx_destvarblk[ix] = tble_idx;
                can[0].cx_myvarblk[ix] = 6;   // txbuf
                can[0].cx_destvaroff[ix] = rxoffset;
                can[0].cx_myvaroff[ix] = 0;
                can[0].cx_dest[ix] = rcv_id;
                
                if(rxnbytes >= 8)
                    can[0].cx_varbyt[ix] = 8;
                  else
                    can[0].cx_varbyt[ix] = (unsigned char)rxnbytes;
                  
                send_can(0);
                
                ltch_CAN = lmms + 3906;   // 0.5 sec timeout to get CAN data
                }  // end if(rxnbytes > 0) 
                else   // rxnbytes = 0 
                {
                // all done receiving; remaining data will be auto-serial-transmitted  
                getCANdat = 0;
                ltch_CAN = 0xFFFFFFFF;
                }   // end else if(rxnbytes > 0)  
                
              break;
            }   // end if(getCANdat && (getCANdat == rcv_id)) ...
            
            else   // not getCANdat
              {
              if(tables[var_blk].addrRam != NULL)  
                {
                // update ram variables with received data
                for(jx = 0;jx < var_byt_rx;jx++)  
                  {
                  *tableByteRam(var_blk, var_off + jx) =
                       *(&CAN_RB_DSR0 + jx);
                  } // end for(jx = ... 
                
              }    // end if(tables[var_blk].addrRam != NULL)
              
              else // tables[var_blk].addrRam == NULL  
              
                {     // burn entire flash-only table with recvd data
                if(!burnCANdat)  {
                  // erase flash sector(s) 
                  burnCANdat = 1; 
                  canrxbytes = 0;
                  flocker = 0xCC; // set semaphore to prevent burning flash through runaway code
                  save_page = PPAGE;
                  PPAGE = tableFPage(var_blk);
                  // erase table
                  nsect = (unsigned char)(tableWords(var_blk)/SECTOR_WORDS);
                  if(tableWords(var_blk) > (nsect*SECTOR_WORDS))nsect++;
                  for (sect = 0; sect < nsect; sect++)  {
                    Flash_Erase_Sector(tableWordFlash(var_blk, 0) + sect * SECTOR_WORDS);
                  }
                  PPAGE = save_page;
                  flocker = 0;
                }
                // update flash table with received data 
                flocker = 0xCC; 
                save_page = PPAGE;
                PPAGE = tableFPage(var_blk);
                for(jx = 0; jx < var_byt_rx;jx += 2)  {
                  tble_word = (*(&CAN_RB_DSR0 + jx) << 8) | 
                               *(&CAN_RB_DSR0 + (jx  + 1));
							   					
                  // write table word
                  ntword = canrxbytes>>1;
                  Flash_Write_Word(tableWordFlash(var_blk, ntword), tble_word);
                  canrxbytes += 2;
                  if(canrxbytes >= tableBytes(var_blk))  {
                    burnCANdat = 0;
                    break;
                  } 
                }
                flocker = 0;
                PPAGE = save_page;
                if(burnCANdat)  {
                  // still more data - set 1 sec timeout
                  ltch_CAN = lmms + 7812;
                } 
                else  {
                  // all done
                  ltch_CAN = 0xFFFFFFFF;
                } 
              } 
            }
            break;
        		
          case MSG_REQ:  // msg to send back current value of variable(s)
                         // Update related parameters for xmt ring buffer
            jx = can[0].cxno_in;
            can[0].cx_msg_type[jx] = MSG_RSP;
            // destination var blk
            can[0].cx_destvarblk[jx] = CAN_RB_DSR0;
            dvar_off_rx = ((unsigned short)CAN_RB_DSR1 << 3) |
                                     ((CAN_RB_DSR2 & 0xE0) >> 5);
            can[0].cx_destvaroff[jx] = dvar_off_rx;
            can[0].cx_dest[jx] = rcv_id;
            var_byt_rx = CAN_RB_DSR2 & 0x1F;   // right most 5 bytes (up to 31)
            
            can[0].cx_varbyt[jx] = var_byt_rx;
            
            // put variable value(s) in transmit ring buffer
            if(tables[var_blk].addrRam != NULL)  
              {
              for(kx = 0;kx < var_byt_rx;kx++)  
                {
                can[0].cx_datbuf[jx][kx] = 
                     *tableByteRam(var_blk, var_off + kx);
                }  // end for(kx
              }
            else // tables[var_blk].addrRam == NULL  
                 // i.e. burn flash
            {
              save_page = PPAGE;
              PPAGE = tableFPage(var_blk);
              for(kx = 0;kx < var_byt_rx;kx++)  {
                can[0].cx_datbuf[jx][kx] = 
                     *tableByteFlash(var_blk, var_off + kx);
              }
              PPAGE = save_page;
            }  // end if(tables[var_blk].addrRam != NULL)
            
            send_can(0);
            break;
        		
          case OUTMSG_REQ:   // ECU receiving msg from gpio device to send back 
                             // various bytes in the outpc structure to destvarblk.
                             // The bytes are as configured in outmsg[msg_no] where 
                             // msg_no = CAN_RB_DSR0 
                             // Update related parameters for xmt ring buffer
            jx = can[0].cxno_in;
            can[0].cx_msg_type[jx] = OUTMSG_RSP;
            msg_no = CAN_RB_DSR0;
            can[0].cx_outmsg_no[jx] = msg_no;
            n_outbytes[msg_no] = outmsg[msg_no].no_bytes; // set the n_outbytes counter 
                                                          // to count down 
                                                          // the number of bytes 
                                                          // remaining for message msg_no
            // destination var blk - for this msg the varblk is really the gpio
            //   varblk to return the data to
            can[0].cx_destvarblk[jx] = var_blk;
            can[0].cx_destvaroff[jx] = 0;   // store data starting at gpio varblk;
              // if > 8 bytes, will have subsequent messages, with varoff provided 
              // by ECU
            can[0].cx_dest[jx] = rcv_id;
            
            if(n_outbytes[msg_no] >= 8)  
              {
              var_byt_rx = 8;
              }
            else   // n_outbytes[msg_no] < 8
              {
              var_byt_rx = n_outbytes[msg_no];
               }   // end if(n_outbytes[msg_no] >= 8)
            
            can[0].cx_varbyt[jx] = var_byt_rx;
            
            // put variable value(s) in transmit ring buffer as follows:
            for(kx = 0;kx < var_byt_rx;kx++)  
               {
               can[0].cx_datbuf[jx][kx] = 
                   *((char *)&outpc + outmsg[msg_no].offset[kx]);
               }    // end for(kx = 0;...
               
            send_can(0);
            break;

          case OUTMSG_RSP:  // msg from ECU in reply to this gpio device's request 
          	// for specific values from ECU outpc to be put in gpio varblk + varoff.
            // update ram var block with received data; don't need n_outbytes[msg_no]
            // counter - ECU provides it via varoff
            for(jx = 0;jx < var_byt_rx;jx++)  {
              *tableByteRam(var_blk, var_off + jx) =
                       *(&CAN_RB_DSR0 + jx);
            }
            break;
        		
          case MSG_XSUB:  // msg to execute a subroutine
            switch(rcv_id)  {
              case 1:           // message to this ecu from device 1
                can_xsub01();		// execute sub immediately here (set
                                // flag if can execute in main loop) 
                break;
            }
        	  break;
        	  
          case MSG_BURN:  // msg to burn ram data table into flash data table
            flocker     = 0xCC;    // set to prevent burning flash thru runaway code
            burn_idx = var_blk;
            burn_flag   = 1;
        	  break;
        }					 // end mtype switch
        
        can_status &= CLR_RCV_ERR;
    
    }    // end if(CANRFLG & 0x01)
    
    if((CANRFLG & 0x72) != 0)  {
        // Rcv error or overrun on receive
        can_status |= RCV_ERR;
        //can_reset = 1;
    }
    /* clear RX buf full, err flags to ready for next rcv int */
    /*  (Note: can't clear err count bits) */
    CANRFLG = 0xC3;
	
    return;
}

void send_can(unsigned char buf)  {
    // buf = 0 for transmission from ISR; = 1 if from non-ISR
    // This is where (in xmt ring buffer) to put next message
    if(can[buf].cxno_in < (NO_CANMSG - 1))
      can[buf].cxno_in++;
    else
      can[buf].cxno_in = 0;	 // overwrite oldest msg in queue
    // increment counter
    if(can[buf].cxno < NO_CANMSG)
      can[buf].cxno++;
    else
      can[buf].cxno = NO_CANMSG;
    if(!(CANTIER & 0x07))  {
      // Following will cause entry to TxIsr without sending msg
      // since when CANTIER = 0, CANTFLG left as buff empty(>0).
      // If CANTIER has at least 1 int buf enabled, will enter
      // TxIsr automatically.
      CANTBSEL = CANTFLG;
      CANTIER = CANTBSEL;
    }
    return;
}

void can_xsub01(void)  {
  return;
}
