/**************************************************************************
**
** SCI Communications
**
** Communications is established when the PC communications program sends
** a command character - the particular character sets the mode:
**
** "a" = send all of the realtime display variables (outpc structure) via tx port.
** "w"+<offset lsb>+<offset msb>+<nobytes>+<newbytes lsb, msb> = 
**    receive updated data parameter(s) and write into offset location
**    relative to start of data block, Can write to RAM only via serial (use 't'
**    below for Flash only tables) or write via CAN to Ram or Flash-only tables.
** "r"+<offset lsb>+<offset msb>+<nobytes lsb, msb> = read and send back values
**     of a data parameter or block in table offset location.  Can be via serial 
**     or CAN. (Will read from RAM if Ram table exists, else will read Flash,) 
** "k" = return crc32 comm validity check
** "b" = jump to flash burner routine and burn a ram data block into a flash 
**    data block.
** "t"+<tble_idx>+<tableWords of data> = receive new data for clt/mat/ego/maf 
**    FLASH tables via serial.
** "c" = Test communications - echo back Seconds
** "Q" = Send over Embedded Code Revision Number
** "S" = Send program title.
**
**************************************************************************/
interrupt void ISR_SCI_Comm(void)  {
  unsigned char dummy,save_page;
  int ix;
  static int xcntr;
  static unsigned char rd_wr,CANid,sendCANdat=0,next_txmode,
   newMsg = 2,rplyMsg,nCRC32,crcStatus,bFlag,tCmd=0;
  unsigned char sect,nsect;
  static unsigned int txptr,tble_word,ntword,crxnbytes,msgSze;
  
#define getCANid             40
#define getTableId           41
#define getCRC32             45
#define setBurningParameters 99
  
  // if RDRF register not set, => transmit interrupt
  if(!(SCI0SR1 & 0x20))goto XMT_INT;

// Receive Interrupt
  // Clear the RDRF bit by reading SCISR1 register (done above), then read data
  //  (in SCIDRL reg).
  // Check if we are receiving new input parameter update
rcv_timeout = 0xFFFFFFFF;
if(SCI0SR1 & 0x08)   	 // check for rcv Overrun error
  goto KILL_COMMS;
dummy = SCI0DRL;
if(!newMsg && (txmode != getCRC32))  {
  save_page = PPAGE;
  PPAGE = 0x3B;
  crcTab = crc_tab[(crc32Val ^ dummy) & 0xFF];
  PPAGE = save_page;
  crc32Val = crcTab ^ (crc32Val >> 8);
}
switch(txmode)  {

case 0:

  if(newMsg)  {
    newMsg--;
    if(newMsg == 1)  {
      if((dummy == 'Q') || (dummy == 'S'))  {
        // non-enveloped
        SCI0CR2 &= ~0x24;   // rcv, rcvint disable
        txcnt = 1;
        if(dummy == 'Q')  {
          txmode = 4;
          txgoal = 20;
          SCI0DRL = RevNum[0];
        } 
        else  {
          txmode = 5;
          txgoal = 32;
          SCI0DRL = Signature[0];
        }
        newMsg = 2;
        rplyMsg = 0;        // no reply
        nCRC32 = 4;         // no reply
        SCI0CR2 |= 0x88;    // xmit enable & xmit interrupt enable
        break;
      }
      msgSze = (dummy << 8); // no bytes in msg payload(msb)
    }
    else  {     // newMsg = 0
      msgSze |= dummy;       // lsb
      crc32Val = 0xFFFFFFFF;
      nCRC32 = 0;
    }
    break;
  }
  bFlag = 0;
  switch(dummy)  {
    case 'a':     // send back all real time ram output variables
  next_txmode = 1;
  txmode = getCANid;
  break;

    case 'w':		  // receive new ram input data and write into offset location;
                  //  also used for forwarding CAN msgs bet. TS & gpio boards.
                  //  With this cmd in CAN pass-thru mode, 
  next_txmode = 5;
  txmode = getCANid;
  rd_wr = 1;
  break;

    case 'r':		  // read and send back ram input data from offset location;
                  //  also used for forwarding CAN msgs bet. TS/ MT & gpio boards,
  next_txmode = 5;    // but *** no_bytes must be < sizeof(txbuf). ***
  txmode = getCANid;
  rd_wr = 0;
  break;
  
    case 'k':         // request crc32
  next_txmode = 5;
  kFlag = -1;
  txmode = getCANid;
  break;

    case 'b':        // burn a block of ram input values into flash;
                  //  also used for forwarding CAN msgs bet. MT & auxilliary boards
  next_txmode = setBurningParameters;
  txmode      = getCANid;
  break;

    case 't':        // update a flash table with following serial data
  txmode = 20;
  tCmd =1;
  CANid = can_id;
  break;

    case '!':        // start receiving reinit/reboot command
  txmode = 30;
  break;

    case 'c':        // send back seconds to test comms
  txcnt = 0;
  next_txmode = 1;
  txgoal = 2;        // seconds is 1st 2 bytes of outpc structure
  tble_idx = 6;
  xcntr = -1;
  txbuf[0] = (char)(outpc.seconds & 0x00FF);
  txbuf[1] = (char)((outpc.seconds & 0xFF00) >> 4);
  txmode = getCRC32;
  break;

    case 'Q':         // send code rev no.
  txcnt = 0;
  next_txmode = 4;
  txgoal = 20;
  txmode = getCRC32;
  break;

    case 'S':         // send program title
  txcnt = 0;
  next_txmode = 5;
  txgoal = 32;
  txmode = getCRC32;
  break;

    default:
  goto KILL_COMMS;
  }     // End of switch for received command
  break;    // end of case 0

case getCANid: // Get CAN id for current command.
   CANid = dummy;
   if (CANid < MAX_CANBOARDS)
     txmode = getTableId;
   else	 {
  	 // CANid wrong - kill comms since don't know where to send data
     goto KILL_COMMS;
   }
   break;
case getTableId: // Get table id for current command.
   tble_idx = dummy;
   if(CANid != can_id)  {
     txmode = next_txmode;
   } 
   else if(tble_idx >= NO_TBLES)  {
  	 // tble index wrong -
     //  kill comms since don't know which/ how much data
     goto KILL_COMMS;
   }
   else
     txmode = next_txmode;
   if (txmode == 1) { 
     txcnt = 0;
     txgoal = sizeof(outpc);
     // load output variables into txbuf to avoid incoherent word data.
     // To work all words in structure must be word aligned, all longs
     // aligned on long boundaries.
     xcntr = -1;
     txptr = 0;
     *tableWordRam(tble_idx,0) = *((unsigned int *)(&outpc) + 0);
     *tableWordRam(tble_idx,1) = *((unsigned int *)(&outpc) + 1);
     next_txmode = 1;
     txmode = getCRC32;
   }
   else if (txmode == setBurningParameters) { // Burn command
     if(CANid != can_id)  {
  	   // set up single CAN message & forward to aux board
  	   ix = can[0].cxno_in;
  	   can[0].cx_msg_type[ix] = MSG_BURN; 
  	   can[0].cx_destvarblk[ix] = tble_idx;
  	   can[0].cx_destvaroff[ix] = 0;
  	   can[0].cx_dest[ix] = CANid;
  	   can[0].cx_varbyt[ix] = 0;		 // no data bytes
  	   send_can(0);
     }
     else 
       bFlag = 1;    // do burn in main loop after reply
     next_txmode = 0;			    // handle burning in main loop
     txmode = getCRC32;
   } 
   break;

case 5:
  	if(CANid != can_id)  {
  	  // TS wants to send data to aux board via CAN
  	  if(rd_wr)
  	    sendCANdat = CANid;
  	  // TS wants to request data from aux board via CAN
  	  else  {
  	    getCANdat = CANid;
  	  }
  	}
  	else  {          // serial only
  	  sendCANdat = 0;
  	  getCANdat = 0;
  	}
  	rxoffset = (dummy << 8); // byte offset(msb) from start of inpram
  	txmode++;
  	break;

case 6:
  	rxoffset |= dummy;       // byte offset(lsb) from start of inpram
  	txmode++;
  	break;
  	
case 7:
  	rxnbytes = (dummy << 8); 	// no. bytes (msb)
  	txmode++;
  	break;

case 8:
  	rxnbytes |= dummy;		    // no. bytes (lsb)
  	rxcnt = 0;
  	if(kFlag == -1)  {
  	  txcnt = 1;
  	  txgoal = 4;
  	  next_txmode = 7;
  	  txmode = getCRC32;
  	  break;
  	}
  	if(CANid == can_id)  {
  	  // check won't blow table arrays
  	  if(rxoffset + rxnbytes > tableBytes(tble_idx))  {
  	    //  kill comms since don't know how much data being sent
  	    goto KILL_COMMS;
  	  }
  	}
  	if(rd_wr == 0)  {		    // read & send back input data
  	  if(getCANdat)  {
  	    // set up single CAN message & forward to aux board
  	    //   TS requesting data from aux board.
  	    //   Note: TS must only deal with 1 board at a time and wait til
  	    //   data back or timeout.
  	    ix = can[0].cxno_in;
  	    can[0].cx_msg_type[ix] = MSG_REQ; 
  	    can[0].cx_destvarblk[ix] = tble_idx;
  	    can[0].cx_myvarblk[ix] = 6;   // txbuf
  	    can[0].cx_destvaroff[ix] = rxoffset;
  	    can[0].cx_myvaroff[ix] = 0;
  	    can[0].cx_dest[ix] = CANid;
  	    // rxnbytes must be <= 255
  	    can[0].cx_varbyt[ix] = (unsigned char)rxnbytes;
  	    send_can(0);
  	    ltch_CAN = lmms + 3906;   // 0.5 sec timeout to get CAN data
  	    ctxboff = 0;
  	    next_txmode = 0;          // done receiving all request data from TS
  	    txmode = getCRC32;
  	  } 
  	  else  {               // serial read only
  	    txcnt = 0;
  	    txgoal = rxnbytes;
  	    next_txmode = 3;
  	    txmode = getCRC32;
  	    // finish reading/sending rest in serial xmit mode
  	  }
  	}
  	else  {        // write data to input/ table data block or via CAN
  	  txmode++;
  	  crxnbytes = 0;
  	}
  	break;

case 9:
  	// Check for data overrun.  Note: this
  	// still not bullet proof because could write half the bytes
  	// then have overrun; need to store and write all at once if
  	// all is valid.
  	if((SCI0SR1 & 0x08) == 0)  {   // none yet- write data.
  	  if(sendCANdat)  {
  	    // want to buffer for 8 bytes for CAN pass thru
  	    txbuf[crxnbytes] = dummy;
  	  } 
  	  else  {   
  	    // write directly to RAM - don't need buffer since pc 
  	    //  data will be coherent
  	    *tableByteRam(tble_idx, rxoffset + rxcnt) = dummy;
  	  }
  	}
  	else  {
  	  goto KILL_COMMS;
  	}
  	rxcnt++;
  	crxnbytes++;
  
  	if(sendCANdat)  {
  	  if((crxnbytes >= 8) || (rxcnt >= rxnbytes))  {
  	    // set up single CAN message & forward to aux board
  	    //   TS sending data to aux board
  	    // CAN faster than serial, so won't get overrun
  	    for(ix = 0; ix < crxnbytes; ix++)  {
  	      can[0].cx_datbuf[can[0].cxno_in][ix] = txbuf[ix];
  	    }
  	    ix = can[0].cxno_in;
  	    can[0].cx_msg_type[ix] = MSG_CMD; 
  	    can[0].cx_destvarblk[ix] = tble_idx;
  	    can[0].cx_destvaroff[ix] = rxoffset;
  	    can[0].cx_dest[ix] = sendCANdat;
  	    can[0].cx_varbyt[ix] = (unsigned char)crxnbytes;
  	    send_can(0);
  	    rxoffset += crxnbytes;
  	    crxnbytes = 0;
  	  }
  	}

  	if(rxcnt >= rxnbytes)  {
  	  next_txmode = 0;      // done receiving/writing all the data
  	  txmode = getCRC32;
  	  sendCANdat = 0;
  	}       // finished all rxnbytes
  	break;

case 20:	 // d/load & burn tables - *** Do while engine NOT turning ***
  	// Note: this type table has no ram copy - its not intended for tuning
  	tble_idx = dummy;    // type of table
  	if(tble_idx > NO_TBLES)	 {
  	  // tble index wrong - kill comms since don't know how much data being sent
  	  goto KILL_COMMS;
  	}
  	ntword = 0;
  	flocker = 0xCC; // set semaphore to prevent burning flash thru runaway code
  	save_page = PPAGE;
  	PPAGE = tableFPage(tble_idx);
	  // erase table
  	nsect = (unsigned char)(tableWords(tble_idx)/SECTOR_WORDS);
  	if(tableWords(tble_idx) > (nsect*SECTOR_WORDS))nsect++;
  	for (sect = 0; sect < nsect; sect++)  {
  	  Flash_Erase_Sector(tableWordFlash(tble_idx, 0) + sect * SECTOR_WORDS);
  	}
  	PPAGE = save_page;
  	flocker = 0;
  	txmode++;
  	break;

case 21:
  	tble_word = (dummy << 8);    // msb
  	txmode++;
  	break;

case 22:
  tble_word |= dummy;            // lsb
  flocker = 0xCC; // set semaphore to prevent burning flash thru runaway code
  save_page = PPAGE;
  PPAGE = tableFPage(tble_idx);
  // write table word
  Flash_Write_Word(tableWordFlash(tble_idx, ntword), tble_word);
  PPAGE = save_page;
  flocker = 0;
  ntword++;
  if(ntword < tableWords(tble_idx))
    txmode = 21;
  else  {
    next_txmode = 0;
    txmode = getCRC32;
  }
  break;

case 30:
  	if(dummy == '!')    // receive 2nd ! for reboot signal
  	  txmode++;
  	else if(dummy == 'x')  {  // received reinit signal (!x)
  	  reinit_flag = 1;
  	  next_txmode = 0;
  	  txmode = getCRC32;
  	}
  	else  {
  	  goto KILL_COMMS;
  	}
  	break;

case 31:
  	if(dummy == 'x')  {   // received complete reboot signal(!!x)
  	  PPAGE = 0x3C;  	    
  	  // go to start of bootload pgm
  	  //asm (jmp $F800);
  	  (void)reboot();
  	}
  	else if(dummy == '!')  {   // received complete reload signal(!!!)
  	  /*
  	  PPAGE = 0x3C;  	    
  	  SCI0CR1 = 0x00;
  	  SCI0CR2 = 0x00;      // disable Tx, Rx
  	  // go to start of monitor code
  	  //asm (jmp $F842);
  	  (void)monitor();
  	  */
  	}
  	next_txmode = 0;
  	txmode = getCRC32;
  	break;
  	
case getCRC32:
    *((char *)&tsCRC32 + nCRC32) = dummy;
    nCRC32++;
    if(nCRC32 >= 4)  {
      SCI0CR2 &= ~0x24;   // rcv, rcvint disable
      crc32Val = crc32Val ^ 0xFFFFFFFF;
      if(tsCRC32 == crc32Val)  {
        if(tCmd)outpc.spare[5] = 7786;
        crcStatus = 0;
      }
      else
        crcStatus = 1;
      txmode = next_txmode;
      if((crcStatus != 0) || (txmode == 0))  {
        // don't send data
        txcnt = 0;
        txgoal = 0;
        if(txmode == 7)kFlag = -2;
      }
      newMsg = 2;
      rplyMsg = 2;
      msgSze = txgoal + 1;   // include status byte
      SCI0DRL = *(unsigned char *)&msgSze;         // msgsze
      SCI0CR2 |= 0x88;    // xmit enable & xmit interrupt enable
    }
  	break;

default:
  break;

  }     // End of case switch for received data
    
  // if another receive expected, set rcv timeout
  if((SCI0CR2 & 0x24) == 0x24)
    rcv_timeout = lmms + 3906;   // 0.5 sec timeout
  return;
  
KILL_COMMS:
   txmode = 0;
   newMsg = 2;
   dummy = SCI0DRL;
   SCI0CR2 &= ~0xAC;   // rcv, xmt disable, interrupt disable
   kill_ser = 1;
   kill_ser_t = outpc.seconds + SER_TOUT;
   outpc.spare[4] = 7793;
   return;

// Transmit Interrupt
XMT_INT:
  // Clear the TDRE bit by reading SCISR1 register (done), then write data
  if(rplyMsg)  {
    rplyMsg--;
    if(rplyMsg == 1) {
      SCI0DRL = *((unsigned char *)&msgSze + 1);   // msgsze
      crc32Val = 0xFFFFFFFF;
      nCRC32 = 0;
    } 
    else {         // crc status
      if(crcStatus != 0)
        dummy = 1;
      else  {
        dummy = 0;
        if(txmode == 7)  {
          kFlag = tble_idx;  // will cause calc of crc32 in main
          SCI0CR2 &= ~0x80;    // xmit interrupt disable
        }
      }
      save_page = PPAGE;
      PPAGE = 0x3B;
      crcTab = crc_tab[(crc32Val ^ dummy) & 0xFF];
      PPAGE = save_page;
      crc32Val = crcTab ^ (crc32Val >> 8);
      SCI0DRL = dummy;
    }
    return;
  }
  // if data, send it
  if(txcnt < txgoal)  {
   switch(txmode)  {
    case 1:        // a, c cmds: send back outpc data via serial 
      xcntr++;
      dummy = *tableByteRam(tble_idx, xcntr);
      if(xcntr >= 3)  {	 // transmitted 4 bytes, load 2 new words
        xcntr = -1;
        txptr += 2;
        *tableWordRam(tble_idx,0) = *((unsigned int *)(&outpc) + txptr);
        *tableWordRam(tble_idx,1) = *((unsigned int *)(&outpc) + txptr+1);
      }
      break;
    case 3:       // r cmd: answer serial data request via ser data xmit
  	  if(tables[tble_idx].addrRam != NULL)
  	    dummy = *tableByteRam(tble_idx, rxoffset + txcnt);
  	  else  {
  	    save_page = PPAGE;
  	    PPAGE = tableFPage(tble_idx);
  	    dummy = *tableByteFlash(tble_idx, rxoffset + txcnt);
  	    PPAGE = save_page;
  	  }
      break;
    case 4:
  	  dummy = RevNum[txcnt];
      break;
    case 5:
  	  dummy = Signature[txcnt];
      break;
    case 6:
  	  // pass on CAN msg to TS/ MT via serial xmit
  	  // make sure serial not ahead of CAN
  	  if(txcnt < ctxboff)  {
  	    dummy = txbuf[txcnt];
  	  }
  	  break;
    case 7:
  	  dummy = *((char *)&crc32KVal + txcnt);
      break;
   }        // end switch(txmode)
   save_page = PPAGE;
   PPAGE = 0x3B;
   crcTab = crc_tab[(crc32Val ^ dummy) & 0xFF];
   PPAGE = save_page;
   crc32Val = crcTab ^ (crc32Val >> 8);
   SCI0DRL = dummy;
   txcnt++;
  }
  else  {   // done transmitting data, xmit crc32
    if(nCRC32 == 0)  {
      crc32Val = crc32Val ^ 0xFFFFFFFF;
      if(tCmd)  {
        outpc.pw1 = *((unsigned int *)&crc32Val + 0);
        outpc.pw2 = *((unsigned int *)&crc32Val + 1);
        tCmd = 0;
      }
    }
    if(nCRC32 < 4)  {
      SCI0DRL = *((unsigned char *)&crc32Val + nCRC32);
      nCRC32++;
    }
    else  {  // completely done
      if(bFlag && (crcStatus == 0) && (flocker == 0))  {
        flocker = 0xCC;    // set semaphore to prevent burning flash thru runaway code
        burn_flag = 1;
        burn_idx = tble_idx;
      }
      txcnt = 0;
      txgoal = 0;
      txmode = 0;
      SCI0CR2 &= ~0x88;    // xmit disable & xmit interrupt disable
      SCI0CR2 |= 0x24;     // rcv, rcvint re-enable
    }
  } 
  return;
}
