/*************************************************************************
		This header must appear on all derivatives of this code.
**************************************************************************
**   MegaShift - V5.xxx
**
**   (C) 2007, 2016 - B. A. Bowling and A. C. Grippo and L. Gardiner
**
**   This header must appear on all derivatives of this code. Any code 
**   modifications made for any reason become the exclusive property of 
**   Bowling and Grippo. 
**   This code is intended for off-road, experimental use under controlled 
**   conditions only.		 
**   All code revisions must be shared with other users free of charge 
**   and without additional restrictions on the forums at 
**   http://www.msgpio.com/forums/index.php 
**   This code, nor any derivative of it, may not be used in any form, 
**   in whole or in part, on non-Bowling and Grippo hardware under any circumstances.
**   No commercial use of this code, in whole or in part, is permitted without 
**   express written permission from Bowling and Grippo.
**									   
***************************************************************************
		This header must appear on all derivatives of this code.
**************************************************************************/


// #define COMMS_382X // uncomment this to use 3.820+ comms code from B&G, otherwise use old code
//                     - must set COMM_382x to 'deactivated' in Project Properties in TS to use old code
//                     - must set COMM_382x to 'activated'   in Project Properties in TS to use new code


/* --- General Code Layout -----------------------------------------------
This is a list of general blocks in the code.

Copyright and other notices
General description of the function of the code
Compiler/Linker/Locator notes
The 9S12C64 Microcontroller Memory Notes
MegaShift GPIO Port Assignments Notes
I/O Port Quick_Start
CAN Comms Notes
Port DDR (data direction register) Configuration Notes
Miscellaneous Notes
Transmission Error Code Notes
Defines, Includes, ISR Definitions and Interrupt Table
Function Prototypes
Table Includes and Table Size Defines (including default values)
Revision Number and Signature
Global Variables
Function MAIN
- define variables
- Initialize PLL (Phase Locked Loop)
- Copy Flash to RAM
- Set Up Input/Output ports
- Set Up Timer Ports
- Set Up CRG RTI Interrupt
- Set up SCI Serial Communications Interface
- Set Up PWM on Timer Pins
- Set Up ADC Channels 
- Set variable block addresses to be used for CAN communications
- Initialize Variables
- MAIN LOOP
-- Update RPM, MAP & Battery Voltage with CAN
-- Determine Current Gear
-- Update Rolling Load Average
-- Update Rolling Shift Switch Inputs
-- Set Pressure Control PWM Duty Cycle
-- Check if Need to Switch to Auto Shift Mode
-- Set the Torque Converter Clutch (TCC)
-- Set SOLC PWM Duty Cycle
-- Read Switches Input1state, Input2state, Input3state, Input4state and lookup manual_gear
-- If check/do shift
	--- direct shift if enabled
			---- downshifts checked first
		---- upshifts second
	--- else auto sequential shift or manual mode
		---- downshifts checked first
		---- upshifts second
-- Burn Flash if flagged
-- Check for reinit command
-- Check for CAN reset
-- Slow Speedometer 
- End MAIN LOOP

Functions:
- tcc_unlock()
- tcc_lock()
- Calc_Initial_LUF()
- Calc_LUF_PWM()
- reset_gear()
- reset_switches()
- VSS_reset()
- reset()
- get_adc()
- switch_page()
- intrp_1dctable()
- intrp_speedo_table()
- intrp_2dctable()
- intrp_2dPWM()
- auto_gear_lookup()
- VSS_timer()
- Timer_Clock_ISR()  -- this is where the millisecond and other timer operations happen
- VPW_transmit_ISR()
- ISR_TimerOverflow()
- Serial_Comm_ISR()
- CanInit()    - #include
- CanTxIsr()   - #include
- CanRxIsr()   - #include
- can_xsub01() - #include
- Output1()
- Output2()
- Output4()
- Output5()
- Output6()
- UnimplementedISR()
- fburner()
- Flash_Init()
- Flash_Write_Word()
- sendCANAdj() - #include
- Output3()
- grabMS2vars()
- checkCANbus()
- Timer0010()
- TimerADC()
- Timer0100()
- Timer1000()
- ShiftNAG()
- getRatio()
- getCurrentGear()
- auto_gear_lookupRPM() - for rpm based shifting
- intrp_RPMtable() - for rpm based shifting

V.4000 Functions - #include "4.000_changes.h"
---------------------------------------------
- shiftTo()
- clutchOffset()
- CANretard()
- afterShiftRevs()
- resetShiftRequests()
- solOffset()
- getGear()
- SetLEDs()
- setOut3PWM()
- PNRset_outputs()
- getLeverDigital(); // in 4.000_changes.h/getLever.h
- getLeverADC();     // in 4.000_changes.h/getLever.h
- SetSparePort();    // in 4.000_changes.h/sparePorts.h
- odometer();        // in 4.000_changes.h/odometer.h 


V4.007 SPI Functions - #include "SPI_code.h"
--------------------------------------------
- SPI_sendByte()
- SPI_sendString()
- SPI_getByte()
- SPI_getString()
- setUp_SPI()
- calcSPIpwm()

*/

// Include port and pin definitions
// --------------------------------
// All MShift specific user I/O labels defined in this header file (not the port set-up)

#include "ports_pins.h"

/* ------------- CAN Communications -----------------------------------------------
CAN data is requested from the MS-II 8 bytes every can_var_rate (.128 ms tics) of the 
Timer_Clock_ISR(). The data is sent in the CanTxIsr() subroutine, and received in the
CanRxIsr() subroutine.

CAN routines now #included from ../../serCanComm_1/ directory as:
- serComm.h  - ISR_SCI_Comm()
- CANComm.h  - CanInit(), CanTxIsr(), CanRxIsr(), send_can(), and can_xsub01()
- serCAN0.h  - serial prototype of ISR_SCI_Comm();
- serCAN1.h  - other CAN prototypes and defines
- serCAN2.h  - typedefs for canmsg and can_outmsg
- serCAN3.h  - global CAN variables
*/

// --- Defines, Includes, ISR Definitions and Interrupt Table ----
// Note that #define statements and other compiler directives do not use a terminating semi-colon

#include "hcs12def.h"   			// This is where things like PORTE are set as names rather than memory addresses.
									// This is done entirely for programmer convenience, so that you can write 
									// PORTE instead of remembering memory address 0x0008

#include "flash.h"         			// flashburner defines, structures

#include <string.h>        			// standard C string handling functions

#define INTERRUPT interrupt         // So we can use upper or lower case

#define ENABLE_INTERRUPTS  asm cli; // 'cli' is an assembly language directive 
									//  meaning 'enable interrupts'
									
#define DISABLE_INTERRUPTS asm sei; // 'sei' is an assembly language directive 
									//  meaning 'disable interrupts'
									
#define NEAR near 					// You can obtain more compact code by using 
									// near types of addressing modes that use relatively
									// small offsets embedded within each instruction,
									// i.e. if we agree to have the compiler generate 
									// code constrained to only access locations "nearby"
									// to some default location or location 0. More 
									// compact code also implies less instruction 
									// fetching per execution hence performance is better. 
									
#define VECT_ATTR @0xFF80          	// vector interrupt space from absolute address 
									// $FF80 to $FFFF (65456 to 65535 decimal)
									
#define SPTEF 0x20                  // SPI data register empty flag bit 5 (0 = not empty, 1 = empty)
#define SPIF 0x80                   // SPI data interrupt request bit 7  (0 = transfer not complete, 
									//                                    1 = new data in SPI0DR)                    

#pragma CODE_SEG NON_BANKED 		// Interrupt section for this module. 
									// Placement will be in NON_BANKED area.
									// See Monitor_linker.prm
							
extern void near _Startup(void);	// Startup routine function prototype

/**** Interrupt Service Routines ********************************** 

When the HC9S12 processor detects an interrupt, you can have it automatically
jump to part of the program which tells it what to do outside of the normal flow 
of the program. The code an interrupt triggers a jump to is called an ISR (Interrupt 
Service Routine). The ISR is located at one of a specific set of memory 
locations called Interrupt Vectors that tell the HCS12 the address of the ISR
for each type of interrupt.

The HC9S12 know where to return to because the return address is pushed onto 
the stack before the HC9S12 jumps to the ISR.

When the processor executes the 'return' instruction at the end of 
the ISR, the registers are pulled off of the stack before returning to
program main loop execution.

To return from the ISR You must return from the ISR using the 'return' instruction in C. 
The return instruction tells the HCS12 to pull all the registers off of the stack and 
return to the address where it was processing when the interrupt occurred.

MShift never calls ISRs directly, ISRs run 'asynchronously' in response to 
interrupt requests. Thus, to the main loop code, each ISR appears to just 
start up on its own. Since the main loop doesn't call the ISR,
it cannot pass arguments to the ISR or accept a return value. Therefore, a 
ISR in C must be a function with no parameters and a void return type. The key 
word INTERRUPT tells the compiler that the code is an interrupt service routine, 
and also not to 'optimize' it right out of existence because it is never called 
in the main loop.
*/
							
INTERRUPT void UnimplementedISR(void); 
INTERRUPT void VSS_timer(void);
INTERRUPT void ISS_timer(void);
INTERRUPT void Timer_Overflow_ISR(void);
INTERRUPT void VPW_transmit_ISR(void);
INTERRUPT void Timer_Clock_ISR(void);
INTERRUPT void CanTxIsr(void);
INTERRUPT void CanRxIsr(void);
void checkCANbus(void);
// timer functions, returns sets timer before it will run again
unsigned int Timer0001(void); 	// milliseconds
unsigned int Timer0010(void); 	// hundredths
unsigned int TimerADC(void); 	// ADC every 15 milliseconds
unsigned int Timer0100(void);	// tenths
unsigned int Timer1000(void); 	// seconds
unsigned char shftNAG(void);
unsigned int getRatio(void);
unsigned char getCurrentGear(unsigned int iRATIO);
signed char auto_gear_lookupRPM(void);
unsigned int intrp_RPMtable(unsigned char row, unsigned char up);


// For new CAN comms
#ifdef COMMS_382X
  #include "..\..\serCANComm_382x\serCAN0.h"
  #include "..\..\serCANComm_382x\serCAN1.h"
  #include "..\..\serCANComm_382x\serCAN2.h"
#else
  #include "..\..\serCANComm_1\serCAN0.h"
  #include "..\..\serCANComm_1\serCAN1.h"
  #include "..\..\serCANComm_1\serCAN2.h"
#endif

void sendCANAdj(void);

typedef void (* NEAR tIsrFunc)(void);
const tIsrFunc _vect[] VECT_ATTR = {      // Interrupt table placed at 0xFF80
		UnimplementedISR,                 // vector 63, 0xFF80 - Reserved
		UnimplementedISR,                 // vector 62, 0xFF82 - Reserved
		UnimplementedISR,                 // vector 61, 0xFF84 - Reserved
		UnimplementedISR,                 // vector 60, 0xFF86 - Reserved
		UnimplementedISR,                 // vector 59, 0xFF88 - Reserved
		UnimplementedISR,                 // vector 58, 0xFF8A - VREG LVI
		UnimplementedISR,                 // vector 57, 0xFF8C - PWM emergency shut-down
		UnimplementedISR,                 // vector 56, 0xFF8E - Port P
		UnimplementedISR,                 // vector 55, 0xFF90 - Reserved
		UnimplementedISR,                 // vector 54, 0xFF92 - Reserved
		UnimplementedISR,                 // vector 53, 0xFF94 - Reserved
		UnimplementedISR,                 // vector 52, 0xFF96 - Reserved
		UnimplementedISR,                 // vector 51, 0xFF98 - Reserved
		UnimplementedISR,                 // vector 50, 0xFF9A - Reserved
		UnimplementedISR,                 // vector 49, 0xFF9C - Reserved
		UnimplementedISR,                 // vector 48, 0xFF9E - Reserved
		UnimplementedISR,                 // vector 47, 0xFFA0 - Reserved
		UnimplementedISR,                 // vector 46, 0xFFA2 - Reserved
		UnimplementedISR,                 // vector 45, 0xFFA4 - Reserved
		UnimplementedISR,                 // vector 44, 0xFFA6 - Reserved
		UnimplementedISR,                 // vector 43, 0xFFA8 - Reserved
		UnimplementedISR,                 // vector 42, 0xFFAA - Reserved
		UnimplementedISR,                 // vector 41, 0xFFAC - Reserved
		UnimplementedISR,                 // vector 40, 0xFFAE - Reserved              
		CanTxIsr,                         // vector 39, 0xFFB0 - CAN transmit
		CanRxIsr,                         // vector 38, 0xFFB2 - CAN receive
		CanRxIsr,                         // vector 37, 0xFFB4 - CAN errors
		UnimplementedISR,                 // vector 36, 0xFFB6 - CAN wake-up
		UnimplementedISR,                 // vector 35, 0xFFB8 - FLASH
		UnimplementedISR,                 // vector 34, 0xFFBA - Reserved
		UnimplementedISR,                 // vector 33, 0xFFBC - Reserved
		UnimplementedISR,                 // vector 32, 0xFFBE - Reserved
		UnimplementedISR,                 // vector 31, 0xFFC0 - Reserved
		UnimplementedISR,                 // vector 30, 0xFFC2 - Reserved
		UnimplementedISR,                 // vector 29, 0xFFC4 - CRG self clock mode
		UnimplementedISR,                 // vector 28, 0xFFC6 - CRG PLL Lock
		UnimplementedISR,                 // vector 27, 0xFFC8 - Reserved
		UnimplementedISR,                 // vector 26, 0xFFCA - Reserved
		UnimplementedISR,                 // vector 25, 0xFFCC - Reserved
		UnimplementedISR,                 // vector 24, 0xFFCE - Port J
		UnimplementedISR,                 // vector 23, 0xFFD0 - Reserved
		UnimplementedISR,                 // vector 22, 0xFFD2 - ATD
		UnimplementedISR,                 // vector 21, 0xFFD4 - Reserved
		ISR_SCI_Comm,                     // vector 20, 0xFFD6 - SCI     
		UnimplementedISR,                 // vector 19, 0xFFD8 - SPI
		UnimplementedISR,                 // vector 18, 0xFFDA - Pulse accumulator input edge
		UnimplementedISR,                 // vector 17, 0xFFBC - Pulse accumulator A overflow
		Timer_Overflow_ISR,               // vector 16, 0xFFDE - Standard Timer Overflow
		VPW_transmit_ISR,	              // vector 15, 0xFFE0 - Standard Timer Channel PT7
		UnimplementedISR,                 // vector 14, 0xFFE2 - Standard Timer Channel PT6
		ISS_timer,                        // vector 13, 0xFFE4 - Standard Timer Channel PT5
		UnimplementedISR,                 // vector 12, 0xFFE6 - Standard Timer Channel PT4
		UnimplementedISR,                 // vector 11, 0xFFE8 - Standard Timer Channel PT3
		UnimplementedISR,                 // vector 10, 0xFFEA - Standard Timer Channel PT2
		UnimplementedISR,                 // vector 09, 0xFFEC - Standard Timer Channel PT1
		VSS_timer,                        // vector 08, 0xFFEE - Standard Timer Channel PT0
		Timer_Clock_ISR,                  // vector 07, 0xFFF0 - Real Time Interrupt
		UnimplementedISR,                 // vector 06, 0xFFF2 - IRQ (PE1)
		UnimplementedISR,                 // vector 05, 0xFFF4 - XIRQ (PE0)
		UnimplementedISR,                 // vector 04, 0xFFF6 - SWI
		UnimplementedISR,                 // vector 03, 0xFFF8 - not used
		UnimplementedISR,                 // vector 02, 0xFFFA - COP failure reset
		UnimplementedISR,                 // vector 01, 0xFFFC - Clock monitor failure reset
		_Startup                          // Reset vector, 0xFFFE
};


// ******************************** FUNCTION PROTOTYPES **********************************   
//
// Functions must be 'prototyped' before they are used. A function prototype
// looks like this:
//
//    void function_name(variable_type variable_name1, variable_type variable_name2, ...);
//
// where:
//       - the first item is the type of the value the function returns 
//         ('void' means the function does not return a value), 
//       - the function_name must be unique and not one of the reserved words,
//       -  a list of the variable passed (preceded by their data type) is 
//          enclosed is round brackets. 
//
// The names in the function must match those in the function itself, but not necessarily those
// that are passed from the main program (or other function).
//
// Note that the variables passed to the function must match the variable type 
// (int, char, unsigned char, etc.) listed in the function prototypes (below).

#pragma CODE_SEG NON_BANKED

#ifdef COMMS_382X
  void switch_page(unsigned char sub_no,char carg1,char carg2);
#else
  void switch_page(unsigned char sub_no);
#endif

// #pragma CODE_SEG ROM_7000
#pragma CODE_SEG NON_BANKED

int intrp_1dctable(char sgnx, int x, unsigned char n, int * x_table, char * z_table);

int intrp_speedo_table(unsigned int x, unsigned int * x_table, int * z_table);
			unsigned char intrp_1d_pwm(unsigned int x, unsigned int * x_table, unsigned char * z_table);

int intrp_2dctable(unsigned int x, int y, unsigned char nx, unsigned char ny,
				unsigned int * x_table, int * y_table, unsigned char * z_table);

unsigned char intrp_2dcPWM(unsigned char nx, unsigned char ny, signed int x, int y, signed int * x_table, signed int * y_table, unsigned char * z_table);

unsigned char auto_gear_lookup(int x, int y, unsigned char nx, unsigned char ny,
				unsigned int * x_table, int * y_table, unsigned char * z_table); 

void blink_gpo3(int flashCount);
void Output3(unsigned char set_to);
void grabMS2vars(void);

#pragma CODE_SEG MAIN_ROM

signed char shiftTo(signed char new_gear);	 	// in 4.000_changes.h
unsigned char clutchOffset(unsigned char gear);	// in 4.000_changes.h
signed int CANretard(char new_gear);			// in 4.000_changes.h
unsigned char afterShiftRevs(char new_gear);	// in 4.000_changes.h
void resetShiftRequests(void);					// in 4.000_changes.h
unsigned int solOffset(signed char gear);   	// in 4.000_changes.h
signed char getGear(void);						// in 4.000_changes.h
void setLEDs(signed char gear);					// in 4.000_changes.h

#pragma CODE_SEG ROM_7000

#pragma CODE_SEG MAIN_ROM

void setOut3PWM(signed char gear);	 			// in 4.000_changes.h
void PNRset_outputs(unsigned char offset);		// in 4.000_changes.h
signed char getLeverDigital(unsigned char num_inputs, unsigned int debounce); // in 4.000_changes.h/getLever.h
signed char getLeverADC(unsigned int input); 	// in 4.000_changes.h/getLever.h
void SetSparePort(unsigned char ix); 			// in 4.000_changes.h/sparePorts.h 
unsigned int odometer(unsigned int newSpeed);	// in 4.000_changes.h/sparePorts.h 
void SPI_send8bits(char cx);
void SPI_send16bits(int cx);
void SPI_sendString(char *ptr);
char SPI_getByte(void);
int SPI_getInt(void);
void SPI_getString(char *ptr, char count);
char setUp_SPI(char device);
unsigned int calcSPIpwm(unsigned int Hz, char PWMpin, unsigned char PWMduty);
signed char checkLever(signed char gear, signed char last_gear);


#pragma CODE_SEG DEFAULT

// --- Table Includes and Table Size Defines ---
// The '#include' preprocessor directive directs the compiler to process 
// external header files.

#include "cltfactor.inc"    // cltfactor.inc defines a 1024 x 1 lookup table.
							// we set the ADC level to be ten bits later: 
							// 2 = 0000 0100 0000 0000 = 1024 = 0x0400 
							// Each value corresponds to one ADC increment, giving the temperature*10 in Fahrenheit 
							// at that ADC value. (if Metric_Units is set, the value is converted to Celsius immediately)
							// The ADC start at 0 Volts for ADC count value of 0, and run to 5.00 Volts at ADC 1024
							// (so each ADC represent a difference of 5.00/1024 or 0.00488 Volts/count)
									
#include "linefactor.inc"   // linefactor.inc defines a 1024 x 1 lookup table that turns a
							// 0 to 5 volt signal into transmission line pressure from 0 to 
							// 500 psi (990 = signal output out-of-range low, 999 = signal 
							// output out of range high)
													
#include "loadfactor.inc"   // loadfactor.inc defines a 1024 x 1 lookup table that takes an  
							// ADC count, and turns it into a 0 to 5 Volt 'load' signal
							// The input can be MAP, TPS, or MAF (used for non-CAN MAP)

#ifdef COMMS_382X
  #pragma NON_BANKED
  const unsigned long crc_tab[256] = {
	0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419, 0x706af48f,
	0xe963a535, 0x9e6495a3,	0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
	0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91, 0x1db71064, 0x6ab020f2,
	0xf3b97148, 0x84be41de,	0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
	0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec,	0x14015c4f, 0x63066cd9,
	0xfa0f3d63, 0x8d080df5,	0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
	0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,	0x35b5a8fa, 0x42b2986c,
	0xdbbbc9d6, 0xacbcf940,	0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
	0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423,
	0xcfba9599, 0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
	0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,	0x76dc4190, 0x01db7106,
	0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
	0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x086d3d2d,
	0x91646c97, 0xe6635c01, 0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
	0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6, 0x12b7e950,
	0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
	0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7,
	0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
	0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9, 0x5005713c, 0x270241aa,
	0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
	0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81,
	0xb7bd5c3b, 0xc0ba6cad, 0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
	0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683, 0xe3630b12, 0x94643b84,
	0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
	0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb,
	0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
	0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5, 0xd6d6a3e8, 0xa1d1937e,
	0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
	0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55,
	0x316e8eef, 0x4669be79, 0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
	0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe, 0xb2bd0b28,
	0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
	0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9, 0xeb0e363f,
	0x72076785, 0x05005713, 0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
	0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21, 0x86d3d2d4, 0xf1d4e242,
	0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
	0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69,
	0x616bffd3, 0x166ccf45, 0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
	0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db, 0xaed16a4a, 0xd9d65adc,
	0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
	0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693,
	0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
	0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
  };
#endif
#pragma CONST_SEG DEFAULT
													
#define NUM_TABLES  16      // tables are cltfactor, sprfactor, loadfactor, 
							// in1flash, in2flash, in3flash, txbuf, outpc (+blank tables)
#define NO_TBLES NUM_TABLES // for serComm.h compatibility
							
#define NUM_LOADS 	9       // Number of loads in shift table and pressure control table
#define NUM_MPH  	16	    // Number of MPH in shift table auto gear table
#define NUM_TEMPS 	10      // number of temps in temperature table(s)
#define NUM_PWM_SP  12      // Number of spare port PWM values
#define NUM_GEARS    9      // Number of Gears (1-4 forward gears), 0 for neutral/reverse
#define NUM_WAITS    5      // Number of independent wait counters.
							// Value is used to declare delay_count[] array size, 
							// as well as the waiting[] array, and is 
							// updated in wait decrementing for loop (in 0.010 second timer).     

// CAN and Serial defined values
#define uint8_t     uchar
#define	MSG_CMD	        0
#define	MSG_REQ	        1
#define	MSG_RSP	        2
#define	MSG_XSUB	    3
#define	MSG_BURN	    4
#define	NUM_VAR_BLKS   16
#define NUM_CANMSG     10
#define MAX_CANBOARDS  16
#define SER_TOUT        5    // Serial time out in seconds (if corrupt data received)
#define MAX_BAUD   921600
#define MIN_BAUD      300

// Error status words: 
//    -bits 0-7 are current errors
//    -bits 8-15 are corresponding latched errors
#define	XMT_ERR		   	0x0101
#define	CLR_XMT_ERR		0xFFFE
#define	XMT_TOUT	  	0x0202
#define	CLR_XMT_TOUT	0xFFFD
#define	RCV_ERR		  	0x0404
#define	CLR_RCV_ERR		0xFFFB
#define	SYS_ERR		    0x0808

// Inpram (INPut RAM definitions)
#include "inpram.h"

//***** RAM copy of inputs *********************************************
// This is where the RAM variables (copied from flash memory) get 
// their inpram., inp2ram, and in3ram. prefixes (inpram = INPut RAM)

inputs1 inpram;

inputs2 in2ram;

inputs3 in3ram;

// Outmsg structure
// Note: *** Must set same outmsg data in ECU and CAN device

#include "outmsg.h"
							
#pragma ROM_VAR DEFAULT

#pragma ROM_VAR OVF_ROM

// array of TC overflow numbers at which to increment seconds counter.
const unsigned int TC_ovfla[60] = {
23,
46,
69,
92,
114,
137,
160,
183,
206,
229,
252,
275,
298,
320,
343,
366,
389,
412,
435,
458,
481,
504,
526,
549,
572,
595,
618,
641,
664,
687,
710,
732,
755,
778,
801,
824,
847,
870,
893,
916,
938,
961,
984,
1007,
1030,
1053,
1076,
1099,
1122,
1144,
1167,
1190,
1213,
1236,
1259,
1282,
1305,
1328,
1350,
1373
};

// --- Revision Number and Signature ---
#include "revision.h"

#include "globals.h"

#ifdef COMMS_382X
	// 256 uniformly distributed (0-100) random integers
	const unsigned char rand_no[256] = {
	58, 84, 62, 39, 51, 13, 63, 56, 51, 21, 
	31, 11, 1, 15, 53, 23, 10, 79, 66, 98, 
	25, 84, 43, 99, 45, 50, 0, 61, 59, 88, 
	34, 20, 13, 73, 0, 26, 50, 50, 50, 77, 
	68, 42, 21, 44, 0, 81, 88, 38, 23, 0, 
	31, 68, 31, 43, 0, 16, 65, 44, 56, 77, 
	91, 23, 82, 64, 70, 2, 8, 21, 1, 8, 
	1, 68, 53, 37, 47, 71, 68, 23, 24, 63, 
	30, 50, 62, 75, 75, 87, 17, 4, 22, 40, 
	3, 86, 6, 9, 70, 39, 60, 4, 10, 62, 
	1, 31, 63, 82, 21, 92, 18, 49, 31, 65, 
	49, 22, 62, 30, 72, 4, 7, 33, 99, 60, 
	30, 90, 53, 41, 64, 17, 5, 49, 94, 50, 
	72, 91, 14, 34, 14, 19, 67, 96, 55, 66, 
	73, 7, 13, 54, 7, 55, 95, 25, 28, 85, 
	73, 29, 52, 78, 16, 26, 77, 85, 97, 81, 
	99, 81, 95, 74, 79, 33, 87, 7, 43, 5, 
	87, 83, 83, 20, 99, 36, 70, 23, 14, 18, 
	17, 24, 53, 31, 97, 57, 66, 56, 86, 65, 
	28, 15, 37, 64, 21, 27, 20, 83, 46, 27, 
	96, 10, 73, 85, 1, 85, 10, 91, 60, 28, 
	70, 52, 60, 60, 94, 23, 86, 10, 12, 27, 
	82, 18, 32, 30, 64, 27, 65, 81, 7, 0, 
	92, 62, 2, 3, 63, 59, 99, 5, 20, 6, 
	49, 62, 0, 43, 97, 84, 83, 53, 48, 65, 
	63, 48, 45, 76, 1, 8};
#endif

// #pragma ROM_VAR DEFAULT

// CAN clocks for auto transmitting data between ECU and GPIO
unsigned short xrate_clks[MAX_OUTMSGS];     // 0.128 tics

// CAN variables
#ifdef COMMS_382X
  #include "..\..\serCANComm_382x\serCAN3.h"
#else
  #include "..\..\serCANComm_1\serCAN3.h"
#endif

unsigned char sentSpkAdj;

// pointers for spare port pins
volatile unsigned char *pPTMpin[8], *pPTTpin[8];

// allocate space in ram for flash burner core
volatile unsigned char RamBurnPgm[36]; 

// include the outpc. structure from the file
#include "outpc.h"

// Create a buffer variable to hold the MS-II output variables 
// received over CAN for use by GPIO
// unsigned char msvar[NUM_MSVAR_BYTES];
unsigned char msvar[512];

#pragma ROM_VAR OVF_ROM

// Define a structure for the table descriptor that gives the address in RAM, the address 
// in Flash (ROM), and the number of bytes
typedef struct {
unsigned int *addrRam;
unsigned int *addrFlash;
unsigned int  n_bytes;
unsigned int  fpage;
} tableDescriptor;

// Use the tableDescriptor structure we just defined to create an array of table descriptions 
// sized NUM_TABLES. They will hold the descriptions of the tables (and the indexes are used 
// in TunerStudio to 'locate' the table with a command like a06, which refers to the 6th table). 
// These are constant. That is, the descriptions - memory locations - of the tables are constant, 
// even if the elements of the tables change values.

const tableDescriptor tables[NUM_TABLES] =  {
  //  RAM copy                    FLASH copy                     Table Size 
  {  NULL,                   (unsigned int *)cltfactor_table,  sizeof(cltfactor_table),    0x3C },  // table  0
  {  NULL,                   (unsigned int *)linefactor_table, sizeof(linefactor_table),   0x3C },  // table  1 
  {  NULL,                   (unsigned int *)loadfactor_table, sizeof(loadfactor_table),   0x3C },  // table  2 
  {  NULL,                    NULL,                            0,                          0x3C },  // table  3
  { (unsigned int *)&inpram, (unsigned int *)&in1flash,        sizeof(inputs1),            0x3C },  // table  4 
  { (unsigned int *)&in2ram, (unsigned int *)&in2flash,        sizeof(inputs2),            0x3C },  // table  5
  { (unsigned int *)&txbuf,   NULL,                            sizeof(txbuf),              0x3C },  // table  6
  { (unsigned int *)&outpc,   NULL,                            sizeof(outpc),              0x3C },  // table  7
  { (unsigned int *)&in3ram, (unsigned int *)&in3flash,        sizeof(inputs3),            0x3C },  // table  8
  { (unsigned int *)msvar,    NULL,                            sizeof(msvar),              0x3C },  // table  9  
  {  NULL,                    NULL,                            0,                          0x3C },  // table 10
  {  NULL,                    NULL,                            0,                          0x3C },  // table 11
  {  NULL,                    NULL,                            0,                          0x3C },  // table 12
  {  NULL,                    NULL,                            0,                          0x3C },  // table 13
  {  NULL,                   (unsigned int *)&Signature,       sizeof(Signature),          0x3C },  // table 14
  {  NULL,                   (unsigned int *)&RevNum,          sizeof(RevNum),             0x3C }   // table 15
};

// In the above, the "&" before a variable name means "at the address of, 
// so &inpram is the address of the inpram table

#pragma ROM_VAR DEFAULT  // Codewarrior generates error message here, is seemingly OK though

// Calculate and define values for working with tables based on the tables array we just created.
#define tableInit(iTable) (void)memcpy(tables[iTable].addrRam, tables[iTable].addrFlash, tables[iTable].n_bytes)
#define tableByteRam(iTable, iByte)   ((unsigned char *)tables[iTable].addrRam + iByte)
#define tableWordRam(iTable, iWord)   (tables[iTable].addrRam + iWord)
#define tableByteFlash(iTable, iByte) ((unsigned char *)tables[iTable].addrFlash + iByte)
#define tableWordFlash(iTable, iWord) (tables[iTable].addrFlash + iWord)
#define tableBytes(iTable)            (tables[iTable].n_bytes)
#define tableWords(iTable)            ((tables[iTable].n_bytes+1)/2) // Round up
#define tableFPage(iTable)            ((unsigned char)tables[iTable].fpage)

#pragma CODE_SEG MAIN_ROM

// Function prototypes - Note: ISRs (interrupt service routines) are prototyped above.
// These are the mandatory function prototypes.
void main(void); // the main function, i.e., where the program starts
char tcc_lock (void); 
char tcc_unlock (void);
char Calc_Initial_LUF(void);
char Calc_LUF_PWM(void);
void reset_gear (void);
void reset_switches(void);

// #include "SPI_code.h"

// some global variables
unsigned char first_instr;
short gpio_ms;
short FuelAdjl,SpkAdjl,IdleAdjl,SprAdjl; 	// Adjustments for fuel(%x10 +/- to ve),
											// spark(degx10),idle step,spare 
											// for GPIO boards. These are previous values, 
											// current are in outpc.
										
#pragma CODE_SEG DEFAULT
void reset(void);
void VSS_reset(void);
signed int get_adc(char chan1); 				// this is the function that reads the ADC channels
void fburner(unsigned int* progAdr, unsigned int* bufferPtr, 
  					 unsigned int no_words); 	// fburner(pointer to destination, pointer to source, number of words)
void Output1(unsigned char);
void Output2(unsigned char);
void Output4(unsigned char);
void Output5(unsigned char);
void Output6(unsigned char);
void Output7(unsigned char);
void Output8(unsigned char);
void Output9(unsigned char);
void spr0(unsigned char);
void spr1(unsigned char);
void spr2(unsigned char);
void spr3(unsigned char);
void Flash_Init(unsigned long oscclk);
void Flash_Erase_Sector(unsigned int *address);
void Flash_Write_Word(unsigned int *address, unsigned int data);


#ifdef  COMMS_382X
  #pragma CODE_SEG NON_BANKED
  void crc32(uchar,uchar,uchar);
#endif

#pragma CODE_SEG DEFAULT

extern void reboot(void);      			// 'extern' means that the code can call the function 
										// defined in another module directly if necessary
extern void monitor(void);
extern void SpSub(void);
extern void NoOp(void);

// ---------------------------------------------------------------------------------------
// ---------------------------------- Function MAIN --------------------------------------
// ---------------------------------------------------------------------------------------

#pragma CODE_SEG MAIN_ROM // Put following code in MAIN_ROM == ppage 0x3C

void main(void) {
// This is where the program actually starts. Everything before this was mainly telling the 
// compiler what and where to put values in memory when we load the code using the serial 
// monitor. So the values above exist in the S19 file we will make (that set up both the 
// program and the defaults) but the above doesn't affect the program on a restart of 
// the code (which is why the flash memory retains its values).

// First set-up variables, etc., then go into an infinite loop
int ix;
unsigned char i;
long ltmp;
signed char tmp1_U08;          // intermediate status of spare port 1 and 2
                               // also used for slip based spark and fuel adjustments
signed int tmp2_S16;           // temp signed int for slip calcs                               
unsigned long ultmp, ultmp1, ultmp2;
unsigned char SP3_table_value, tempPC1;
long tempPC2;                  // temp variable to hold PC solenoid duty cycle while we adjust and error check it
unsigned int scount, lcount;   // Loop counts for LOAD
char lock_TCC = 0;             // set TCC unlocked
int rev_distance;              // tire revs/mile or revs/km

// Set-up processor (I/O, etc.)
#include "InitProcessor.h"

// --------------------- MegaShift ----------------------------------------------------
// Initialize MShift variable values to 'safe' state
#include "InitMShift.h"

// MShift MAIN LOOP starts here ---------------------------------------------------------------------------
// Start an infinite FOR loop to keep cycling through the main loop

MAIN_LOOP:
for (;;)  { // start main loop

// Bound target gear
if (outpc.target_gear > inpram.num_gears)
	{
	outpc.target_gear = inpram.num_gears; // set to 'D'
	}

// Bound manual gear
if (outpc.manual_gear > inpram.num_gears)
	{
	outpc.switches &= 0x80; // set flag for datalog
	outpc.manual_gear = inpram.num_gears; // set to 'D'
	}
	
// Set modes
switch (inpram.NAGmode)
	{
	case 0: // no mode
	// do nothing
	break;
	
	case 1: // NAG mode
	in2ram.ignoreLever = 1;
	break;
		
	case 2: // 4L60e
	// set the relevant parameters:
	inpram.Output1ptrn = 79;	 // 1001111
	inpram.Output2ptrn = 31;	 // 0011111
	inpram.Output3ptrn = 96;	 // 1100000
	inpram.Inputptrn[0] = 115;	 // 1110011
	inpram.Inputptrn[1] = 56;	 // 0111000
	inpram.Inputptrn[2] = 31;	 // 0011111
	in2ram.ignoreLever = 0;		 // Use inputs
	inpram.num_teeth = 40;		 // VSS teeth
	inpram.gear_table[0] = 2290; // reverse (x1000)
	inpram.gear_table[1] = 3060; // first   (x1000)
	inpram.gear_table[2] = 1630; // second  (x1000)
	inpram.gear_table[3] = 1000; // third   (x1000)
	inpram.gear_table[4] = 700;	 // fourth  (x1000)
	// in2ram.inter_ratio = 1000;	 // no additional gear - what about quick change rear-ends?
	inpram.VSS_location = 0;	 // on output shaft
	
	break;
		
	case 3: // 4L80e
	inpram.Output1ptrn = 79;
	inpram.Output2ptrn = 96;
	inpram.Output3ptrn = 96;
	inpram.Inputptrn[0] = 115;
	inpram.Inputptrn[1] = 56;
	inpram.Inputptrn[2] = 31;
	in2ram.ignoreLever = 0;		 // Use inputs
	inpram.num_teeth = 40;
	inpram.iss_divider = 31;
	inpram.gear_table[0] = 2077;
	inpram.gear_table[1] = 2482;
	inpram.gear_table[2] = 1482;
	inpram.gear_table[3] = 1000;
	inpram.gear_table[4] = 750;
	inpram.VSS_location = 0;	 // on output shaft
	break;
	
	}

// check for Current Gear status error
if ((outpc.current_gear < 1) && (outpc.manual_gear > 0) && (inpram.leverErr == 0)) 
	{
	outpc.current_gear = shiftTo(outpc.target_gear);
	}

// do the Timer details (will also do in delay loops)
last0001 = Timer0001();	 // set millisecond countdown timer
last0010 = Timer0010();
lastADC  = TimerADC();
last0100 = Timer0100();
last1000 = Timer1000();

// Send whole port status' to datalog 
outpc.porta   = PORTA;
outpc.porte   = PORTE;

// check the CANbus flag
checkCANbus();

// set datalog values for current test states
outpc.testing_inputs = in2ram.test_mode;

// Set period for VSS/ISS test modes
if ((in2ram.test_mode & 0x01) && (in2ram.test_mode & 0x20)) // if testing VSS
	{
	for (i=1; i <21; i++)
		{
		// Set speed from 0 to 102.3 mph
		ic1_period[i] = (int) (((53*(vss_divide/inpram.num_teeth))+27)/(in2ram.VSSdbug * 10));
		restart = 1; // so that speedo will work when we go back to normal input
		}  // end for (i=1; i <21; i++) 
	} // end if ((in2ram.test_mode & 0x01) ...
	
if ((inpram.ic2_usage != 0) && (in2ram.test_mode & 0x01) && (in2ram.test_mode & 0x40)) // if testing ISS
	{
	for (i=1; i <21; i++)
		{
		for (i=1; i <21; i++)
			{
			// There are 65,536 tics in 0.349 seconds, or 0.000005333 sec/tic (= 187500 tics/sec)
			// tics = (187500 tics/sec) / (revs/minute * 1 minute/60 seconds  * tach events/rev)
			// tics = (187500*60 tics/min) / (revs/minute  * tach events/rev)
			// but our rpm (VSSdbug) is /50, so:
			//
			// tics = (187500*60 tics/min) / ((VSSdbug * 50)  * tach events/rev)
			//
			// collecting the numeric factors =>
			//
			// tics = (225000 tics/min) / (VSSdbug  * tach events/rev)
			//
			// Ex. 8 cyl tach at 6000 rpm (VSSdbug would be 6000/50 = 120) 
			//			= 100 rev/sec = 400 tach events/sec 
			//			= 2.5msec/event
			//			= 2500/5.3333 = 469 tics/event
			//			== 225000(120*4)
			//
			// and 600 rpm is  4688 tics/event
			// (4 cyl is twice as many tics)
			// 
			// ic2_period[i] = ((200/(inpram.iss_divider)) * (1125/(in2ram.ISSdbug)));
			
			if ((in2ram.ISSdbug > 20) && (inpram.iss_divider > 20))	 // will be dividing by at least 400
				{
				ic2_period[i] = ((400/(inpram.iss_divider)) * (2250/(in2ram.ISSdbug)))/4;
				}
				else if ((in2ram.ISSdbug < 8) && (inpram.iss_divider < 8)) // will be dividing by no more than 49
				{
				ic2_period[i] = ((50/(inpram.iss_divider)) * (281/(in2ram.ISSdbug)))*16;
				}
				else
				{
				ic2_period[i] = ((100/(inpram.iss_divider)) * (2250/(in2ram.ISSdbug)));
				}	
				
			// ISS period = 60/rpm * teeth/rev * 1000/5.333
			// Ex. ISS with 32 teeth at 6000 rpm is
			// 			= 100rev/sec*32teeth/rev = 3200 teeth/sec
			// 			= 3.125msec/tooth = 
			//			= 59 tics/tooth
			// 
			// Ideally we would like the tics to go from 59 to 9000
			// So 9000/255 = 35
			}  // end for (i=1; i <21; i++)	
		}   
	} // end if (in2ram.test_mode & 0x01) ...	

outpc.mloop++;  // incement loop counter

// Determine decel mode
if  ((in2ram.decel_Load) && !(inpram.shift_mode == 3) && (outpc.speedo > inpram.min_dcl_spd))  // not if RPM based shifting
	{
	// decel mode allowed
	if ((outpc.speedo < last_decel_mph) && (outpc.rawTPS < inpram.decel_TPS))
		{
		outpc.decel_flag = in2ram.decel_Delay; // will decrement in 10ths section of timerISR
		}
		else if (outpc.speedo > last_decel_mph)
		{
		outpc.decel_flag = 0;  // clear the flag if not decelerating
		}
	}
	else  // no decel mode
	{
	outpc.decel_flag = 0;  // clear the flag
	} // end else if  ((in2ram.decel_Load) && !(inpram.shift_mode == 3)

// Update Odometer
outpc.odo = outpc.odo + odometer(outpc.speedo);

// if using a custom pressure sensor, derive values from 2 points
if (inpram.line_units == 3)
	{
	// y = mx + b
	// x is volts, y is pressure
	// volt1/2 in Volts in TS, but in ADC in code
	// press1/2 is PSI (or any other consistent unit)
	// slope m = rise/run
	Pslope = ((10*inpram.press2 - 10*inpram.press1) + ((inpram.volt2 - inpram.volt1)/2))/(inpram.volt2 - inpram.volt1);	
	// slope x 100 * 33.333, 
	// because press *33.333, PSI is x10, added factor of 10 
	// (otherwise value can be low single digits)
	
	// b = y - mx
	// y - b = press2 - (Pslope * volt2)
	Pintcpt = (10*inpram.press2 - (Pslope * inpram.volt2));
	}

// Send gear retard timing and slip timing/fuel adjustments to MS-II
if ((inpram.CAN_enabled) && (outpc.current_gear > 1)) 
  { // outpc.SpkAdj = CANretard(outpc.current_gear);
  tmp2_S16 = 0;
	if (
	  	  (in2ram.SlipOptn & 0x01) &&  // ISS enabled
	      ((inpram.ic2_usage & 0x01) == 1) && 
	      (outpc.LOAD >= inpram.Sretard_load) && 
	      ((outpc.slip > inpram.Slip_threshold)) &&
	      // if slip is line than spark, wait until delay has expired
	      (((inpram.SlipConditions == 1) || (inpram.SlipConditions == 2) || (inpram.SlipConditions == 3) || ((inpram.SlipConditions == 5 )) && (Sflag == 1)))
	      )
	  	{ // adjust value and set flag
	  	tmp2_S16 = inpram.Slip_adv;  			// adjust advance
	  	outpc.FuelAdj = inpram.Slip_fuel; 	// adjust fuel (this is  currently the only fuel adjustment)
	  	if (Sflag == 0) Sflag = inpram.slipDelay;  		// set slip flag if we haven't already  - will count down to 1 in timer
	  	}
	  	else
	  	{
	  	tmp2_S16 = 0;        // clear the adjusted advance value
	  	outpc.FuelAdj = 0; // clear the adjusted fuel value
	  	Sflag = 0;
	  	}

  // add timing adjustment to gear retard	
  if (outpc.LOAD > inpram.Tretard_load)
  	{ // use slip and gear retard
    outpc.SpkAdj = CANretard(outpc.current_gear) + tmp2_S16;
  	}
  	else
  	{ // use only slip retard
  	outpc.SpkAdj = tmp2_S16;
  	}
  	
  // send the value if in a new gear or slip factors have been set
  if ((sentSpkAdj != outpc.current_gear) || (Sflag == 1)) 
  	{
  	sendCANAdj();
  	sentSpkAdj = outpc.current_gear;
  	}
    else // same gear and Sflag not set
   	{
    if (sentSpkAdj != 0) 
    	{  // if haven't already set a zero to clear ECU, send now
    	sendCANAdj();
    	sentSpkAdj = 0;
    	Sflag = 0; // clear slip flag
    	}    	
   	} // end else if ((sentSpkAdj
  } // end else if CAN_enabled
  else  // non-CAN or below 2nd gear
  {
  outpc.SpkAdj = 0;
  if (sentSpkAdj != 0) 
  	{
  	if (inpram.CAN_enabled) sendCANAdj();  // below second gear
   	sentSpkAdj = 0;
   	}
  }// end if (inpram.CAN_enabled) 

// Check to see if forced shift needs to be cleared
if ((inpram.rpm_check) && (rev_shift_flag) && !(shifting_flag))
	{
	if ((outpc.engine_rpm < in2ram.rpm_limit[outpc.current_gear - 1]) && (outpc.engine_rpm > in2ram.under_rev_limit[outpc.current_gear - 2]))
		{
		rev_shift_flag = 0; 		// clear flag
		outpc.miscStatus &= ~0x18;	// clear forced shift bits for datalog
		}
	}  // end if ((inpram.rpm_check) && (rev_shift_flag))

if ((inpram.rpm_check) && (!rev_shift_flag) && !(shifting_flag))
	{ // rpm checking enabled but rev forced shift not already requested 
	  // here we check if the running rpm has exceed hi or low values
	
	// Check for under rev and downshift
	// (post shift rpm checked later) 	
	if (   ((outpc.manual_gear > inpram.follow_maxgear) || (inpram.follow_mode == 0)) // if not in follow mode
		&& (!(outpc.shift_action & 0x04))  // AND not already downshift request
		&& ((outpc.current_gear > 1) && (outpc.current_gear <= inpram.num_gears)) // AND current gear is valid
		&& ((inpram.forceAUTOoptn == 1) || (inpram.forceAUTOoptn == 3) || (inpram.forceAUTOoptn == 5) || (inpram.forceAUTOoptn == 6)) 
		&&  (outpc.engine_rpm <= in2ram.under_rev_limit[outpc.current_gear - 2]) // OR if revs are too low
		) // if revs are too low, request downshift                          
		{
		if (outpc.current_gear > 1) 
			{
			// request downshift immediately
			outpc.shift_action |=  0x04;   // set downshift request 
			outpc.shift_action &= ~0x08;   // clear upshift request
			rev_shift_flag = 2;            // will reset after shift
			outpc.target_gear = outpc.current_gear - 1; // set target gear to current gear plus one for datalog
			outpc.miscStatus |= 0x10; 	   // set forced downshift bit for datalog 
			outpc.shiftStat =  3;
			shiftStatcounter = 5;
			outpc.hyst_timer = 0;		   // clear the hyst flag to allow immediate shift
			shiftSpeedHyst_flag = 0;
			outpc.decel_flag = 0;
			outpc.cruise_flag = 0;
			}
		} // end if UP_avg > DWN_avg ...         

	// Check for over rev and upshift
	// (post shift rpm checked later) 
	if (   ((outpc.manual_gear > inpram.follow_maxgear) && (inpram.follow_mode == 0)) // if not in follow mode 
		&& (!(outpc.shift_action & 0x08))  // not already upshift request
		&& ((outpc.current_gear < inpram.num_gears) && (outpc.current_gear > 0)) // if gear is valid
		&& ((inpram.forceAUTOoptn == 1) || (inpram.forceAUTOoptn == 2) || (inpram.forceAUTOoptn == 5) || (inpram.forceAUTOoptn == 6)) 
		&&  (outpc.engine_rpm >= in2ram.rpm_limit[outpc.current_gear - 1]) // OR revs are too high
		) // if revs are too high, request upshift                          
		{
		if (outpc.current_gear < inpram.num_gears) 
			{
			// request upshift
			outpc.shift_action |=  0x08;   // set upshift request
			outpc.shift_action &= ~0x04;   // clear downshift request
			rev_shift_flag = 1;            // will reset after shift
			outpc.target_gear = outpc.current_gear + 1; // set target gear to current gear plus one for datalog
			outpc.miscStatus |= 0x08; 	   // set forced upshift bit for datalog
			outpc.shiftStat =  3;
			shiftStatcounter = 5;
			outpc.hyst_timer = 0;		   // clear the hyst flag to allow immediate shift
			shiftSpeedHyst_flag = 0;
			outpc.decel_flag = 0;
			outpc.cruise_flag = 0;
			} // end if ((outpc.target_gear <= inpram.num_gears) ...
		} // end if UP_avg > DWN_avg ... 
	}   // end if ((inpram.rpm_check == 1)...

// COMPUTE VEHICLE SPEED
// ---------------------
// We do pre-calculations here and then do interrupt calcs in timer ISR 
// so that the speedo updates regardless of any delays in the code
// (like the shift and pressure delays).

// For imperial units: 

// Vehicle speed = VSS_rpm * tire_diam/(5280*12) * 60 minutes/hour * PI/axle_ratio  (miles/hour)

// delta X is the distance travelled by 20 teeth (in inches x100)
// = tire_distance/rev * 1/axle_ratio * 20/num_teeth
// = PI*tire_diam * 1/axle_ratio * 20/num_teeth == vss_divide
//
// so speedo = (pi*tire_diam * 1/axle_ratio * 20/num_teeth)/(ic_per_acc * constant)
//           = constant * vss_divide/ic_per_acc

// use averaging above at 1.5x minSpeed
tmpVSF = 1; // initialize
if (outpc.speedo < ((inpram.min_speed * 3)/2)) tmpVSF = 1; else tmpVSF = inpram.VSSsmoothFactor;

// speedoFactor is all terms except /(ic1_per_avg)  and /tmpVSF
// speedo factor is /100, so we divide ic1_per_avg by 100 as well 
// vss_divide is the number of inches travel for 20 outputs shaft revs (x100)

if (Inpt4_active || FWD_LogOnly || (outpc.miscStatus & 0x04)) // If not using table switching or pin is high use normal table
		{
		speedoFactor = (unsigned long) ((1420/inpram.num_teeth) * vss_divide);
		// Ex. vss_divide == 53040 in default case
		// so speedoFactor = 1420/40 * 53040 = 35* 53040 = 1856400 
		} 
		else        // if 4WD and speedo switching allowed, divide by speedo factor (x1000)
		{
		if (!Inpt4_active && !FWD_LogOnly) speedoFactor = (unsigned long) ((((1420*vss_divide/inpram.num_teeth))*1000)/inpram.FWD_factor);
		}

// SPARE PORTS
// -----------
// get the spare port index values

// sp0 first
if (in2ram.sp0_mode != 0)
	{ // if not using the spare port as a speedo output 
	switch (in2ram.sp0_9index_mode) 
		{
		case 0:
		  // Load
		  outpc.sp09inx = outpc.LOAD;
		  break;
		case 1:
		  // Speed
		  outpc.sp09inx = outpc.speedo;
		  break;
		case 2:
		  // RPM
		  outpc.sp09inx = outpc.engine_rpm;
		  break;
		case 3:
		  // Temperature
		  outpc.sp09inx = outpc.clt;
		  break;
		case 4:
		  // gear
		  outpc.sp09inx = outpc.current_gear * 10; // because will divide by 10 in function call
		  break;
		case 5:
		  // MAP
		  outpc.sp09inx = outpc.rawMAP;
		  break;
		case 6:
		  // TPS
		  outpc.sp09inx = outpc.rawTPS;
		  break;
		case 7:
		  // Aux channel
		  outpc.sp09inx = outpc.aux_volts;
		  break;
		default:
		  // Use load
		  outpc.sp09inx = outpc.LOAD;
		  break;
		}

	switch (in2ram.sp0_16index_mode) 
		{
		case 0:
		  // Load
		  outpc.sp016inx = outpc.LOAD;
		  break;
		case 1:
		  // Speed
		  outpc.sp016inx = outpc.speedo;
		  break;
		case 2:
		  // RPM
		  outpc.sp016inx = outpc.engine_rpm;
		  break;
		case 3:
		  // Temperature
		  outpc.sp016inx = outpc.clt;
		  break;
		case 4:
		  // gear
		  outpc.sp016inx = outpc.current_gear * 10; // because will divide by 10 in function call
		  break;
		case 5:
		  // MAP
		  outpc.sp016inx = outpc.rawMAP;
		  break;
		case 6:
		  // TPS
		  outpc.sp016inx = outpc.rawTPS;
		  break;
		case 7:
		  // Aux channel
		  outpc.sp016inx = outpc.aux_volts;
		  break;			
		default:
		  // Use load
		  outpc.sp016inx = outpc.LOAD;
		  break;
		}
	}

// sp1 next	
switch (inpram.sp1_index_mode) 
{
case 0:
  // Load
  sp1_indicial = outpc.LOAD;
  break;
case 1:
  // Speed
  sp1_indicial = outpc.speedo;
  break;
case 2:
  // RPM
  sp1_indicial = outpc.engine_rpm;
  break;
case 3:
  // Temperature
  sp1_indicial = outpc.clt;
  break;
case 4:
  // gear
  sp1_indicial = outpc.current_gear * 10; // because will divide by 10 in function call
  break;
case 5:
  // MAP
  sp1_indicial = outpc.rawMAP;
  break;
case 6:
  // TPS
  sp1_indicial = outpc.rawTPS;
  break;
case 7:
  // Aux Volts
  sp1_indicial = outpc.aux_volts;
  break;
default:
  // Use load
  sp1_indicial = outpc.LOAD;
  break;
}

// then sp2
switch (inpram.sp2_index_mode) 
{
case 0:
  // Load
  sp2_indicial = outpc.LOAD;
  break;
case 1:
  // Speed
  sp2_indicial = outpc.speedo;
  break;
case 2:
  // RPM
  sp2_indicial = outpc.engine_rpm;
  break;
case 3:
  // Temperature
  sp2_indicial = outpc.clt;
  break;
case 4:
  // gear
  sp2_indicial = outpc.current_gear * 10; // because will divide by 10 in function call
  break;
case 5:
  // MAP
  sp2_indicial = outpc.rawMAP;
  break;
case 6:
  // TPS
  sp2_indicial = outpc.rawTPS;
  break;
case 7:
  // Aux Volts
  sp2_indicial = outpc.aux_volts;
  break;
default:
  // Use load
  sp2_indicial = outpc.LOAD;
  break;
}

// sp3 last
if (in2ram.sp3_mode != 0)
	{ // if not using the spare port as a line pressure 
	switch (in2ram.sp3_9index_mode) 
		{
		case 0:
		  // Load
		  outpc.sp39inx = outpc.LOAD;
		  break;
		case 1:
		  // Speed
		  outpc.sp39inx = outpc.speedo;
		  break;
		case 2:
		  // RPM
		  outpc.sp39inx = outpc.engine_rpm;
		  break;
		case 3:
		  // Temperature
		  outpc.sp39inx = outpc.clt;
		  break;
		case 4:
		  // gear
		  outpc.sp39inx = outpc.current_gear * 10; // because will divide by 10 in function call
		  break;
		case 5:
		  // MAP
		  outpc.sp39inx = outpc.rawMAP;
		  break;
		case 6:
		  // TPS
		  outpc.sp39inx = outpc.rawTPS;
		  break;
		case 7:
		  // Aux Volts
		  outpc.sp39inx = outpc.aux_volts;
		  break;
		default:
		  // Use load
		  outpc.sp39inx = outpc.LOAD;
		  break;
		}

	switch (in2ram.sp3_16index_mode) 
		{
		case 0:
		  // Load
		  outpc.sp316inx = outpc.LOAD;
		  break;
		case 1:
		  // Speed
		  outpc.sp316inx = outpc.speedo;
		  break;
		case 2:
		  // RPM
		  outpc.sp316inx = outpc.engine_rpm;
		  break;
		case 3:
		  // Temperature
		  outpc.sp316inx = outpc.clt;
		  break;
		case 4:
		  // gear
		  outpc.sp316inx = outpc.current_gear * 10; // because will divide by 10 in function call
		  break;
		case 5:
		  // MAP
		  outpc.sp316inx = outpc.rawMAP;
		  break;
		case 6:
		  // TPS
		  outpc.sp316inx = outpc.rawTPS;
		  break;
		case 7:
		  // Aux Volts
		  outpc.sp316inx = outpc.aux_volts;
		  break;			
		default:
		  // Use load
		  outpc.sp316inx = outpc.LOAD;
		  break;
		}
	}

// If using spare port 2 in gear indicator mode
// Counting and switching done in timer interrupt service routine
    
if (inpram.sp_mode[1] == 5)
	{
	switch (in2ram.gearIndctrMode)
    	{	
		case 0:
			tmp1_U08 = outpc.current_gear;
			break;
	
		case 1:
			tmp1_U08 = outpc.manual_gear;
			break;
	
		default:
			break;
    } // end switch (in2ram.gearIndctrMode)
	
	if (tmp1_U08 > 0)  // forward gears
		{
		// Get total count (in 1/4000th of a second)
		sp2_pwm_total = in2ram.gear_indic_per[tmp1_U08+2];
		// Calc toggle point from PWM percentage
		sp2_pwm_hi = ((in2ram.gear_indic_dc[tmp1_U08+2] * sp2_pwm_total) / 100);
		if ((in2ram.gear_indic_dc[tmp1_U08+2] > 0) && (sp2_pwm_hi == 0)) sp2_pwm_hi = 1; // make sure we get a signal
		}
		else if (tmp1_U08 == -1)	 // reverse
		{ 
		sp2_pwm_total = in2ram.gear_indic_per[2];
		sp2_pwm_hi = ((in2ram.gear_indic_dc[2] * sp2_pwm_total) / 100);
		if ((in2ram.gear_indic_dc[2] > 0) && (sp2_pwm_hi == 0)) sp2_pwm_hi = 1; // make sure we get a signal
		}
		else  // Park or Neutral
		{
		if (parkNeutral == 1) // Park
		 	{
		 	sp2_pwm_total = in2ram.gear_indic_per[0];
		 	sp2_pwm_hi = ((in2ram.gear_indic_dc[0] * sp2_pwm_total) / 100);
		 	if ((in2ram.gear_indic_dc[0] > 0) && (sp2_pwm_hi == 0)) sp2_pwm_hi = 1; // make sure we get a signal
			}
			else if (parkNeutral == 2) // Neutral
			{
			sp2_pwm_total = in2ram.gear_indic_per[1];
			sp2_pwm_hi = ((in2ram.gear_indic_dc[1] * sp2_pwm_total) / 100);	
			if ((in2ram.gear_indic_dc[1] > 0) && (sp2_pwm_hi == 0)) sp2_pwm_hi = 1; // make sure we get a signal
			} // end if (parkNeutral == 1)
		}
	// Update PWM% for datalog
	outpc.sp2PWM = (unsigned char) ((sp2_pwm_hi*100)/sp2_pwm_total);
	}  // end if (inpram.sp_mode[1] == 5)



// Get spare port PWM% from table
// also switched in shiftTo function if spX_active == 1 or 3 or in TCC mode (tcc_lock() and tcc_unlock()) 
// spare port 0 PWM mode
if ((in2ram.sp0_mode == 1) && ((in2ram.sp0_active == 0) || (in2ram.sp0_active == 1))) spr0(1);  // sp0 always ON, or ON except while shifting
if ((in2ram.sp0_mode == 1) &&  (in2ram.sp0_active == 2)) spr0(0); // OFF except while shifting
// spare port 3	PWM mode
if ((in2ram.sp3_mode == 1) && ((in2ram.sp3_active == 0) || ((in2ram.sp3_active == 1) && (outpc.engine_rpm >= 300)))) spr3(1); // sp3 always ON, or ON if RPM greater than 300
if ((in2ram.sp3_mode == 1) && ((in2ram.sp3_active == 1) && (outpc.engine_rpm < 300))) spr3(0); // OFF if RPM less than 300
// spare ports 1 and 2
for (ix = 0; ix < 2; ix++)
	{
	if ((inpram.sp_mode[ix] == 1) && ((inpram.sp_active[ix] == 0) || (inpram.sp_active[ix] == 1))) if (ix == 0) spr1(1); else spr2(1); // sp1 always ON, or ON except while shifting
	if ((inpram.sp_mode[ix] == 1) && (inpram.sp_active[ix] == 2)) if (ix == 0) spr1(0); else spr2(0); // OFF except while shifting
	}  // end for (ix = 0; ...
	
if (outpc.current_gear < 1)	 // if PNR
	{
	if ((inpram.shift_mode > 0) && ((inpram.forceAUTOoptn == 1) || (inpram.forceAUTOoptn == 4))) outpc.auto_mode = inpram.shift_mode;
	// reset shift button counters
	UP_avg   = 500;
	DWN_avg  = 500;
	MODE_avg = 500;	
	}

// Determine if we are using Input4 for lever input
if (inpram.stdin_cfg & 0x08) Inpt4_active = 1; else Inpt4_active = 0; // capture state of bit 3

if ((inpram.stdin_cfg & 0x01) == 0x00) FWD_LogOnly = 1; else FWD_LogOnly = 0; // capture state of bit 0

// Error codes
outpc.error = 0x01; // always set first bit - error code must be odd otherwise 
					// a comms error has occurred                      
/*
bit --  & with  --  meaning
---     ------      -------
0  --   0x01   --  1 = okay, 0 = serial comms error
1  --   0x02   --  0 = okay, 1 = no CAN comms (load = 0.0)
2  --   0x04   --  0 = okay, 1 = under/over rev shift attempted
3  --   0x08   --  0 = okay, 1 = VSS reset occurred
4  --   0x10   --  0 = okay, 1 = other VSS error - no signal, excessive noisy/drop-outs (check tooth error counter)
5  --   0x20   --  0 = okay, 1 = gear out-of-range
6  --   0x40   --  0 = okay, 1 = low voltage (<10)
7  --   0x80   --  0 = okay, 1 = no brake signal  
*/

error_flags = 0x80; // == 10000000 - set brakes flag, will clear when brake signal received

if (inpram.error_check && (outpc.seconds > 30))
	{
	if (inpram.CAN_enabled)
		{
		// CAN comms error
		if (CANrx < 5) outpc.error |= 0x02;   // set CAN error flag
		// Low voltage error
		if (outpc.vBatt < 100) outpc.error |= 0x40; // set low voltage flag
		} // end if (inpram.CAN_enabled)
	} // end if (inpram.error_check...

// VSS error 
if (VSS_error > inpram.vss_error_max) 
{
outpc.error |= 0x10; // set VSS tooth count reset error 
VSS_reset();         // reset VSS
}
	
loop_count++;     // increment main loop counter

outpc.dbug = VSS_error; // set dbug to VSS error count unless used for 
						// logging elsewhere in code

// Capture the highest mph seen since last stop, use to determine if we ought
// to switch to auto_mode (otherwise we would swap immediately on selecting 
// first if stopped)
if (outpc.speedo > high_mph) 
	{
	high_mph = outpc.speedo; // capture the highest mph seen since last stop
	}
	else
	{
	if (outpc.speedo < 10) high_mph = 0; // reset high MPH if have stopped
	}

// Clear COP timeout counter
// When COP is enabled, the program must write 0x0055 and 0x00AA (in this order) to the ARMCOP
// register during the selected time-out period. As soon as this is done, the COP time-out period
// is restarted. If the program fails to do this and the COP times out, the processor will reset. 
// Also, if any value other than 0x0055 or 0x00AA is written, the processor is immediately reset.
// ARMCOP = 0x0055;
// ARMCOP = 0x00AA;	


/*-------------------------- DETERMINE CURRENT GEAR --------------------------------*/
// Determine current gear from state of SolA, SolB, etc.:

if (!inpram.NAGmode == 1) outpc.current_gear = getGear(); // if not NAG mode, check gear, otherwise leave it as set in shiftTo()

// (and set gear indicator LEDs)
setLEDs(outpc.current_gear);

/*-- UPDATE ROLLING LOAD AVERAGE 

The LOAD average is kept on two levels:

- a short term LOAD average (LOAD_short) created by adding the 
difference between the current LOAD_short and measured LOAD to the previous LOAD 
short term average (updated every ), used to smooth 
the LOAD signal, HOWEVER the load averaging is only for decreasing loads

- a long term LOAD average (LOAD_long), used to determine driving mode 
created by adding 1/LOADlongcount of the difference between the 
current LOAD_short and LOAD_long to the previous LOAD_long term average 
(updated every time through the main loop) IF braking is not applied  

The LOAD averages are used to determine the 'aggressiveness' of the driving, 
and hence the shift strategy. Higher values indicate more aggressive driving. 
This is not currently implemented, but the intent is to add a factor to the 
short term load average based on the long term average, perhaps:

LOAD_short = LOADshort + LOAD_long/10;

or possibly a look-up table value based on LOAD_long

--*/

scount++; // increment short count each time through loop

if (scount >= (inpram.LOADshortcount)) 
	{ // skip until reach LOADshortcount
	// reset short counter
	scount = 0;
	// increment long counter
	lcount++; // increment lcount (long count) 

	if (lcount > 65000) lcount = 0; // cycle the lcount value           
	
	// Update short LOAD
	if (!(outpc.miscStatus & 0x01))	  // Not braking
		{ 
		// LOAD_short is kpa x 100
		if (outpc.LOAD < outpc.LOAD_short) // if current LOAD is less than average 
			{
		 	if (outpc.LOAD_short > 0) outpc.LOAD_short -= ((outpc.LOAD_short - (uint)outpc.LOAD)/10) + 1; // lower the average by 1/10th the difference
			// if (outpc.LOAD_short > 0) outpc.LOAD_short--; // lower the average by 1/10th (smoothing factor sets the rate)
			}
			else // if current LOAD is higher than average
			{
			if (outpc.LOAD > 0)
				{
				if (inpram.load_adj & 0x04) 
					{
					outpc.LOAD_short = outpc.LOAD; // jump to current LOAD
					}
					else  // use normal rate
					{
					outpc.LOAD_short += (((uint)outpc.LOAD - outpc.LOAD_short)/10) + 1; // increase the average by 1/10th the difference
					}
				}     
			} // end if ((outpc.LOAD) < ...
		
		// RPM_short is kpa x 100
		if ((outpc.engine_rpm) < outpc.RPM_short) // if current LOAD is less than average 
			{
			if (outpc.RPM_short > 0) outpc.RPM_short -= (outpc.RPM_short-(uint)outpc.engine_rpm)/10; // lower the average by 1/10th the difference
			}
			else // if current RPM is higher than average
			{
			if (outpc.engine_rpm > 0) outpc.RPM_short += ((uint)outpc.engine_rpm-outpc.RPM_short)/10; // increase the average by 1/10th the difference     
			} // end if ((outpc.engine_rpm) < ...
		
		}  // End If brake == 0
		else   // braking
		{
		// LOADshort is kpa x 100
		if ((outpc.LOAD < outpc.LOAD_short)) // if braking and current LOAD is less than average 
			{
			if (inpram.load_adj & 0x01)
				{
				if ((outpc.LOAD_short > 0) && (inpram.brake_load)) outpc.LOAD_short -= (((outpc.LOAD_short - (uint) outpc.LOAD) * (inpram.brake_load))/1000) + 1; // lower the average by 0.1 kPa
				}
				else
				{
				outpc.LOAD_short -= ((outpc.LOAD_short - (uint)outpc.LOAD)/10) + 1; // lower the average by 1/10th the difference
				}
			}
			else // if braking and current LOAD is higher than average
			{
			if (outpc.LOAD > 0)	 // if load is sensible
				{
				if (inpram.load_adj & 0x01) // decrease with braking
					{
					if ((outpc.LOAD_short > 0) && (inpram.brake_load)) outpc.LOAD_short += (((outpc.LOAD_short - (uint) outpc.LOAD) * (inpram.brake_load))/1000); // increase the average by 0.1 kPa
					}
					else
					{
					outpc.LOAD_short += (((uint)outpc.LOAD - outpc.LOAD_short)/10) + 1; // increase the average by 1/10th the difference
					}
				}
			} // end if ((outpc.LOAD) < ...  
			
		// RPMshort is kpa x 100
		if (outpc.engine_rpm > 0) outpc.RPM_short = outpc.engine_rpm; // jump to current LOAD     
	    outpc.cruise_flag = 0;
	    
		}  // end if (outpc.brake == 0 )

	// When long count reached, update LOAD_long from LOAD_short
	if (lcount > (inpram.LOADlongcount-1)) 
		{ // skip until reach LOADlongcount
		// reset long counter
		lcount = 0;
		// Update long LOAD average  
		LOAD_long += (outpc.LOAD_short-LOAD_long)/10;
		} // End If (lcount > LOADlongcount-1)
	}  // End If (scount > (LOADshortcount))

/*-- UPDATE ROLLING SHIFT SWITCH INPUTS 

The switch average is:

- upshift button average (UP_avg) is calculated by adding or subtracting 1 
(depending on the input state) to the previous UP_avg. This is updated 
every time through the main loop (so it is NOT updated during a shift - by design!).

- downshift button average (DWN_avg) is calculated by adding adding or 
subtracting 1 (depending on the input state) to the previous DWN_avg. This is 
updated every time through the main loop.

A shift occurs when the button average exceeds the threshold value - assuming 
'manual mode' is selected, both are not pressed, etc. (this provides a software debounce, 
and sets the 'responsiveness' of the shift command)

--*/

// -------------------- Check Shift Buttons ------------------------------------------------ 
if (inpram.shftbut_mode == 2) // if log only shift buttons
    {
    // set UP button status for datalog
    // DWN button is done in get_adc()
	if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x80))) // not testing paddle up
		{
		if (PADDLE_UP & PADDLE_UPpin)
			{
			outpc.shift_action   |= 0x02;
			}
			else
			{
			outpc.shift_action   &= ~0x02;
			}
		}
		else  // testing paddle up
		{
		if (in2ram.TSTdbug & 0x01) // if Pup bit is high
			{
			outpc.shift_action   |= 0x02;
			}
			else
			{
			outpc.shift_action   &= ~0x02;
			}		
		}
		
	// reset shift button counters
	UP_avg   = 500;
	DWN_avg  = 500;
	MODE_avg = 500;
    }
	else if (inpram.shftbut_mode == 1) // if voltage shift buttons
	{ 
	// set UP button status for datalog
	// DWN button is done in get_adc()
	    // set UP button status for datalog
	    // DWN button is done in get_adc()
		if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x80))) // not testing paddle up
			{
			if (PADDLE_UP & PADDLE_UPpin)
				{
				outpc.shift_action   |= 0x02;
				}
				else
				{
				outpc.shift_action   &= ~0x02;
				}
			}
			else  // testing paddle up
			{
			if (in2ram.TSTdbug & 0x01)
				{
				outpc.shift_action   |= 0x02;
				}
				else
				{
				outpc.shift_action   &= ~0x02;
				}		
			}

		// check no buttons first
		if (outpc.butADC > inpram.no_button_V) 
			{
			if ((outpc.butADC - inpram.no_button_V) < inpram.tol_button_V) 
				{
				if (shift_release > 0) 
					{
					shift_release--;
					goto GOT_BUTTONS;
					}
					else
					{
					if (UP_avg   )   UP_avg--;     // decrement upshift average
					if (DWN_avg  )  DWN_avg--;     // decrement downshift average
					if (MODE_avg > 0) MODE_avg--;  // decrement mode change average 
					outpc.shift_action &= ~0x03;   // clear the shift buttons
					goto GOT_BUTTONS;
					}
				}
			}
			else
			{
			if ((inpram.no_button_V - outpc.butADC) < inpram.tol_button_V) 
				{
				if (shift_release > 0) 
					{
					shift_release--;
					}
					else
					{
					if (UP_avg   )   UP_avg--;     // decrement upshift average
					if (DWN_avg  )  DWN_avg--;     // decrement downshift average
					if (MODE_avg > 0) MODE_avg--;  // decrement mode change average  				
					outpc.shift_action &= ~0x03;   // clear the shift buttons
					goto GOT_BUTTONS;
					}
				}
			}  // done with no buttons
	
		// check both buttons next
		if (outpc.butADC > inpram.both_button_V) 
			{
			if ((outpc.butADC - inpram.both_button_V) < inpram.tol_button_V) 
				{
				MODE_avg++;  					// increment mode change average 
				if (UP_avg   )   UP_avg--;     	// decrement upshift average
				if (DWN_avg  )  DWN_avg--;     	// decrement downshift average
				outpc.shift_action |= 0x03;     // set both shift buttons 
				goto GOT_BUTTONS;
				}
			}
			else
			{
			if ((inpram.both_button_V - outpc.butADC) < inpram.tol_button_V) 
				{
				MODE_avg++;  					// increment mode change average
				if (UP_avg   )   UP_avg--;     	// decrement upshift average
				if (DWN_avg  )  DWN_avg--;     	// decrement downshift average  				
				outpc.shift_action |= 0x03;     // set both shift buttons
				goto GOT_BUTTONS;
				}
			}  // done with both buttons
		
		// check up button next
		if (outpc.butADC > inpram.up_button_V) 
			{
			if ((outpc.butADC - inpram.up_button_V) < inpram.tol_button_V) 
				{
				UP_avg++;     					// increment upshift average
				if (DWN_avg  )  DWN_avg--;     	// decrement downshift average
				if (MODE_avg ) MODE_avg--;  	// decrement mode change average 
				outpc.shift_action |=  0x02;    // set upshift button
				outpc.shift_action &= ~0x01;    // clear downshift button
				goto GOT_BUTTONS;
				}
			}
			else
			{
			if ((inpram.up_button_V - outpc.butADC) < inpram.tol_button_V) 
				{
				UP_avg++;     					// increment upshift average
				if (DWN_avg  )  DWN_avg--;     	// decrement downshift average
				if (MODE_avg > 0) MODE_avg--;  	// decrement mode change average  				
				outpc.shift_action |=  0x02;    // set upshift button
				outpc.shift_action &= ~0x01;    // clear downshift button
				goto GOT_BUTTONS;
				}
			}  // done with up button 	

		// check down button next
		if (outpc.butADC > inpram.dwn_button_V) 
			{
			if ((outpc.butADC - inpram.dwn_button_V) < inpram.tol_button_V) 
				{
				DWN_avg++;     					// increment downshift average
				if (UP_avg   )   UP_avg--;     	// decrement upshift average
				if (MODE_avg ) MODE_avg--;  	// decrement mode change average 
				outpc.shift_action |=  0x01;    // set downshift button
				outpc.shift_action &= ~0x02;    // clear upshift button
				goto GOT_BUTTONS;
				}
			}
			else	
			{
			if ((inpram.dwn_button_V - outpc.butADC) < inpram.tol_button_V) 
				{
				DWN_avg++;     				   	// increment downshift average
				if (UP_avg   )   UP_avg--;     	// decrement upshift average
				if (MODE_avg > 0) MODE_avg--;  	// decrement mode change average  				
				outpc.shift_action &= ~0x02;    // clear upshift button
				outpc.shift_action |=  0x01;    // set downshift button
				goto GOT_BUTTONS;
				}
			}  // done with down button 
		} // end if (inpram.shftbut_mode == 1) ...	
		else if (inpram.shftbut_mode == 0) // if digital shift buttons
		{
		// get shift button states
		// DWN button is done in get_adc()
		if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x80))) // not testing paddle UP
			{
			if (PADDLE_UP & PADDLE_UPpin)       // pin high
			    {  // not grounded
			    if ( inpram.shft_btn_polarity == 0) // active low
			    	{
			    	outpc.shift_action &= ~0x02;    // clear bit 1
			    	}  // end if (inpram.shft_btn_polarity == 0)
			    	else
			    	{
			    	outpc.shift_action |=  0x02;    // set bit 1 
			    	} // end else if (inpram.shft_btn_polarity == 0)
			    }
			    else
			    { // grounded
			    if ( inpram.shft_btn_polarity == 0) // active low
			    	{
			    	outpc.shift_action |=  0x02;    // set bit 1
	    			} // end if (inpram.shft_btn_polarity == 0)
	    			else
	    			{
	    			outpc.shift_action &= ~0x02;    // clear bit 1 
	    			} // end else if (inpram.shft_btn_polarity == 0)	    		
	    		} // end if (PADDLE_UP & PADDLE_UPpin) 
			}
			else // testing paddle up
			{
			if (in2ram.TSTdbug & 0x01)       // pin high
			    {  // not grounded
			    if (inpram.shft_btn_polarity == 0) // active low
			    	{
			    	outpc.shift_action &= ~0x02;    // clear bit 1
			    	}
			    	else
			    	{
			    	outpc.shift_action |=  0x02;    // set bit 1 
			    	} // end if (inpram.shft_btn_polarity == 0)
			    }
			    else
			    { // grounded
			    if (inpram.shft_btn_polarity == 0) // active low
			    	{
			    	outpc.shift_action |=  0x02;    // set bit 1
	    			}
	    			else
	    			{
	    			outpc.shift_action &= ~0x02;    // clear bit 1 
	    			} // end else if (inpram.shftbut_mode == 0)	    		
	    		} // end if (PADDLE_UP & PADDLE_UPpin)
			} // end else if (!(inpram.test_mode & 0x80))
		}  // end else if (inpram.shftbut_mode == 0)
		
	switch (outpc.shift_action & 0x03)
		{
		case 0:	 // neither button pressed
		    if (shift_release)
				{
				// wait out shift release before altering shift averages
    			shift_release--;
				}
				else
				{	
				if (UP_avg   )   UP_avg--; // decrement upshift average
				if (DWN_avg  )  DWN_avg--; // decrement downshift average
				if (MODE_avg > 0) MODE_avg--; // decrement mode change average	
				}
			break;
			
		case 1:	  // downshift button pressed
			if (shift_release) break;
			if (UP_avg   )   UP_avg--;  // decrement upshift average
			if (DWN_avg  < 19999)  DWN_avg++; // increment downshift average
			if (MODE_avg ) MODE_avg--;  // decrement mode change average				
			break;
		
		case 2:	  // upshift button pressed
			if (shift_release) break;
			if (UP_avg   < 19999) UP_avg++; // decrement upshift average
			if (DWN_avg  )  DWN_avg--;  // increment downshift average
			if (MODE_avg ) MODE_avg--;  // decrement mode change average	
			break;
		
		case 3:		// both buttons pressed
			if (shift_release) break;
			if (UP_avg   )     UP_avg--;  // decrement upshift average
			if (DWN_avg  )    DWN_avg--;  // increment downshift average
			if (MODE_avg < 19999) MODE_avg++; // decrement mode change average	
			break;			
		}	 // end switch(...

GOT_BUTTONS:

// Rail the shift button counts
if (UP_avg  > 19999)    UP_avg  = 19999;     // rail the UP_avg count
if (DWN_avg > 19999)   DWN_avg  = 19999;     // rail the DWN_avg count 
if (MODE_avg > 19999) MODE_avg  = 19999;     // rail the DWN_avg count

//-- Act on switch states --------------------------------------------

if ((outpc.current_gear > 0) && !(rev_shift_flag) && (inpram.shftbut_mode != 2)) // if log only shift buttons) // only if not forcing a shift due to revs
{	
// check states of both buttons first to allow mode changes
if (MODE_avg > (3*inpram.shftr_accum))      // if both buttons have been pressed for a while
	{
	if (outpc.auto_mode == 0)                 // if not auto shifting
		{
		if (inpram.shift_mode > 0) // if default is an auto mode
			{
			outpc.auto_mode = inpram.shift_mode;  // set to default
			} // end if inpram.shift_mode>0
			else
			{
			outpc.auto_mode = 1;   // otherwise set to sequential auto mode
			} // end else if inpram.shift_mode>0
		
		outpc.shift_action &= ~0x0C; // clear both up and down shift requests (0x0C = 1100)
	
		// reset shift button counters
		UP_avg   = 500;
		DWN_avg  = 500;
		MODE_avg = 500;
		shift_release = inpram.shftr_accum/4;          
		} // end if auto_mode == 0
	} // end if UP_avg ...

// Check upshift  
if (UP_avg > (DWN_avg+inpram.shftr_accum)) // if we have 400 more ups than downs, request upshift                          
	{
	// request shift
	if (outpc.auto_mode == 0)              // if already in manual mode
		{
		outpc.shift_action &= ~0x04; // clear downshift request
		outpc.shift_action |=  0x08; // set upshift request
		}
		else  // auto_mode != 0
		{
		outpc.auto_mode = 0;         // switch to manual mode
		outpc.shift_action &= ~0x0C; // clear all requests
		// reset shift button counters
		UP_avg   = 500;
		DWN_avg  = 500;
		MODE_avg = 500;
		shift_release = inpram.shftr_accum/4;     	
		}
	} // end if UP_avg > DWN_avg ...
	
// Check downshift next
if (DWN_avg > (UP_avg+inpram.shftr_accum)) // if we have 400 more downs than ups, OR revs are too low, request downshift
{
	// request shift  
	if (outpc.auto_mode == 0) 	           // if already in manual mode
		{
		outpc.shift_action |=  0x04; // set downshift request
		outpc.shift_action &= ~0x08; // clear upshift request
		}
		else   // auto_mode != 0
		{
		outpc.auto_mode = 0;         // switch to manual mode
		outpc.shift_action &= ~0x0C; // clear all requests
		// reset shift button counters
		UP_avg   = 500;
		DWN_avg  = 500;
		MODE_avg = 500;
		shift_release = inpram.shftr_accum/4;    	
		}
	} // end if DWN_avg > UP_avg ...
	
if (inpram.min_manual_speed >= 0 )
	{
	if ((UP_avg < 100) && (DWN_avg < 100) && (MODE_avg < 100)) // if no buttons pressed
	{
		// if speed is less than minimum and haven't shifted to first in 2WD, 
		// and have actually moved (exceeded 20 mph)
		if ((outpc.speedo < inpram.min_manual_speed) && (outpc.current_gear > 1)  && (high_mph > 200) && (outpc.miscStatus & 0x04)) 
			{                     
			// if have not received a brake signal, set flag
			if (error_flags & 0x80) outpc.error |= 0x80; // set bit 7 in error code if flag has 
														// not been cleared by brake signal
			}   // end IF outpc.speedo <= inpram.min_manual_speed
	}   // end IF UP_avg < 20
	} // end if (inpram.min_manual_speed >= 0 ) 
} // end if (outpc.current_gear > 0)

// Get and Adjust PC Table Value
// -----------------------------
// Will assign PC_DUTY later
// Look-up pressure control solenoid (PC) duty cycle in pc_table[16x9] and set tempPC1
// as a fractional count of PWMPER2 (tempPC1 is the look-up value, tempPC2 is the 
// voltage/temperature adjustment, these are added to get the duty cycle count)
    	
PC_table_value = 0; // variable to hold PC table % until all calcs done
tempPC1  = 0; // variable to hold on-going PC % as it is adjusted
tempPC2  = 0;
		
// outpc.LOAD is always the latest MAP, TPS lookup, or blended load value
PC_table_value = (uchar)(intrp_2dctable((uint) (outpc.speedo/10),  // in PWM%
										(int) (outpc.LOAD/10),
										(uchar) NUM_MPH, 
										(uchar) NUM_LOADS, 
										(uint *)&in2ram.vss_PCtable[0], 
										(int *)&inpram.LOAD_PCtable[0], 
										(uchar *)&inpram.pc_table[0][0]));



// ------------- Line Pressure Slip Adjustment First -------------------------------------------------
if ((inpram.ic2_usage & 0x01) && (inpram.SlipConditions)) // if ISS enabled
	{
	if (!(shifting_flag) && (outpc.current_gear > 0) && (outpc.speedo > 30) && (outpc.LOAD >= inpram.Sretard_load)) 	// not shifting and opting to adjust PC PWM and in forward gear
		{
		if (
			(outpc.slip >= inpram.Slip_threshold)	// if slipping while accelerating
			&& ((inpram.SlipConditions == 2)    				// spark AND line pressure adjustments
			|| ((inpram.SlipConditions == 3) && (Sflag == 1)) 	// spark THEN line pressure adjust
			||  (inpram.SlipConditions == 4)						// line pressure adjustment only
			||  (inpram.SlipConditions == 5)
			))
			{
			if (outpc.Slip_Adjustment == 0)  	// if this is the first time adjusting
				{
				Slip_msec = 0; // increment in timer loop
				if (inpram.SlipConditions == 5) Sflag = inpram.slipDelay;
				if ((PC_table_value - outpc.Slip_Adjustment) > in2ram.Slip_limit[outpc.current_gear - 1]) outpc.Slip_Adjustment++;
				}
			if (Slip_msec > (100 * outpc.Slip_Adjustment))  // only if 1/10th of a seconds has passed since last change to adjustment value
				{
				if ((PC_table_value - outpc.Slip_Adjustment) > in2ram.Slip_limit[outpc.current_gear - 1]) 
					{
					outpc.Slip_Adjustment++;
					}
			    }
			}
			else // (outpc.slip < inpram.Slip_threshold)
			{
			Slip_msec = 0; // increment in timer loop
			Sflag = 0;
			// outpc.Slip_Adjustment will be decremented 1% per sec in timer ISR
			}
		} // end if (not shifting_flag)
		else
		{ // shifting or option not to adjust or neutral
		outpc.Slip_Adjustment = 0; // reset adjustment
		}
	}   // end if (inpram.ic2_usage & 0x01)
	else // ((ic2 not used for ISS))
	{
	outpc.Slip_Adjustment = 0; // reset adjustment
	}
		
PC_table_value = PC_table_value - outpc.Slip_Adjustment;	// make the adjustment
															// subtract because we need PWM% to go down 
															// for pressure to go up
	    
// ------------- PC Voltage Correction -------------------------------------------------
// The voltage difference is (120-vBatt)/10, since vBatt is x10
// the correction is the voltage difference multiplied by the 
// battery correction factor inpram BatFac/10
// Note that BatFac is negative in INI, but positive in this code.
// We move the divisors around a bit to avoid over/underflows.
if (outpc.vBatt < 120)                           // if less than 12.0 Volts, decrease 
											// the DC (increase pressure), 
											// base is 12.0 volts x 10
	{
	tempPC2 = (long) ((((120 - outpc.vBatt)*inpram.BatFac) + 50)/100);   // adjust for battery supply voltage less than 12 volts                                 
	if (PC_table_value > tempPC2)      // make sure result will be positive
		{   
		tempPC1 = (uchar)((int) PC_table_value - (int) tempPC2);   // set the value by subtracting DC (increase pressure)
		} 
	}
	else  // battery voltage greater than 12.0 V
	{
	tempPC2 = (long) ((((outpc.vBatt - 120)*inpram.BatFac) + 50)/100); 	// adjust for battery supply voltage greater than 12 volts
																	// and convert to number of PWM tics
	if ((PC_table_value + (char) tempPC2) < 100)  // make sure we don't try to set PWM DC to greater than 100%
		{ 
		tempPC1 = (uchar)((int) PC_table_value + (int) tempPC2);    // set the value by adding DC (decrease pressure) 
		}
	} // end if (vBatt <= 120)																																					

// tempPC1 now holds PC table value adjusted for voltage    		
tempPC2 = 0; // reinitialize
  	
// ------------- PC Temperature Correction -------------------------------------------------
// From PCtemp[] table
// We move the divisors around a bit to avoid over/underflows.
// Get the value and convert it to counts of PC duty period:																													
if (inpram.stdin_cfg & 0x02) tempPC2 = intrp_1dctable(	0,
														outpc.clt,
														NUM_TEMPS,
														inpram.temp_table,
														inpram.PCtemp)* PC_PER/100; // 1-D table lookup	
// 1D Interpolation Routine
// intrp_1dctable(char sgnx, int x, unsigned char n, int * x_table, char * z_table)
// where: 
// sgnx     = 
// x        = the index value for the lookup
// n		    = the number of rows in (size of) the lookup table
// x_table  = the name of the lookup table
// z_table  = the 'looked up' values
// return the interpolated value as an integer
   	
if (tempPC1 > tempPC2)                   // make sure result will be positive
	{		  
	tempPC1 = (uchar)(tempPC1 - tempPC2); // set the value by subtracting DC (increase pressure)		
	}
	else                                    // result less than zero
	{
	tempPC1 = 0;                         // do not let value go negative (or wrap around)
	}                                                                                                                                                   

PC_table_value = tempPC1; // final adjusted value


// get the SP3 table value
// -----------------------

SP3_table_value = (unsigned char) (intrp_2dcPWM(
										16,
										9,
										(int) (outpc.sp316inx/10), 
										(int) (outpc.sp39inx/10), 
										(int *) in2ram.sp3_16Bin, 
										(int *) in2ram.sp3_9Bin, 
										(uchar *) in2ram.sp3_table));

if (in2ram.sp3_mode == 0)   // spare port in line pressure mode
	{
	if ((jake_brake && (inpram.line_units == 2)) || (((inpram.CAN_enabled) || (!inpram.CAN_enabled && (inpram.ic2_usage & 0x03) == 2)) && (outpc.engine_rpm < 300)) || dither_flag)
		{
		PC_DUTY = PC_OFF;
		}
		else if (outpc.current_gear > 0)  	// not jake brake AND rpm > 300 AND not dither and in forward gear, 
											// so use table
		{
		PC_DUTY = (uchar)(((PC_table_value * PC_PER) + 50)/100); // set PWM 'tics'
		}
		else if (outpc.current_gear == 0) // Neutral 
		{
		if (inpram.PCOptn & 0x02) // bit 1 = 1: use table
			{
			PC_DUTY = (uchar)(((PC_table_value * PC_PER) + 50)/100); // set PWM 'tics'
			}
			else // bit 1 = 0: use fixed setting
			{
			PC_DUTY = (uchar)(((inpram.PCneutral * PC_PER) + 50)/100); // set PWM 'tics'
			}
		}  // end else if (outpc.current_gear == 0) 
		else if (outpc.current_gear < 0) //reverse
		{
		if (inpram.PCOptn & 0x01) // bit 0 = 1: use table
			{
			PC_DUTY = (uchar)(((PC_table_value * PC_PER) + 50)/100); // set PWM 'tics'
			}
			else // bit 0 = 0: use fixed setting
			{
			PC_DUTY = (uchar)(((inpram.PCreverse * PC_PER) + 50)/100); // set PWM 'tics'
			}
		}
		outpc.PC_duty = (uchar)(100-((int) (PC_DUTY*100)/PC_PER)); 	// set datalog value (100-value to have datalog value same as PC table)                                                              																								// input/table values are 0-60%  - adjusted by TS                               
		outpc.sp3PWM  = (uchar)((int) (PC_DUTY*100)/PC_PER);  // set value for datalog even if we haven't used it
	} // end if (in2ram.sp3_mode == 0) i.e. line pressure mode
    else // sp3_mode > 0
    {
    if ((in2ram.sp3_active == 1) && (outpc.engine_rpm < 300)) 
    	{// use sp3_off
    	PC_DUTY = (uchar)((in2ram.sp3_off * PC_PER)/100); // set PWM 'tics'
    	}
		else if (outpc.current_gear > 0) // forward gear and rpm > 300 so use table
		{
		PC_DUTY = (uchar)(((SP3_table_value * PC_PER) + 50)/100); // set PWM 'tics'
		}
		else if (outpc.current_gear == 0) // Neutral 
		{
		if (inpram.PCOptn & 0x02) // bit 1 = 1: use table
			{
			PC_DUTY = (uchar)(((SP3_table_value * PC_PER) + 50)/100); // set PWM 'tics'
			}
			else // bit 1 = 0: use fixed setting
			{
			PC_DUTY = (uchar)((((inpram.PCneutral) * PC_PER) + 50)/100); // set PWM 'tics'
			}
		}  // end else if (outpc.current_gear == 0) 
		else if (outpc.current_gear < 0) //reverse
		{
		if (inpram.PCOptn & 0x01) // bit 0 = 1: use table
			{
			PC_DUTY = (uchar)(((SP3_table_value * PC_PER) + 50)/100); // set PWM 'tics'
			}
			else // bit 0 = 0: use fixed setting
			{
			PC_DUTY = (uchar)((((inpram.PCreverse) * PC_PER) + 50)/100); // set PWM 'tics'
			}
		}
		outpc.PC_duty = PC_table_value;                         
		outpc.sp3PWM = 	(uchar)((PC_DUTY*100)/PC_PER); 	// set datalog values
    }  // end else if sp3_mode > 0 ...

// --------------- Set the Torque Converter Clutch (TCC) ----------------------------------------
// if  current_gear > minTCC_gear && clt > noTCC_temp OR jake_brake set
// We will check brakes and mph in function call.

// first do full time lock-up mode checks
if ((outpc.engine_rpm >= inpram.STCC_rpm) && (outpc.current_gear >= inpram.STCC_gear)) // if full time lock-up mode rpm and gear conditions met
	{
	outpc.TCC_count = 0; // lock immediately, disable hysteresis
	outpc.lock_TCC = tcc_lock();
	STCC_flag = 1;
	}
	else  // below TCC temp and above TCC load and rpm below full time lock up or gear below full time lock up
	{
	// outpc.lock_TCC = tcc_unlock();
	if ((STCC_flag == 1) && ((outpc.engine_rpm < inpram.STCC_ULrpm) || (outpc.current_gear < inpram.STCC_gear)) && !((inpram.TCC_highGear & 0x01) && (outpc.current_gear == inpram.num_gears))) 
		{
		STCC_flag = 0;
		outpc.lock_TCC = tcc_unlock();
		} 
	} // end if ((outpc.engine_rpm ...

if ((!STCC_flag) && !((inpram.TCC_highGear & 0x01) && (outpc.current_gear == inpram.num_gears)))
	{
	if (outpc.current_gear >= inpram.minTCC_gear)     // not jake brake, above minimum gear
		{
		if ((inpram.line_units == 2) && (jake_brake)) // if jake brake active
			{
			outpc.lock_TCC = tcc_lock();			  // lock TCC for jake brake
			}
			else if ((outpc.clt < inpram.noTCC_temp) && (inpram.stdin_cfg & 0x02))   // no jake brake and low temp and not 'log only'
			{
			outpc.lock_TCC = tcc_unlock();
			}
			else // not jake brake and above minimum temp
			{
			// if ((outpc.LOAD < inpram.noTCC_load) || (outpc.clt > inpram.trans_temp_limit))
			if (outpc.clt >= inpram.trans_temp_limit)
				{  // if below load or above temp limit
				outpc.lock_TCC = tcc_lock();
				} 
				else  // below temp limit
				{
				if (outpc.LOAD < (inpram.noTCC_load-in2ram.TCC_hyst)) outpc.lock_TCC = tcc_lock();
				else if (outpc.LOAD > inpram.noTCC_load) outpc.lock_TCC = tcc_unlock();
				}
			}  // end else if  (outpc.clt < inpram.noTCC_temp)
		} // end if (outpc.current_gear >= inpram.minTCC_gear)
		else // too low gear
		{
		// unlock TCC
		 if (outpc.current_gear != -99) outpc.lock_TCC = tcc_unlock();  // function call to unlock TCC, returns 0
		}
	}   // end if (!STCC_flag ...


//----- Set SOL32 PWM Duty Cycle ----------------------------------
// This is user configurable. Set based on bits in Output3 (x4321RNP)

setOut3PWM(outpc.current_gear);

//**** Determine Manual Shift Lever Position *********************************

// Each time through the main loop, determine input states to find 
// the current manual gear

// Switch           Port         Circuit   AMPseal
// ------           ----         -------   -------
// Input1           PAD00          EGT4       26
// Input2	        PAD01          GPI2        6
// Input3           PAD03          EGT3       25
// Input4           PE1            GPI1        5

// Input1 is high if In1adc > digital_threshold
// Input2 is high if In2adc > digital_threshold
// Input3 is high if In3adc > digital_threshold
// Input4 is digital

// ******** Condition SWx inputs ********************************
// Set Input1acc, Input2acc, Input3acc, Input4acc to mid point 
// of user range (inpram.debounce) on start up or after a shift
// then increment/decrement by one each time through the loop 
// until all are <10 or >90% of debounce value
// then set Input1, Input2, Input3 (& Input4) to zero if <10% and 1 if >90%

FIND_GEAR:    // If manual gear is in error state (gear_error == 99), loop to here until we get to current gear

if ((outpc.manual_gear > -2) && (outpc.manual_gear <= inpram.num_gears)) prev_gearlever = outpc.manual_gear;  // store current manual gear position - will check later to see if lever has been 'shifted' 

gear_error = 99; // will reset this to zero if we find a valid gear
				// if it is still 99 when we get to the end of gear checking, then loop to here

// call getLever...() functions
if (inpram.mlever_mode == 0) // If digital input (GM) scheme - 3 or 4 inputs
  { // function uses globals variables directly to access and store ADC, accumulators, etc.
  // Determine if we are using 3 or 4 digital inputs
  if (Inpt4_active) tmp1_U08 = 4; else tmp1_U08=3; // tmp1_U08 holds the number of active digital inputs

  // set the manual_gear value to the returned value
  outpc.manual_gear = getLeverDigital(tmp1_U08,inpram.debounce); 
  }
  else // if voltage based manual gear lever position -> mlever_mode == 1 (Ford)
  {
  // set the manual_gear to the returned value
  outpc.manual_gear = getLeverADC(outpc.Inadc[0]); // calculate from ADC on Input1
  } // end if (inpram.mlever_mode == 0)

// If manual lever in park, reverse or neutral, unlock TCC and skip gear setting.
// Done *after* getting auto gear from table, for diagnostic reasons, 
// otherwise target gear looks 'dead' in TS

if ((outpc.manual_gear < 1) && (outpc.manual_gear != -99)) // neutral, park = 0, reverse = -1
	{ 
	outpc.lock_TCC = tcc_unlock(); // unlock the TCC if in P,N,R
	outpc.hyst_timer = 0;	   // clear the hysteresis flag
	// STCC_flag = 0;
	outpc.current_gear = shiftTo(outpc.manual_gear); // shift to the manual lever lever position
	goto SKIP_SHIFT;               // skip gear selection code
	} // end if (outpc.manual_gear < 1)

if ((outpc.auto_mode > 0 ) && !(rev_shift_flag) && (outpc.manual_gear > 0))
	{
	// Shift in auto mode = 1 (sequential) or 2 (direct) or 3 (Tach-based) if target gear is not the current gear
	// if in auto mode but not having a forced shift
	outpc.shift_action &= ~0x0C; // clear all requests
	// set shift requests
	if ((outpc.target_gear < outpc.current_gear) || (outpc.current_gear > outpc.manual_gear) || (rev_shift_flag == 2))   
		{
		// if target gear is lower than current gear, or if have shifted manually to a lower gear
		// downshift
		outpc.shift_action |=  0x04; // set downshift request
		outpc.shift_action &= ~0x08; // clear upshift request
		// Always Unlock TCC for downshift
		// if (!STCC_flag) outpc.lock_TCC = tcc_unlock(); // function call returns 0 
		}  // End if target_gear < current_gear 

	if (outpc.target_gear > outpc.current_gear) 
		{ 
		// if the requested gear is greater than the current gear and less/equal than the manual position
		// upshift
		if ((outpc.target_gear <= outpc.manual_gear) || (rev_shift_flag == 1))
		{
		outpc.shift_action &= ~0x04; // clear downshift request
		outpc.shift_action |=  0x08; // set upshift request
		outpc.shiftStat = 4;
		}
		else  // do not shift above manual gear position
		{
		outpc.shift_action &= ~0x0C; // clear all requests
		} // end else
			
		}  // End if target_gear > current_gear
		
	} // End If outpc.auto_mode 

// ------------- Direct Jump to Target Gear --------------------------------------------
// If auto mode = 2, and shift required, skip direct to desired gear
// do not sequentially shift

if ((shifting_flag == 0) && (outpc.auto_mode == 2) && !(outpc.current_gear == outpc.target_gear)) 
	{
	if ((outpc.target_gear > outpc.current_gear) && (outpc.target_gear <= inpram.num_gears)) outpc.current_gear = shiftTo(outpc.target_gear);
	if ((outpc.target_gear < outpc.current_gear) && (outpc.target_gear >= -1)) outpc.current_gear = shiftTo(outpc.target_gear);
	} // End shift_mode==2


/*-- SHIFT_UP (sequential or manual shift) ----------------------------------------------
check upshift status to see if threshold has be reached, upshift if it has
if current_gear < 4 then target_gear = current gear + 1
Check if new gear will under-rev the engine (unless rpm_check = 0)
If required, unlock TCC
- TCC unlocks under certain conditions during upshifts, depending on the vehicle speed 
and engine LOAD.  It is safe to upshift under light load while the TCC is locked. 
Most OEM systems keep the torque converter clutch applied during light throttle upshifts.
LEDs are set near beginning of main loop.
--*/

if ((shifting_flag == 0) && (outpc.shift_action & 0x08) && !(outpc.shift_action & 0x04) && ((outpc.auto_mode < 2) || (outpc.auto_mode == 3))) 
	{
	if ((outpc.current_gear < inpram.num_gears) && (outpc.current_gear > 0)) outpc.current_gear = shiftTo(outpc.current_gear + 1);
	} //end if upshift_request


/*-- SHIFT_DOWN (sequential and manual) -----------------------------------------
check downshift status to see if threshold has be reached, upshift if it has
if current_gear > 1 then target_gear = current gear - 1
Check if new gear will over-rev the engine
If required, unlock TCC
- TCC always unlocks under during downshifts.
LEDs are set near beginning of main loop. 
--*/

if ((shifting_flag == 0) && (outpc.shift_action & 0x04) && !(outpc.shift_action & 0x08) && ((outpc.auto_mode < 2) || (outpc.auto_mode == 3))) 
	{
	if ((outpc.current_gear <= inpram.num_gears) && (outpc.current_gear > 1)) outpc.current_gear = shiftTo(outpc.current_gear - 1);
	} //end if downshift_request

// Set LEDs
setLEDs(outpc.current_gear);

SKIP_SHIFT:

DISABLE_INTERRUPTS
ultmp = lmms;
ENABLE_INTERRUPTS 

/***************************************************************************
**
** Check for reinit command
**
**************************************************************************/
     if(reinit_flag)  
     {
      reinit_flag = 0;
      // update initialization of variables/ registers dependent on
      // user config inputs
      reset();
      // Set up SCI (rs232): SCI BR reg= BusFreq(=24MHz)/16/baudrate
      SCI0BDL = (unsigned char)(1500000/inpram.baud);
      ltmp = (150000000/inpram.baud) - ((long)SCI0BDL*100);
      if(ltmp > 50)SCI0BDL++;   // round up
      can_id = inpram.mycan_id;
      can_reset = 1;     // do this to reset id in CAN message mask register
      }   // End if(reinit_flag)



/***************************************************************************
**
**  Check for serial, CAN receiver timeout, crc32 comm check
**
**************************************************************************/

    DISABLE_INTERRUPTS
    ultmp = lmms;
    ultmp2 = rcv_timeout;
    ENABLE_INTERRUPTS
    if(ultmp > ultmp2)  {
      txmode = 0;    // break out of current receive sequence
#ifdef COMMS_382X
      newMsg = 2;
      noEnvelope = 0;
#endif
      rcv_timeout = 0xFFFFFFFF;
    }
    DISABLE_INTERRUPTS
    ultmp2 = ltch_CAN;
    ENABLE_INTERRUPTS
    if(ultmp > ultmp2)  {
      getCANdat = 0;    // break out of current receive sequence
      ltch_CAN = 0xFFFFFFFF;
    }
    if(kill_ser)  {
      if(outpc.seconds > kill_ser_t)  
      {
        kill_ser = 0;
        SCI0CR2 |= 0x24;     // rcv, rcvint re-enable
      }

    }  
    
//    4.110 changes
#ifdef COMMS_382X
      checkCANbus();
#else
      if(vfy_flg)  {
      vfy_flg = 0;
      for(ix = 0; ix < tables[tble_idx].n_bytes; ix++)  {
        if(*tableByteFlash(tble_idx,ix) != *tableByteRam(tble_idx,ix))
        vfy_fail++;
        }
      txcnt = 0;
      txgoal = 2;
      SCI0DRL = *(char *)&vfy_fail;   
      SCI0CR2 &= ~0x24;   // rcv, rcvint disable
      SCI0CR2 |= 0x88;    // xmit enable & xmit interrupt enable   */
      }
#endif          
  
  
/***************************************************************************
**
**  Check for CAN reset
**
**************************************************************************/
	  if(can_reset)  
	    {
		  /* Re-initialize CAN comms */
		  CanInit();
		  can_reset = 0;
	    }  // end if (can_reset) ...
	    

if(inpram.can_var_rate)  
	{  // inpram.can_var_rate is the equivalent of in2ram.activate_xrate in MS2 code
	// send various bytes in the outpc structure to destvarblk.
	// The bytes are as configured in outmsg[msg_no] where 
	// msg_no = CAN_RB_DSR0
	uctmp = 1; 
	for(mx = 0; mx < MAX_OUTMSGS; mx++)  
		{   // mx = msg_no
	    if(outmsg[mx].xrate && (inpram.can_var_rate & uctmp))  
	    	{
	      	// this message activated - check if time to send
	      	if(xrate_clks[mx] > outmsg[mx].xrate)  
	      		{     // time to send
	        	xrate_clks[mx] = 0;           // zero clk
   	     		// Update related parameters for xmt ring buffer
   	     		ix = can[1].cxno_in;
   	     		can[1].cx_msg_type[ix] = OUTMSG_RSP;
   	     		can[1].cx_outmsg_no[ix] = mx;
   	     		n_outbytes[mx] = outmsg[mx].no_bytes;
   	     		// destination var blk - for this msg the varblk is really the gpio
  	     		//   varblk to return the data to
   	     		can[1].cx_destvarblk[ix] = outmsg[mx].dest_varblk & 0x0F;   
   	                                                          // gpio var_blk
    	 		can[1].cx_destvaroff[ix] = 0;   // store data starting at gpio varblk;
    	 		// if > 8 bytes, will have subsequent messages, with varoff provided 
    	 		// by ECU
    	 		can[1].cx_dest[ix] = outmsg[mx].dest_varblk >> 4;    // gpio id
    	 		if(n_outbytes[mx] >= 8)  
    	   			{
    	   			kx = 8;
    				}
    				else  
    				{
    				kx = n_outbytes[mx];
    				}  // end if(n_outbytes[mx] >= 8)
    			
    			can[1].cx_varbyt[ix] = kx;
        		
        		// put variable value(s) in xmit ring buffer as follows:
        		for(jx = 0;jx < kx;jx++)  
        			{
          			can[1].cx_datbuf[ix][jx] = *((char *)&outpc + outmsg[mx].offset[jx]);
        			} // end for (jx,...
        		send_can(1);
      			}  // end if(xrate_clks[mx] > outmsg[mx].xrate) 
    		}  // end if(outmsg[mx].xrate && (inpram.can_var_rate & uctmp))
    		uctmp = uctmp << 1;    // = 2 ^^ mx
  		}     // end for(mx = 0; mx < MAX_OUTMSGS; mx++) message loop
	}       // end if(inpram.can_var_rate) CAN outmsg send code

// Set User Spare Ports
// --------------------	    
for (i=0; i < 2; i++)
	{
	if (inpram.sp_mode[i] == 0) // spare port[ix] is a 'user spare port'
		{
 		SetSparePort(i); 	// function call to set user spare port i
 							// defined in SetSparePorts.h
		}  // end if (inpram.sp_mode[i]
	}   // end for (i=0; ...	   

} // end for(;;) main loop

} // end main() function

/** =========== FUNCTIONS ===================================================== **/

// ------------ Unlock TCC --------------------------------------------------------
char tcc_unlock(void)  
{
// If want to stay locked in high gear, and we are actually in high gear, 
// then return without doing anything
if ((inpram.TCC_highGear & 0x01) && ((outpc.current_gear == inpram.num_gears) || (shifting_flag && (outpc.target_gear == inpram.num_gears))))
	{
	outpc.lock_TCC = tcc_lock();
	outpc.TCC_count = inpram.TCC_hyst; // set flag for TCC hysteresis, will count down in 0.100 second timer   
	return ((unsigned char) ((TCC_OUT*100)/TCC_ON));
	}

// spare ports first

// spare port 0
if (in2ram.sp0_mode == 2) spr0(0);

// spare port 1
if (inpram.sp_mode[0] == 1) 
	{ // SP1 in PWM mode
	if (inpram.sp_active[0] == 0) // always on
		{
		spr1(1);
		sp[0] = 1;
		}  // end if (inpram.sp_active[0]
	} // end if (inpram.sp0_mode
	
if (inpram.sp_mode[0] == 2)
	{ // SP1 in TCC mode
	spr1(0); // Set PWM% to target
	sp[0] = 0;
	}

if ((!shifting_flag) && (inpram.sp_mode[0] == 4) && (inpram.TCC_follow_sp1 & (0x01 << (outpc.current_gear-1))))
	{ // SP1 in clutch mode	
	TCC_follow_count1 = 0; // reset
	TCC_follow_flag1 = 0;
	spr1(0);
	sp[0] = 0;
	}

// spare port 2	
if (inpram.sp_mode[1] == 1) 
	{ // SP2 in PWM mode
	if (inpram.sp_active[1] == 0) // always on
		{
		spr2(1);
		sp[1] = 1;
		}  // end if (inpram.sp_active[0]
	} // end if (inpram.sp1_mode

if (inpram.sp_mode[1] == 2)
	{ // SP2 in TCC mode
    spr2(0);	// Set PWM% to target
    sp[1] = 0;	
	}

// need to differentiate between a shift and the TCC causing spr2 to activate
// need to set count to 0 for all shifts (so we reset if a shift interrupts count down)
// need to set counter to input value for initial TCC apply only	
// need to count down TCC count only while TCC is active
// if TCC becomes inactive, need to set count to zero
// in clutch mode (== 4) sp2 is ON only while:
// - shifting and bit set, OR
// - TCC is ON and count timer is > 1
 
if ((!shifting_flag) && (inpram.sp_mode[1] == 4) && (inpram.TCC_follow_sp2 & (0x01 << (outpc.current_gear-1))))
	{ // SP2 in clutch mode	
	TCC_follow_count2 = 0; // reset
	TCC_follow_flag2 = 0;
	spr2(0);
	sp[1] = 0;
	}

// spare port 3
if (in2ram.sp3_mode == 2) spr3(0);

if (STCC_flag == 0) 
	{
	// Unlock
	if (inpram.TCC_PWM == 0)
		{
		if (!jake_brake) TCC_OUT = TCC_OFF;    // set PT3 low (DC = 0%)
		} 
	else // TCC PWM'd, so set PWMDTY3 for LUF (and set SPR2 low if TCC on spr2+LUF)
		{
		TCC_OUT = (unsigned char) ((inpram.LUF_off * TCC_ON)/100);  // set PT3 to on state
		outpc.lock_TCC = (unsigned char) ((TCC_OUT*100)/TCC_ON); // set value for datalog
		LUF_counter = 0;
	}
	outpc.TCC_count = inpram.TCC_hyst; // set flag for TCC hysteresis, will count down in 0.100 second timer   
	return ((unsigned char) ((TCC_OUT*100)/TCC_ON));          // return for outpc.lock_TCC flag
	} 
else // don't unlock because TCC is in full-time apply mode 
	{
	return ((unsigned char) ((TCC_OUT*100)/TCC_ON));
	}
} // end tcc_unlock

// ------------ Lock TCC ----------------------------------------------------------
char tcc_lock(void)  {

uchar high_gear = 0;

// apply spare port 1/2 if TCC_follow mode for this gear
// set timer to deactivate sp1/2

if ((!shifting_flag) && (inpram.sp_mode[0] == 4) && (inpram.TCC_follow_sp1 & (0x01 << (outpc.current_gear - 1))))
	{
	if (TCC_follow_flag1 == 0) 
		{
		TCC_follow_flag1 = 1;
		TCC_follow_count1 = inpram.TCC_follow_sp1dur; // set timer
		spr1(1); // set spare port 1 to ON
		sp[0] = 1;
		}
	}
	 
// Lock if we have 'lock always in top gear' and we are in top gear
if ((inpram.TCC_highGear & 0x01) &&  ((outpc.current_gear == inpram.num_gears) || (shifting_flag && (outpc.target_gear == inpram.num_gears))))
	{
	high_gear = 1;		 // set the flag
	}
	
// Only lock if TCC_count has reached zero (decremented in 0.100 sec timer)
if (outpc.TCC_count == 0) 
	{
	// Most conditions are checked before requesting function call,
	// however here we check for braking and speed. *Never* lock while braking (unless jake_brake active).
	if ((high_gear) || ((((!(outpc.miscStatus & 0x01)) || ((jake_brake == 1) && (inpram.line_units == 2))) || (STCC_flag == 1)) && (outpc.speedo > inpram.minTCCspeed)))
		{ // lock TCC
		// deal with spare port TCC functions first before returning 
		if ((!shifting_flag) && (TCC_follow_flag1 == 0) && (inpram.sp_mode[0] == 4) && (inpram.TCC_follow_sp1 & (0x01 << (outpc.current_gear - 1))))
			{
			TCC_follow_flag1 = 1;
			TCC_follow_count1 = inpram.TCC_follow_sp1dur; // set timer
			spr1(1); // set spare port 1 to ON
			sp[0] = 1;
			}
		if ((!shifting_flag) && (TCC_follow_flag2 == 0) && (inpram.sp_mode[1] == 4) && (inpram.TCC_follow_sp2 & (0x01 << (outpc.current_gear - 1))))
			{
			TCC_follow_flag2 = 1;
			TCC_follow_count2 = inpram.TCC_follow_sp2dur; // set timer
			spr2(1); // set spare port 2 to ON
			sp[1] = 1;
			}
		
		if (inpram.TCC_PWM == 0)   // no TCC PWM
			{
			TCC_OUT = TCC_ON;        // set PT3 if not using PWM
			if (in2ram.sp0_mode   == 2) spr0(1);
			if (inpram.sp_mode[0] == 2) spr1(1);
			if (inpram.sp_mode[1] == 2) spr2(1);
			return 100;               // return 100% for the datalog
			} 
			else 
			{	 // TCC PWM'd				
			if (LUF_counter == 0) 
				{
				LUF_start = Calc_Initial_LUF();                // Capture initial LUF PWM% value
				TCC_OUT = (unsigned char) ((LUF_start * TCC_ON)/100);  // Set LUF DC the first time, otherwise do in timer - we will update this every 0.010 seconds
				if (in2ram.sp0_mode == 2) spr0(1);
				if (in2ram.sp3_mode == 2) spr3(1);
				if (inpram.sp_mode[0] == 2) spr1(1);
				if (inpram.sp_mode[1] == 2) spr2(1);
				}
			return ((unsigned char) ((TCC_OUT * 100)/TCC_ON));                 // return PWM% for outpc.lock_TCC flag for datalog
			}   // end else if (TCC_PWM ...)
		} // end if ((high_gear) || ...)
		else   // else if ((highGear) || ...
		{
		TCC_OUT = TCC_OFF;
		if (in2ram.sp0_mode == 2) spr0(0);
		if (in2ram.sp3_mode == 2) spr3(0);
		if (inpram.sp_mode[0] == 2) spr1(0);
		if (inpram.sp_mode[1] == 2) spr2(0);
		return 0;                 // return 0 for outpc.lock_TCC flag for datalog
		}   // end else if ((highGear) || ...
	} // end if (outpc.TCC_count == 0) ...
	else 	     // TCC_count != 0
	{
	// TCC_count != 0, do not lock-up, return 0 for datalog
	TCC_OUT = TCC_OFF; 
	if (in2ram.sp0_mode == 2) spr0(0);
	if (in2ram.sp3_mode == 2) spr3(0);
	if (inpram.sp_mode[0] == 2) spr1(0);
	if (inpram.sp_mode[1] == 2) spr2(0);
	return 0;
	}
}   // end tcc_lock()


// Calculate the initial LUF PWM Percentage
char Calc_Initial_LUF(void)
{
char percent = 0;
// percent is calculated as:
//  - equal to LUF_loPWM if load is less than or equal to LUF_loLoad
//  - equal to LUF_hiPWM if load is greater than or equal to LUF_hiLoad
//  - linear interpolation if load is between LUF_loLoad and LUF_hiLoad

if (outpc.LOAD < inpram.LUF_loLoad)
	{
	percent = inpram.LUF_loPWM;
	} 
else if (outpc.LOAD > inpram.LUF_hiLoad)
	{    
	percent = inpram.LUF_hiPWM;
	} 
else // between high and lo load, so interpolate
	{
	if (inpram.LUF_hiPWM > inpram.LUF_loPWM)   // positive slope to interpolation
	{ 
	percent = inpram.LUF_loPWM+(inpram.LUF_hiPWM-inpram.LUF_loPWM)*(outpc.LOAD-inpram.LUF_loLoad)/(inpram.LUF_hiLoad-inpram.LUF_loLoad);
	} 
	else   // negative slope
	{
	percent = inpram.LUF_loPWM-(inpram.LUF_loPWM-inpram.LUF_hiPWM)*(outpc.LOAD-inpram.LUF_loLoad)/(inpram.LUF_hiLoad-inpram.LUF_loLoad);       
	}  // end if (inpram.LUF_hiPWM > ...
	}   // end if (outpc.LOAD <
outpc.lock_TCC = percent; // set value for datalog
return percent;
}  // end Calc_Initial_LUF()

char Calc_LUF_PWM(void)
{
// this function computes and returns the current TCC PWM percent based on:
//  - the initial value (LUF_Start),
//  - the end value (inpram.LUF_on),
//  - the taper time (inpram.LUF_taper), and
//  - the current LUF_counter value.

char tempx;
if ((jake_brake && (inpram.line_units == 2)) || (LUF_counter == inpram.LUF_taper)) 
	{
	tempx = inpram.LUF_on;
	}
	else
	{ 
	tempx =  LUF_start + ((LUF_counter * (inpram.LUF_on - LUF_start)) / inpram.LUF_taper);
	}
outpc.lock_TCC = tempx; // set value for datalog
return tempx;
}  // end  Calc_LUF_PWM()

void reset_gear(void) 
{
	if (inpram.error_check) outpc.error |= 0x20; // out of range, set 00X00000 in error code 
	gear_error = 99; // set flag so we poll until have a gear lever position
	return;
}   // end reset_gear(

// ------------ Reset input switch manifold --------------------------------------------
void reset_switches(void)
{                
	// reset switch accumulator counts re-establish current gear
	Inputacc[0] = inpram.debounce/2;
	Inputacc[1] = inpram.debounce/2; 
	Inputacc[2] = inpram.debounce/2;
	Inputacc[3] = inpram.debounce/2; 
	return; 
}  // end reset_switches()

//--------------- vss_reset -------------------------------------------------------
#pragma CODE_SEG DEFAULT

void VSS_reset(void)  
{ 
// reinitialize
if (inpram.error_check) outpc.error |= 0x08;   // set 4th bit of error code 
												// (0x08 = 00001000)
// outpc.VSS_ovflow = 0;
TC_ov_ix = 0;

// Reset the error counter
VSS_error = 0;    // set the error value to zero 
restart = 1;      // set restart flag

return;
} // end vss_reset()

void reset(void)  
{
outpc.FuelAdj = 0;
outpc.SpkAdj = 0;
outpc.IdleAdj = 0;
outpc.SprAdj = 0;
FuelAdjl = 0;
SpkAdjl = 0;
IdleAdjl = 0;
SprAdjl = 0;
first_instr = 1;
gpio_ms = 0;
return;
} // end reset()

#pragma CODE_SEG DEFAULT

//--------------- get_adc -------------------------------------------------------
signed int get_adc(char chan1)  
{
// chan1 = 0 is manual gear lever Input1
// chan1 = 1 is manual gear lever Input2
// chan1 = 2 is transmission temperature
// chan1 = 3 is manual gear lever Input3
// chan1 = 4 is line pressure
// chan1 = 5 is non-CAN MAP
// chan1 = 6 is voltage shift buttons (if used)

signed int adc_tmp; // temp value

unsigned char atd0dien,perad,ddrad;

// Set flag
adc_flag = 1;

// Save configuration
atd0dien = ATD0DIEN; 	// 0 = adc, 1 = digital
perad = PERAD;	   	
ddrad = DDRAD;			// 0=input, 1=output

// Set digital buffers so we can enable pull-ups if required (for miroTCU)
ATD0DIEN = 0x01 << chan1;  // enable only the buffer for the pin we wish to sample

// Set Pull-up Enable Register for the AD port
if (inpram.hardware != 0) 
	{
	switch (chan1) 
		{
		case SWITCHA_ADC:	// IN1 = PAD00
			ATD0DIEN = 0x01; // digital buffer on PAD00; 0 - adc, 1 - digital buffer 
			if (!(inpram.mlever_mode & 0x01)) PERAD = 0x01; 	// enable pull-up (0 is digital, 1 is voltage based)
			DDRAD = ~0x01; 	// 0 - input, 1- output; set to all outputs, except current pin
			break;
			
		case SWITCHB_ADC:	// IN2 = PAD01
			ATD0DIEN = 0x02; // digital buffer on PAD01; 0 = adc, 1 - digital buffer   
			PERAD = 0x02; 	// enable pull-up
			DDRAD = ~0x02; 	// 0 - input, 1- output; set to all outputs, except current pin
			break;
			
		case SWITCHC_ADC:	// IN3 = PAD03
			ATD0DIEN = 0x08; // digital buffer on PAD03; 0 = adc, 1 - digital buffer   
			PERAD = 0x08; 	// enable pull-up
			DDRAD = ~0x08; 	// 0 - input, 1- output; set to all outputs, except current pin
			break;
			
		default:
			ATD0DIEN = atd0dien;
			PERAD = perad;
			DDRAD = ddrad;
			break;
		} // end switch (chan 1)
	} // end if (inpram.hardware != 0) 


// the ATD0DRx ADC result register contains the ten bit result of the 
// ADC conversion for channel x, we use this as an index to the 'lookup'_table 
// to get the actual value when appropriate

switch(chan1)  {
	
	case SWITCHA_ADC:
		if (inpram.hardware == 0)   // GPIO
			{
			if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x02)))  
				{
				adc_tmp = ATD0DR0; 
				}
				else
				{
				adc_tmp = in2ram.IN1dbug * 4; // set value to ADC count for swA (manual shift lever position)
				}
			}
			else   // microTCU PAD00
			{
			if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x02)))
				{
				if (PORTAD0 & 0x01) adc_tmp = 1023; else adc_tmp = 0;
				}
			 	else
			 	{
			 	if (in2ram.IN1dbug) adc_tmp = 1023; else adc_tmp = 0;
			 	}
			}
		break;

	case SWITCHB_ADC:
		if (inpram.hardware == 0)   // GPIO
			{
			if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x04)))
				{
				adc_tmp = ATD0DR1; // set value to ADC count for swB (manual shift lever position)
				}
				else
				{
				adc_tmp = in2ram.IN2dbug * 4;
				}
			}
			else   // microTCU PAD01
			{
			if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x04)))
				{
				if (PORTAD0 & 0x02) adc_tmp = 1023; else adc_tmp = 0;
				}
				else
			    {
			    if (in2ram.IN2dbug) adc_tmp = 1023; else adc_tmp = 0;
			    }
			}
		break;
	
	case TEMPER_ADC:
		// transmission temperature sensor         
		if (((inpram.stdin_cfg & 0x04) == 0) && ((inpram.stdin_cfg & 0x02) == 0)) // must be both 'log only' and 'log ADC'
			{	
			if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x10))) 
				{
				adc_tmp = ATD0DR2;
				}
				else
				{
				adc_tmp = in2ram.TMPdbug * 4;
				}
			}
			else   // use temperature or log temperature
			{
			if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x10))) 
				{	
				adc_tmp = cltfactor_table[ATD0DR2]; // look up temperature in degrees F x 10
				}
				else
				{
				adc_tmp = cltfactor_table[in2ram.TMPdbug * 4]; // look up temperature in degrees F x 10
				}
			}// end if if (((inpram.stdin_cfg & 0x04) == 0) &&
		break;
	
	case SWITCHC_ADC:
	if (inpram.hardware == 0)   // GPIO
			{
			if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x08)))
				{
				adc_tmp = ATD0DR3; // set value to ADC count for swC (manual shift lever position)
				}
				else
				{
				adc_tmp = in2ram.IN3dbug * 4;
				}
			}
			else   // microTCU PAD03
			{
			if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x08)))
				{
				if (PORTAD0 & 0x08) adc_tmp = 1023; else adc_tmp = 0;
				}
				else
				{
				if (in2ram.IN3dbug) adc_tmp = 1023; else adc_tmp = 0;
				}
			}
		break;
	
	case PRESSURE_ADC:
		// line pressure table value if line_units = 0
		// ADC count if inpram.line_units          = 1 - obsolete
		// engine brake input if line_units        = 2
		// 2-point calcs if line_units             = 3 
		
		switch (inpram.line_units)
			{
			case 0:    // use pressure table
				if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x400)))
					{	
					adc_tmp = (linefactor_table[ATD0DR4]);
					}
					else
					{
					adc_tmp = (linefactor_table[in2ram.LPdbug * 4]);
					}
				jake_brake = 0;		// jake brake and open switch
				outpc.miscStatus &= ~0x20; // clear Jake bit for datalog
				break;
		
		// Removed from INI and code because of pressure psi/bar conversions in INI	
   		// Use 2 point calcs instead for ADC 
			case 1:   // report ADC count
			//	adc_tmp = ATD0DR4;
				break;
			
			case 2:  // use as engine brake input
				if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x400)))
					{
					if ((ATD0DR4 < 50) && (inpram.line_units == 2))   	// if below 0.244 Volts
						{ // jake brake and grounded switch	
						jake_brake = 1;
						outpc.miscStatus |= 0x20; // set bit for datalog 
						}
						else
						{	
						jake_brake = 0;		// jake brake and open switch
						outpc.miscStatus &= ~0x20; // clear bit for datalog
						}
					adc_tmp = ATD0DR4; // report ADC count
					}
					else
					{
					if (((in2ram.LPdbug * 4) < 50) && (inpram.line_units == 2))   	// if below 0.244 Volts
						{ // jake brake and grounded switch	
						jake_brake = 1;
						outpc.miscStatus |= 0x20; // set bit for datalog 
						}
						else
						{	
						jake_brake = 0;		// jake brake and open switch
						outpc.miscStatus &= ~0x20; // clear bit for datalog
						}
					adc_tmp = in2ram.LPdbug * 4; // report ADC count
					}
				break;
			
			case 3:	 // use 2-point calcs
				if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x400)))
					{
					adc_tmp = (int) ((((long) Pslope * (long) ATD0DR4) + Pintcpt)/100);
					}
					else
					{
					adc_tmp = (int) ((((long) Pslope * (long) (in2ram.LPdbug * 4)) + Pintcpt)/100);
					}
				jake_brake = 0;		// jake brake and open switch
				outpc.miscStatus &= ~0x20; // clear bit for datalog
				break;
				 
			default:
				break;
		}    // end switch (inpram.line_units)
		
		break;
	
	case NC_LOAD_ADC:
	// non-CAN MAP
		if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x1000)))
			{
			adc_tmp = ATD0DR5; // don't use the table at all
			}
			else
			{
			adc_tmp = in2ram.nCANdbug * 4;
			}
		break;
	
	case VOLT_SHIFT_ADC:
	// Downshift/Voltage shift buttons
	    if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x100)))
			{
			adc_tmp = ATD0DR6;
			}
			else
			{
			adc_tmp = in2ram.PDNdbug * 4;
			}
		break;
	
	case BRAKE_ADC:
		if (inpram.hardware == 0 ) // GPIO
			{
			if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x200)))
				{
				adc_tmp = ATD0DR7;
				}
				else
				{
				adc_tmp = in2ram.BRKdbug * 4;
				}
			}
			else // microTCU
			{
			if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x200)))
				{
				if (PORTAD0 & 0x80) adc_tmp = 1024; else adc_tmp = 0;
				}
				else
				{
				if (in2ram.BRKdbug) adc_tmp = 1024; else adc_tmp = 0;
				}
			}
		break;
	
	default:
		break;
	}			 // end of switch
	

// Restore Configuration
ATD0DIEN = atd0dien;
PERAD = perad;
DDRAD = ddrad;

// Clear flag
adc_flag = 0;

return(adc_tmp);
}  // end get_adc()

#pragma CODE_SEG NON_BANKED

// 4.110 changes
#ifdef COMMS_382X
	 						  
void switch_page(unsigned char sub_no, char carg1, char carg2)  
{
// This resides in non-banked memory and is used to call subroutines
//  from another page.
  char save_page;

  save_page = PPAGE;
  switch(sub_no)  {
          
    case 0:
      PPAGE = 0x3D;
      break;
      
    case 1:
      PPAGE = 0x3C;
      break;

    case 2:
      PPAGE = 0x3D;
      break;
/*    
    case 3:
      PPAGE = 0x3D;
//      tpsaen();
      break;
      
    case 4:
      PPAGE = 0x3D;
//      xtau();
      break;
      
    case 5:
      PPAGE = 0x38;
//      rgmfactor = rgm1factor_table[rgmtble_ix];
      break;
      
    case 6:
      PPAGE = 0x38;
//      rgmfactor = rgm1afactor_table[rgmtble_ix];
      break;
      
    case 7:
      PPAGE = 0x38;
//      rgmfactor = rgm2factor_table[rgmtble_ix];
      break;
      
    case 8:
      PPAGE = 0x3C;
//      set_spr_port(carg1,carg2);
      break;

    case 9:
      PPAGE = 0x3C;
//      (void)move_IACmotor();
      break;
      
    case 10:
      PPAGE = 0x38;
      (void)ecu_init();
      break;
      
    case 11:
      PPAGE = 0x38;
//      (void)calcTrig(carg1);
      break;
      
    case 12:
      PPAGE = 0x38;
//      (void)injector_test();
      break;
*/      
    case 13:
      PPAGE = 0x3B;
      if(carg1 == 0)
        (void)crc32(13,0,0);
      else
        crcTab = crc_tab[(crc32Val ^ carg2) & 0xFF];
      break;
      
    case 14:
      PPAGE = 0x3B;
      break;

    default:
      break;
  }
  PPAGE = save_page;
  
  return;
}   // end  switch_page(unsigned char sub_no, char carg1, char carg2) 

#else

void switch_page(unsigned char sub_no)  
  {
  // This resides in non-banked memory and is used to call subroutines
  //  from another page.

  switch(sub_no)  
  	{	
	case 0:
	PPAGE = 0x3D;
	PPAGE = 0x3C;
	break;
	
	case 1:
	PPAGE = 0x3C;
	PPAGE = 0x3D;
	break;

	case 2:
	PPAGE = 0x3D;
	PPAGE = 0x3C;
	break;
	
	case 3:
	PPAGE = 0x3D;
	PPAGE = 0x3C;
	break;
	
	default:
	break;
	} 	 // end switch(sub_no)
	
return;
} // end  switch_page(unsigned char sub_no)


#endif



// #pragma CODE_SEG ROM_7000
#pragma CODE_SEG NON_BANKED

// --------------- intrp_1dctable -------------------------------------------------------
//
// 1D Interpolation Routine
// ------------------------
// sgnx     = 
// x        = the index value for the lookup
// n		= the number of rows in (size of) the lookup table
// x_table  = the name of the lookup table
// sgny     =
// z_table  = the 'looked up' values
// return the interpolated value as an integer

int intrp_1dctable(char sgnx, int x, unsigned char n, int * x_table, char * z_table)  
{
int ix;
long interp, interp3;
// bound input arguments
if((sgnx && (x > x_table[n-1])) ||
	(!sgnx && ((unsigned int)x > (unsigned int)x_table[n-1])))  
	{
	return((int)z_table[n -1]);
	}
	
if((sgnx && (x < x_table[0])) ||
	(!sgnx && ((unsigned int)x < (unsigned int)x_table[0])))  
	{
	return((int)z_table[0]);
	}
	
for(ix = n - 2; ix > -1; ix--)  
	{ 
	if((sgnx && (x > x_table[ix])) ||
	(!sgnx && ((unsigned int)x > (unsigned int)x_table[ix])))  
		{
		break;
		}
	}
if(ix < 0)ix = 0;

interp =	(unsigned int)x_table[ix + 1] - (unsigned int)x_table[ix];
if(interp != 0)  
	{
	interp3 = (unsigned int)x - (unsigned int)x_table[ix];
	interp3 = (100 * interp3);
	interp = interp3 / interp;
	}

return((int)((char)z_table[ix] +
		interp * ((char)z_table[ix+1] - (char)z_table[ix])/ 100));
}   // end intrp_1dctable()



// --------------- intrp_speedo_table -------------------------------------------------------
//
// 1D Interpolation Routine
// ------------------------
// sgnx     = 
// x        = the index value for the lookup
// n		= the number of rows in (size of) the lookup table
// x_table  = the name of the lookup table
// sgny     =
// z_table  = the 'looked up' values
// return the interpolated value as an integer

int intrp_speedo_table(unsigned int x, unsigned int * x_table, int * z_table)  
{
int ix;
long interp, interp3;
char n;

n = NUM_MPH;

// bound input arguments
if (x > x_table[n-1])  
	{
	return((int)z_table[n -1]);
	}
	
if (x < x_table[0])  
	{
	return((int)z_table[0]);
	}
	
for(ix = n - 2; ix > -1; ix--)  
	{ 
	if (x > x_table[ix])  
		{
		break;
		}
	}
	
if(ix < 0)ix = 0;

interp =	(unsigned int) x_table[ix + 1] - (unsigned int) x_table[ix];

if(interp != 0)  
	{
	interp3 = (unsigned int) x - (unsigned int) x_table[ix];
	interp3 = (100 * interp3);
	interp = interp3 / interp;
	}

return((int)((int) z_table[ix] +
		interp * ((int) z_table[ix+1] - (int) z_table[ix])/ 100));
} // end  intrp_speedo_table()

unsigned char intrp_1d_pwm(unsigned int x, unsigned int * x_table, unsigned char * z_table)  
{
// spare port pwm lookup
char ix;
int interp, interp3;

// bound input arguments
if (x > x_table[NUM_PWM_SP-1]) 	 // if greater than largest bin
  {
  return((unsigned char) z_table[NUM_PWM_SP-1]); // return largest bin value
  }

if(x < x_table[0])     // if less than smallest bin
  {
  return((unsigned char) z_table[0]);  // return smallest bin value
  }

for(ix = NUM_PWM_SP - 2; ix > -1; ix--)   // find next lower value
  { 
  if(x > x_table[ix])  
    {
    break;
    }
  }

if(ix < 0) ix = 0;	  // make sure index is not negative

interp = x_table[ix + 1] - x_table[ix]; // interp is the difference between the next lower and next higher bins

if(interp != 0)  
  {
  interp3 = 100*(x - x_table[ix]); // interp3 is 100 times the difference between the index value and the next higher bin value
  interp = interp3 / interp;  // 100x the ratio of the differences
  }

return((unsigned char) ((int) z_table[ix] + (int) interp * (int) (z_table[ix+1] - z_table[ix])/ 100));

} // end intrp_1d_pwm()

// --------------- intrp_2dctable -------------------------------------------------------

int intrp_2dctable(unsigned int x, int y, unsigned char nx, unsigned char ny,
unsigned int * x_table, int * y_table, unsigned char * z_table)  
{
int ix,hx;
long interp1, interp2, interp3;
// bound input arguments
if(x > x_table[nx-1])
	x = x_table[nx-1];
else if(x < x_table[0])
	x = x_table[0];

if(y > y_table[ny-1])
	y = y_table[ny-1];
else if(y < y_table[0])
	y = y_table[0];

// Find bounding indices in table
for(ix = ny - 2; ix > -1; ix--)  
	{	// Start w highest index
		//  because will generally have least time for calculations at hi y
	if(y > y_table[ix])  
		{
		break;
		}
	}
if(ix < 0) ix = 0;
for(hx = nx - 2; hx > -1; hx--)  
	{  // Start w highest index
		// because will generally have least time for calculations at hi x
	if(x > x_table[hx])  
		{
		break;
		}
	}
	
if(hx < 0) hx = 0;
// do 2D interpolate
interp1 = y_table[ix + 1] - y_table[ix];
if(interp1 != 0)  
	{
	interp3 = (y - y_table[ix]); 
	interp3 = (100 * interp3); 
	interp1 = interp3 / interp1; 
	}
	
interp2 =	x_table[hx + 1] - x_table[jx];

if(interp2 != 0)  
	{
	interp3 = (x - x_table[hx]); 
	interp3 = (100 * interp3); 
	interp2 = interp3 / interp2; 
	}

return((int)(((100 - interp1) * (100 - interp2) * z_table[ix*nx+hx]
	+ interp1 * (100 - interp2) * z_table[(ix+1)*nx+hx]
	+ interp2 * (100 - interp1) * z_table[ix*nx+hx+1]
	+ interp1 * interp2 * z_table[(ix+1)*nx+hx+1]) / 10000));
}  // end intrp_2dctable()

// --------------- intrp_2dcPWM -------------------------------------------------------

unsigned char intrp_2dcPWM(unsigned char nx, unsigned char ny, signed int x, int y, signed int * x_table, signed int * y_table, unsigned char * z_table)  
{
int ix=0,jx=0;
long interp1=0, interp2=0, interp3=0;

// bound input arguments
if(x > x_table[nx-1]) x = x_table[nx-1];
 else if(x < x_table[0]) x = x_table[0];

if(y > y_table[ny-1]) y = y_table[ny-1];
 else if(y < y_table[0]) y = y_table[0];

// Find bounding indices in table
for(ix = ny - 2; ix > -1; ix--)  {  // Start with highest index
	//  because will generally have least time for calculations at hi y
	if(y > y_table[ix])  {
		break;
	}
}
if(ix < 0) ix = 0;

for(jx = nx - 2; jx > -1; jx--)  {  // Start with highest index
	// because will generally have least time for calculations at hi x
	if(x > x_table[jx])  {
		break;
	}
}

if(jx < 0) jx = 0;

// do 2D interpolate

interp1 = y_table[ix + 1] - y_table[ix];  // ix+1 is the row
if(interp1 != 0)  {
	interp3 = (y - y_table[ix]);  // current value minus the bin for the row below (=1 for gears)
	interp3 = (100 * interp3); 	  // 100 times the difference
	interp1 = interp3/interp1; 	  // 100 times the [y-y_bin(below)]/[y_bin(above)-y_bin(below)] (=100 for gears)
	}

interp2 =	x_table[jx + 1] - x_table[jx];
if(interp2 != 0)  {
	interp3 = (x - x_table[jx]);  // current value minus the bin for the column to the left
	interp3 = (100 * interp3); 	  // 100 times the difference
	interp2 = interp3/interp2; 	  // 100 times the [x-x_bin(left)]/[y_bin(right)-y_bin(left)]
	}

return((unsigned char) (((100 - interp1) * (100 - interp2) * z_table[ix*nx+jx]  // the value below and to the right
		+ interp1 * (100 - interp2) * z_table[(ix+1)*nx+jx]	  // the portion to add due to the Y axis difference
		+ interp2 * (100 - interp1) * z_table[ix*nx+jx+1]	  // the portion to add due to the X axis difference
		+ interp1 * interp2 * z_table[(ix+1)*nx+jx+1]) / 10000));
} // end intrp_2dcPWM(...)

// --------------- auto_gear_lookup -------------------------------------------------------

unsigned char auto_gear_lookup(int x, int y, unsigned char nx, unsigned char ny, unsigned int * x_table, int * y_table, unsigned char * z_table)  
{
// This function has NO INTERPOLATION, it is designed to look up target gears only 
// (we don't want an interpolated gear, since we might interpolate to the wrong gear 
// when the integer is truncated and returned).
// This function is simpler than the intrp_2dctable() function above (which is used 
// in this code to interpolate the PC table values), but auto_gear_lookup uses 
// similar naming conventions, etc..
// x is mph x10, (outpc.speedo)
// y is kpa x10, (outpc.LOAD_short)
// nx = 12 = NUM_MPH = number of speed bins
// ny = 12 = NUM_MAP = number of kpa bins
// * x_table is pointer to the speed bins
// * y_table is pointer to the kpa bins
// * x_table is pointer to the gear table

char ix,jx,row,col;
// ix is the index that contains the bin 
//    number of the mph bin immediately 
//    below the actual mph value
// jx is the index that contains the bin 
//    number of the kpa bin immediately 
//    below the actual kpa value

unsigned char gear; // returned gear

if ((outpc.manual_gear == -99) || (outpc.hyst_timer)) // manual gear equal to -99, so return current value
	{
	if ((last_target_gear <= inpram.num_gears) && (last_target_gear >= -1)) 
		{
		// if last target gear is valid, use it
		gear = last_target_gear;
		}
		else
		{
		// otherwise use 'D' if last target gear is not valid
		gear = inpram.num_gears;
		}
	return gear;	
	}
	
if ((outpc.manual_gear <= inpram.follow_maxgear) && (outpc.manual_gear > 0) && (inpram.follow_mode == 1)) // burnout mode active and lever in 2nd 
	{
	gear = outpc.manual_gear; // set to lever position only
	return gear;
	}

if (!(WOT_flag) && (outpc.hyst_timer))
	{   
	gear = last_target_gear; // set existing target gear as default
	return gear;
	}	
	else // WOT_flag > 0 OR hyst_flag == 0	
	{    // if we have just had WOT or hysteresis condition met
	     // then go ahead and look up new gear
	// bound input arguments
	// multiply mph and load bins by 10 below 
	// (because values are x10, but bins are not)
	if (x >= (x_table[nx-1]*10))   x = (x_table[nx-1]*10); // use maximum mph (nx-1 because indices are 0-15 for a 16x9 table)
	else if (x <= (x_table[0]*10)) x = (x_table[0]*10);    // use minimum mph

	if (y >= (y_table[ny-1]*10))   y = (y_table[ny-1]*10); // use max load
	else if (y <= (y_table[0]*10)) y = (y_table[0]*10);    // use min load

	// ----- Find indices in table -----

	// First find y bin (kPa) - this is the 'row'
	for (ix = 0; ix <= ny-1; ix++)
		{
		row = ix;
		if(y <= (y_table[ix]*10))  
			{
			break;  // break out of for loop if y is greater than bin value, save row
			} // end if y < (y_table[... 
		} // end for loop
	
	// Now find x bin (speed) - this is the 'column'
	for (jx = 0; jx <= nx-1; jx++)  
		{
		col = jx;
		if(x <= (x_table[jx+1]*10))  
			{
			break;  // break out of for loop if x is greater than bin value, save col
			} // end if
		} // end for
	
	// if at the boundaries of the table
	if (x >= (x_table[nx-1]*10)) col = nx-1;
	if (y >= (y_table[ny-1]*10)) row = ny-1;
	
	// ------ Look Up Target Gear -------------------------------------------
	gear = z_table[(row*nx)+col]; // index runs from 0 to 143 for 16x9 table           
	
	} // end else if ((WOT_flag == 0)
	
	if (gear > inpram.num_gears) gear = inpram.num_gears; // only highest gear
	
	if ((gear > outpc.manual_gear) && (outpc.manual_gear > 0)) 
		{
		gear = outpc.manual_gear; // only <= to gear lever
		}
	
	if (gear < 1) gear = 1; // only forward gears (can't be -99 since we eliminated this above)

	// check for downshifts in decel mode
	if ((outpc.decel_flag) && (gear < outpc.current_gear)) gear = outpc.current_gear; // prevent downshifts

	return(gear);
	} // end auto_gear_lookup()

#pragma CODE_SEG NON_BANKED // Interrupts must be in non-banked memory, because the vectors are only 16 bit and 
							// will not accommodate the appended 6 bit ppage register. 
// 4.110 changes
#ifdef COMMS_382X
  #include "..\..\serCANComm_382x\serComm.h"
  #include "..\..\serCANComm_382x\CANComm.h"
#else
  #include "..\..\serCANComm_1\serComm.h"
  #include "..\..\serCANComm_1\CANComm.h"
#endif
// --------------- VSS Timer ----------------------------------------
// Time the period on the VSS input, set in array ic_period[]

// #ifdef NEW_VSS

INTERRUPT void VSS_timer(void)  
{
unsigned int TC0_save;
unsigned long TC0_now;
unsigned char rising_edge;
#ifdef COMMS_382X
unsigned char savTIE,savCANTIER,savCANRIER;
#endif

if (in2ram.VSS_ISR == 1) // if new VSS
	{
	TC0_now = TC0;
	
	rising_edge = (TCTL4 & 0x01); // 0x01 if TLTL4 = 01, 0x00 if TCTL4 = 10 
	TCTL4 ^= 0x03; // 0x03 = 00000011
	
	// Last interrupts times are:
	// - TC0_up 	(oldest valid rising edge) 
	// - TC0_dwn    (oldest valid falling edge)
	// - TC0_last
	// - TC0_now 	(current)

	if ((TC0_now - TC0_last) > (vssTics)) // if long enough, otherwise ignore
		{
		if (!inpram.InputCaptureEdge) // use rising edge, 0 = rising edge, 1 = falling edge
			{
			if (rising_edge == 1)
				{
				// Optimize:
				period1 = (((VSS_ovflow - (TC0_up > TC0_now)) * 65535) + (TC0_now - TC0_up)); // Last rising edge to now	
				TC0_dwn  = TC0_last; // declare last falling edge as valid
				TC0_last = TC0_now;	 // capture current time for next event
				} // end if rising_edge == 1
				else // falling edge
				{
				TC0_up = TC0_last; // declare last rising edge as valid
				}
			}  // end if (inpram.InputCaptureEdge == 0)
			else // inpram.InputCaptureEdge != 0, use falling edge
			{
			if (rising_edge == 0)
				{
				period1 = (((VSS_ovflow - (TC0_dwn > TC0_now)) * 65535) + (TC0_now - TC0_dwn)); // Last falling edge to now 
				TC0_up = TC0_last; // declare last rising edge as valid
				TC0_last = TC0_now;
				} // end if rising_edge == 0
				else // rising edge == 1
				{
				TC0_dwn = TC0_last; // declare last falling edge as valid
				}
			}  // end else if (inpram.InputCaptureEdge == 1)
			VSS_ovflow = 0;   	// Reset overflow count for next interrupt
		} // end if ((TC0_now - TC0_last) > (minTics/4))
		else // too short
		{
		// TC0_last = TC0_now; // reset last edge time
							// don't reset overflows, etc.
		goto DONE;
		}
	
	#ifdef COMMS_382X  
	   savTIE = 0;
	#endif 
	           
	// set interrupt received flag
	VSS_rcvd++;                      // will clear flag (set to zero) in 0.010 second section
	msecVSS = 0;
	
	// increment tooth counter (regardless of filtering)
	vss_teeth++;
	
	// set interrupt received flag
	VSS_rcvd++;     // will clear flag (set to zero) in 0.010 second section
	
	#ifdef COMMS_382X
	// This is because this ISR is very long and can mess up TS when 
	// using Android comms.  So what is done after entry to the ISR is to disable the 
	// spark + fuel timer interrupt itself + the CAN ISR + the .128 ms timer clock so that 
	// if a serial request is made it will jump to that and then come back to the spark/fuel. 
	// Evidently getting data through a smart phone at very high speed is more important to 
	// nerd users than their engine, even though they have no clue how to use the data.
	
	if(in2ram.nestINTS)  
	  {
	  // start nested interrupts
  	  // disable all ISRs except SCI
	  savTIE = TIE;
	  TIE = 0;
	  TSCR2 &= ~0x80;    // disable timer overflow interrupt (TOI)
	  savCANTIER = CANTIER;
	  CANTIER = 0;
	  savCANRIER = CANRIER;
	  CANRIER = 0;
	  CRGINT &= ~0x80;  // disable interrupt
	  ENABLE_INTERRUPTS
	  }
	#endif
	
	// reject if outside mask range of current average
	// ic_per_avg is ~100 times the period
	// We calculate long_period and short_period in 0.010 section of timer

	if ((VSS_error > inpram.vss_error_max) || restart) goto DONE;

	if (period1 > (long_period_vss)) 
		{
		// period1 = long_period_vss; // don't limit slow speed value
		VSS_error++; // increment the error count because we had a missed tooth 
		goto DONE;					  // but don't save period
		}
		else if (period1 < (short_period_vss))
		{
		// period1 = short_period_vss; // don't limit slow speed value, just don't record it
		VSS_error++;  // increment the error count because we had an extra tooth 		
		goto DONE;					   // but don't save period
		}
	
	// no exclusion conditions met
	if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x20))) // if not testing VSS
		{
		ic1_period[ic1_count] = period1;    // 5.333 sec ticks
		ic1_count++;   // sample counter for last 20 input capture periods
		if (ic1_count > 20) ic1_count=1;
		}
		
	TC0_last = TC0_now; // reset last edge time
	
	// Increment (or reset) the counter for the speedo output
	if  (in2ram.sp0_mode == 0)
	  { // if using speedo output
	  if (speedo_counter < speedo_total) speedo_counter++; else speedo_counter = 0;
	  if (speedo_counter < speedo_toggle)  
		{
		// Set pin PT4 high
		PTT |= 0x10;
		}
		else  
		{
		// set pin PT4 low
		PTT &= ~0x10;   
		} // end if  (speedo_counter <= speedo_toggle)  	
	  }	 // end if  (in2ram.sp0_mode == 0)
	
	DONE:
	
	// reset interrupt
	TFLG1 = 0x01;           // write a one (00000001) to TFLG to clear flag for timer channel 0 
	
	#ifdef COMMS_382X
	  if(savTIE)  
	    {
	    // end nested interrupts
	    // re-enable all ISRs
	    DISABLE_INTERRUPTS  
	    TIE = savTIE;
	    TSCR2 |= 0x80;    // enable timer overflow interrupt (TOI)
	    CANTIER = savCANTIER;
	    CANRIER = savCANRIER;
	    CRGINT |= 0x80;   // enable interrupt
	    }
	#endif
	
	return;
	}  // end VSS_timer()
	else // VSS_ISR == 0 == old code
	{
	// save interrupt time for use later in routine
	TC0_save = TC0;  // from 0x0000 to 0xFFFF
	#ifdef COMMS_382X  
	   savTIE = 0;
	#endif            
	// set interrupt received flag
	VSS_rcvd++;                      // will clear flag (set to zero) in 0.010 second section
	msecVSS = 0;
	
	// reset interrupt
	TFLG1 = 0x01;                    // write a one (00000001) to TFLG1 to clear flag for timer channel 0 
	// period1 = (TC0_save - tcsav1);   	// tcsav1 is a global variable - it is the time of the last interrupt
	
	if (tcsav1 > TC0_save) VSS_ovflow = VSS_ovflow - 1;
	period1 = ((VSS_ovflow)*0xFFFF) + (TC0_save - tcsav1);
		
	VSS_ovflow = 0;   // Reset overflow count for next interrupt

	// if too short, do nothing (don't update tc1sav) and return
	if (period1 < minTics) return;

	// increment tooth counter (regardless of filtering)
	vss_teeth++;

	// set interrupt received flag
	VSS_rcvd++;     // will clear flag (set to zero) in 0.010 second section

	#ifdef COMMS_382X
	// This is because this ISR is very long and can mess up TS when 
	// using Android comms.  So what is done after entry to the ISR is to disable the 
	// spark + fuel timer interrupt itself + the CAN ISR + the .128 ms timer clock so that 
	// if a serial request is made it will jump to that and then come back to the spk/fuel. 
	// Evidentially getting data through a smart phone at very high speed is more important to 
	// nerd users than their engine, even though they have no clue how to use the data.
	if(in2ram.nestINTS)  
	  {
	  // start nested interrupts
	  // disable all ISRs except SCI
	  savTIE = TIE;
	  TIE = 0;
	  TSCR2 &= ~0x80;    // disable timer overflow interrupt (TOI)
	  savCANTIER = CANTIER;
	  CANTIER = 0;
	  savCANRIER = CANRIER;
	  CANRIER = 0;
	  CRGINT &= ~0x80;  // disable interrupt
	  ENABLE_INTERRUPTS
	  }
	#endif
	
	if ((VSS_error > inpram.vss_error_max) || restart) goto DONE2;
	
	if (period1 > long_period_vss) 
		{
		if (VSS_error < inpram.vss_error_max) VSS_error++; // increment the error count because we had a missed tooth 
		return;
		}
		else if (period1 < short_period_vss)
		{
		period1 = short_period_vss;
		if (VSS_error < inpram.vss_error_max) VSS_error++; // increment the error count because we had a short tooth 		
		return;
		}
	
	DONE2:
	
	// no exclusion conditions met
	if ((!(in2ram.test_mode & 0x01)) || (!(in2ram.test_mode & 0x20))) // if not testing VSS
		{	
		ic1_period[ic1_count] = period1;    // 5.333 sec ticks
		ic1_count++;   // sample counter for last 20 input capture periods
		if (ic1_count > 20) ic1_count=1;
		}
		
	tcsav1 = TC0_save;                  // save time of the latest trigger edge
		
	// Increment (or reset) the counter for the speedo output
	if  (in2ram.sp0_mode == 0)
	  { // if using speedo output
	  if (speedo_counter < speedo_total) speedo_counter++; else speedo_counter = 0;
	  if (speedo_counter < speedo_toggle)  
		{
		// Set pin PT4 high
		PTT |= 0x10;
		}
		else  
		{
		// set pin PT4 low
		PTT &= ~0x10;   
		} // end if  (speedo_counter <= speedo_toggle)  	
	  }	 // end if  (in2ram.sp0_mode == 0)
	
	// reset interrupt
	TFLG1 = 0x01;           // write a one (00000001) to TFLG to clear flag for timer channel 0 
	
	// 4.110 changes
	#ifdef COMMS_382X
	  if(savTIE)  
	    {
	    // end nested interrupts
	    // re-enable all ISRs
	    DISABLE_INTERRUPTS  
	    TIE = savTIE;
	    TSCR2 |= 0x80;    // enable timer overflow interrupt (TOI)
	    CANTIER = savCANTIER;
	    CANRIER = savCANRIER;
	    CRGINT |= 0x80;   // enable interrupt
	    }
	#endif  
	return;
	}
}  // end VSS_timer()

// --------------- ISS/non-CAN Tach Timer ----------------------------------------
// Time the period on the VSS input, set in array ic_period[]

INTERRUPT void ISS_timer(void)  
{
unsigned long period2;
unsigned int TC5_save;

if ((in2ram.test_mode & 0x01) && (in2ram.test_mode & 0x40)) // if testing ISS skip ISR
	{
	return;	
	}

// set interrupt received flag
// ISS_rcvd++;     // will clear flag (set to zero) in 0.010 second section
// reset interrupt
TC5_save = TC5;  // save initial value
period2 = (TC5_save - tcsav2);   // tcsav2 is a global variable - it is the time of the last interrupt
TFLG1 |= 0x20;     // write a one to TFLG bit 5 (00100000) to clear flag for timer channel 5     

// reject if outside mask range of current average
// We calculate long_period and short_period in 0.010 section of timer

if (!restart)
	{	
	if (inpram.iss_mask != 0)
		{
		if (period2 < short_period_iss) period2 = short_period_iss;
		}
	} // end if !restart

ic2_period[ic2_count] = period2; // set the period

tcsav2 = TC5_save;               // save time of new trigger edge for next interrupt

ic2_count++;   // sample counter for last 20 input capture periods
if (ic2_count > 20) ic2_count=1;  
return;
}  // end ISS_timer()

#pragma CODE_SEG  NON_BANKED
INTERRUPT void Timer_Clock_ISR(void)  
{
// .128 ms clock interrupt - clear flag immediately to resume count
CRGFLG = 0x80;  // clear RTI interrupt flag
// also generate 1.024 ms, .10035 sec and 1.0035 sec clocks
lmms++;         // free running clock(.128 ms tics) good for ~ 110 hours
can_mmsclk++;
mms++;          // in .128 ms tics - reset every 8 tics = 1.024 ms
	
// check mms to generate other clocks
if (mms < 8) goto CLK_DONE;

mms = 0;

//////////////////////////// 0.001 'millisecond' section ///////////////////////

millisec++;     // actually 1.024 ms, for seconds count
ODOmsec++;
msecVSS++;
lastCANcheck++;
if (last0001) last0001--; // count down to zero
if (last0010) last0010--; // count down to zero
if (last0100) last0100--; // count down to zero
if (last1000) last1000--; // count down to zero
if (DelayTimer1) DelayTimer1--;
if (DelayTimer2) DelayTimer2--;
if (Slip_msec < 65535) Slip_msec++;

// milliseconds
if (millisec >= 976) 
   	{
   	millisec = 0;
   	// update seconds to send back to TunerStudio
	outpc.seconds++;
   	}

CLK_DONE:   // Statements after this label are executed every 0.128 millisecond clock 'tic'

// Check if spare port in use as PWM and active - increment counter if they are
if (sp[0] == 1) 
	{
    // get PWM table value in main loop	
    sp1_pwm_count++;
    if (sp1_pwm_count <= sp1_pwm_hi) 
    	{
    	SPARE1_OUTport |= SPARE1pin;
    	}
    	else 
    	{   		
    	SPARE1_OUTport &= ~SPARE1pin;  // write to output directly to avoid function call overhead
    	}
	if (sp1_pwm_count > (inpram.sp1_pwm_per - 1)) sp1_pwm_count = 0; // reset counter when total is reached
	}   // end sp1_pwm
	else // sp[0] == 0
	{
	if (inpram.sp1_off) sp1_pwm_hi = (unsigned char) ((((int) inpram.sp1_off * (int) inpram.sp1_pwm_per) + 50)/100); else sp1_pwm_hi = 0;
	sp1_pwm_count++;
    if ((sp1_pwm_count <= sp1_pwm_hi) && (sp1_pwm_hi)) 
    	{
    	SPARE1_OUTport |= SPARE1pin;
    	}
    	else 
    	{   		
    	SPARE1_OUTport &= ~SPARE1pin;  // write to output directly to avoid function call overhead
    	}
	}
	
////////////////////////////////////////////////

if (sp[1] == 1) 
	{
    // get PWM table value in main loop	
    sp2_pwm_count++;
    if (sp2_pwm_count <= sp2_pwm_hi) 
    	{
    	SPARE2_OUTport |= SPARE2pin;
    	}
    	else 
    	{   		
    	SPARE2_OUTport &= ~SPARE2pin;  // write to output directly to avoid function call overhead
    	}
	if (sp2_pwm_count > (inpram.sp2_pwm_per - 1)) sp2_pwm_count = 0; // reset counter when total is reached
	}   // end sp1_pwm
	else
	{
	if (inpram.sp2_off) sp2_pwm_hi = (unsigned char) ((((int) inpram.sp2_off * (int) inpram.sp2_pwm_per) + 50)/100); else sp2_pwm_hi = 0;
	sp2_pwm_count++;
    if ((sp2_pwm_count <= sp2_pwm_hi) && (sp2_pwm_hi)) 
    	{
    	SPARE2_OUTport |= SPARE2pin;
    	}
    	else 
    	{   		
    	SPARE2_OUTport &= ~SPARE2pin;  // write to output directly to avoid function call overhead
    	}
	}	

if (inpram.out1_pwm_off)     // if OUTPUT1/5/6 PWM set
{
// Bit-bang PWM on output1 (PE4) and output5/6
out1_pwm_count++;       // increment PWM counter

if (!(dither_flag & 0x02))  // if not on a refresh cycle on output1 (solA)
	{
	// Toggle solA (aka. output1)
	if (Output1state && (out1_pwm_count < inpram.out1_pwm_on))    // counter below out_pwm_on value (start with on portion of cycle)
		{
		OUTPUT1port |= OUTPUT1pin;      // if solA is active, set PE4 high
		}
	else     // counter between out_pwm_on and (out_pwm_off+out_pwm_on)
		{
		if (inpram.PWM_enable & 0x01) OUTPUT1port &= ~OUTPUT1pin;     // set PE4 low
		}    // end if  (Output1state && ...
	}  // end if (!dither_flag...
		
if (!(dither_flag & 0x08))  // if not on a refresh cycle on output5 (solE)
	{ 
	// Toggle solE
	if (Output5state && (out1_pwm_count < inpram.out1_pwm_on))     // counter below out_pwm_on value
		{
		if (inpram.spareInputs > 0) OUTPUT5port |= OUTPUT5pin; // if solE is active, set solD high
		}
	else     // counter between out_pwm_on and (out_pwm_off+out_pwm_on)
		{
		if ((inpram.spareInputs > 0) && (inpram.PWM_enable & 0x10)) OUTPUT5port &= ~OUTPUT5pin;// set solE low
		}                     // end if  (Output5state && ...
	}  // end if (!dither_flag...
	
	
if (!(dither_flag & 0x08))  // if not on a refresh cycle on output6 (solF)
	{ 
	// Toggle solF
	if (Output6state && (out1_pwm_count < inpram.out1_pwm_on))     // counter below out_pwm_on value
		{
		if (inpram.spareInputs == 2) OUTPUT6port |= OUTPUT6pin; // if solF is active, set solD high
		}
		else     // counter between out_pwm_on and (out_pwm_off+out_pwm_on)
		{
		if ((inpram.spareInputs == 2) && (inpram.PWM_enable & 0x20)) OUTPUT6port &= ~OUTPUT6pin;// set solF low
		}                     // end if  (Output6state && ...
	}  // end if (!dither_flag...

					
if (out1_pwm_count > (inpram.out1_pwm_off + inpram.out1_pwm_on - 1)) out1_pwm_count = 0; // reset counter when total is reached

} // end if (inpram.out12_pwm_off)


if (inpram.out2_pwm_off)     // if OUTPUT2 PWM set
	{
	// Bit-bang PWM on output2 (PM2)
	out2_pwm_count++;       // increment PWM counter

	if (!(dither_flag & 0x04))  // if not on a refresh cycle on output2 (solB)
		{ 
		// Toggle solB (aka. output2)
		if (Output2state && (out2_pwm_count < inpram.out1_pwm_on))     // counter below out_pwm_on value
			{
			OUTPUT2port |= OUTPUT2pin; // if solB is active, set PM2 high
			}
		else     // counter between out_pwm_on and (out_pwm_off+out_pwm_on)
			{
			if (inpram.PWM_enable & 0x02) OUTPUT2port &= ~OUTPUT2pin;// set PM2 low
			}                     // end if  (Output2state && ...
		}  // end if (!dither_flag...
		
	if (out2_pwm_count > (inpram.out1_pwm_off + inpram.out2_pwm_on - 1)) out2_pwm_count = 0; // reset counter when total is reached

	} // end if (inpram.out12_pwm_off)

return;
} // end Timer_Clock_ISR

// Variable Pulse Width Transmit timer ISR
INTERRUPT void VPW_transmit_ISR(void)
{

}

// --------------- ISR_TimerOverflow ------------------------------------

INTERRUPT void Timer_Overflow_ISR(void) 
// Timer interrupt overflow service routine
// for TOI (bit 7 of TSCR2).
// Interrupts when TOF flag set if TOI == 1 
// The clock ticks are counted in the variable TCNT (which is defined in the 
// hsc12def.h file at memory location 0x0044). 
// There are 0xFFFF 'tics' (65,536 counts).
// In our case each 'tic' is 1/0.1875MHz  = 5.333 s, so the overflow occurs 
// every 65535 * 5.3333 = 349.53 millisecond.
{
// Count VSS overflows
VSS_ovflow++;

// Get display seconds from continuously running TCNT
TC_ovflow++;
if(TC_ovflow >= TC_ovfla[TC_ov_ix])  
	{
	if(TC_ov_ix >= 59)  
	{
	TC_ov_ix = 0;
	TC_ovflow = 0;
	}    // end if if(TC_ov_ix
	else
	TC_ov_ix++;
	} // end if(TC_ovflow >=...

// clear timer overflow interrupt flag (TOF)
TFLG2 = 0x80;
return;
} // end Timer_Overflow_ISR()

#ifdef COMMS_382X
// #pragma CODE_SEG ROM_3BC
#pragma CODE_SEG NON_BANKED

void crc32(unsigned char sub_no, unsigned char carg1, unsigned char carg2)  
{
	unsigned char *p;
	unsigned short kbytes;
	unsigned long crc;
	
	if (sub_no != 13) return;
	
	if(carg1 == 0)
		{
		// pick ram or flash table
		if(tables[kFlag].addrRam != 0)  // kFlag is the serCOM table index from Transmit Interrupt in serComm.h
	  		p = tableByteRam(kFlag,0);
		else
	  		p = tableByteFlash(kFlag,0);  // kFlag is the serCOM table index from Transmit Interrupt in serComm.h
	
		kbytes = tableBytes(tble_idx);
	
		// initialize
		crc = 0xFFFFFFFF;
  
  		// do calculation
		while (kbytes--)
			crc = crc_tab[(crc ^ *p++) & 0xFF] ^ (crc >> 8);

		crc32KVal = crc ^ 0xFFFFFFFF;
		return;
		}  // end if (carg1 == 0)
		else
		{  // carg1 != 0
		crcTab = crc_tab[(crc32Val ^ carg2) & 0xFF];
		}
}  // end crc32()

#endif
#pragma CODE_SEG DEFAULT

// ------ Solenoid functions -------------------------------------------------

void Output1(unsigned char set_to) 
{
// set off if set_to=0, on otherwise
if (set_to == 1)
	{ // SolA state in first is 3rd bit in Output1 
	OUTPUT1port |= OUTPUT1pin; // turn on SolA
	Output1state = 1;            // Set state indicator
	outpc.solst |= 0x01;   // set first bit for datalog
	}
	else
	{
	OUTPUT1port &= ~OUTPUT1pin;// turn off SolA
	Output1state =0;             // Set state indicator
	outpc.solst &= ~0x01;  // set first bit for datalog
	}
return;
}  // end Output1()

void Output2(unsigned char set_to) 
{
// set off if set_to=0, on otherwise
if (set_to)
	{ // SolB state in first is 3rd bit in Output1 
	OUTPUT2port |= OUTPUT2pin; // turn on Sol B, 0x04 = 00000100, 
						   // so set port M pin 2 to 1 
	Output2state = 1;            // Set state indicator
	outpc.solst |= 0x02;   // set second bit for datalog
	}
	else
	{
	OUTPUT2port &= ~OUTPUT2pin;// turn off Sol B, ~0x04 = 11111011, 
						   // so set port M pin 2 to 0 
	Output2state = 0;            // Set state indicator
	outpc.solst &= ~0x02;  // set second bit for datalog
	}
return;
}   // end Output2()

void Output4(unsigned char set_to) 
{
// set off if set_to=0, on otherwise
if (set_to)
	{ // SolD state in first is 3rd bit in Output1 
	sp2_pwm_hi = (unsigned char) (((int) intrp_1d_pwm(sp2_indicial, inpram.sp1_pwm_indx, inpram.sp2_pwm_table) * (int) inpram.sp2_pwm_per)/100);
	OUTPUT4port |= OUTPUT4pin; // turn on Sol D, 
					 	   // so set port M pin 2 to 1 
	Output4state = 1;            // Set state indicator
	outpc.solst |= 0x08;   // set 4th bit for datalog
	sp[1] = 1;         // set datalog value
	}
	else
	{
	sp2_pwm_hi = (unsigned char) ((((int) inpram.sp2_off * (int) inpram.sp2_pwm_per) + 50)/100);
	OUTPUT4port |= OUTPUT4pin; // turn off Sol D,
						   // so set port M pin 2 to 0 
	Output4state = 0;            // Set state indicator
	outpc.solst &= ~0x08;  // set 4th bit for datalog
	sp[1] = 0;         // set datalog value
	} // end if (set_to)
outpc.sp2PWM = (uchar) ((((int) sp2_pwm_hi*100) + (inpram.sp2_pwm_per/2))/(int) inpram.sp2_pwm_per);
return;
}    // end Output4()

void Output5(unsigned char set_to) 
{
// set off if set_to=0, on otherwise
if (set_to)
	{ // SolE state in first is 4th bit in Output1 
	OUTPUT5port |= OUTPUT5pin; // turn on Sol E
	Output5state = 1;            // Set state indicator
	outpc.solst |= 0x10;   // set 5th bit for datalog
	}
	else
	{
	OUTPUT5port &= ~OUTPUT5pin;// turn off Sol E 
	Output5state = 0;            // Set state indicator
	outpc.solst &= ~0x10;  // set 5th bit for datalog
	}
return;
}   // end Output5()

void Output6(unsigned char set_to) 
{
// set off if set_to=0, on otherwise
if (set_to)
	{ // SolF state in first is 5th bit in Output1 
	OUTPUT6port |= OUTPUT6pin; // turn on Sol F
	Output6state = 1;            // Set state indicator
	outpc.solst |= 0x20;   // set 6th bit for datalog
	}
else
	{
	OUTPUT6port &= ~OUTPUT6pin;// turn off Sol F 
	Output6state = 0;            // Set state indicator
	outpc.solst &= ~0x20;  // set 6th bit for datalog
	}
return;
}  // end Output6()

void Output7(unsigned char set_to) 
{
// set off if set_to=0, on otherwise
if (set_to)
	{ // SolG state in first is 6th bit in Output1 
	sp1_pwm_hi = (unsigned char) (((int) intrp_1d_pwm(sp1_indicial, inpram.sp1_pwm_indx, inpram.sp1_pwm_table) * (int) inpram.sp1_pwm_per)/100);
	OUTPUT7port |= OUTPUT7pin; // turn on output 7
	Output7state = 1;            // Set state indicator
	outpc.solst |= 0x40;   // set 7th bit for datalog
    sp[0] = 1;         // set datalog value
	}
else
	{
    sp1_pwm_hi = (unsigned char) ((((int) inpram.sp1_off * (int) inpram.sp1_pwm_per) + 50)/100);
    OUTPUT7port &= ~OUTPUT7pin;// turn on output 7
	Output7state = 0;            // Set state indicator
	outpc.solst &= ~0x40;  // set 6th bit for datalog
    sp[0] = 0;         // set datalog value   
	}
outpc.sp1PWM = (uchar) ((((int) sp1_pwm_hi*100) + (inpram.sp1_pwm_per/2))/(int) inpram.sp1_pwm_per);
return;
}   // end Output7()


void Output8(unsigned char set_to) 
{ // If speedo output used as shift solenoid
// set off if set_to=0, on otherwise
if (in2ram.sp0_mode == 0) return; // return if used as speedo output
if (set_to)
	{ 
	SP0_DTY = (unsigned char) (((intrp_2dcPWM(16, 9, (outpc.sp016inx/10), (outpc.sp09inx/10), (int *) in2ram.sp0_16Bin, (int *) in2ram.sp0_9Bin, (uchar *) in2ram.sp0_table) * SP0_PER) + 50)/100);
	Output8state = 1; 		// Set state indicator
	outpc.solst |= 0x80;	// set 7th bit for datalog
    sp0 = 1;				// set datalog value
	}
else
	{
    SP0_DTY = (unsigned char) ((((int) in2ram.sp0_off * (int) SP0_PER) + 50)/100);
	Output8state = 0;		// Set state indicator
	outpc.solst &= ~0x80;	// set 7th bit for datalog
    sp0 = 0;				// set datalog value   
	}
outpc.sp0PWM = (uchar) ((((int) SP0_DTY*100) + (SP0_PER/2))/(int) SP0_PER);
return;
}     // end Output8()

void Output9(unsigned char set_to) 
{ // If line pressure control (PC) used as shift solenoid
// set off if set_to = 0, on otherwise
if (in2ram.sp3_mode != 3) return; // return if not in shift solenoid mode
if ((set_to) && (sp3 != 1))
	{ 
	PC_DUTY = (uchar) (((intrp_2dcPWM(16, 9, (outpc.sp316inx/10), (outpc.sp39inx/10), (int *) in2ram.sp3_16Bin, (int *) in2ram.sp3_9Bin, (uchar *) in2ram.sp3_table) * PC_PER) + 50)/100);
	outpc.sp3PWM = (uchar)((PC_DUTY*100)/PC_PER);
	Output9state = 1; 		// Set state indicator
	outpc.solst |= 0x100;	// set 8th bit for datalog
    sp3 = 1;				// set datalog value
	}
else if (!(set_to) && (sp3 != 0))
	{
    PC_DUTY = (uchar) (((in2ram.sp3_off * PC_PER) + 50)/100);
    outpc.sp3PWM = (uchar)((PC_DUTY*100)/PC_PER);
	Output9state = 0;		// Set state indicator
	outpc.solst &= ~0x100;	// set 8th bit for datalog
    sp3 = 0;				// set datalog value   
	}
outpc.sp3PWM = (uchar) ((((int) PC_DUTY*100) + (PC_PER/2))/(int) PC_PER);
return;
}     // end Output8()

// -------------- Spare Port functions ---------------------------------
void spr0(unsigned char set_to) 
{
unsigned char tmp_uchar;

if (in2ram.sp0_mode == 0) return; // return if used as speedo output

// set off if set_to=0, on otherwise
if (set_to == 1)
  {
  tmp_uchar = (unsigned char) (((intrp_2dcPWM(16, 9, (outpc.sp016inx/10), (outpc.sp09inx/10), (int *) in2ram.sp0_16Bin, (int *) in2ram.sp0_9Bin, (uchar *) in2ram.sp0_table) * SP0_PER) + 50)/100);
  if (tmp_uchar > SP0_PER) SP0_DTY = SP0_PER; else SP0_DTY = tmp_uchar;
  outpc.sp0PWM = (unsigned char) (((SP0_DTY*100) + (SP0_PER/2))/SP0_PER);
  sp0 = 1;        // set datalog values
  }
 else // set_to == 0 
  {
  SP0_DTY = (unsigned char) ((((int) in2ram.sp0_off * (int) SP0_PER) + 50)/100);
  outpc.sp0PWM = (unsigned char) (((SP0_DTY*100) + (SP0_PER/2))/SP0_PER);
  sp0 = 0;        // set datalog value
  }
return;
} // end void spr0(unsigned char set_to) 
 
void spr1(unsigned char set_to) 
{
// set off if set_to=0, on otherwise
if (set_to)
  {
  sp1_pwm_hi = (unsigned char) (((int) intrp_1d_pwm(sp1_indicial, inpram.sp1_pwm_indx, inpram.sp1_pwm_table) * (int) inpram.sp1_pwm_per)/100);
  sp[0] = 1;        // set datalog values
  }
 else  // set_to  == 0
  {
  if (inpram.sp1_off) sp1_pwm_hi = (unsigned char) ((((int) inpram.sp1_off * (int) inpram.sp1_pwm_per) + 50)/100); else sp1_pwm_hi = 0;
  sp[0] = 0;        // set datalog value
  }
outpc.sp1PWM = (unsigned char) ((((int) sp1_pwm_hi*100) + (inpram.sp1_pwm_per/2))/(int) inpram.sp1_pwm_per);
return;
} // end void spr1(unsigned char set_to) 

void spr2(unsigned char set_to) 
{						
if (inpram.sp_mode[1] == 5) return; // return if using gear indicator mode

// set off if set_to=0, on otherwise
if (set_to)
  {
  // SPARE2_OUTport |= SPARE2pin;        // turn on PA0
  sp2_pwm_hi = (unsigned char) ((((int) intrp_1d_pwm(sp2_indicial, inpram.sp2_pwm_indx, inpram.sp2_pwm_table) * (int) inpram.sp2_pwm_per) + 50)/100);
  sp[1] = 1;        // set datalog value
  }
 else
  {
  // SPARE2_OUTport &= ~SPARE2pin;       // turn on PA0
  if (inpram.sp2_off) sp2_pwm_hi = (unsigned char) (((((int) inpram.sp2_off * (int) inpram.sp2_pwm_per)) + 50)/100); else sp2_pwm_hi = 0;
  sp[1] = 0;        // set datalog value
  }
outpc.sp2PWM = (unsigned char) ((((int) sp2_pwm_hi*100) + (inpram.sp2_pwm_per/2))/(int) inpram.sp2_pwm_per);
return;
}  // end void spr2(unsigned char set_to) 

void spr3(unsigned char set_to) 
{
unsigned char tmp_uchar;
// set off if set_to=0, on otherwise
if (set_to == 1)
  {
  if (sp3 == 0)
  	{
  	// if not already on, set initial PWM%, otherwise update from 16x9 table 
  	// in main loop to get temperature and pressure adjustments 
    tmp_uchar = (unsigned char) (((intrp_2dcPWM(16, 9, (outpc.sp316inx/10), (outpc.sp39inx/10), (int *) in2ram.sp3_16Bin, (int *) in2ram.sp3_9Bin, (uchar *) in2ram.sp3_table) * SP3_PER) + 50)/100);
    if (tmp_uchar > SP3_PER) SP3_DTY = SP3_PER; else SP3_DTY = tmp_uchar;
  	}
  sp3 = 1;        // set datalog values
  }
 else // set_to == 0 
  {
  if (sp3 == 1) SP3_DTY = (unsigned char) ((((int) in2ram.sp3_off * (int) SP3_PER) + 50)/100);
  sp3 = 0;        // set datalog value
  }
outpc.sp3PWM = (uchar)((SP3_DTY*100)/SP3_PER);
return;
} // end void spr0(unsigned char set_to) 

// ------ Unimplemented ISRs -------------------------------------------------
#pragma CODE_SEG  NON_BANKED
INTERRUPT void UnimplementedISR(void) 
{
// Unimplemented ISRs trap.
return;
} // end UnimplementedISR()

#pragma CODE_SEG DEFAULT 

#include "flashburn.h"

// CAN Adjustments for Fuel/Spark/Idle on MS-II
#pragma CODE_SEG NON_BANKED

#include "..\..\serCANComm_1\sendCANAdj.h"

// #pragma CODE_SEG ROM_7000
#pragma CODE_SEG NON_BANKED

void Output3(unsigned char set_to)
{
if (inpram.NAGmode != 1)
	{
	
	// set state indicators
	if (OUTPUT3pwm > 0)
		{
		outpc.solst |= 0x04;
		Output3state = 1;
		}
		else
		{
		outpc.solst &= ~0x04;
		Output3state = 0;
		}
		return;
	}
	else
	{ // NAG mode
	if (set_to > 0)
		{
		outpc.solst |= 0x04;
		Output3state = 1;
		}
		else
		{
		outpc.solst &= ~0x04;
		Output3state = 0;
		return;
		}
	}
	
} // end Output3()

//////////////////////////////////////////////////////////////////////  

void grabMS2vars(void)
{
unsigned char ix;

if (!inpram.can_testout)
{
// Get ECU output variables.
// Do below if want entire outpc array - very slow
// Grab 8 bytes every can_var_rate (.128 ms tics)
if(can_mmsclk > inpram.can_var_rate)  
 	{
 	// load transmit(xmt) interrupt ring buffer(can[0]) - request data
 	ix = can[1].cxno_in;
 	can[1].cx_msg_type[ix] = MSG_REQ; 
 	// Get all data from var. block outpc in ECU)      
 	can[1].cx_destvarblk[ix] = inpram.ms2varBLK; // (was 7) is outpc on MS-II
 	// Put data rcvd from ECU into var block msvar in GPIO)      
 	can[1].cx_myvarblk[ix] = 9;   // msvar table number for this GPIO (from tableDescriptor)
 	can[1].cx_myvaroff[ix] = can_varoff;
 	can[1].cx_dest[ix] = inpram.ms2canID; // (was 0) send to ECU (board 0)
 	// below is offset from start of outpc in ECU
 	can[1].cx_destvaroff[ix] = can_varoff;
 	// calculate no bytes, and offset for next time
 	can_varoff += 8;
	
 	if(can_varoff < in2ram.msOutpcBytes)
		{
		can[1].cx_varbyt[ix] = 8;	 // get 8 bytes
		}
		else  
		{  // get remaining bytes                       
		can[1].cx_varbyt[ix] = in2ram.msOutpcBytes - (can_varoff - 8);
		can_varoff = 0; 
		}
 	send_can(1);
 	can_mmsclk = 0;
   	} // end if(can_mmsclk > inpram.can_var_rate)
 }     // end if (!inpram.can_testout)
 else // if (inpram.can_testout)
 {
 // Do below if want specific, non-contiguous vars from outpc array                                                           _
 // Grab all bytes every can_var_rate (.128 ms tics)
 // Note: *** Must set same outmsg data in ECU and CAN device, or at 
 // least have the same no_bytes and offsets.
	if(can_mmsclk > (inpram.can_var_rate*2)) // double can_var_rate for outmsg  
	{
		ix = can[1].cxno_in;
		can[1].cx_msg_type[ix] = OUTMSG_REQ; 
		// Put data from ECU in this gpio block      
		// can[1].cx_destvarblk[ix] = 3;  // from tableDescriptor
		can[1].cx_destvarblk[ix] = 9;  // into msvar
		// above really myvarblk, destvarblk always outpc
		can[1].cx_dest[ix] = 0;		// send to ECU (board 0)
		// Below is offset from start of gpio block - it should
		// always be 0. If more than 1 msg, ECU will send offset
		// to count byte sent
		can[1].cx_destvaroff[ix] = 0;
		can[1].cx_outmsg_no[ix] = 0;  // 0 to < MAX_OUTMSGS
		send_can(1);
	can_mmsclk = 0;
	}
 } // end else (!inpram.can_testout)

return;
}  // end grabMS2var()


#pragma CODE_SEG  NON_BANKED

void checkCANbus (void)
{
 if (lastCANcheck < 3) return; // only perform every few milliseconds
 // do something keep CAN comms alive
 #ifdef COMMS_382X
 	lastCANcheck = 0; // rest timer
	if(kFlag >= 0)  
		{  	// kFlag is the serCOM table index from Transmit Interrupt in serComm.h
			// kFlag = -2 -> waiting for TS request
			// kFlag = -2 -> processing request
			// kFlag = 0..15 -> table descriptor value
		// calculate crc32
		crc32(13,0,0); // calculate crc32KVal
		kFlag = -2;	  // reset kFlag
		uctmp = *(unsigned char *)&crc32KVal;
		crc32(13,99,uctmp);	 // calculates crcTab = crc_tab[(crc32Val ^ uctmp) & 0xFF]
		crc32Val = crcTab ^ (crc32Val >> 8);
		jx = SCI0SR1;       // clears TDRE bit
		SCI0DRL = uctmp;
		SCI0CR2 |= 0x80;    // xmit interrupt enable
		} // end if (kFlag...
#else
	// do something pointless
    outpc.PC_duty = (unsigned char)(100-(((int) PC_DUTY*100 + (PC_PER/2))/(int) PC_PER)); // set datalog value (100-value to have datalog value same as PC table)
#endif      

}	 // end void checkCANbus (void)

// timer functions, returns msec at last call
unsigned int Timer0001(void) // milliseconds
	{
	// make sure we haven't called routine in last msec
	if (last0001 == millisec) return last0001; else last0001 = millisec;
	
	//////////////////////////// 0.001 'millisecond' section ///////////////////////

	// Pressure Control Output Dither Pulse
	dither_counter++;          // increment the counter

	burn_msec_flag = 0;        // reset burn timer to allow next step of flash burn

	// slow down the flash burns to every ~6 milliseconds
	if (!(millisec%6))
	  {   
	  // burn flash 512 byte(256 word) sector(s)
	  if(burn_flag >= 1)  
	    {
	    burn_msec_flag = 1;
	    fburner(tableWordFlash(burn_idx, 0), tableWordRam(burn_idx, 0), tableWords(burn_idx));
	    if(burn_flag >= tableWords(burn_idx))  
	      {
	      burn_flag = 0;
	      flocker = 0;
	      }
	      else
	      burn_flag++;
	      } 
	    }
    
	if (burn_flag) outpc.miscStatus |= 0x02; else outpc.miscStatus &= ~0x02; // set burn flag for datalog
	
	// Dither and PWM Refresh control
	// ------------------------------  
	if (inpram.dither_durPC) // only if using dither
		{
		if (!(shifting_flag) && (dither_counter > inpram.dither_intPC))
			{
			if (outpc.auto_mode && (!shifting_flag) && (outpc.LOAD < 750))
				{
				dither_flag |= 0x01;  // set the dither_flag first bit to one
				PC_DUTY = PC_OFF;    // set duty cycle to 0% == 100% pressure
				}  // end if (outpc.auto_mode...
			}  // end if dither_counter ...
		if (dither_counter > (inpram.dither_intPC + inpram.dither_durPC - 1))
			{
			dither_counter = 0;     // reset the dither counter to zero
			dither_flag &= ~0x01;   // reset the dither_flag first bit to zero
			}  // end if dither_counter ...
		} // end if (dither_durPC) ...
		
	if (inpram.refresh_dur)	   // only if using refresh
		{
		// Output1/2/4/5/6/7 Refresh Pulse
		refresh_counter++;         // increment the counter
		if (refresh_counter > inpram.refresh_int)
			{ // we handle the outputs separately, since they won't both necessarily be on at the same time
			if (Output1state)              // if output1 (solA) is active
			  {
		  	  dither_flag |= 0x02;  // set the dither_flag second bit to one
		      OUTPUT1port     |= OUTPUT1pin;  // set PE4 high
			  }   // end if (Output1state)...
			if (Output2state)              // if output2 (solB) is active
			  {
			  dither_flag |= 0x04;  // set the dither_flag third bit to one
			  OUTPUT2port     |= OUTPUT2pin;  // set PM2 high
			  }  // end if (Output1state)...
			if ((Output4state) && (inpram.sp_mode[0] == 3)) // if output4 (solD) is active
			  {
			  dither_flag |= 0x08;  // set the dither_flag 4th bit to one
			  spr1(1);   // set solD high
			  }  // end if (Output1state)...	
			}    // end if refresh_counter ...
		if (refresh_counter > (inpram.refresh_int + inpram.refresh_dur - 1))
			{
			refresh_counter = 0;    // reset the refresh counter to zero
			dither_flag &= ~0xE;    // reset the dither_flag second, third and 4th bits to zero
			}  // end if refresh_counter ...
		} // end if (refresh_dur) ...

	if ((!inpram.CAN_enabled) && ((millisec+31) % 42 == 0)) // if no CAN connection
	  	{
  		// Calculate load from auxiliary channel voltage (external MAP, TPS, MAF, etc.)   
  		// LOAD = (LOADmult * aux_volts) + LOADzero  
  		// LOADmult is  kpa/volts x100 and LOADzero is kpa x100, get_adc(5) is adc count, LOAD is x10 
  		outpc.LOAD = (uint) ((((long) inpram.LOADmult * (long) outpc.aux_volts)/2047) + (inpram.LOADzero)/10); // non-CAN MAP (volts to kPa)
  		}
  
  
	//////////////////////////// END of 0.001 second section ///////////////////////
	return millisec;
	}  // end Timer0001()
	
unsigned int Timer0010 (void) // hundredths
	{
	unsigned short ix,icx;
	unsigned long tmp1_U32;
	// call with last0010 = Timer0010();
	 
	// make sure we haven't called routine in last 10 msec
    if (last0010) return last0010;
    
    // Deal with TCC follow mode spare ports
    if (inpram.sp_mode[0] == 4) 
    	{
    	if ((TCC_follow_count1 > 0) && (TCC_follow_flag1 == 1)) 
   			{
   			TCC_follow_count1--; 	// decrement the timer count down to 1
   									// only reset to zero in tcc_unlock()
   									// so we know if we have already applied it when TCC is on
   									// and is re-applied
   			}
   			else
   			{
   			if (!shifting_flag) 
   				{
   				TCC_follow_count1 = 0;
   				spr1(0); // turn off spare port
   				sp[0] = 0;
   				}
   			}
   		}
 
    if (inpram.sp_mode[1] == 4) 
    	{
    	if ((TCC_follow_count2 > 0) && (TCC_follow_flag2 == 1))
   			{
   			TCC_follow_count2--; 	// decrement the timer count down to 1
   									// only reset to zero in tcc_unlock()
   									// so we know if we have already applied it when TCC is on
   									// and is re-applied
   			}
   			else // count == 0 OR flag == 0
   			{
			if (!shifting_flag)	  // don't interfere with clutch functions
				{
				TCC_follow_count2 = 0;
				spr2(0); // turn off spare port
				sp[1] = 0;
				}
   			}
   		}
    	
    // Get current instant gear ratio and estimate gear
    outpc.instant_ratio = getRatio();
    outpc.estimated_gear = getCurrentGear(outpc.instant_ratio);

	// Get Table switching pin (PE1/PE0) status
	if (!Inpt4_active) // Use PE1 if not using PE1 for 4th input
		{
		if (!(in2ram.test_mode & 0x01) || (!(in2ram.test_mode & 0x800))) // not testing 4WD/table switching
				{
				tmp_U08 = PORTE;
				}
				else   // testing 4WD/table switching
				{
				if (!(in2ram.TSTdbug & 0x02)) tmp_U08 = 0; else tmp_U08 = 0x02;
				}
					
		if (tmp_U08 & 0x02) // PE1 High - Use Table 1
			{
			if (!(outpc.miscStatus & 0x04))  // if have changed 
				{
				// clear all flags
				outpc.hyst_timer = 0;
				shiftSpeedHyst_flag = 0;
				outpc.decel_flag = 0;
				outpc.cruise_flag = 0;
				outpc.shift_action &= ~0x0C; // clear both up and down shift requests (0x0C = 1100)
				}
			outpc.miscStatus |= 0x04; // Set bit 4 to 1  
			}
			else  // PE1 low - Use Table 2
			{
			if (outpc.miscStatus & 0x04)  // if have changed 
				{
				// clear all flags
				outpc.hyst_timer = 0;
				shiftSpeedHyst_flag = 0;
				outpc.decel_flag = 0;
				outpc.cruise_flag = 0;
				outpc.shift_action &= ~0x0C; // clear both up and down shift requests (0x0C = 1100)
				}
			outpc.miscStatus &= ~0x04; // Set bit 4 to 0
			}
		}
		else // Use PE0 if using PE1 for 4th input
		{
		if (tmp_U08 & 0x01) 
			{
			if (!(outpc.miscStatus & 0x04))  // if have changed 
				{
				// clear all flags
				outpc.hyst_timer = 0;
				shiftSpeedHyst_flag = 0;
				outpc.decel_flag = 0;
				outpc.cruise_flag = 0;
				outpc.shift_action &= ~0x0C; // clear both up and down shift requests (0x0C = 1100)
				}
			outpc.miscStatus |= 0x04;
			}
			else 
			{
			if (outpc.miscStatus & 0x04)  // if have changed 
				{
				// clear all flags
				outpc.hyst_timer = 0;
				shiftSpeedHyst_flag = 0;
				outpc.decel_flag = 0;
				outpc.cruise_flag = 0;
				outpc.shift_action &= ~0x0C; // clear both up and down shift requests (0x0C = 1100)
				}
			outpc.miscStatus &= ~0x04;
			}
		}
	
	// Set up CAN transmission
	if(inpram.activate_xrate > 0)  
		{
		// send various bytes in the outpc structure to destvarblk.
		// The bytes are as configured in outmsg[msg_no] where 
		// msg_no = CAN_RB_DSR0
		uctmp = 1; 
		for (mx = 0; mx < MAX_OUTMSGS; mx++)  
		{   // mx = msg_no
		if(outmsg[mx].xrate && (inpram.activate_xrate & uctmp))  
			{
			// this message activated - check if time to send
			if(xrate_clks[mx] > outmsg[mx].xrate)  
			{     // time to send
			xrate_clks[mx] = 0;           // zero clk
			// Update related parameters for xmt ring buffer
			ix = can[1].cxno_in;
			can[1].cx_msg_type[ix] = OUTMSG_RSP;
			can[1].cx_outmsg_no[ix] = mx;
			n_outbytes[mx] = outmsg[mx].no_bytes;
			// destination var blk - for this msg the varblk is really the gpio
			// varblk to return the data to
			can[1].cx_destvarblk[ix] = outmsg[mx].dest_varblk & 0x0F;   
																// gpio var_blk
			can[1].cx_destvaroff[ix] = 0;   // store data starting at gpio varblk;
			// if > 8 bytes, will have subsequent messages, with varoff provided 
			// by ECU
			can[1].cx_dest[ix] = outmsg[mx].dest_varblk >> 4;    // gpio id
			if(n_outbytes[mx] >= 8)  
				{
				kx = 8;
				}
				else  
				{
				kx = n_outbytes[mx];
				}  // end if (n_outbytes...
			
			can[1].cx_varbyt[ix] = kx;
			// put variable value(s) in xmit ring buffer as follows:
			for(jx = 0;jx < kx;jx++)  
				{
				can[1].cx_datbuf[ix][jx] = 
					*((char *)&outpc + outmsg[mx].offset[jx]);
				}  // end for(jx = 0 ...
			send_can(1);
			}   // end if(xrate_clks[mx]...
		} //  end if(outmsg[mx].xrate &&...
		uctmp = uctmp << 1;    // = 2 ^^ mx
	   }      // end for (mx = 0;...
	  } // end if(inpram.activate_xrate)

	//////////////////////////////////////////////////////////////////////
	// Send CAN to get outpc variables from MS-II
	
	grabMS2vars();
	
	//////////////////////////////////////////////////////////////////////
	
	//---------------- Slow Speedometer Decay ------------------------------------------
	// If the VSS rate drops to zero very quickly, the speedo will be frozen at a finite 
	// value (large period = small frequency) that is greater than zero, since the 
	// VSS_timer only works if the interrupt is called, so zeros are never filled 
	// into the ic_period[] array.
	// So instead, we force this array to zero gradually - not fast enough to 
	// affect the speedo at any reasonable input rate but enough that it eventually 
	// goes to zero if the interrupt is not called.
	// Estimate ic1_per_avg if using VSS from another controller
	if ((inpram.CAN_enabled) && (in2ram.canVSSoffset != 0))
		{
		if (outpc.speedo == 0) 
			{
			ic1_per_avg = 0xFFFFF;
			}
			else 	
			{
			// ic1_per_avg is time for 20 teeth
			ic1_per_avg = ((71 * (long) vss_divide)/(long) outpc.speedo)*20;
			}
		} // end if ((inpram.CAN_enabled) && (in2ram.canVSSoffset))  
		else  // not CAN enabled OR not VSS_offset
		{ 
		// average last 20 periods for both VSS and ISS/tach for filtering
		ic1_per_acc = 0;  // initialize accumulator variables
	
		for (icx=1;icx < 21;icx++) // runs 1 to 20
			{
			ic1_per_acc = ic1_per_acc + ic1_period[icx];
			}
		 
		ic1_per_avg = (ic1_per_acc*16)/3; // the time for 20 VSS teeth (microseconds)
		
		outpc.VSS_per = ic1_per_avg/20;   // Save average period value for datalog
			
		// Calculate periods for VSS/ISS input masking:
		//
    	// short_period_mask = average*(100 - mask%/2)/100
    	//                   = (icx_per_acc/20) * (100 - (xss_mask/2))/100
		//                   = icx_per_acc * (100 - (xss_mask/2))/2000
		//
		// long_period_mask  = average*(100 + mask%)/100
		//                   = (icx_per_acc/20) * (100 + xss_mask)/100
		//                   = icx_per_acc * (100 + xss_mask)/2000
		// 
		// For example, if you enter 33%, and the current count for 
		// the 20 periods is 1440 (=1440/20 = 72 tics/tooth), then the 
		// mask is set to accept periods from 
		// 72-(33%/2) = 72-(0.165*72) = 60 tics to 
		// 72+33% = 96 tics. 
		// So the range is not based on 100% being no filtering. 
		// Instead, low numbers mean more filtering, high numbers mean less 
		// filtering (the INI limits the range to plausible values, you can 
		// edit the INI to expand the range, of course).
		//
		// Dividing the short period mask by two makes a bigger difference as the 
		// mask increases. For example, if you put in a mask of 100%, it allows 
		// periods from 36 to 144 in the above example, instead of from 
		// 0 to 144 if the 'divide by 2' wasn't used. 0 is any input frequency, 
		// no matter how fast! Instead, the range is half as fast to twice as fast.
		// The permissible mask% is also increased from 100% to 200%. This change
		// should help to more effectively filter the very short 'noise' pulses 
		// at low speeds.		

		tmp1_U32 = inpram.vss_mask;

		if (inpram.vss_mask > 0) 
			{
			if (inpram.vss_mask > 200) tmp1_U32 = 200; 		// limit vss masking range
			else if (inpram.vss_mask < 55) tmp1_U32 = 55; 	// limit vss masking range		
			}

		if ((inpram.vss_mask) && (ic1_per_avg < 0x1FFF00))  // don't use masking at very slow speeds
			{  // use masking
			short_period_vss = ((ic1_per_acc*(100-(tmp1_U32/2)))/2000);
			long_period_vss =  ((ic1_per_acc*(100+tmp1_U32))/2000);
			}
			else
			{  // no masking
			short_period_vss = 0;
			long_period_vss = 0xFFFFF;	
			}

		} // end else if ((inpram.CAN_enabled) && (in2ram.canVSSoffset))

	ic2_per_acc =  0;
	
	for (icx=1;icx < 21;icx++) // runs 1 to 20
		{
		ic2_per_acc = ic2_per_acc + ic2_period[icx];
		}

	ic2_per_avg = (ic2_per_acc/3)*16; // the time for 20 ISS teeth (microseconds)
		
	tmp1_U32 = inpram.iss_mask;

	if (inpram.iss_mask > 0) 
		{
		if (inpram.iss_mask > 200) tmp1_U32 = 200; 		// limit ISS masking range
		else if (inpram.iss_mask < 55) tmp1_U32 = 55; 	// limit ISS masking range		
		}
					
	if ((inpram.iss_mask)  && (ic2_per_avg < 0x8FFFFFFF))
		{  // use masking
		short_period_iss = ((ic2_per_acc*(100-(tmp1_U32/2)))/2000);
		long_period_iss =  ((ic2_per_acc*(100+tmp1_U32))/2000);
		}
		else
		{ // no masking	
		short_period_iss = 0;
		long_period_iss = 0xFFFFF;
		}

	// Check brakes, unlock TCC if brakes on 
	// Check if braking 100 times per second.
	// Brake input is on GPI4 = AD7.
	// if not braking, update LOAD short term average. 
	// Jake Brake (diesel engine compression brake)
	if ((jake_brake && (inpram.line_units == 2)) && (outpc.current_gear >= inpram.minTCC_gear) && (outpc.speedo > inpram.minTCCspeed) && !(shifting_flag))
		{
		// 100% line pressure
		PC_DUTY = PC_OFF;
		// lock TCC
		outpc.lock_TCC = tcc_lock();
		LUF_counter = inpram.LUF_taper; // prevent PWM taper
		}
	
	// Update the LUF duty cycle if TCC is on and LUF enabled
	if  (outpc.TCC_count == 0)
		{
		if (inpram.TCC_PWM)
			{
			if (outpc.lock_TCC >= 1) // TCC locked
				{
				if (LUF_counter < inpram.LUF_taper)  // in taper period
					{
					TCC_OUT = ((unsigned char) ((Calc_LUF_PWM() * TCC_ON)/100));  // Set LUF DC - we will update this every 0.010 seconds
					outpc.lock_TCC = (unsigned char) ((TCC_OUT*100)/TCC_ON); // set value for datalog
					}
					else // taper period over
					{
					TCC_OUT = (unsigned char) ((inpram.LUF_on * TCC_ON)/100);
					outpc.lock_TCC = (unsigned char) ((TCC_OUT*100)/TCC_ON); // set value for datalog
					}
				} 
				else  // TCC not locked
				{
				TCC_OUT = (unsigned char) ((inpram.LUF_off * TCC_ON)/100);  // Set LUF DC - we will update this every 0.010 seconds
				outpc.lock_TCC = (unsigned char) ((TCC_OUT*100)/TCC_ON); // set value for datalog
				LUF_counter = 0;  
				}  // end if (outpc.lock_TCC)
			} // end if ((inpram.TCC_PWM ...
		}
	
	// Check for messed up target_gear inherited from multiple gear errors
	if ((outpc.target_gear <= inpram.num_gears) && (outpc.target_gear > 0)) 
		{
		last_target_gear = outpc.target_gear;  // save the gear if it is valid
		}
		else
		{
		if ((last_target_gear > 1) && (last_target_gear <= inpram.num_gears)) 
			{
			outpc.target_gear = last_target_gear;  // revert to last good gear if invalid
			}
			else
			{
			outpc.target_gear = inpram.num_gears;  // Drive
			}
		}
	
	
	if ((!inpram.CAN_enabled) || (!in2ram.canVSSoffset))
		{	
		// COMPUTE VEHICLE SPEED
		// ---------------------
		// We do this here so that the speedo updates regardless of any delays in the code
		// (like the shift and pressure delays).
	
		// For imperial units: 
	
		// Vehicle speed = VSS_rpm * tire_diam/(5280*12) * 60 minutes/hour * PI/axle_ratio  (miles/hour)
	
		// vehicle speed is delta X / delta T
		// delta T is the time for icx == 20 teeth = ic_per_acc, in 5.333 sec tics
	
		// delta X is the distance travelled by 20 teeth (in inches x100)
		// = tire_distance/rev * 1/axle_ratio * 20/num_teeth
		// = PI*tire_diam * 1/axle_ratio * 20/num_teeth == vss_divide
		//
		// so speedo = (pi*tire_diam * 1/axle_ratio * 20/num_teeth)/(ic_per_acc * constant)
		//           = constant * vss_divide/ic_per_acc
	
		// use averaging above at 1.5x minSpeed
		// done in main loop (doesn't hurt if averaging delayed by low speed shift)
	    
		if (outpc.speedo < ((inpram.min_speed * 3)/2)) tmpVSF = 1; else tmpVSF = inpram.VSSsmoothFactor;
	
		// use unadjusted speedo if 2WD (i.e. if FWD_LogOnly and not Inpt4_active and outpc.miscStatus & 0x04)
	
		// rawSpeedo = 4*(((outpc.speedo/4 * (tmpVSF-1)) + (unsigned int) ((1420*vss_divide/inpram.num_teeth)/(ic1_per_avg)))/tmpVSF);
		// Put averaging AFTER speedoCorr adjustment, otherwise value bounces 
		outpc.rawSpeedo = (unsigned int) ((5680*vss_divide/inpram.num_teeth)/(ic1_per_avg));
	
	
		if  (!FWD_LogOnly)       // if 4WD and speedo switching allowed, divide by speedo factor (x1000)
			{
			if (outpc.miscStatus & 0x04)
				{
				outpc.rawSpeedo = (unsigned int) ((((5680*vss_divide/inpram.num_teeth)/(ic1_per_avg))*1000)/inpram.FWD_factor);	
				}
				else
				{
				outpc.rawSpeedo = (unsigned int) ((5680*vss_divide/inpram.num_teeth)/(ic1_per_avg));
				} 
			}
	
		// Convert raw speedo to corrected speedo
		tmp1_U32 = (1000 + intrp_speedo_table((outpc.rawSpeedo/10), in2ram.speedoCorrBin, in2ram.speedoCorr)); // get adjustment factor
	
		if (((msecVSS > 350) && (outpc.speedo < 80)) || (msecVSS > 1500))
			{
			outpc.speedo = 0;
			}
			else
			{
			tmp1_U32 = (unsigned int) (((long) tmp1_U32 * (long) outpc.rawSpeedo)/1000);
			outpc.speedo = (unsigned int) (((outpc.speedo * (tmpVSF-1)) + (tmp1_U32))/tmpVSF);
			}
		}	 // end if (!in2ram.canVSSoffset) 
		
	if (outpc.speedo > inpram.maxSpeed) outpc.speedo = inpram.maxSpeed; // rail the speedometer
	if ((outpc.speedo < inpram.min_speed) && (outpc.current_gear == 1)) outpc.speedo = 0; // set the speedo to zero
	
	if (inpram.ic2_usage & 0x01) // if IC2 used for input shaft sensor
		{   // Calculate input shaft RPM from last 20 teeth
		if ((outpc.current_gear > 0) && (outpc.current_gear <= inpram.num_gears))
			{
			outpc.is_rpm = (unsigned int) ((in2ram.ISSratio[outpc.current_gear] * ((long) 1200/(long) inpram.iss_divider) * (1000000/(long) ic2_per_avg))/1000);
			}
			else if (outpc.current_gear == -1)
			{
			outpc.is_rpm = (unsigned int) ((in2ram.ISSratio[0] * ((long) 1200/(long) inpram.iss_divider) * (1000000/(long) ic2_per_avg))/1000);		
			}
		}
		else if (inpram.ic2_usage & 0x02) // tach
		{
		outpc.engine_rpm = (unsigned int) (((long) 1200/(long) inpram.iss_divider) * (1000000/(long) ic2_per_avg));				
		outpc.is_rpm  = (unsigned int) ((((long)os_rpm * (long)inpram.gear_table[outpc.current_gear]) + 500)/1000);
		} 
		else   // ic2_usage == 0; calculate input shaft speed from output shaft speed and gear ratios 
		{
		outpc.is_rpm  = (unsigned int) ((((long)os_rpm * (long)inpram.gear_table[outpc.current_gear]) + 500)/1000);
		} // end if ic2_usage
    	
	// Calculate Slip
	if ((in2ram.SlipOptn & 0x01) && ((inpram.ic2_usage & 0x01) == 1)) // if option to report internal slip chosen and ISS is active (not estimated)
		{  // calculate and report internal slip
		if ((!shifting_flag) && (os_rpm > 0)) outpc.slip = (uchar)((outpc.is_rpm*100)/(os_rpm)); else outpc.slip = 0;
		}
		else
		{  // calculate and report TCC slip
		if (outpc.is_rpm) outpc.slip = (uchar)((long)outpc.engine_rpm*100/(long)outpc.is_rpm); else outpc.slip = 0; 	
		}	// end else if ((in2ram.SlipOptn & 0x01) && ...
	
	/*  -----  Update RPM, MAP & Battery Voltage with CAN ----------------------------------- */
	// Update outpc. rpm, battery voltage, MAP, etc. for calculations using CAN data
	// received from MS-II
	// (that is, all the info is already there in the msvar array, the CAN data is 
	// grabbed in sync with the timer updates in Timer_Clock_ISR(), but not all of 
	// info that is assigned to specific variables).
	if (inpram.CAN_enabled) 
		{
		if (!inpram.can_testout)
			{    
			// outpc.seconds        = *(int *)(msvar + 0);   // 2 bytes (int) at offset 0 in msvar[]
			pulsewidth1      = *(unsigned int *)(msvar + in2ram.canPW1outpc);   // 2 bytes (int) at offset 2 in msvar[]
			// outpc.pw2            = *(int *)(msvar + 4);   // 2 bytes (int) at offset 4 in msvar[]
			if (!(inpram.ic2_usage & 0x02)) // not using ISS for tach 
				{
				if (!(in2ram.test_mode & 0x01) || !(in2ram.test_mode & 0x4000)) // not testing CANbus RPM
					{
					outpc.engine_rpm = *(unsigned int *)(msvar + in2ram.canRPMoutpc);   // rpm is 2 bytes (int) at offset 6 in msvar[]
					}
					else    // testing CANbus RPM
					{
					outpc.engine_rpm = in2ram.CANrpmdbug;
					}
				}
			// adv_deg              = *(int *)(msvar + 8);   // 2 bytes (int) at offset 8 in msvar[]
			// outpc.squirt         = *(char *)(msvar + 10); // 1 byte (char) at offset 10 in msvar[]
			// outpc.engine         = *(char *)(msvar + 11); // 1 byte (char) at offset 11 in msvar[]
			// outpc.afrtgt1        = *(char *)(msvar + 12); // 1 byte (char) at offset 12 in msvar[]
			// outpc.afrtgt1        = *(char *)(msvar + 13); // 1 byte (char) at offset 13 in msvar[]
			// outpc.wbo2_en1       = *(char *)(msvar + 14); // 1 byte (char) at offset 14 in msvar[]
			// outpc.wbo2_en2       = *(char *)(msvar + 15); // 1 byte (char) at offset 15 in msvar[]
			// outpc.barometer      = *(int *)(msvar + 16);  // 2 bytes (int) at offset 16 in msvar[]
			// Ensure that the TPS is in the 0.0 to 100.0% range    
			// TPS first
			tmpLOAD = (unsigned int)(((long)((*(signed int *)(msvar + in2ram.canTPSoutpc))/10) * (long)(100-in2ram.TPSmap))/10);  // use CAN TPS value  
			if (tmpLOAD > 1000)
				{
				tmpLOAD = 1000;  // rail at 100.0%
				}
				else if (tmpLOAD < 0)
				{
				tmpLOAD = 0;     // rail at 0.0  
				}
			// int intrp_2dctable(
			//		unsigned int x, 
			//		int y, 
			//		unsigned char nx, 
			//		unsigned char ny,
			//		unsigned int * x_table, 
			//		int * y_table, 
			//		unsigned char * z_table) 
			tmpLOAD = intrp_2dctable(	(uint) (outpc.engine_rpm), 
								(int) (tmpLOAD),
								(uchar) 6, 
								(uchar) 6, 
								(uint *)&in2ram.TPSrpm[0], 
								(int *)&in2ram.TPStps[0], 
								(uchar *)&in2ram.TPStable[0][0]) * 10;
			tmpLOAD = (unsigned int)(((long)((tmpLOAD)/10) * (long)(100-in2ram.TPSmap))/10);  // use CAN TPS value  					
			// MAP next	
			outpc.rawMAP = (*(signed int *)(msvar + in2ram.canMAPoutpc));	
			tmpLOAD = tmpLOAD + (unsigned int)(((long)((outpc.rawMAP)/10) * (long)(in2ram.TPSmap))/10);  // MAP kPa is 2 bytes (int) at offset 18 in msvar[]
			
			// set load value for decel
			if (inpram.load_adj & 0x02)
				{
				// if using decel load limiting
				if ((outpc.auto_mode != 3) && ((outpc.decel_flag) && (tmpLOAD > in2ram.decel_Load) && (in2ram.decel_Load))) tmpLOAD = in2ram.decel_Load;
				}
			// assign the adjusted load value
			if (!(in2ram.test_mode & 0x01) || !(in2ram.test_mode & 0x2000))	 // not testing load
				{
				outpc.LOAD = tmpLOAD;
				}
				else // testing load
				{
				outpc.LOAD = in2ram.CANloaddbug;
				}
					
			// outpc.mat            = *(int *)(msvar + 20);  // 2 bytes (int) at offset 20 in msvar[]
			// outpc.coolant        = *(int *)(msvar + 22);  // 2 bytes (int) at offset 22 in msvar[]
	
			// grab TPS for throttle factors
			tmpLOAD          = *(int *)(msvar + in2ram.canTPSoutpc);  // TPS - 2 bytes (int) at offset 24 in msvar[]
			if (tmpLOAD < 0)
				{
				outpc.rawTPS = 0;
				}
				else if (tmpLOAD > 1000)
				{
				outpc.rawTPS = 1000;
				}
				else
				{
				outpc.rawTPS = tmpLOAD;  	    		
				}

			if (!(in2ram.test_mode & 0x01) || !(in2ram.test_mode & 0x8000))	 // not testing battery voltage
				{		
				outpc.vBatt = *(int *)(msvar + in2ram.canVBATToutpc);  // 2 bytes (int) at offset 26 in msvar[]
				}
				else
				{
				outpc.vBatt = in2ram.CANbattdbug;
				}
			// outpc.afr1           = *(int *)(msvar + 28);  // 2 bytes (int) at offset 28 in msvar[]
			// outpc.afr2           = *(int *)(msvar + 30);  // 2 bytes (int) at offset 30 in msvar[]
			// outpc.knock          = *(int *)(msvar + 32);  // 2 bytes (int) at offset 32 in msvar[]
			// outpc.egoCorrection1 = *(int *)(msvar + 34);  // 2 bytes (int) at offset 34 in msvar[]
			// outpc.egoCorrection1 = *(int *)(msvar + 36);  // 2 bytes (int) at offset 36 in msvar[]
			// outpc.airCorrection  = *(int *)(msvar + 38);  // 2 bytes (int) at offset 38 in msvar[]
			// outpc.warmupEnrich   = *(int *)(msvar + 40);  // 2 bytes (int) at offset 40 in msvar[]
			// outpc.accelEnrich    = *(int *)(msvar + 42);  // 2 bytes (int) at offset 42 in msvar[]
			// outpc.tpsfuelcut     = *(int *)(msvar + 44);  // 2 bytes (int) at offset 44 in msvar[]
			// outpc.baroCorrection = *(int *)(msvar + 46);  // 2 bytes (int) at offset 46 in msvar[]
			// outpc.gammaEnrich    = *(int *)(msvar + 48);  // 2 bytes (int) at offset 48 in msvar[]
			// outpc.veCurr1        = *(int *)(msvar + 50);  // 2 bytes (int) at offset 50 in msvar[]
			// outpc.veCurr2        = *(int *)(msvar + 52);  // 2 bytes (int) at offset 52 in msvar[]
			// outpc.iacstep        = *(int *)(msvar + 54);  // 2 bytes (int) at offset 54 in msvar[]
			// outpc.coldAdvDeg     = *(int *)(msvar + 56);  // 2 bytes (int) at offset 56 in msvar[]
			// outpc.tpsDOT         = *(int *)(msvar + 58);  // 2 bytes (int) at offset 58 in msvar[]
			// outpc.mapDOT         = *(int *)(msvar + 60);  // 2 bytes (int) at offset 60 in msvar[]
			// outpc.dwell          = *(int *)(msvar + 62);  // 2 bytes (int) at offset 62 in msvar[]
			// outpc.maf            = *(int *)(msvar + 64);  // 2 bytes (int) at offset 64 in msvar[]
			// outpc.calcMAP        = *(int *)(msvar + 66);  // 2 bytes (int) at offset 66 in msvar[]
			// outpc.fuelCorrection = *(int *)(msvar + 68);  // 2 bytes (int) at offset 68 in msvar[]
			// outpc.portStatus     = *(char *)(msvar + 70); // 1 byte (char) at offset 70 in msvar[]
			// outpc.knockRetard    = *(char *)(msvar + 71); // 1 byte (char) at offset 71 in msvar[]
			// outpc.xTauFuelCorr1  = *(int *)(msvar + 72);  // 2 bytes (int) at offset 72 in msvar[]
			// outpc.egoV1          = *(int *)(msvar + 74);  // 2 bytes (int) at offset 74 in msvar[]
			// outpc.egoV2          = *(int *)(msvar + 76);  // 2 bytes (int) at offset 76 in msvar[]
			// outpc.amcUpdates     = *(int *)(msvar + 78);  // 2 bytes (int) at offset 78 in msvar[]
			// outpc.aux_voltsix    = *(int *)(msvar + 80);  // 2 bytes (int) at offset 80 in msvar[]
			// outpc.xTauFuelCorr2  = *(int *)(msvar + 82);  // 2 bytes (int) at offset 82 in msvar[]
			// outpc.spare1         = *(int *)(msvar + 84);  // 2 bytes (int) at offset 84 in msvar[]
			// outpc.spare2         = *(int *)(msvar + 86);  // 2 bytes (int) at offset 86 in msvar[]
			// outpc.spare3         = *(int *)(msvar + 88);  // 2 bytes (int) at offset 88 in msvar[]
			// outpc.spare4         = *(int *)(msvar + 90);  // 2 bytes (int) at offset 90 in msvar[]
			// outpc.spare5         = *(int *)(msvar + 92);  // 2 bytes (int) at offset 92 in msvar[]
			// outpc.spare6         = *(int *)(msvar + 94);  // 2 bytes (int) at offset 94 in msvar[]
			// outpc.spare7         = *(int *)(msvar + 96);  // 2 bytes (int) at offset 96 in msvar[]
			// outpc.spare8         = *(int *)(msvar + 98);  // 2 bytes (int) at offset 98 in msvar[]
			// outpc.spare9         = *(int *)(msvar + 100); // 2 bytes (int) at offset 100 in msvar[]
			// outpc.spare10        = *(int *)(msvar + 102); // 2 bytes (int) at offset 102 in msvar[]
			// outpc.spare11        = *(int *)(msvar + 104); // 2 bytes (int) at offset 104 in msvar[]
			// outpc.ospare         = *(char *)(msvar + 106);// 1 byte (char) at offset 106 in msvar[]
			// outpc.chksum         = *(char *)(msvar + 107);// 1 byte (char) at offset 107 in msvar[]
			// outpc.deltaT         = *(long *)(msvar + 108);// 4 bytes (long) at offset 108 in msvar[]
			if (in2ram.canVSSoffset) outpc.speedo = (unsigned int) (((long) in2ram.CANVSSadj * (long) (*(unsigned int *)(msvar + in2ram.canVSSoffset)))/1000);  // 2 bytes (int) at canVSSoffset in msvar[]
			}   // end if (!inpram.can_testout)
			else   // using outmsg
			{
			//  offset 0 is PW1
			//  offset 2 is rpm
			//  offset 4 is MAP (x10)
			//  offset 6 is tps (x10)
			//  offset 8 is vBatt (x10)

			pulsewidth1      = *(unsigned int *)(msvar + in2ram.canPW1outmsg);   // 2 bytes (int) at offset 0 in outmsg
			outpc.engine_rpm = *(unsigned int *)(msvar + in2ram.canRPMoutmsg);

			if (in2ram.canSPEEDOoutmsg) outpc.speedo = (unsigned int) (((long) in2ram.CANVSSadj * (long) (*(unsigned int *)(msvar + in2ram.canRPMoutmsg)))/1000);  // 2 bytes (int) at in2ram.canSPEEDOoutmsg in outmsg

			// TPS first
			tmpLOAD = (unsigned int)(((long)((*(signed int *)(msvar + in2ram.canTPSoutmsg))/10) * (long)(100-in2ram.TPSmap))/10);  // use CAN TPS value  
			if (tmpLOAD > 1000)
				{
				tmpLOAD = 1000;  // rail at 100.0%
				}
				else if (tmpLOAD < 0)
				{
				tmpLOAD = 0;     // rail at 0.0  
				}
			// int intrp_2dctable(
			//		unsigned int x, 
			//		int y, 
			//		unsigned char nx, 
			//		unsigned char ny,
			//		unsigned int * x_table, 
			//		int * y_table, 
			//		unsigned char * z_table) 
			tmpLOAD = intrp_2dctable(	(uint) (outpc.engine_rpm), 
								(int) (tmpLOAD),
								(uchar) 6, 
								(uchar) 6, 
								(uint *)&in2ram.TPSrpm[0], 
								(int *)&in2ram.TPStps[0], 
								(uchar *)&in2ram.TPStable[0][0]) * 10;
			tmpLOAD = (unsigned int)(((long)((tmpLOAD)/10) * (long)(100-in2ram.TPSmap))/10);  // use CAN TPS value  					
			// MAP next	
			tmpLOAD = tmpLOAD + (unsigned int)(((long)((*(signed int *)(msvar + in2ram.canMAPoutmsg))/10) * (long)(in2ram.TPSmap))/10);  // MAP kPa is 2 bytes (int) at offset 18 in msvar[]

			// set load value for decel
			if (inpram.load_adj & 0x02)
				{
				// if using decel loaf limiting
				if ((outpc.auto_mode != 3) && ((outpc.decel_flag) && (tmpLOAD > in2ram.decel_Load) && (in2ram.decel_Load))) tmpLOAD = in2ram.decel_Load;
				}
			// assign the load value
			outpc.LOAD = tmpLOAD;

			// get TPS		
			tmpLOAD          = *(int *)(msvar + in2ram.canTPSoutmsg);  // TPS - 2 bytes (int) at offset 6 in msvar[]
			if (tmpLOAD <0)
				{
				outpc.rawTPS = 0;
				}
				else if (tmpLOAD > 1000)
				{
				outpc.rawTPS = 1000;
				}
				else
				{
				outpc.rawTPS = tmpLOAD;  	    		
				}  
		
			// get battery voltage	
			outpc.vBatt = *(int *)(msvar + in2ram.canVBATToutmsg);  
			} // end else (!inpram.can_testout)
		}
		else // end if (inpram.CAN_enabled) 
		{
		// not CAN enabled
		if (inpram.ic2_usage & 0x02) // if IC2 usage for tach signal
			{ // calculate rpm from IC2 period average
			outpc.engine_rpm = (unsigned int) (((120000000/(long) ic2_per_avg) * (1000/(long) inpram.iss_divider))/100);	
			} 
			else // ic2 not used, or used for ISS
			{  // calculate rpm from input shaft speed (which may be measured or may be calc'd from output shaft speed)
			outpc.engine_rpm = ((3*outpc.engine_rpm) + outpc.is_rpm + 2)/4;   // set engine rpm to input shaft speed
			inpram.rpm_check = 0;  // don't do rpm checks  
			}  // end else if ((inpram.ic2_usage

		outpc.vBatt = 120;                 // set default voltage to 12.0V
		}  // end else (inpram.CAN_enabled) 

		// if TPS is high set short-term load to current load (to force immediate downshift, etc.)
		if (inpram.CAN_enabled && (outpc.rawTPS > inpram.WOT_tps_threshold) && (inpram.load_adj & 0x08)) 
			{
			if ((inpram.load_adj & 0x08) && (outpc.LOAD > 0)) outpc.LOAD_short = outpc.LOAD;     // reset load to current value (i.e., near maximum)
			WOT_flag = inpram.WOT_dur;   // set hysteresis flag
			if (outpc.hyst_timer) outpc.hyst_timer--;
			} // end if (CAN_enabled && (tps >
	
		// Capture initial speed at shift for hysteresis calcs
		// Set shiftSpeedHyst_flag so we only do this once per shift
		if ((shifting_flag > 0) && (shiftSpeedHyst_flag == 1)) 
			{
			// currently shifting but flag not yet set
			// last_mph = outpc.speedo;      // capture speed
			// last_kpa = outpc.LOAD_short;  // capture load
			// last_tps = outpc.rawTPS;
			} 
			else 
			{ // not shifting or flag set
			// shifting_flag == 0 OR shiftSpeedHyst_flag == 1
			// Once hysteresis is broken reset flag	
			if (!(outpc.hyst_timer)) shiftSpeedHyst_flag = 0;      // no longer shifting, reset flag but retain speeds
			}   // end if (shifting_flag && !shiftSpeedHyst_flag) 

		if (outpc.lever_count > 2) 
			{
			outpc.lever_count--; // decrement the lever timer count
			}
			else if (outpc.lever_count == 1) 
			{
			outpc.error |= 0x20; // set bit for datalog
			outpc.manual_gear = -99;
			}
			else if (outpc.lever_count == 0)
    		{
    		outpc.error &= ~0x20; // clear bit for datalog
    		}

		// if shifting and in auto mode, set flag 
		// if ((outpc.auto_mode > 0) && (shifting_flag > 0)) outpc.hyst_flag |= 0x01;
		
	/////////////////////////// END of 0.010 second section 3 ////////////////////////////

	return 10;
	}   // end Timer0010()
	

unsigned int TimerADC(void) // hundredths
	{
	// Called every 40 milliseconds to deal with ADCs	
	adc_channel++;
	if (adc_channel > 7) adc_channel = 0;
	switch (adc_channel)
		{	
		case SWITCHA_ADC:	  // Input 1 (channel 0)
			outpc.Inadc[SWITCHA_ADC] = get_adc(SWITCHA_ADC);
			if (outpc.Inadc[SWITCHA_ADC] >=  inpram.digital_threshold)
				{	 
				outpc.switches |= 0x01;
				}
				else 
				{	
				outpc.switches &= ~0x01;
				}

			break;
			
		case SWITCHB_ADC:	  /// Input 2 (channel 1)
			outpc.Inadc[SWITCHB_ADC] = get_adc(SWITCHB_ADC);
			// if ((inpram.mlever_mode == 0) && (inpram.hardware != 0))
			if (inpram.mlever_mode == 0)
				{
				// if pin1 is high, else if it is low	
				// if (PORTAD0 & 0x02) outpc.Inadc[SWITCHB_ADC] = 1023; else outpc.Inadc[1] = 0;
				if (outpc.Inadc[SWITCHB_ADC] >=  inpram.digital_threshold)
					{
					outpc.switches |= 0x02;
					}
					else 
					{
					outpc.switches &= ~0x02;  
					}
				}  // end if ((inpram.mlever_mode == 1) ...
			break;
			
		case TEMPER_ADC:	  // Trans Temperature (channel 2)
			// Convert to temp or ADC in get_adc() function 
			// based on inpram.stdin_cfg
			outpc.clt = get_adc(TEMPER_ADC); // Set ADC
			break;
			
		case SWITCHC_ADC:	  // Input 3		(channel 3)
			outpc.Inadc[2] = get_adc(SWITCHC_ADC); // swC for manual shift lever position load
			if (outpc.Inadc[2] >=  inpram.digital_threshold)
				{	 
				outpc.switches |= 0x04;
				}
				else 
				{	
				outpc.switches &= ~0x04;
				}
			break;
			
		case PRESSURE_ADC:	  // Line Pressure	 (channel 4)
			outpc.line_pressure = get_adc(PRESSURE_ADC); // get line pressure
			break;
			
		case NC_LOAD_ADC:	  // Non-CANbus Load	  (channel 5)
			outpc.aux_volts     = get_adc(NC_LOAD_ADC);      // aux. data channel (0-5 volts), also used for non-CAN load
			break;  
			
		case VOLT_SHIFT_ADC:   
			outpc.butADC = get_adc(VOLT_SHIFT_ADC); // voltage based shifter on downshift paddle
			if (outpc.butADC >=  inpram.digital_threshold)			
  				{ // button high	
				if (inpram.shft_btn_polarity == 0)  // active low
					{
					outpc.shift_action &= ~0x01;    // clear bit 0
					}
					else		                    // active high
					{
					outpc.shift_action |=  0x01;    // set bit 0 
					}
				}
				else // button low								// pin grounded (low)
				{
				if (inpram.shft_btn_polarity == 0)  // active low
					{
					outpc.shift_action |=  0x01;    // set bit 0
					}
					else						    // active low
					{
					outpc.shift_action &= ~0x01;    // clear bit 0 
					}
				}   // end else // button low
			break;
				
		case BRAKE_ADC:  // Brake sense (channel 7)
			if (get_adc(BRAKE_ADC) >=  inpram.digital_threshold)   // signal high
				{
				if (inpram.brake_ON_polarity == 1) // 1 is active high
						{
						outpc.miscStatus |= 0x01;   // Set brake ON bit
						if (!jake_brake && !STCC_flag && !((inpram.TCC_highGear & 0x01) && (outpc.current_gear == inpram.num_gears))) outpc.lock_TCC = tcc_unlock();
						error_flags = error_flags & ~0x80; // clear brake flag
						}
						else    // signal low
						{
						outpc.miscStatus &= ~0x01;	// Set brake OFF bit
						}
				}
				else    // signal low
				{
				if (inpram.brake_ON_polarity == 1) // 1 is active high
						{
						outpc.miscStatus &= ~0x01;	 // Set brake OFF bit
						}
						else    // signal low
						{
						outpc.miscStatus |= 0x01;	 // Set brake ON bit
						if (!jake_brake && !STCC_flag) outpc.lock_TCC = tcc_unlock();
						error_flags = error_flags & ~0x80; // clear brake flag
						}
				}	
			break;
		default:
			break;
		} // end switch
	return 15;
	}
	
unsigned int Timer0100(void) // tenths
	{
	unsigned long tmp1_U32;
	unsigned long tmp2_U32;

	if (last0100) return last0100;

	// Decrement hyst_flag if it is active (> 0) 
	if (outpc.hyst_timer) 
		{
		if ((outpc.seconds > 4) && (outpc.speedo > inpram.min_speed)) outpc.hyst_timer--; else outpc.hyst_timer = 0;
		}
	
	// Decrement P to D counter if active	
//	if (P2Dtimer) P2Dtimer--;
		
	// Decrement the decel mode timer if it is active
	last_decel_mph = outpc.speedo; // check every 1/10th of a seconds for deceleration
	if (outpc.decel_flag) outpc.decel_flag--;
	
	// decrement counter on slip flag
	if (Sflag > 1) 
		{
		if ((inpram.SlipConditions < 5) 
			|| 	((inpram.SlipConditions == 5) && 
				(((PC_table_value - outpc.Slip_Adjustment) <= in2ram.Slip_limit[outpc.current_gear - 1])))) 
				{
				Sflag--;
				}
		} // end if (Sflag > 1)
	
	// Check Shift Hysteresis Conditions
	// ---------------------------------
	// last_mph/mph were set in shiftTo() when shifting_flag == 1 (shift in progress)
	// we don't look up another gear until at least one hysteresis condition is met
	
	if  ((outpc.hyst_timer) && (shifting_flag == 0))
		{
		if ((outpc.speedo < (last_mph - inpram.gear_hyst)) // if any hysteresis conditions met
		||  (outpc.speedo > (last_mph + inpram.gear_hyst))
		||  (outpc.speedo < inpram.hyst_enable_speed))    // always look up gear if mph is less than user setting      
			{
			outpc.hyst_timer = 0;	   // clear outpc.hyst_flag (was set at last shift)
			} // end if (outpc.speedo <= ...
		}  // end if  (hyst_timer > 0 && ...)
		
			
	// Calculate Fuel Efficiency
	// - Use current speed and averaged pulse width to calculate miles per gallon
	// - fuel density - gasoline = 6.073 lb/USgal = 7.290 lb/Imp.gal                                    
	if ((inpram.CAN_enabled) && (outpc.speedo > 0) && (outpc.engine_rpm > 300))  // if engine running and vehicle moving
		{
		tmp2_U32 = ((long) outpc.speedo * (long) inpram.fuel_density);
		tmp1_U32 = (120000 / (long) outpc.engine_rpm);
		outpc.mileage =  (short) (tmp2_U32/((long) inpram.inj_flow)*100*tmp1_U32/(long) inpram.nSquirts/((long) pulsewidth1-((long) inpram.open_time*10)));
		} 
		else // else mileage is zero
		{
		outpc.mileage = 0;
		}  // end if ((inpram.CAN_enabled) && ...
	
	// unlock the TCC if the speed has fallen 5 mph/kph below user set lock-up threshold
	if (outpc.speedo < (inpram.minTCCspeed - 50)) tcc_unlock();
	
	// Calculate converter slippage (100% is no slip)
	if (in2ram.canVSSoffset == 0)
		{
		if (inpram.VSS_location == 0)
			{
			// use original RWD calcs
			os_rpm = (unsigned int)((((1000000/ic1_per_avg) * (2000/inpram.num_teeth)) * 600)/1050);
			}
			else
			{
			// use new FWD calcs	
			os_rpm = (unsigned int)((((1000000/ic1_per_avg) * (2000/inpram.num_teeth)) * effective_axle_ratio)/1750);	
			}
		}
		else  // Using CANbus VSS
		{
		os_rpm = (unsigned int) ((((long) outpc.speedo * (long) inpram.axle_ratio)/((long) inpram.tire_diam))/3);
		}
		
	// decrement TCC timer if active
	if (outpc.TCC_count > 0) outpc.TCC_count--;
	
	// -------------- Auto Shift Mode -------------------------------------------------------
	// Get target gear from shift table
	
	// SET TARGET GEAR
	// and request shift as necessary 
	// from auto_gear_lookup(
	//      x = speed - int, will be divided by 10 in subroutine to match bins
	//      y = kpa - int, will be divided by 10 in subroutine to match bins
	//      nx = [NUM_MPH],
	//      ny = [NUM_LOADS],
	//      x_table = vss_table - unsigned int,
	//      y_table = LOAD_table - int,
	//      z_table) = auto_table - int.
	
	// set load value for decel
	if (inpram.load_adj & 0x02)
		{
		// if using decel loaf limiting
		if ((outpc.auto_mode != 3) && ((outpc.decel_flag) && (tmpLOAD > in2ram.decel_Load) && (in2ram.decel_Load))) outpc.LOAD_short = in2ram.decel_Load;
		}
	
	// Get target gear if not shifting
	if (!(shifting_flag) && !(rev_shift_flag) && (!(outpc.decel_flag) || !(in2ram.decel_Load)))   // if not waiting during a shift
		{
  		if ((outpc.miscStatus & 0x04) || (FWD_LogOnly))
			{   // 2WD
			// Only get new target gear if we aren't already dealing with a RPM shift
			if (!(outpc.shift_action & 0x04) && !(outpc.shift_action & 0x08)) 
				{
				if ((outpc.auto_mode == 1) || (outpc.auto_mode == 2)) 
					{
					outpc.target_gear = auto_gear_lookup(outpc.speedo,outpc.LOAD_short,NUM_MPH,NUM_LOADS,&inpram.vss_table1[0],&inpram.LOAD_table1[0],&inpram.auto_table1[0][0]);
					}
					else if (inpram.shift_mode == 3)	 // RPM-based shifting
					{
					outpc.auto_mode = 3;
					outpc.target_gear = auto_gear_lookupRPM();
					}
				}
			} // end if (outpc.miscStatus & 0x04)
			else
			{   // 4WD
			if (!(outpc.shift_action & 0x04) && !(outpc.shift_action & 0x08)) 
				{
				if ((outpc.auto_mode == 1) || (outpc.auto_mode == 2)) 
					{
					outpc.target_gear = auto_gear_lookup(outpc.speedo,outpc.LOAD_short,NUM_MPH,NUM_LOADS,&inpram.vss_table2[0],&inpram.LOAD_table2[0],&inpram.auto_table2[0][0]);			
					}
					else if (inpram.shift_mode == 3)	 // RPM-based shifting
					{
					outpc.auto_mode = 3;
					outpc.target_gear = auto_gear_lookupRPM();
					}
				}
				
			}  // end else (outpc.miscStatus & 0x04)
		} // end if (!(shifting_flag) &&
		
	// Idle Speed/Timing Advance Adjustment on MS-II if Stopped in Gear
	if (inpram.CAN_enabled)
		{  
		if ((outpc.speedo < 10) && (outpc.rawTPS < inpram.idle_tps_adc) && (outpc.current_gear==1))
			{
			if (!IdleFlag) 
				{
				outpc.IdleAdj = inpram.IACAdjngear;
				outpc.SpkAdj = inpram.AdvAdjngear;
				IdleFlag = 1;
				sendCANAdj();
				}  // end if (!IdleFlag)
			}
			else   // if ((outpc.speedo < 10)...
			{
			if (IdleFlag) 
				{
				outpc.IdleAdj = 0;
				outpc.SpkAdj = 0;
				IdleFlag = 0;
				sendCANAdj();
				}  // end if (IdleFlag)
			}  // end if ((outpc.speedo < 10) && ...
		}  // end if (inpram.CAN_enabled)
	
	// Set VSS interrupt received flag to false 10 times/sec (will be set to true in interrupt)
	VSS_rcvd =  0;            // set VSS interrupt flag to zero
		
	// VSS error - reset if get to 1000
	if (VSS_error >= inpram.vss_error_max) 
		{
		outpc.error |= 0x10; // set VSS tooth count reset error 
		VSS_reset();         // reset VSS
		}
		
	// Increment LUF counter if enabled
	if (outpc.lock_TCC >= 1)   // do not increment if TCC not ON
								// or if taper set to zero 
								// (keep PWM at initial value)
		{
		if(LUF_counter < inpram.LUF_taper) LUF_counter++;
		}
		else
		{
		LUF_counter = 0;
		}

  	
	// Check if have 'down shifted' manually using the gear lever
	if ((shifting_flag == 0) && (prev_gearlever > 0) && (outpc.manual_gear > 0))
	   { // if not shifting at the moment
		  if (prev_gearlever > outpc.manual_gear)  // have 'downshifted' manual lever
		  { // check if should down shift (in any mode)
			if (outpc.current_gear > outpc.manual_gear)
			{
	        outpc.shift_action &= ~0x08; // clear upshift request
	        outpc.shift_action |=  0x04; // set downshift request
			}   // end if (outpc.current_gear >
		  }
	   }
	
	if (inpram.CAN_enabled && outpc.SpkAdj && (outpc.LOAD < inpram.Tretard_load) && (shifting_flag == 0))
		{
		outpc.SpkAdj = 0;
		if (sentSpkAdj != 0) sendCANAdj();
		sentSpkAdj = 0;            
		} // end if (inpram.CAN_enabled && ...
		
	if (((outpc.seconds*10)+(millisec/100)) >= inpram.resetTime) restart = 0; // clear reboot flag at after user specified time
	
	// Set LEDs and verify LEDs for datalog
	setLEDs(outpc.current_gear);
	if (LED1_OUT & LED1pin)
		{
		outpc.LEDstatus |= 0x01; // set bit
		}
		else
		{	 
		outpc.LEDstatus &= ~0x01; // clear bit
		}
		
	if (LED2_OUT & LED2pin)
		{
		outpc.LEDstatus |= 0x02; // set bit
		}
		else
		{	 
		outpc.LEDstatus &= ~0x02; // clear bit
		}
		
	if (LED3_OUT & LED3pin)
		{
		outpc.LEDstatus |= 0x04; // set bit
		}
		else
		{	 
		outpc.LEDstatus &= ~0x04; // clear bit
		}
		
	if (LED4_OUT & LED4pin)
		{
		outpc.LEDstatus |= 0x08; // set bit
		}
		else
		{	 
		outpc.LEDstatus &= ~0x08; // clear bit
		}
	
	if (shifting_flag) outpc.cruise_flag = 0; // clear the cruise flag when shifting
	
	// update spare port PWM%
	if (sp0) spr0(1); else spr0(0);
	if (sp[0]) spr1(1); else spr1(0);
	if (sp[1]) spr2(1); else spr2(0);
	if (sp3) spr3(1); else spr3(0);
	
	return 98; // return wait time in msec
	}    // end Timer0100()
	
unsigned int Timer1000(void) // seconds
	{
	// unsigned char ix;
	// call with last1000 = Timer1000();
	
	// make sure we haven't called routine in last 1000 msec
	//if ((last1000+1) < outpc.seconds) return last1000; else last1000 = outpc.seconds;
    if (last1000) return last1000;
    
    // Check if we are cruising
    // if load ~= short term load (+/- 5 kPa)
    // AND
    // if engine rpm ~= rpm avg (+/- 200 rpm)
    // AND
    // if not braking recently
    // AND
    // if no recent gear change	 (reset cruise_flag to 0 when shifting)
    //
    // then increment cruising variable
    
    if    (((outpc.LOAD > outpc.LOAD_short) && ((outpc.LOAD - outpc.LOAD_short) < 50)) 
       ||  ((outpc.LOAD < outpc.LOAD_short) && ((outpc.LOAD_short - outpc.LOAD) < 50))) 
       	{
       	// increment the cruise flag
       	// which we use to determine if the speedo should be changing rapidly or not
        if     (((outpc.engine_rpm > outpc.RPM_short) && ((outpc.engine_rpm - outpc.RPM_short) < 200)) 
       		||  ((outpc.engine_rpm < outpc.RPM_short) && ((outpc.RPM_short - outpc.engine_rpm) < 200)))
        	{
        	// increment the cruise_flag
        	// we will declare cruising status (not allow drastic changes in speedo)
        	// when cruise_flag >= 10 seconds (since braking or shifting)
       		outpc.cruise_flag++;
        	}
       	}
    
	if (outpc.Slip_Adjustment) outpc.Slip_Adjustment--; // decrement unslip_counter

	
	if (shiftStatcounter) shiftStatcounter--;
	if (!shiftStatcounter) outpc.shiftStat = 0; // reset the shift status

	if (outpc.seconds%3) 
		{
		jake_brake = 0; // cancel jake brake  (let ADC reset it if still active)
		}
	
	if (WOT_flag > 0) WOT_flag--;       // decrement the WOT flag
	
	if ((outpc.seconds % 5) == 0) outpc.miscStatus &= ~0x40; // clear 6th bit (baud error) for datalog after 5 second
	
	return 976;
	}    // end Timer1000()
	
signed char ShiftNAG(signed char UPorDOWN)
{
// This function performs shifts on the NAG1 transmission
// by switching solenoids for the duration of a shift,
// and keeping them off otherwise.

// Only call this function if in2ram.NAGmode == 1

// This function takes one signed char as an argument. The argument is:
// positive value = shift up
// negative value = shift down
// 0 = report current gear

// The intended usage is:
//  if (inpram.NAGmode)
//		{
// 		outpc.current_gear = ShiftNAG(outpc.target_gear - outpc.current_gear);
//  	}

// The function returns the current_gear (whether or not the shift was successful)

// The duration of the shift (and the solenoid ON time)
// Is determined by the length of time it takes the actual rpm to reach the 
// calculated post-shift rpm (+/- a user specified percentage).

// The post-shift rpm is calculated in the afterShiftRPM() function. 
// The shift will not be performed if the post shift rpm is calculated to be 
// above the maximum or below the minimum. The current gear will be reported.

// The solenoid states for each shift are set in UpNAG[] and DownNAG[] 
// which are the first 10 elements of a 16 bit int as follows:

// This is the shift logic:

// Engaged 		Shift Solenoid	 	Resulting 
// Gear   		Enabled			   	Gear
//  1 			Output1 (y3) 		2
//  2 			Output1 (y3) 		1
//  2 			Output3 (y5) 		3
//  3 			Output3 (y5) 		2
//  3 			Output2 (y4) 		4
//  4 			Output2 (y4) 		3
//  4 			Output1 (y3) 		5
//  5 			Output1 (y3) 		4

// y3 aka. 1-2/4-5 (pulsed during cranking)
// y5 aka. 2-3	   
// y4 aka. 3-4	   (pulsed in Park or Neutral at 40% duty cycle)

// Shifts
// [0]: P <-> R
// [1]: R <-> N
// [2]: N <-> 1
// [3]: 1 <-> 2
// [4]: 2 <-> 3
// [5]: 3 <-> 4
// [6]: 4 <-> 5
// [7]: 5 <-> 6
// [8]: 6 <-> 7
// [9]: 7 <-> 8

// So we have NAGout1 = [0001001000] = 0x48 = 72
// So we have NAGout2 = [0000010000] = 0x20 = 32
// So we have NAGout3 = [0000100000] = 0x10 = 16
// where the index is:
// the current_gear+2 if we are upshifting
// the current gear+1 if we are downshifting 

signed int targetRPM, afterShiftRPM;

if (inpram.NAGmode != 1) return outpc.current_gear;

// Calculate expected rpm after shift
afterShiftRPM = (outpc.engine_rpm/20) * (inpram.gear_table[outpc.current_gear + UPorDOWN]/10)/ (inpram.gear_table[outpc.current_gear]/10)*20;

// Check for possible over-rev
if (afterShiftRPM >= in2ram.rpm_limit[outpc.current_gear - 1]) return outpc.current_gear;

// Check for possible under-rev
if (afterShiftRPM <= in2ram.under_rev_limit[outpc.current_gear - 2]) return outpc.current_gear;

// Check for PNR
if (outpc.current_gear < 1) return outpc.current_gear;

// Make sure shift is not out of range
if ((UPorDOWN ==0) || (UPorDOWN > inpram.num_gears) || (UPorDOWN < (-1*inpram.num_gears)))
	{
	return outpc.current_gear;
	}

// We will shift, so set MPC on Output4
Output4(1); // We will look up PWM percentage in called function

DelayTimer1 = in2ram.NAGwait;

if (UPorDOWN == 1) 
	{
	// upshift
	
	// determine target rpm
	targetRPM = (afterShiftRPM * (100+in2ram.NAGrpmTol))/100;
	
	// Send shift timing retard to MS-II
	if ((inpram.CAN_enabled) && ((in2ram.NAGretard >> (outpc.current_gear+2)) & 0x01))
	  {
	  if (outpc.LOAD > inpram.Tretard_load)
	    {
	    outpc.SpkAdj = inpram.upshift_retard;   // up shift
	    outpc.FuelAdj = 0; // reset in case we set for slip reasons
	    sendCANAdj(); 
	    } // end if (outpc.LOAD > ...
	    else	 // LOAD less than Tretard_load
	    {  // reset spark and fuel adjustments in case we have set them for slip reasons
	    outpc.SpkAdj = 0;
	    outpc.FuelAdj = 0;
  	  sendCANAdj();
  	  }
  	} // end if (inpram.CAN_enabled)
	
	// bit position for outpc.current_gear+2 is
	// shift bit to right by 
	// (outpc.current_gear+2)
	// then AND with 0x01
	// to get 1 or 0


	while ((outpc.engine_rpm > targetRPM) && DelayTimer1)
		{
		Output1(((in2ram.NAGout1 >> (outpc.current_gear+2)) & 0x01)); 
		Output2(((in2ram.NAGout2 >> (outpc.current_gear+2)) & 0x01));
		Output3(((in2ram.NAGout3 >> (outpc.current_gear+2)) & 0x01));
		checkCANbus();
		// do the Timer details
		last0001 = Timer0001();
		lastADC = TimerADC();
		last0100 = Timer0100();
		last1000 = Timer1000();
		}
	}
	else
	{
	// downshift
	
	// determine target rpm
	targetRPM = (afterShiftRPM * (100-in2ram.NAGrpmTol))/100;
	
	// Send shift timing retard to MS-II
	if ((inpram.CAN_enabled) && ((in2ram.NAGretard >> (outpc.current_gear+1)) & 0x01))
	  {
	  if (outpc.LOAD > inpram.Tretard_load)
	    {
	    outpc.SpkAdj = inpram.dwnshift_retard;  // down shift
	    outpc.FuelAdj = 0; // reset in case we set for slip reasons
	    sendCANAdj(); 
	    } // end if (outpc.LOAD > ...
	    else	 // LOAD less than Tretard_load
	    {  // reset spark and fuel adjustments in case we have set them for slip reasons
	    outpc.SpkAdj = 0;
	    outpc.FuelAdj = 0;
  	  sendCANAdj();
  	  }
  	} // end if (inpram.CAN_enabled)
	
	while ((outpc.engine_rpm < targetRPM) && DelayTimer1)
		{
		Output1(((in2ram.NAGout1 >> (outpc.current_gear+1)) & 0x01)); 
		Output2(((in2ram.NAGout2 >> (outpc.current_gear+1)) & 0x01));
		Output3(((in2ram.NAGout3 >> (outpc.current_gear+1)) & 0x01));
		checkCANbus();
		// do the Timer details
		last0001 = Timer0001();
		last0010 = Timer0010();
		lastADC = TimerADC();
		last0100 = Timer0100();
		last1000 = Timer1000();
		}
	}

// shut off MPC
Output4(0);

// Send gear retard timing to MS-II
if (inpram.CAN_enabled) 
  {
  if (outpc.LOAD > inpram.Tretard_load)
    {
    outpc.SpkAdj = CANretard(outpc.current_gear + UPorDOWN); 
    sendCANAdj();
    sentSpkAdj = outpc.current_gear + UPorDOWN;
    }
    else
    {
    outpc.SpkAdj = 0;
    sendCANAdj();
    sentSpkAdj = 0;              
    }
  } // end if (inpram.CAN_enabled)

return outpc.current_gear + UPorDOWN;

} // end shiftNAG

unsigned int getRatio(void)
	{
    // Get the current effective ratio
    unsigned int ratio;
    
	if (outpc.manual_gear < 1) 	 // PNR
		{  // if neutral
		if (!outpc.manual_gear) ratio = 0; else ratio = inpram.gear_table[0];
		return ratio;
		}
	 
	// Generate current overall transmission ratio
	// Overall ratio = VSS signal RPM / Input Shaft RPM     ...OR...
	// Overall ratio = VSS signal RPM / (Engine RPM/TCC ratio),
	// where TCC ratio == 1 when the converter is locked or the RPM > lockup speed
	// if the transmission is not in overrun
	
	if (outpc.engine_rpm > 300)	 // only if running
		{
		if (inpram.ic2_usage & 0x01)
			{
			// ISS measuring input speed directly
			ratio = (uint) (((long) outpc.is_rpm * 1000)/(long) os_rpm);
			}
			else if ((inpram.ic2_usage & 0x02) || (inpram.CAN_enabled)) // using tach, so estimating is_rpm
			{
			ratio = (uint) (((long) outpc.engine_rpm * 1000)/(long) os_rpm);
			}
		}
		else // no engine rpm
		{
		ratio = 0;
		}
		
	return ratio;
	} 

unsigned char getCurrentGear(unsigned int iRATIO)
	{
	signed char i,gear;
	
	gear = outpc.estimated_gear;  // Initialize
	
	// if shifting, do nothing and return
	if (shifting_flag) return gear; 
	
	// Use current effective ratio to get gear number
	if (outpc.lock_TCC) // *should* be a match but must account for rounding, input errors/uncertainty, etc.
		{
		for (i=1; i<=inpram.num_gears; i++) // for each gear in  turn
			{
			if (iRATIO > inpram.gear_table[i])
				{
				// if ((iRATIO - inpram.gear_table[i]) < (((long) inpram.gear_table[i] * (long) (100 + in2ram.gear_tolerance))/100))
				if (iRATIO < (((long) inpram.gear_table[i] * (long) (100 + in2ram.gear_tolerance))/100))
					{
					gear = i;
					break;
					}
				}
				else // outpc.instant_ratio <= inpram.gear_ratio[i] 
				{
				//	if ((inpram.gear_table[i] - iRATIO) < (((long) inpram.gear_table[i] * (long) (100 - in2ram.gear_tolerance))/100))
				if (iRATIO > (((long) inpram.gear_table[i] * (long) (100 - in2ram.gear_tolerance))/100))
						
					{
					gear = i;
					break;
					}
				}
			}
		}
		else  // TCC unlocked
		// gear is likely greater than an actual gear ratio
		// due to TCC slippage
		// so start from lowest gear and select the last gear that is 
		// less than or equal to current overall ratio
		// example: 
		// actual ratios 3.85, 2.20, 1.00, 0.70
		// with current overall ratio of 1.86
		// loop through -
		// 3.85 is greater than 1.86 so loop again
		// 2.20 is greater than 1.86 so loop
		// 1.00 is not greater than 1.86, so we must be in third (with TCC slippage)	
		{
		for (i=1; i<=inpram.num_gears; i++)
			{
			if (iRATIO >= ((uint) (((long) inpram.gear_table[i] * (long) (100 - in2ram.gear_tolerance))/100))) 
				{
				gear = i;
				break;
				} // end if (outpc.overall_ratio ...
			} // end for (i=1;...
		} // end else // TCC unlocked
		return gear;
	} // end getCurrentGear()
	
	
signed char auto_gear_lookupRPM(void)
{
signed char gear;

gear = outpc.current_gear; // initialize

// This function looks up the current gear's load x rpm shift tables
// if the auto mode is three - RPM based shifting.

// If the rpm is greater than or equal to the upshift rpm,
// the function returns the target gear = current gear + 1

// If the rpm is less than or equal to the downshift rpm
// the the function returns the target gear = current gear - 1

// We have 2 x (num_gears - 1) tables (num_gears 4...8). So up to 14 tables. 
// If each has 8 signed char elements  - the gears (so 8 bits), this is 112 bytes. 
// Plus an additional 16 bits for the load bins. 128 bytes overall.

// Set these up as a 14 x 8 array called shiftRPM[(numgears*2-1),8], where
// rows 0 - 6  are the upshift   values for shifts 1/2 ... 7/8 (i.e. current_gear - 1)
// rows 7 - 13 are the downshift values for shifts 2/1 ... 8/7 (i.e. current_gear + 5)


if ((inpram.shift_mode != 3) || (outpc.manual_gear < 1) || (outpc.current_gear < 1) || (shifting_flag)) return gear;

if (outpc.hyst_timer) return gear;

// upshift first
// int intrp_1dctable(char sgnx, int x, unsigned char n, int * x_table, char * z_table);
// ex. if (inpram.stdin_cfg & 0x02) tempPC2 = intrp_1dctable(	0,
//														outpc.clt,
//														NUM_TEMPS,
//														inpram.temp_table,
//														inpram.PCtemp)* PC_PER/100; // 1-D table lookup	


// Upshift first
// -------------
// if rpm is greater than this gear's upshift value
// AND only if rpm is less than next higher gears downshift value (so will not result in immediate downshift)
if (outpc.current_gear < inpram.num_gears) 
	{
	// only if not already at max gear
	if ((outpc.engine_rpm >= intrp_RPMtable(outpc.current_gear - 1, 1)))	
		{
		gear = outpc.current_gear + 1;
		return gear;
		}
	}

// Downshift next
// --------------
// if rpm is less than this gear's downshift value
// AND only if rpm will not result in immediate upshift
if (outpc.current_gear > 1) 
	{
	// only if not already at lowest gear
	if ((outpc.engine_rpm <= intrp_RPMtable((outpc.current_gear + 5), 0)))	
		{
		gear = outpc.current_gear - 1;
		}
	}
return gear;
}	 // end auto_gear_lookupRPM()

unsigned int intrp_RPMtable(unsigned char row, unsigned char up)  
{
// This function is used in the rpm based shifting mode.
// It takes the target gear and returns the up or downshift RPM based on the 
// RPMshiftTable[14 rows x 8 columns = 112 elements] table. 
// Up or down part of the table is set by the row.
// The RPM to shift up/down at for the load bins 
// Rows 0 - 6  are the upshift values for shifts 1/2 ... 7/8  (i.e. current_gear - 1)
// Rows 7 - 13 are the downshift values for shifts 2/1 ... 8/7 (i.e. current_gear + 5)

int ix;
long interp, interp3;

if (up == 1)
	{
	
	// bound input arguments
	if(outpc.LOAD_short > in2ram.RPMloadUPbins[7]) // if greater than largest bin  
		{
		return(in2ram.RPMshiftTable[(row*8)+7]); // rows run 0 to 14
		}
		
	if(outpc.LOAD_short < in2ram.RPMloadUPbins[0])  
		{
		return(in2ram.RPMshiftTable[(row*8) + 0]);
		}
	
	for(ix = 7; ix > -1; ix--)  
		{ 
		if(outpc.LOAD_short > in2ram.RPMloadUPbins[ix])  
			{
			break;
			}
		}

	if(ix < 0) ix = 0;
	
	interp =	in2ram.RPMloadUPbins[ix + 1] - in2ram.RPMloadUPbins[ix];
	if(interp != 0)  
		{
		interp3 = outpc.LOAD_short - in2ram.RPMloadUPbins[ix];
		interp3 = (100 * interp3);
		interp = interp3 / interp;
		}

	return((uint) (in2ram.RPMshiftTable[(row*8)+ix] +
		interp * (in2ram.RPMshiftTable[(row*8)+ix+1] - in2ram.RPMshiftTable[(row*8)+ix])/ 100));
	}
	else // up == 0
	{
	// Downshift
	// bound input arguments
	if(outpc.LOAD_short > in2ram.RPMloadDWNbins[7]) // if greater than largest bin  
		{
		return(in2ram.RPMshiftTable[(row*8)+7]); // rows run 0 to 14
		}
		
	if(outpc.LOAD_short < in2ram.RPMloadDWNbins[0])  
		{
		return(in2ram.RPMshiftTable[(row*8) + 0]);
		}
	
	for(ix = 7; ix > -1; ix--)  
		{ 
		if(outpc.LOAD_short > in2ram.RPMloadDWNbins[ix])  
			{
			break;
			}
		}

	if(ix < 0) ix = 0;
	
	interp =	in2ram.RPMloadDWNbins[ix + 1] - in2ram.RPMloadDWNbins[ix];
	if(interp != 0)  
		{
		interp3 = outpc.LOAD_short - in2ram.RPMloadDWNbins[ix];
		interp3 = (100 * interp3);
		interp = interp3 / interp;
		}

	return((uint) (in2ram.RPMshiftTable[(row*8) + ix] +
		interp * (in2ram.RPMshiftTable[(row*8) + ix+1] - in2ram.RPMshiftTable[(row*8) + ix])/ 100));
	}
	
}   // end intrp_RPMtable()
	
#pragma CODE_SEG ROM_7000

#include "4.000_changes.h"
